/*****************************************************************************
 *------------------------ Utility  -----------------------------------------*
 *****************************************************************************/
#include <math.h>
#include "rsvna.h"
#include "rscore.h"

#if defined (_CVI_)
    #include <utility.h>
#endif

#pragma warning( disable : 4152 )

extern ViString dataFormatArr[];
/*****************************************************************************
 *- Callback Declarations ---------------------------------------------------*
 *****************************************************************************/

/*****************************************************************************
 *- Common Callback Declarations --------------------------------------------*
 *****************************************************************************/

 /// HIFN  Checks the range of a ViInt32 value against a specified range table.
 /// HIRET Returns the status code of this operation.
 /// HIPAR instrSession/The ViSession handle that you obtain from the rsvna_init or
 /// HIPAR instrSession/rsvna_InitWithOptions function. The handle identifies a particular
 /// HIPAR instrSession/instrument session.
 /// HIPAR rangeTable/RsCoreRangeTableRangedPtr to the range table.
 /// HIPAR value/The value to range check.
ViStatus rsvna_CheckRangeTableRangeViInt32(ViSession                 instrSession,
                                           RsCoreRangeTableRangedPtr rangeTable,
                                           ViInt32                   value)
{
    return RsCore_InvalidViInt32Range(instrSession, value,
        (ViInt32)(rangeTable->min), (ViInt32)(rangeTable->max));
}


/*****************************************************************************
 * Function: rsvna_InitAttributes
 * Purpose:  This function inits attributes to the desired values if needed.
 *****************************************************************************/
ViStatus rsvna_InitAttributes(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    // - Class Driver Identification
    //checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_CLASS_DRIVER_DESCRIPTION, 0, RSVNA_CLASS_DRIVER_DESCRIPTION));
    //checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_CLASS_DRIVER_PREFIX, 0, RSVNA_CLASS_DRIVER_PREFIX));
    //checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_CLASS_DRIVER_VENDOR, 0, RSVNA_CLASS_DRIVER_VENDOR));
    //checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_CLASS_DRIVER_REVISION, 0, RSVNA_CLASS_DRIVER_REVISION));
    //checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION, 0, RSVNA_CLASS_SPEC_MAJOR_VERSION));
    //checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION, 0, RSVNA_CLASS_SPEC_MINOR_VERSION));

    // - Driver Identification
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_DESCRIPTION, 0, RSVNA_SPECIFIC_DRIVER_DESCRIPTION));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_PREFIX, 0, RSVNA_SPECIFIC_DRIVER_PREFIX));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_LOCATOR, 0, RSVNA_SPECIFIC_DRIVER_LOCATOR));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_VENDOR, 0, RSVNA_SPECIFIC_DRIVER_VENDOR));

    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION, 0, RSVNA_CLASS_SPEC_MAJOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION, 0, RSVNA_CLASS_SPEC_MINOR_VERSION));

    // - Driver Capabilities
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SUPPORTED_INSTRUMENT_MODELS, 0, RSVNA_SUPPORTED_INSTRUMENT_MODELS));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_GROUP_CAPABILITIES, 0, RSVNA_GROUP_CAPABILITIES));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_FUNCTION_CAPABILITIES, 0, RSVNA_FUNCTION_CAPABILITIES));

    // - Version Info
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MAJOR_VERSION, 0, RSVNA_MAJOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MINOR_VERSION, 0, RSVNA_MINOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MINOR_MINOR_VERSION, 0, RSVNA_MINOR_MINOR_VERSION));

    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_CLASS_DRIVER_MAJOR_VERSION, 0, RSVNA_CLASS_SPEC_MAJOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_CLASS_DRIVER_MINOR_VERSION, 0, RSVNA_CLASS_SPEC_MINOR_VERSION));

    checkErr(RsCore_SetSpecificDriverRevision(instrSession, RS_ATTR_SPECIFIC_DRIVER_REVISION));

Error:
    return error;
}

/*****************************************************************************
 * Function: rsvna_DefaultInstrSetup
 * Purpose:  This function sends a default setup to the instrument. The
 *           rsvna_reset function calls this function. The
 *           rsvna_RsInit function calls this function when the
 *           user passes VI_FALSE for the reset parameter. This function is
 *           useful for configuring settings that other instrument driver
 *           functions require.
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
ViStatus rsvna_DefaultInstrSetup(ViSession instrSession)
{
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession = NULL;
    ViInt32          portCount;

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));
    // Set all the attributes to the default state. Do not update inherent attributes!
    checkErr(RsCore_ApplyAttributeDefaults(instrSession, VI_FALSE));

    checkErr(rsvna_InitAttributes(instrSession));
    rsSession->fastSweepInstrument = 0;
    checkErr(RsCore_ResetRegistersEseSre(instrSession));

    if (!RsCore_Simulating(instrSession))
    {
        checkErr(RsCore_Write(instrSession, "*CLS"));
        // Instrument identification

        // Number of ports
        checkErr(RsCore_QueryViInt32(instrSession, "INST:PORT:COUN?", &portCount));
        checkErr(RsCore_SetAttributeViInt32(instrSession, "", RSVNA_ATTR_NUMBER_OF_PORTS, 0, portCount));
    }

Error:
    return error;
}

/*****************************************************************************
 * Function: rsvna_CheckStatus
 * Purpose:  This function checks the status of the instrument to detect
 *           whether the instrument has encountered an error. This function
 *           is called at the end of most exported driver functions. If the
 *           instrument reports an error, this function returns
 *           RS_ERROR_INSTRUMENT_SPECIFIC. The user can set the
 *           rsvna_ATTR_QUERY_INSTRUMENT_STATUS attribute to VI_FALSE to disable this
 *           check and increase execution speed.
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
ViStatus rsvna_CheckStatus(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    if (RsCore_QueryInstrStatus(instrSession) && !RsCore_Simulating(instrSession))
    {
        checkErr(RsCore_CheckStatus(instrSession, VI_SUCCESS));
    }

Error:
    return error;
}

/*****************************************************************************
 * Function: Parse Limit Line Catalog
 * Purpose:  Parse Limit Line Catalog returned from instrument and removes
 *           from it all unneeded parameters from this list.
*****************************************************************************/
ViStatus rsvna_ParseLimitLineCatalog(ViChar* buffer, ViInt32 size, ViChar* catalogList, ViInt32* nbValues)
{
    ViStatus rsvna_status = VI_SUCCESS;
    ViChar* plist_of_files = 0;
    ViChar* ptag = NULL;
    ViChar* pend = NULL;
    ViInt32 i = 0;

    if ((buffer) && (*buffer != '\0'))
    {
        (void)sscanf(buffer, "%*ld,%*ld,%[^\r\n]", buffer);
        plist_of_files = (ViChar*)malloc(strlen(buffer) + 1);
        if (plist_of_files == NULL)
        {
            rsvna_status = RS_ERROR_ALLOC;
            goto Error;
        }
        pend = plist_of_files;
        *pend = '\0';
        ptag = strtok(buffer, ",");
        while (ptag != NULL && rsvna_status == VI_SUCCESS)
        {
            i++;
            pend += sprintf(pend, "%s,", ptag + 1);
            ptag = strtok(NULL, ",");
            if (ptag == NULL)
                rsvna_status = RS_ERROR_INVALID_VALUE;
            else
                ptag = strtok(NULL, ",");
        }

        *--pend = '\0';

        strncpy(catalogList, plist_of_files, size);
    }

    if (nbValues) *nbValues = i;

Error:
    if (plist_of_files) free(plist_of_files);
    return rsvna_status;
}

/*****************************************************************************
 * Function: Write Memory Trace Data Extended
 * Purpose:  This function reads the stimulus values of the active data or
 *           memory trace.
 *****************************************************************************/
ViStatus rsvna_WriteMemoryTraceDataExt(ViSession instrSession,
                                       ViInt32   channel_Trace,
                                       ViInt32   dataFormat,
                                       ViInt32   no_ofPoints,
                                       ViReal64  traceData[])
{  
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession;
    ViChar           cmd[RS_MAX_MESSAGE_BUF_SIZE];
    ViInt32          dataSize = 0;
    ViByte*          data = NULL;

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, no_ofPoints, 1, 250000),
            4, "No of Points");

    if (rsSession->sessionType == RS_INTF_ASRL)
    {
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, ":FORMAT ASCII;:CALC%ld:DATA %s,", channel_Trace, dataFormatArr[dataFormat]);
        checkErr(RsCore_WriteAsciiViReal64Array(instrSession, cmd, traceData, no_ofPoints));
    }
    else
    {
        if (rsSession->sessionType == RS_INTF_GPIB)
        {
            checkErr(RsCore_Write(instrSession, ":SYST:COMM:GPIB:SELF:RTER EOI"));
        }

        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, ":FORM REAL,32;:CALC%ld:DATA %s,", channel_Trace, dataFormatArr[dataFormat]);
        checkErr(RsCore_Convert_Float32Array_To_BinaryDataBlock(instrSession, traceData, no_ofPoints, &data, &dataSize, VI_FALSE));
        checkErr(RsCore_WriteBinaryDataBlock(instrSession, cmd, data, dataSize));

        if (rsSession->sessionType == RS_INTF_GPIB)
        {
            checkErr(RsCore_Write(instrSession, ":SYST:COMM:GPIB:SELF:RTER LFEOI"));
        }
    }

    checkErr(rsvna_CheckStatus(instrSession));

Error:
    if (data)
        free(data);

    RsCore_Write(instrSession, "FORM ASC");

    return error;
}

/*****************************************************************************
 * Function: Get ViReal64 Array Window
*****************************************************************************/
ViStatus rsvna_getViReal64ArrayWindow(ViSession     instrSession,
                                      ViConstString cmdFormatString,
                                      ViInt32       window,
                                      ViReal64      array[],
                                      ViInt32       bufferSize,
                                      ViStatus      parameterErrorCode)
{  
    ViChar cmd[RS_MAX_MESSAGE_BUF_SIZE];

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, cmdFormatString, window);  // NOLINT(clang-diagnostic-format-nonliteral)

    return rsvna_getViReal64Array(instrSession, cmd, array, bufferSize, parameterErrorCode);
}

/*****************************************************************************
 * Function: Get ViReal64 Array
*****************************************************************************/
ViStatus rsvna_getViReal64Array(ViSession     instrSession,
                                ViConstString cmd,
                                ViReal64      array[],
                                ViInt32       bufferSize,
                                ViStatus      parameterErrorCode)
{  
    ViStatus error = VI_SUCCESS;

    if (array == NULL)
        return parameterErrorCode;

    checkErr(RsCore_QueryFloatArrayToUserBuffer(instrSession, cmd, bufferSize, array, NULL));
    checkErr(rsvna_CheckStatus(instrSession));

Error:
    return error;
}

/*****************************************************************************
 * Function: Format Parameter Command
 * Purpose:  This function select appropriate command according to whether the
 *           trace already exists or not
 *****************************************************************************/
ViStatus rsvna_traceCommandSelect(ViSession instrSession,
                                  ViInt32   channel,
                                  ViString  traceName,
                                  ViChar    buffer[],
                                  ViInt32*  offset)
{  
    ViChar   rd_buffer [RS_MAX_MESSAGE_BUF_SIZE];
    ViStatus error;

    checkErr(rsvna_TraceList(instrSession, channel, rd_buffer, RS_MAX_MESSAGE_BUF_SIZE));

    if (strstr(rd_buffer, traceName))
        *offset = sprintf(buffer, "CALC%ld:PAR:MEAS ", channel);
    else
        *offset = sprintf(buffer, "CALC%ld:PAR:SDEF ", channel);

Error:
    return error;
}

ViStatus rsvna_GetOrAddSequenceRepCap(ViSession instrSession,
                                      ViAttr    attribute,
                                      ViInt32   sequence,
                                      ViInt32   bufferSize,
                                      ViChar    repCapBuffer[])
{  
    ViStatus      error = VI_SUCCESS;
    ViChar        repCapCmd[RS_REPCAP_BUF_SIZE] = "";
    ViChar        repCapId[RS_REPCAP_BUF_SIZE] = "";
    ViStatus      sequence_status = VI_SUCCESS;
    ViConstString rcNameId = "Sequence";

    sequence_status = RsCore_GetAttributeRepCapName(instrSession,
                      attribute,
                      "Sequence",
                      sequence - 1,
                      RS_REPCAP_BUF_SIZE,
        repCapId);

    if (sequence_status == VI_SUCCESS)
    {
        //sequence number already exists
    }
    else
    {
        //sequence number does not exist, add it
        snprintf(repCapId, RS_REPCAP_BUF_SIZE, "Seq%ld", sequence);
        snprintf(repCapCmd, RS_REPCAP_BUF_SIZE, "%ld", sequence);
        checkErr(RsCore_AddRepCap(instrSession, rcNameId, repCapId, repCapCmd));
    }

    strncpy(repCapBuffer, repCapId, bufferSize);

Error:
    return error;
}

ViStatus rsvna_utility_error_message(ViSession instrSession,
                                     ViStatus  errorCode,
                                     ViChar    errorMessage[])
{
    ViStatus error = VI_SUCCESS;
    static   RsCoreStringValueTable errorTable =
             {
             RSVNA_ERROR_CODES_AND_MSGS,
    {0,      NULL}
             };

    (void)(RsCore_LockSession(instrSession));

    // all VISA and RS error codes are handled as well as codes in the table
    viCheckParm(RsCore_InvalidNullPointer(instrSession, errorMessage), 3, "Null address for Error Message");

    checkErr(RsCore_GetSpecificDriverStatusDesc(instrSession, errorCode, errorMessage, 256, errorTable));

Error:  
    (void)RsCore_UnlockSession(instrSession);
    return error;
}