/****************************************************************************
 *
 *  Rohde & Schwarz Vector Network Analyzer instrument driver include file
 *
 ****************************************************************************/

#ifndef __RSVNA_HEADER
#define __RSVNA_HEADER

#include "rscore.h"
#include "rsvna_attributes.h"
#include "rsvna_utility.h"
#include "rsvna_internal.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#if defined(__cplusplus) || defined(__cplusplus__)
extern "C" {
#endif

    /*****************************************************************************
     *- Useful Macros & Constants -----------------------------------------------*
     *****************************************************************************/
#define RSVNA_SIMULATION_ID_QUERY      "Rohde-Schwarz,VNA,123456/789,1.23"  /* Default definition of ID Query Response for simulation */
#define RSVNA_VALID_ID_RESPONSE_STRING "Rohde-Schwarz"      /* Valid response for identification query */
#define RSVNA_SIMULATION_OPT_QUERY     ""      /* Simulated response for *OPT? command */

#define RSVNA_OPC_TIMEOUT              5000L   /* Maximum time to wait for OPC in milliseconds */
#define RS_VAL_NAN_VI_REAL64                 9.91e+37

#define CHAN_RC_FORMAT "Channel%ld"

    /****************************************************************************
     *----------------- Instrument Driver Revision Information -----------------*
     ****************************************************************************/

    /* Class Driver Identification */

#define RSVNA_CLASS_DRIVER_DESCRIPTION           ""
#define RSVNA_CLASS_DRIVER_PREFIX                ""
#define RSVNA_CLASS_DRIVER_VENDOR                ""
#define RSVNA_CLASS_DRIVER_REVISION              ""
#define RSVNA_CLASS_SPEC_MAJOR_VERSION           1L    /* Class specification major version */
#define RSVNA_CLASS_SPEC_MINOR_VERSION           0L    /* Class specification minor version */

    /* Driver Identification */

#define RSVNA_SPECIFIC_DRIVER_DESCRIPTION        ""
#define RSVNA_SPECIFIC_DRIVER_PREFIX             "RSVNA"
#define RSVNA_SPECIFIC_DRIVER_LOCATOR            ""
#define RSVNA_SPECIFIC_DRIVER_VENDOR             ""
#define RSVNA_MAJOR_VERSION                      3L    /* Instrument driver major version          */
#define RSVNA_MINOR_VERSION                      46L   /* Instrument driver minor version          */
#define RSVNA_MINOR_MINOR_VERSION                0L    /* Instrument driver minor minor version    */

    /* Driver Capabilities */

#define RSVNA_SUPPORTED_INSTRUMENT_MODELS        "ZNB,ZNC,ZNBT,ZND,ZNL,ZNLE,ZNA"    /* Instrument driver supported model(s)     */
#define RSVNA_GROUP_CAPABILITIES                 ""
#define RSVNA_FUNCTION_CAPABILITIES              ""

    /****************************************************************************
     *---------------------------- Attribute Defines ---------------------------*
     ****************************************************************************/

    /*- Inherent Instrument Specific Attributes ----------------------------*/

    /* User Options */

#define RSVNA_ATTR_RANGE_CHECK                                           RS_ATTR_RANGE_CHECK                                     /* ViBoolean */
#define RSVNA_ATTR_QUERY_INSTRUMENT_STATUS                               RS_ATTR_QUERY_INSTRUMENT_STATUS                         /* ViBoolean */
#define RSVNA_ATTR_CACHE                                                 RS_ATTR_CACHE                                           /* ViBoolean */
#define RSVNA_ATTR_SIMULATE                                              RS_ATTR_SIMULATE                                        /* ViBoolean */
#define RSVNA_ATTR_RECORD_COERCIONS                                      RS_ATTR_RECORD_COERCIONS                                /* ViBoolean */
#define RSVNA_ATTR_INTERCHANGE_CHECK                                     RS_ATTR_INTERCHANGE_CHECK                               /* ViBoolean */
#define RSVNA_ATTR_SPY                                                   RS_ATTR_SPY                                             /* ViBoolean */
#define RSVNA_ATTR_USE_SPECIFIC_SIMULATION                               RS_ATTR_USE_SPECIFIC_SIMULATION                         /* ViBoolean */

    /* Class Driver Identification */

#define RSVNA_ATTR_CLASS_DRIVER_DESCRIPTION                              RS_ATTR_CLASS_DRIVER_DESCRIPTION                        /* ViString, read-only */
#define RSVNA_ATTR_CLASS_DRIVER_PREFIX                                   RS_ATTR_CLASS_DRIVER_PREFIX                             /* ViString, read-only */
#define RSVNA_ATTR_CLASS_DRIVER_VENDOR                                   RS_ATTR_CLASS_DRIVER_VENDOR                             /* ViString, read-only */
#define RSVNA_ATTR_CLASS_DRIVER_REVISION                                 RS_ATTR_CLASS_DRIVER_REVISION                           /* ViString, read-only */
#define RSVNA_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION                 RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION           /* ViInt32,  read-only */
#define RSVNA_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION                 RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION           /* ViInt32,  read-only */

    /* Driver Identification */

#define RSVNA_ATTR_SPECIFIC_DRIVER_DESCRIPTION                           RS_ATTR_SPECIFIC_DRIVER_DESCRIPTION                     /* ViString, read-only */
#define RSVNA_ATTR_SPECIFIC_DRIVER_PREFIX                                RS_ATTR_SPECIFIC_DRIVER_PREFIX                          /* ViString, read-only */
#define RSVNA_ATTR_SPECIFIC_DRIVER_LOCATOR                               RS_ATTR_SPECIFIC_DRIVER_LOCATOR                         /* ViString, read-only */
#define RSVNA_ATTR_SPECIFIC_DRIVER_VENDOR                                RS_ATTR_SPECIFIC_DRIVER_VENDOR                          /* ViString, read-only */
#define RSVNA_ATTR_SPECIFIC_DRIVER_REVISION                              RS_ATTR_SPECIFIC_DRIVER_REVISION                        /* ViString, read-only */
#define RSVNA_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION              RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION        /* ViInt32,  read-only */
#define RSVNA_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION              RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION        /* ViInt32,  read-only */

    /* Driver Capabilities */

#define RSVNA_ATTR_SUPPORTED_INSTRUMENT_MODELS                           RS_ATTR_SUPPORTED_INSTRUMENT_MODELS                     /* ViString, read-only */
#define RSVNA_ATTR_GROUP_CAPABILITIES                                    RS_ATTR_GROUP_CAPABILITIES                              /* ViString, read-only */
#define RSVNA_ATTR_FUNCTION_CAPABILITIES                                 RS_ATTR_FUNCTION_CAPABILITIES                           /* ViString, read-only */
#define RSVNA_ATTR_CHANNEL_COUNT                                         RS_ATTR_CHANNEL_COUNT                                   /* ViInt32,  read-only */

    /* Driver Setup */

#define RSVNA_ATTR_DRIVER_SETUP                                          RS_ATTR_DRIVER_SETUP                                    /* ViString            */

    /* Instrument Identification */

#define RSVNA_ATTR_INSTRUMENT_MANUFACTURER                               RS_ATTR_INSTRUMENT_MANUFACTURER                         /* ViString, read-only */
#define RSVNA_ATTR_INSTRUMENT_MODEL                                      RS_ATTR_INSTRUMENT_MODEL                                /* ViString, read-only */
#define RSVNA_ATTR_INSTRUMENT_FIRMWARE_REVISION                          RS_ATTR_INSTRUMENT_FIRMWARE_REVISION                    /* ViString, read-only */

    /* Advanced Session Information */

#define RSVNA_ATTR_IO_RESOURCE_DESCRIPTOR                                RS_ATTR_IO_RESOURCE_DESCRIPTOR                          /* ViString, read-only */
#define RSVNA_ATTR_LOGICAL_NAME                                          RS_ATTR_LOGICAL_NAME

    /* ViString, read-only */

    /*- Instrument-Specific Attributes -------------------------------------*/

//#define RSVNA_ATTR_ID_QUERY_RESPONSE                                     (RS_SPECIFIC_PUBLIC_ATTR_BASE + 1L)                     /* ViString (Read Only)              */

    /* BEGIN GENERATE */
#define RSVNA_ATTR_INIT_SW_TRIGGER                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 386
#define RSVNA_ATTR_FILE_RECALL_SET_NEW                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 387
#define RSVNA_ATTR_FILE_RECALL_SET_ACTIVE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 388
#define RSVNA_ATTR_FILE_RECALL_SET_CLOSE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 389
#define RSVNA_ATTR_FILE_RECALL_SET_CLOSE_ALL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 390
#define RSVNA_ATTR_FILE_RECALL_SET_CATALOG                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 391
#define RSVNA_ATTR_FILE_RECALL_SET_STORE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 392
#define RSVNA_ATTR_FILE_RECALL_SET_LOAD                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 393
#define RSVNA_ATTR_FILE_RECALL_SET_COUNT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 842
#define RSVNA_ATTR_FILE_CHANGE_DRIVE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 394
#define RSVNA_ATTR_FILE_CHANGE_DIRECTORY                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 395
#define RSVNA_ATTR_FILE_CREATE_DIRECTORY                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 396
#define RSVNA_ATTR_FILE_REMOVE_DIRECTORY                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 397
#define RSVNA_ATTR_FILE_REMOVE                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 398
#define RSVNA_ATTR_FILE_LIST_OF_FAVORITE_RECALL_SET                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 877
#define RSVNA_ATTR_FILE_LOAD_DISTANCE_TO_FAULT_CABLE_TYPES                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 878
#define RSVNA_ATTR_FILE_STORE_DISTANCE_TO_FAULT_CABLE_TYPES                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 879
#define RSVNA_ATTR_FILE_ADDED_COMMENT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 880
#define RSVNA_ATTR_FILE_SYMMETRIC_PARAMS                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 881
#define RSVNA_ATTR_FILE_DATA_DECIMAL_PLACES                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1203
#define RSVNA_ATTR_FILE_STIMULUS_DECIMAL_PLACES                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1204
#define RSVNA_ATTR_FILE_TOUCHSTONE_EXPORT_FORMAT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1205
#define RSVNA_ATTR_PRINT_START                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 399
#define RSVNA_ATTR_PRINT_PRINTER_NAME                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 400
#define RSVNA_ATTR_PRINT_FILE_NAME                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 401
#define RSVNA_ATTR_PRINT_FILE_FORMAT                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 402
#define RSVNA_ATTR_PRINT_ALL_HEADERS                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 403
#define RSVNA_ATTR_PRINT_LOGO                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 404
#define RSVNA_ATTR_PRINT_MARKER_LIST                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 405
#define RSVNA_ATTR_PRINT_TIME                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 406
#define RSVNA_ATTR_PRINT_DIAGRAM_AREA                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 407
#define RSVNA_ATTR_PRINT_USE_COLOR                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 408
#define RSVNA_ATTR_PRINT_COLOR_SCHEME                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 882
#define RSVNA_ATTR_PRINT_PAGE_ORIENTATION                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 409
#define RSVNA_ATTR_PRINT_PAGE_MARGIN                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 410
#define RSVNA_ATTR_PARAMETER_CONVERSION                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 411
#define RSVNA_ATTR_WAVE_QUANTITIES_UNIT_TYPE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 412
#define RSVNA_ATTR_WAVE_QUANTITIES_PHASE_NORMALIZATION_ENABLED                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1362
#define RSVNA_ATTR_DC_RANGE                                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 413
#define RSVNA_ATTR_PAE_MEASUREMENT_TYPE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 414
#define RSVNA_ATTR_PAE_DC_INPUT                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 415
#define RSVNA_ATTR_PAE_PARAMETERS_VALUE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 416
#define RSVNA_ATTR_DELETE_BALANCED_PORT                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 417
#define RSVNA_ATTR_DELETE_ALL_BALANCED_PORTS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 418
#define RSVNA_ATTR_LOGICAL_PORT_REFERENCE_IMPEDANCE_RENORMALIZATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 419
#define RSVNA_ATTR_LOGICAL_PORT_REFERENCE_IMPEDANCE_USE_DEFAULT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 843
#define RSVNA_ATTR_TRACE_FORMAT                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 420
#define RSVNA_ATTR_TRACE_APERTURE_POINTS                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 421
#define RSVNA_ATTR_TRACE_ASSIGN_NAME                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 422
#define RSVNA_ATTR_TRACE_ASSIGN_DIAGRAM                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 423
#define RSVNA_ATTR_TRACE_DIAGRAM_STATE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 424
#define RSVNA_ATTR_TRACE_SELECT_ACTIVE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 425
#define RSVNA_ATTR_TRACE_DELETE                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 426
#define RSVNA_ATTR_TRACE_DELETE_ALL                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 427
#define RSVNA_ATTR_TRACE_DELETE_ALL_CHANNELS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 428
#define RSVNA_ATTR_TRACE_CHANNEL_CATALOG                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 429
#define RSVNA_ATTR_TRACE_CHANNEL_CATALOG_BALANCED_PORTS                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 821
#define RSVNA_ATTR_TRACE_CHANNEL_RENAME                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 430
#define RSVNA_ATTR_TRACE_CATALOG                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 431
#define RSVNA_ATTR_TRACE_NAME_CATALOG                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 432
#define RSVNA_ATTR_TRACE_DATA_COPY_TO_MEMMORY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 433
#define RSVNA_ATTR_MEMORY_TRACE_DELETE_ALL                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 844
#define RSVNA_ATTR_MEMORY_TRACE_DELETE_ALL_CHANNELS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 845
#define RSVNA_ATTR_TRACE_MATH_EXPRESSION                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 434
#define RSVNA_ATTR_TRACE_MATH_FORMATTED_EXPRESSION                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 883
#define RSVNA_ATTR_TRACE_MATH_MODE_STATE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 435
#define RSVNA_ATTR_TRACE_MATH_FORMATTED_MODE_ENABLED                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 884
#define RSVNA_ATTR_TRACE_MATH_FUNCTION                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 436
#define RSVNA_ATTR_TRACE_MATH_FORMATTED_FUNCTION                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 885
#define RSVNA_ATTR_TRACE_MATH_WAVE_QUANTITY_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 437
#define RSVNA_ATTR_TRACE_RESPONSE_S_DATA_CATALOG                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 438
#define RSVNA_ATTR_QUERY_TRACE_COMPLETE_SWEEP_COUNT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 439
#define RSVNA_ATTR_TRACE_SHIFT_STIMULUS_VALUE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 440
#define RSVNA_ATTR_SMOOTHING_STATE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 441
#define RSVNA_ATTR_SMOOTHING_APERTURE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 442
#define RSVNA_ATTR_HOLD_TRACE                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 443
#define RSVNA_ATTR_IMPORT_TRACE_DATA_AUTO_DISTRIBUTE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 826
#define RSVNA_ATTR_DEFINE_SWEEP_AND_MEASUREMENT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 886
#define RSVNA_ATTR_QUERY_SWEEP_AND_MEASUREMENT_DEFINITION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 889
#define RSVNA_ATTR_TIME_DOMAIN_STATE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 444
#define RSVNA_ATTR_TIME_DOMAIN_TYPE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 445
#define RSVNA_ATTR_TIME_DOMAIN_STIMULUS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 446
#define RSVNA_ATTR_TIME_DOMAIN_IMPULSE_RESPONSE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 447
#define RSVNA_ATTR_TIME_DOMAIN_SIDE_LOBE_LEVEL                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 448
#define RSVNA_ATTR_TIME_DOMAIN_RESOLUTION_ENHANCEMENT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 449
#define RSVNA_ATTR_TIME_DOMAIN_HARMONIC_GRID_AND_KEEP                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 450
#define RSVNA_ATTR_TIME_DOMAIN_DC_VALUE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 451
#define RSVNA_ATTR_TIME_DOMAIN_EXTRAPOLATE_DC_VALUE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 452
#define RSVNA_ATTR_TIME_DOMAIN_CONTINUOUS_EXTRAPOLATION                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 453
#define RSVNA_ATTR_TIME_DOMAIN_CALCULATE_HARMONIC_GRID                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 454
#define RSVNA_ATTR_TIME_DOMAIN_VSWR_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 894
#define RSVNA_ATTR_TIME_GATE_STATE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 455
#define RSVNA_ATTR_TIME_GATE_TYPE                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 456
#define RSVNA_ATTR_TIME_GATE_SHOW_RANGE_LINES                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 457
#define RSVNA_ATTR_TIME_GATE_FILTER_TYPE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 458
#define RSVNA_ATTR_TIME_GATE_SIDEBAND_SUPRESSION                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 459
#define RSVNA_ATTR_TIME_GATE_START_TIME                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 460
#define RSVNA_ATTR_TIME_GATE_STOP_TIME                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 461
#define RSVNA_ATTR_TIME_GATE_CENTER_TIME                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 462
#define RSVNA_ATTR_TIME_GATE_SPAN                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 463
#define RSVNA_ATTR_TIME_GATE_OFFSET_STATE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 798
#define RSVNA_ATTR_TIME_GATE_LINKING                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1680
#define RSVNA_ATTR_DISTANCE_TO_FAULT_STATE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 895
#define RSVNA_ATTR_DISTANCE_TO_FAULT_WINDOW_START_DISTANCE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 896
#define RSVNA_ATTR_DISTANCE_TO_FAULT_WINDOW_STOP_DISTANCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 897
#define RSVNA_ATTR_DISTANCE_TO_FAULT_WINDOW_CENTER                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 898
#define RSVNA_ATTR_DISTANCE_TO_FAULT_WINDOW_SPAN                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 899
#define RSVNA_ATTR_DISTANCE_TO_FAULT_AUTO_NUMBER_OF_POINTS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 900
#define RSVNA_ATTR_DISTANCE_TO_FAULT_CABLE_TYPE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 901
#define RSVNA_ATTR_DISTANCE_TO_FAULT_LIMIT_CHECK_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 902
#define RSVNA_ATTR_DISTANCE_TO_FAULT_LIMIT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 903
#define RSVNA_ATTR_DISTANCE_TO_FAULT_DEFINE_CABLE_TYPE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 904
#define RSVNA_ATTR_DISTANCE_TO_FAULT_DELETE_CABLE_TYPE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 906
#define RSVNA_ATTR_DISTANCE_TO_FAULT_PEAK_LIST_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 907
#define RSVNA_ATTR_STATISTICS_DISPLAY_RESULTS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 464
#define RSVNA_ATTR_STATISTICS_DISPLAY_ALL_RESULTS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 465
#define RSVNA_ATTR_STATISTICS_EVALUATION_RANGE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 466
#define RSVNA_ATTR_STATISTICS_EVALUATION_RANGE_START                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 467
#define RSVNA_ATTR_STATISTICS_EVALUATION_RANGE_STOP                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 468
#define RSVNA_ATTR_STATISTICS_EVALUATION_RANGE_SHOW_LIMIT_LINES                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 469
#define RSVNA_ATTR_STATISTICAL_RESULTS                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 470
#define RSVNA_ATTR_STATISTICAL_FORMAT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 846
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_VALUE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 471
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_POINT_RESULT_STATE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 472
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_PHASE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 909
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_VALUE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 910
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_LEVEL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 911
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_MARKER                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 912
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_PHASE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 913
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_RANGE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 914
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_RANGE_SHOW                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 915
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_RANGE_START                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 916
#define RSVNA_ATTR_STATISTICAL_COMPRESSION_REFERENCE_RANGE_STOP                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 917
#define RSVNA_ATTR_SHOW_LIMIT_LINE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 473
#define RSVNA_ATTR_LIMIT_CHECK                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 474
#define RSVNA_ATTR_LIMIT_CHECK_RESULT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 475
#define RSVNA_ATTR_GLOBAL_LIMIT_CHECK_RESULT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 476
#define RSVNA_ATTR_LIMIT_CHECK_CLEAR                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 799
#define RSVNA_ATTR_LIMIT_LINE_FAIL_BEEP                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 477
#define RSVNA_ATTR_LIMIT_LINE_SEGMENT_TYPE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 478
#define RSVNA_ATTR_LIMIT_LINE_SEGMENT_VALUES                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 479
#define RSVNA_ATTR_LIMIT_LINE_SEGMENT_COUNT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 480
#define RSVNA_ATTR_LIMIT_LINE_SEGMENT_FORMULA_ENABLED                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 918
#define RSVNA_ATTR_LIMIT_LINE_SEGMENT_FORMULA                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 919
#define RSVNA_ATTR_LIMIT_LINE_SEGMENT_INTERPOLATION                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 920
#define RSVNA_ATTR_LIMIT_LINE_HORIZONTAL_SHIFT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 481
#define RSVNA_ATTR_LIMIT_LINE_VERTICAL_SHIFT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 482
#define RSVNA_ATTR_LIMIT_LINE_DELETE_ALL_SEGMENTS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 483
#define RSVNA_ATTR_LIMIT_LINE_TTL_PASS                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 484
#define RSVNA_ATTR_LIMIT_LINE_TTL_PASS_DEFAULT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1206
#define RSVNA_ATTR_HORIZONTAL_LINE_STATE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 485
#define RSVNA_ATTR_HORIZONTAL_LINE_POSITION                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 486
#define RSVNA_ATTR_SHOW_RIPPLE_LIMITS                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 487
#define RSVNA_ATTR_RIPPLE_LIMIT_CHECK                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 488
#define RSVNA_ATTR_RIPPLE_LIMIT_GLOBAL_CHECK_RESULT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 489
#define RSVNA_ATTR_RIPPLE_LIMIT_CHECK_CLEAR                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 800
#define RSVNA_ATTR_RIPPLE_FAIL_BEEP                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 490
#define RSVNA_ATTR_RIPPLE_LIMIT_RANGE_SEGMENT_CHECK                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 491
#define RSVNA_ATTR_RIPPLE_LIMIT_LINE_SEGMENT_VALUES                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 492
#define RSVNA_ATTR_RIPPLE_LIMIT_LINE_SEGMENT_LIMIT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 493
#define RSVNA_ATTR_RIPPLE_LIMIT_RANGES_COUNT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 494
#define RSVNA_ATTR_RIPPLE_LIMIT_DELETE_ALL_RANGES                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 495
#define RSVNA_ATTR_RIPPLE_LIMIT_PHYSICAL_UNITS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 496
#define RSVNA_ATTR_RIPPLE_LIMIT_RESPONSE_DOMAIN_FORMAT_UNITS                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 497
#define RSVNA_ATTR_SHOW_CIRCLE_LIMITS                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 767
#define RSVNA_ATTR_CIRCLE_LIMIT_CHECK                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 768
#define RSVNA_ATTR_CIRCLE_LIMIT_CHECK_RESULT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 769
#define RSVNA_ATTR_CIRCLE_LIMIT_CHECK_CLEAR                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 801
#define RSVNA_ATTR_CIRCLE_FAIL_BEEP                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 770
#define RSVNA_ATTR_CIRCLE_CLEAR                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 923
#define RSVNA_ATTR_CIRCLE_SHOW_BORDER                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 924
#define RSVNA_ATTR_LIMIT_TO_CIRCLE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 925
#define RSVNA_ATTR_CIRCLE_DATA                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 927
#define RSVNA_ATTR_INFINITE_AVERAGING_ENABLED                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 928
#define RSVNA_ATTR_INFINITE_AVERAGING_MODE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 929
#define RSVNA_ATTR_EXTERNAL_DLL_LIST                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1207
#define RSVNA_ATTR_ADD_EXTERNAL_DLL                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1217
#define RSVNA_ATTR_REMOVE_EXTERNAL_DLL                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1218
#define RSVNA_ATTR_LINEARITY_DEVIATION_MODE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1363
#define RSVNA_ATTR_RECALCULATE_LINEARITY_DEVIATION                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1364
#define RSVNA_ATTR_LINEARITY_DEVIATION_CONSTANT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1365
#define RSVNA_ATTR_LINEARITY_DEVIATION_ELECTRICAL_LENGTH                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1366
#define RSVNA_ATTR_LINEARITY_DEVIATION_SLOPE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1367
#define RSVNA_ATTR_MARKER_STATE                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 498
#define RSVNA_ATTR_MARKER_STIMULUS                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 499
#define RSVNA_ATTR_QUERY_MARKER_RESPONSE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 500
#define RSVNA_ATTR_REFERENCE_MARKER_STATE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 501
#define RSVNA_ATTR_REFERENCE_MARKER_STIMULUS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 502
#define RSVNA_ATTR_QUERY_REFERENCE_MARKER_RESPONSE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 503
#define RSVNA_ATTR_DELTA_MARKER_STATE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 504
#define RSVNA_ATTR_MARKER_FORMAT                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 505
#define RSVNA_ATTR_MARKER_REFERENCE_FORMAT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 930
#define RSVNA_ATTR_MARKER_DEFAULT_FORMAT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1219
#define RSVNA_ATTR_MARKER_DISABLE_ALL                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 506
#define RSVNA_ATTR_MARKER_EXPORT                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 507
#define RSVNA_ATTR_MARKER_COUPLING                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 508
#define RSVNA_ATTR_MARKER_NAME                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 509
#define RSVNA_ATTR_MARKER_MODE                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 510
#define RSVNA_ATTR_MARKER_TYPE                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 511
#define RSVNA_ATTR_MARKER_THRESHOLD                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 931
#define RSVNA_ATTR_MARKER_THRESHOLD_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 932
#define RSVNA_ATTR_MARKER_COUPLED_MARKERS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 933
#define RSVNA_ATTR_MARKER_COUPLING_TYPE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 934
#define RSVNA_ATTR_MARKER_MAXIMUM_REFERENCE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1600
#define RSVNA_ATTR_MARKER_MINIMUM_REFERENCE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1601
#define RSVNA_ATTR_MARKER_EXCURSION_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 935
#define RSVNA_ATTR_MARKER_EXCURSION_MINIMUM                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 936
#define RSVNA_ATTR_MARKER_SPECTRUM_EQUALS_MARKER                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1368
#define RSVNA_ATTR_MARKER_MULTIPLE_PEAK_EXCURSION_ENABLED                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 937
#define RSVNA_ATTR_MARKER_MULTIPLE_PEAK_EXCURSION_MINIMUM                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 938
#define RSVNA_ATTR_MARKER_MULTIPLE_PEAK_THRESHOLD                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 939
#define RSVNA_ATTR_MARKER_MULTIPLE_PEAK_THRESHOLD_ENABLED                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 940
#define RSVNA_ATTR_CENTER_SWEEP_RANGE_TO_MARKER_REFERENCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1602
#define RSVNA_ATTR_START_SWEEP_RANGE_TO_MARKER_REFERENCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1603
#define RSVNA_ATTR_STOP_SWEEP_RANGE_TO_MARKER_REFERENCE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1604
#define RSVNA_ATTR_MARKER_THRESHOLD_REFERENCE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1605
#define RSVNA_ATTR_MARKER_THRESHOLD_ENABLED_REFERENCE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1606
#define RSVNA_ATTR_MARKER_EXCURSION_REFERENCE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1607
#define RSVNA_ATTR_MARKER_EXCURSION_REFERENCE_ENABLED                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1608
#define RSVNA_ATTR_MARKER_SEARCH                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 512
#define RSVNA_ATTR_MARKER_SEARCH_REFERENCE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1609
#define RSVNA_ATTR_MARKER_SEARCH_MODE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 941
#define RSVNA_ATTR_MARKER_TARGET_VALUE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 513
#define RSVNA_ATTR_MARKER_TARGET_VALUE_REFERENCE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1610
#define RSVNA_ATTR_MARKER_TARGET_FORMAT                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 815
#define RSVNA_ATTR_MARKER_TARGET_FORMAT_REFERENCE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1611
#define RSVNA_ATTR_MARKER_BANDFILTER_SEARCH_MODE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 514
#define RSVNA_ATTR_MARKER_X_DB_BANDWIDTH                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 515
#define RSVNA_ATTR_MARKER_BANDFILTER_SEARCH_RESULT_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 516
#define RSVNA_ATTR_MARKER_TRACKING                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 517
#define RSVNA_ATTR_MARKER_TRACKING_REFERENCE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1612
#define RSVNA_ATTR_MARKER_SEARCH_RANGE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 518
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_REFERENCE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1613
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_START                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 519
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_START_REFERENCE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1614
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_STOP                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 520
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_STOP_REFERENCE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1615
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_SHOW_LIMIT_LINES                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 521
#define RSVNA_ATTR_MARKER_SEARCH_RANGE_SHOW_LIMIT_LINES_REFERENCE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1616
#define RSVNA_ATTR_MARKER_SEARCH_LEFT_REFERENCE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1617
#define RSVNA_ATTR_MARKER_SEARCH_NEXT_REFERENCE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1618
#define RSVNA_ATTR_MARKER_SEARCH_RIGHT_REFERENCE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1619
#define RSVNA_ATTR_MARKER_SEARCH_IMMEDIATE_REFERENCE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1620
#define RSVNA_ATTR_SWEEP_RANGE_TO_MARKER                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 522
#define RSVNA_ATTR_STIMULUS_FREQUENCY_START                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 523
#define RSVNA_ATTR_STIMULUS_FREQUENCY_STOP                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 524
#define RSVNA_ATTR_STIMULUS_FREQUENCY_CENTER                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 525
#define RSVNA_ATTR_STIMULUS_FREQUENCY_SPAN                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 526
#define RSVNA_ATTR_STIMULUS_INTERNAL_SOURCE_POWER                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 527
#define RSVNA_ATTR_STIMULUS_FREQUENCY_CW                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 528
#define RSVNA_ATTR_STIMULUS_POWER_START_STOP                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 529
#define RSVNA_ATTR_STIMULUS_SOURCE_PORT                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 530
#define RSVNA_ATTR_STIMULUS_SOURCE_STEP_ATTENUATOR                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1369
#define RSVNA_ATTR_STIMULUS_FREQUENCY_RANGE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 531
#define RSVNA_ATTR_STIMULUS_FREQUENCY_START_MIN_MAX_DEFAULT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 874
#define RSVNA_ATTR_STIMULUS_FREQUENCY_STOP_MIN_MAX_DEFAULT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 875
#define RSVNA_ATTR_TIME_DOMAIN_STIMULUS_START_TIME                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 532
#define RSVNA_ATTR_TIME_DOMAIN_STIMULUS_STOP_TIME                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 533
#define RSVNA_ATTR_TIME_DOMAIN_STIMULUS_CENTER_TIME                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 534
#define RSVNA_ATTR_TIME_DOMAIN_STIMULUS_TIME_SPAN                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 535
#define RSVNA_ATTR_TIME_DOMAIN_STIMULUS_X_AXIS_SCALING                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 536
#define RSVNA_ATTR_PREAMPLIFIER_ENABLED                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1370
#define RSVNA_ATTR_PREAMPLIFIER_GAIN                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1371
#define RSVNA_ATTR_POWER_ALC_AUTOMATIC_IF_BANDWIDTH                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1372
#define RSVNA_ATTR_POWER_ALC_BANDWIDTH                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1373
#define RSVNA_ATTR_POWER_ALC_ENABLED                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1374
#define RSVNA_ATTR_POWER_ALC_PORT_COUPLE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1375
#define RSVNA_ATTR_POWER_ALC_AUTO_SETTLING                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1376
#define RSVNA_ATTR_POWER_ALC_SETTLING_TOLERANCE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1377
#define RSVNA_ATTR_POWER_ALC_CONTROL_TIME                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1378
#define RSVNA_ATTR_POWER_ALC_PORT_OPTIMIZE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1379
#define RSVNA_ATTR_POWER_ALC_PORT_PI                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1380
#define RSVNA_ATTR_POWER_ALC_PORT_PI_GAIN                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1381
#define RSVNA_ATTR_POWER_ALC_PORT_PI_INTEGRATION_TIME                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1382
#define RSVNA_ATTR_POWER_ALC_PORT_CONTROL_RANGE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1383
#define RSVNA_ATTR_POWER_ALC_PORT_RECEIVER                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1384
#define RSVNA_ATTR_POWER_ALC_PORT_START_OFFSET                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1385
#define RSVNA_ATTR_POWER_ALC_PORT_TABLE_AREA_ON                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1386
#define RSVNA_ATTR_POWER_ALC_PORT_TARGET_POWER                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1387
#define RSVNA_ATTR_POWER_ALC_PORT_TARGET_POWER_DEFAULT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1388
#define RSVNA_ATTR_POWER_ALC_PORT_DEEMBEDING_NETWORK_USAGE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1389
#define RSVNA_ATTR_POWER_ALC_CLAMP                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1681
#define RSVNA_ATTR_CHANNEL_RF_STATE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 537
#define RSVNA_ATTR_CHANNEL_RESOLUTION_BANDWIDTH                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 538
#define RSVNA_ATTR_CHANNEL_FILTER_SELECTIVITY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 539
#define RSVNA_ATTR_CHANNEL_SWEEP_AVERAGE_STATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 540
#define RSVNA_ATTR_CHANNEL_SWEEP_AVERAGE_FACTOR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 541
#define RSVNA_ATTR_CHANNEL_SWEEP_AVERAGE_MODE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 542
#define RSVNA_ATTR_CHANNEL_SWEEP_AVERAGE_RESTART                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 543
#define RSVNA_ATTR_CHANNEL_DYNAMIC_BW_AT_LOW_FREQUENCIES                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1390
#define RSVNA_ATTR_CHANNEL_ANALOG_IF_FILTER                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1391
#define RSVNA_ATTR_SWEEP_POINTS_COUNT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 544
#define RSVNA_ATTR_SWEEP_FREQUENCY_STEP_SIZE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 545
#define RSVNA_ATTR_SWEEP_TIME_AUTO                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 546
#define RSVNA_ATTR_SWEEP_TIME                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 547
#define RSVNA_ATTR_SWEEP_MEASUREMENT_DELAY                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 548
#define RSVNA_ATTR_SWEEP_MEASUREMENT_DELAY_INSERTION_POINT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 771
#define RSVNA_ATTR_SWEEP_TYPE                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 549
#define RSVNA_ATTR_SWEEP_STIMULUS_AXIS                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 550
#define RSVNA_ATTR_SWEEP_START                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 551
#define RSVNA_ATTR_SWEEP_START_ALL                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 552
#define RSVNA_ATTR_SWEEP_IN_ALL_CHANNELS                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 553
#define RSVNA_ATTR_SWEEP_IN_SINGLE_CHANNEL                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 554
#define RSVNA_ATTR_SWEEP_OPTIMIZE_CHANNEL_SWITCHING_AUTO                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1220
#define RSVNA_ATTR_SWEEP_SEGMENT_INSERT_NEW_SEGMENT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 942
#define RSVNA_ATTR_SWEEP_SEGMENT_INSERT_TIME_MODE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 555
#define RSVNA_ATTR_SWEEP_SEGMENT_REDEFINE_SEGMENT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 943
#define RSVNA_ATTR_SWEEP_SEGMENT_REDEFINE_TIME_MODE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 556
#define RSVNA_ATTR_SWEEP_SEGMENT_ADD                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 557
#define RSVNA_ATTR_SWEEP_SEGMENT_DELETE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 558
#define RSVNA_ATTR_SWEEP_SEGMENT_DELETE_ALL                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 559
#define RSVNA_ATTR_SWEEP_SEGMENT_STATE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 560
#define RSVNA_ATTR_SWEEP_SEGMENT_X_AXIS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 561
#define RSVNA_ATTR_SWEEP_SEGMENT_FREQUENCY_START                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 562
#define RSVNA_ATTR_SWEEP_SEGMENT_FREQUENCY_STOP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 563
#define RSVNA_ATTR_SWEEP_SEGMENT_NUMBER_OF_POINTS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 564
#define RSVNA_ATTR_SWEEP_SEGMENT_POWER                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 565
#define RSVNA_ATTR_SWEEP_SEGMENT_TIME                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 566
#define RSVNA_ATTR_SWEEP_SEGMENT_POINT_DELAY                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 567
#define RSVNA_ATTR_SWEEP_SEGMENT_MEAS_BANDWIDTH                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 568
#define RSVNA_ATTR_SWEEP_SEGMENT_SELECTIVITY                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 569
#define RSVNA_ATTR_SWEEP_SEGMENT_INDEPENDENT_POWER_STATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 570
#define RSVNA_ATTR_SWEEP_SEGMENT_INDEPENDENT_MEAS_BANDWIDTH_STATE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 571
#define RSVNA_ATTR_SWEEP_SEGMENT_INDEPENDENT_SELECTIVITY_STATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 572
#define RSVNA_ATTR_SWEEP_SEGMENT_INDEPENDENT_TIME_STATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 573
#define RSVNA_ATTR_SWEEP_SEGMENT_INDEPENDENT_POINT_DELAY_STATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 574
#define RSVNA_ATTR_QUERY_SWEEP_SEGMENT_COUNT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 575
#define RSVNA_ATTR_QUERY_SWEEP_SEGMENT_FREQUENCY_CENTER                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 576
#define RSVNA_ATTR_QUERY_SWEEP_SEGMENT_FREQUENCY_SPAN                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 577
#define RSVNA_ATTR_QUERY_SWEEP_SEGMENTS_SUM_TIME                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 578
#define RSVNA_ATTR_SWEEP_MODE_CONTINUOUS                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 579
#define RSVNA_ATTR_SWEEP_MODE_CONTINUOUS_ALL_CHANNEL                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 580
#define RSVNA_ATTR_SWEEP_MODE_CONTROLLED_TIMING                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1392
#define RSVNA_ATTR_SWEEP_COUNT                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 581
#define RSVNA_ATTR_SWEEP_COUNT_ALL_CHANNELS                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 582
#define RSVNA_ATTR_SWEEP_SCOPE                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 583
#define RSVNA_ATTR_SWEEP_DETECTOR_TIME                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 584
#define RSVNA_ATTR_SWEEP_SEGMENT_PORT_POWER_LEVEL                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 944
#define RSVNA_ATTR_SWEEP_SEGMENT_PORT_LEVEL_ENABLED                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 945
#define RSVNA_ATTR_SWEEP_WAIT_TIME_PER_POINT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1393
#define RSVNA_ATTR_SWEEP_TIME_MODE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1394
#define RSVNA_ATTR_SWEEP_FAST_POWER_SWEEP                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1395
#define RSVNA_ATTR_SWEEP_REVERSE                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1396
#define RSVNA_ATTR_SWEEP_TIMER_MODE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1397
#define RSVNA_ATTR_SWEEP_ONLY_ACTIVE_CHANNEL_ON                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1682
#define RSVNA_ATTR_SWEEP_MINIMUM_WAIT_TIME                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1683
#define RSVNA_ATTR_SWEEP_MINIMUM_WAIT_TIME_ENABLED                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1684
#define RSVNA_ATTR_SWEEP_FIXED_WAIT_TIME                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1685
#define RSVNA_ATTR_SWEEP_FIXED_WAIT_TIME_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1686
#define RSVNA_ATTR_SWEEP_MANUAL_WAIT_TIME_ENABLED                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1687
#define RSVNA_ATTR_SWEEP_MINIMUM_WAIT_TIME_INITIALIZATION_CHANNEL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1688
#define RSVNA_ATTR_SWEEP_MINIMUM_WAIT_TIME_INITIALIZATION_CHANNEL_ENABLED                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1689
#define RSVNA_ATTR_SWEEP_MAXIMUM_WAIT_TIME_PORT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1690
#define RSVNA_ATTR_SWEEP_MAXIMUM_WAIT_TIME_PORT_ENABLED                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1691
#define RSVNA_ATTR_SWEEP_MINIMUM_WAIT_TIME_PORT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1692
#define RSVNA_ATTR_SWEEP_MINIMUM_WAIT_TIME_PORT_ENABLED                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1693
#define RSVNA_ATTR_SWEEP_SEGMENT_WAIT_TIME_FIXED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1694
#define RSVNA_ATTR_SWEEP_SEGMENT_WAIT_TIME_FIXED_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1695
#define RSVNA_ATTR_SWEEP_SEGMENT_WAIT_TIME_FIXED_MAXIMUM                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1696
#define RSVNA_ATTR_SWEEP_SEGMENT_WAIT_TIME_FIXED_MINIMUM                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1697
#define RSVNA_ATTR_SWEEP_TIMING_INFO_BOX_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1698
#define RSVNA_ATTR_SWEEP_GENERATION_MODE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1699
#define RSVNA_ATTR_SELECT_CALIBRATION_TYPE_SIMPLE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 946
#define RSVNA_ATTR_SYSTEM_ERROR_CORRECTION_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 585
#define RSVNA_ATTR_LOAD_MATCH_CORRECTION_STATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 586
#define RSVNA_ATTR_CALIBRATION_AUTO_POWER                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 587
#define RSVNA_ATTR_CALIBRATION_ALL_CHANNELS_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 588
#define RSVNA_ATTR_DELETE_CALIBRATION_GROUP_FILE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 589
#define RSVNA_ATTR_INITIATE_CALIBRATION_AUTO_ASSIGNMENT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 780
#define RSVNA_ATTR_DELETE_ALL_CALIBRATION_AUTO_ASSIGNMENTS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 781
#define RSVNA_ATTR_SAVE_CALIBRATION_AUTO_ASSIGNMENT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 782
#define RSVNA_ATTR_GENERATE_DEFAULT_CALIBRATION_AUTO_ASSIGNMENT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 783
#define RSVNA_ATTR_CALIBRATION_AUTO_ASSIGNMENT_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 785
#define RSVNA_ATTR_CALIBRATION_AUTO_ASSIGNMENT_ALL_COUNT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 816
#define RSVNA_ATTR_MULTICALIBRATION_TYPE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 947
#define RSVNA_ATTR_CALIBRATION_AUTO_MINIMUM_CONNECT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 948
#define RSVNA_ATTR_CALIBRATION_AUTO_COMMON_PORT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 949
#define RSVNA_ATTR_TERMINATE_SYSTEM_ERROR_CORRECTION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 950
#define RSVNA_ATTR_CALIBRATION_CLEAN                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 951
#define RSVNA_ATTR_WAVE_CORRECTION_ENABLED                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 952
#define RSVNA_ATTR_CALCULATE_AFTER_DEEMBED_ENABLED                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 953
#define RSVNA_ATTR_CALIBRATION_DATE_AND_TIME                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 590
#define RSVNA_ATTR_CALIBRATION_STATUS                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 591
#define RSVNA_ATTR_REPEAT_PREVIOUS_CALIBRATION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 592
#define RSVNA_ATTR_REPEAT_PREVIOUS_CALIBRATION_CURRENT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 593
#define RSVNA_ATTR_SELECT_CALIBRATION_UNIT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 594
#define RSVNA_ATTR_CALIBRATION_UNIT_LIST                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 595
#define RSVNA_ATTR_CALIBRATION_AUTOMATIC_POWER_REDUCTION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 596
#define RSVNA_ATTR_CALIBRATION_AUTOMATIC_MEMORY_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 597
#define RSVNA_ATTR_CALIBRATION_UNIT_CHARACTERIZATIONS_LIST                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 598
#define RSVNA_ATTR_CALIBRATION_UNIT_WARMUP_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 599
#define RSVNA_ATTR_CALIBRATION_CONNECTOR                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 600
#define RSVNA_ATTR_CALIBRATION_SAME_CONNECTOR_AT_ALL_PORTS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 601
#define RSVNA_ATTR_CALIBRATION_SAME_GENDER_AT_ALL_PORTS                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 602
#define RSVNA_ATTR_CALIBRATION_DELETE_USER_CONNECTOR_TYPE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 603
#define RSVNA_ATTR_PORT_CONNECTOR_TYPE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 803
#define RSVNA_ATTR_CALIBRATION_CONNECTOR_TYPE_CATALOG                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 604
#define RSVNA_ATTR_SAVE_CALIBRATION_DATA                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 605
#define RSVNA_ATTR_GENERATE_DEFAULT_CALIBRATION_DATA                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 606
#define RSVNA_ATTR_DELETE_CALIBRATION_DATA                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 607
#define RSVNA_ATTR_DELETE_ALL_CALIBRATION_DATA                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 608
#define RSVNA_ATTR_EXPORT_FACTORY_CALIBRATION_DATA_TO_TOUCHSTONE_FILE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 609
#define RSVNA_ATTR_QUERY_CALIBRATION_DATA_PARAMETERS_COUNT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 804
#define RSVNA_ATTR_SELECT_CALIBRATION_KIT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 610
#define RSVNA_ATTR_CALIBRATION_KIT_PASSWORD                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 611
#define RSVNA_ATTR_SAVE_CALIBRATION_KIT                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 612
#define RSVNA_ATTR_DELETE_CALIBRATION_KIT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 613
#define RSVNA_ATTR_IMPORT_CAL_KIT                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 614
#define RSVNA_ATTR_CALIBRATION_KIT_ADDITIONAL_DIRECTORY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 615
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 616
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_OTHER_SOURCES                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 617
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_NUMBER_OF_READINGS                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 618
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_TOLERANCE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 619
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_CONVERGENCE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 620
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_POWER_METER                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 621
#define RSVNA_ATTR_QUERY_SOURCE_POWER_CALIBRATION_DATA_PARAMETERS_COUNT                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 805
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_OFFSET                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 623
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_GENERATOR_STATE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 624
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_GENERATOR_OFFSET                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 625
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_POWER_SELECT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 626
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_REFERENCE_POWER                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 627
#define RSVNA_ATTR_DUMMY_SOURCE_POWER_CALIBRATION_ENABLED                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1399
#define RSVNA_ATTR_FAST_SOURCE_POWER_CALIBRATION_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1400
#define RSVNA_ATTR_ELECTRONIC_SOURCE_ATTENUATION                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1401
#define RSVNA_ATTR_ELECTRONIC_SOURCE_ATTENUATION_MODE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1402
#define RSVNA_ATTR_EXTERNAL_GENERATOR_MINIMUM_SETTLING_DELAY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 955
#define RSVNA_ATTR_RECEIVER_POWER_ATTENUATION                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 806
#define RSVNA_ATTR_RECIEVER_POWER_CALIBRATION_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 628
#define RSVNA_ATTR_RECIEVER_POWER_CALIBRATION_AWAVE_STATE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 629
#define RSVNA_ATTR_POWER_METER_TWO_PORT_POSITION                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 631
#define RSVNA_ATTR_POWER_METER_DELETE_COEFFICIENT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 632
#define RSVNA_ATTR_POWER_METER_DELETE_ALL_COEFFICIENTS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 633
#define RSVNA_ATTR_POWER_METER_SAVE_POWER_LOSS_LIST                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 636
#define RSVNA_ATTR_POWER_METER_LOAD_POWER_LOSS_LIST                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 956
#define RSVNA_ATTR_ACTIVATE_POWER_CALIBRATION                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 637
#define RSVNA_ATTR_START_CALIBRATION_INTERMODULATION_SOURCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 638
#define RSVNA_ATTR_START_CALIBRATION_INTERMODULATION_RECEIVER                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 639
#define RSVNA_ATTR_POWER_SWEEP_END_MODE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 957
#define RSVNA_ATTR_POWER_SWEEP_END_SETTLING_DELAY                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 958
#define RSVNA_ATTR_POWER_PORT_FLATNESS_CALIBRATION                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 961
#define RSVNA_ATTR_POWER_CALIBRATION_METHOD                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 962
#define RSVNA_ATTR_POWER_CALIBRATION_REFERENCE_RECEIVER_ENABLED                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 963
#define RSVNA_ATTR_CALIBRATION_NOISE_DATA_RECEIVING_PORT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1404
#define RSVNA_ATTR_CALIBRATION_SAME_SWEEP_SETUP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1405
#define RSVNA_ATTR_CALIBRATION_SWEEP_SETUP                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1406
#define RSVNA_ATTR_CALIBRATION_REFERENCE_RECEIVER_ATTENUATION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1407
#define RSVNA_ATTR_CALIBRATION_IMD_METHOD                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1408
#define RSVNA_ATTR_START_HARMONIC_SOURCE_CALIBRATION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1409
#define RSVNA_ATTR_CALIBRATION_INTERPOLATION_METHOD                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1410
#define RSVNA_ATTR_CALIBRATION_FAST_MULTIPORT_CORRECTION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1411
#define RSVNA_ATTR_START_HARMONIC_RECEIVER_CALIBRATION                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1412
#define RSVNA_ATTR_CALIBRATION_ICS_REPEAT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1621
#define RSVNA_ATTR_CALIBRATION_ICS_TEMEPRATURE_COMPENSATION                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1622
#define RSVNA_ATTR_CALIBRATION_INTERNAL_COMBINER_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1701
#define RSVNA_ATTR_CONFIGURE_CALIBRATION_STANDARD                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 965
#define RSVNA_ATTR_CONFIGURE_CALIBRATION_STANDARD_WITH_LABEL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 967
#define RSVNA_ATTR_RUN_CALIBRATION_VALIDATION                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1228
#define RSVNA_ATTR_CALIBRATION_VALIDATION_STANDARD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1229
#define RSVNA_ATTR_CALIBRATION_VALIDATION_CHARACTERIZATION                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1230
#define RSVNA_ATTR_CALIBRATION_VALIDATION_AVERAGE_ENABLED                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1703
#define RSVNA_ATTR_CALIBRATION_VALIDATION_FORMAT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1704
#define RSVNA_ATTR_CALIBRATION_VALIDATION_PRESET                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1705
#define RSVNA_ATTR_CALIBRATION_VALIDATION_RESULT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1706
#define RSVNA_ATTR_METAS_CALIBRATION_ACTIVE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1237
#define RSVNA_ATTR_METAS_CABLE_CATALOG                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1238
#define RSVNA_ATTR_METAS_CLEAR_CABLE_RECONNECTION_HISTORY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1243
#define RSVNA_ATTR_METAS_CLEAR_CABLE_MOVEMENT_HISTORY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1246
#define RSVNA_ATTR_METAS_UNCERTAINLY_BAND                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1711
#define RSVNA_ATTR_ADVANCED_HARMONIC_ORDERS                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1413
#define RSVNA_ATTR_ADVANCED_HARMONIC_ORDER_ADD                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1414
#define RSVNA_ATTR_ADVANCED_HARMONIC_ORDER_REMOVE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1415
#define RSVNA_ATTR_ADVANCED_POWER                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1416
#define RSVNA_ATTR_ADVANCED_POWER_ENABLED                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1417
#define RSVNA_ATTR_ADVANCED_POWER_ATTENUATOR                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1418
#define RSVNA_ATTR_ADVANCED_POWER_ATTENUATOR_ENABLED                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1419
#define RSVNA_ATTR_ADVANCED_POWER_SOURCE_ATTENUATOR                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1420
#define RSVNA_ATTR_ADVANCED_POWER_SOURCE_ATTENUATOR_ENABLED                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1421
#define RSVNA_ATTR_ADVANCED_POWER_PREAMPLIFIER                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1422
#define RSVNA_ATTR_ADVANCED_POWER_PREAMPLIFIER_ENABLED                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1423
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_TERMINATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1424
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SAVE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1425
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SWEEP                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1427
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_ENABLED                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1428
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_NOISE_TIME                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1429
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_GAIN_TIME                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1430
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_AVG_DETECTOR_TIME                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1432
#define RSVNA_ATTR_NOISE_FIGURE_SIDEBAND_CORRECTION_ENABLED                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1433
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SIDEBAND_CORRECTION_MINIMUM_FREQUENCY        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1434
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SIDEBAND_CORRECTION_FREQUENCY_LIMIT          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1435
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SIDEBAND_CORRECTION_MAXIMUM_ORDER            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1436
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_IF_FREQUENCY                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1437
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_IF_FREQUENCY_ENABLED                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1438
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_RMS_DETECTOR_TIME                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1439
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_DUT_CORRECTION_ENABLED                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1440
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_AMBIENT_TEMPERATURE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1441
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_VIRTUAL_IMAGE_REJECTION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1712
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SEGMENT_VIRTUAL_IMAGE_REJECTION              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1713
#define RSVNA_ATTR_NOISE_FIGURE_CALIBRATION_SEGMENT_VIRTUAL_IMAGE_CONTROL_ENABLED        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1714
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_EXTERNAL_ATTENUATOR_RECEIVER                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1715
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_EXTERNAL_ATTENUATOR_SOURCE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1716
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_TRACE_NOISE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1717
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_DUT_EXPECTED_NOISE_FIGURE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1718
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_DUT_EXPECTED_GAIN                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1719
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_DUT_POWER                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1720
#define RSVNA_ATTR_NOISE_FIGURE_QUICKSET_EXECUTE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1721
#define RSVNA_ATTR_CALIBRATION_MIXER_PHASE_IF_START_CALIBRATION                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1442
#define RSVNA_ATTR_CALIBRATION_MIXER_PHASE_IF_START_SECOND_STEP                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1443
#define RSVNA_ATTR_CALIBRATION_MIXER_PHASE_START_RF_CALIBRATION                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1444
#define RSVNA_ATTR_CALIBRATION_MIXER_PHASE_START_LO_CALIBRATION                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1445
#define RSVNA_ATTR_SOURCE_LEVELING_CLEAR_ALL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1722
#define RSVNA_ATTR_SOURCE_LEVELING_CLEAR                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1723
#define RSVNA_ATTR_SOURCE_LEVELING_REMOVE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1724
#define RSVNA_ATTR_SOURCE_LEVELING_LOAD                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1725
#define RSVNA_ATTR_SOURCE_LEVELING_STORE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1726
#define RSVNA_ATTR_SOURCE_LEVELING_NUMBER_OF_POINTS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1727
#define RSVNA_ATTR_SOURCE_LEVELING_START_FREQUENCY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1728
#define RSVNA_ATTR_SOURCE_LEVELING_STOP_FREQUENCY                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1729
#define RSVNA_ATTR_SOURCE_LEVELING_START_POWER                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1730
#define RSVNA_ATTR_SOURCE_LEVELING_STOP_POWER                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1731
#define RSVNA_ATTR_SOURCE_LEVELING_ALC_STEP_SIZE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1732
#define RSVNA_ATTR_SOURCE_LEVELING_POWER_STEP_SIZE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1733
#define RSVNA_ATTR_SOURCE_LEVELING_STATE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1734
#define RSVNA_ATTR_SOURCE_LEVELING_START_CAL_SWEEP                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1735
#define RSVNA_ATTR_SOURCE_LEVELING_PORT_ENABLED                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1736
#define RSVNA_ATTR_PREMEASUREMENT_ENABLED                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1737
#define RSVNA_ATTR_PREMEASUREMENT_MODE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1738
#define RSVNA_ATTR_PREMEASUREMENT_PORT_ENABLED                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1739
#define RSVNA_ATTR_PREMEASUREMENT_CONVERGENCE_FACTOR                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1740
#define RSVNA_ATTR_PREMEASUREMENT_DETECTOR                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1741
#define RSVNA_ATTR_PREMEASUREMENT_MAX_PRECAL_SWEEPS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1742
#define RSVNA_ATTR_PREMEASUREMENT_TOLERANCE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1743
#define RSVNA_ATTR_PREMEASUREMENT_RECEIVER                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1744
#define RSVNA_ATTR_CHANNEL_STATE                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 640
#define RSVNA_ATTR_CHANNEL_NAME                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 641
#define RSVNA_ATTR_CHANNEL_CATALOG                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 642
#define RSVNA_ATTR_ACTIVE_CHANNEL                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 643
#define RSVNA_ATTR_NUMBER_OF_TEST_PORTS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 644
#define RSVNA_ATTR_DRIVING_MODE                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 645
#define RSVNA_ATTR_DRIVING_MODE_ALTERNATED_BLOCKS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1247
#define RSVNA_ATTR_TRACK_LO                                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1446
#define RSVNA_ATTR_TRACK_LO_FREQUENCY_DIFFERENCE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1447
#define RSVNA_ATTR_LO_USAGE                                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1448
#define RSVNA_ATTR_PHASE_MODE                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1449
#define RSVNA_ATTR_SPURIOUS_AVOIDANCE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 646
#define RSVNA_ATTR_FREQUENCY_CONVERSION_MEASUREMENT_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 649
#define RSVNA_ATTR_FREQUENCY_CONVERSION_MIXER_FIXED_IF_FREQUENCY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 969
#define RSVNA_ATTR_FREQUENCY_CONVERSION_MIXER_FIXED_LO_FREQUENCY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 970
#define RSVNA_ATTR_FREQUENCY_CONVERSION_MIXER_FIXED_RF_FREQUENCY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 971
#define RSVNA_ATTR_RF_SIGNAL_SOURCE_STATE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 650
#define RSVNA_ATTR_EXTERNAL_GENERATOR_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 651
#define RSVNA_ATTR_PERMANENT_SIGNAL_SOURCE_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 652
#define RSVNA_ATTR_PERMANENT_SIGNAL_GENERATOR_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 653
#define RSVNA_ATTR_SOURCE_POWER_CALIBRATION_SLOPE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 654
#define RSVNA_ATTR_AUTOMATIC_GAIN_CONTROL_FOR_MEASUREMENT_CHANNEL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 647
#define RSVNA_ATTR_DELETE_LOGICAL_PORT                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 778
#define RSVNA_ATTR_DELETE_ALL_LOGICAL_PORTS                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 779
#define RSVNA_ATTR_DELETE_PORT_GROUP                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 795
#define RSVNA_ATTR_DELETE_ALL_PORT_GROUPS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 796
#define RSVNA_ATTR_PORT_GROUPS_COUNT                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 797
#define RSVNA_ATTR_PORT_GROUPS_PORT_ORDER                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 972
#define RSVNA_ATTR_PORT_GROUPS_STATE_INDICATION_WARNING                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 973
#define RSVNA_ATTR_PORT_GROUPS_PARALLEL_MEASUREMENT_WITH_FREQUENCY_OFFSET_ENABLED        RS_SPECIFIC_PUBLIC_ATTR_BASE + 974
#define RSVNA_ATTR_PORT_GROUPS_PARALLEL_MEASUREMENT_WITH_FREQUENCY_OFFSET_BANDWIDTH      RS_SPECIFIC_PUBLIC_ATTR_BASE + 975
#define RSVNA_ATTR_PORT_GROUPS_PARALLEL_MEASUREMENT_MINIMUM_OFFSET                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 976
#define RSVNA_ATTR_PORT_GROUPS_PARALLEL_MEASUREMENT_MINIMUM_OFFSET_MODE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 977
#define RSVNA_ATTR_PARALLEL_MEASUREMENT_PORT_GROUP_NAME                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 847
#define RSVNA_ATTR_PARALLEL_MEASUREMENT_NUMBER_OF_PORTS                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 848
#define RSVNA_ATTR_DATA_EXTRAPOLATION                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 978
#define RSVNA_ATTR_ACTIVATES_LO_SUPRESSION                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1626
#define RSVNA_ATTR_SOURCE_POWER_THRU_MEAS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1745
#define RSVNA_ATTR_SOURCE_POWER_THRU_MEAS_ENABLED                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1746
#define RSVNA_ATTR_RFFE_STATE                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 850
#define RSVNA_ATTR_RFFE_FREQUENCY                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 851
#define RSVNA_ATTR_RFFE_IO_VOLTAGE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 852
#define RSVNA_ATTR_RFFE_LOW_VOLTAGE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 853
#define RSVNA_ATTR_RFFE_HIGH_VOLTAGE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 854
#define RSVNA_ATTR_RFFE_COMMAND                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 855
#define RSVNA_ATTR_RFFE_SEND_COMMAND                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 856
#define RSVNA_ATTR_GPIO_STATE                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 857
#define RSVNA_ATTR_GPIO_VOLTAGE                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 858
#define RSVNA_ATTR_GPIO_RANGE                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 979
#define RSVNA_ATTR_GPIO_SHUNT                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 980
#define RSVNA_ATTR_GPIO_APPLY                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 859
#define RSVNA_ATTR_GPIO_START_ALL_MEASUREMENTS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 981
#define RSVNA_ATTR_GPIO_MEASUREMENT_TIME                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 982
#define RSVNA_ATTR_GPIO_CURRENT_MEASUREMENT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 985
#define RSVNA_ATTR_GPIO_CURRENT_MEASUREMENT_ALL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 986
#define RSVNA_ATTR_GPIO_VOLTAGE_MEASUREMENT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 988
#define RSVNA_ATTR_GPIO_VOLTAGE_MEASUREMENT_ALL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 989
#define RSVNA_ATTR_SEQUENCER_UNSEGMENTED_SWEEP_COUNT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 860
#define RSVNA_ATTR_SEQUENCER_UNSEGMENTED_SWEEP_DELETE_ALL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 861
#define RSVNA_ATTR_SEQUENCER_UNSEGMENTED_SWEEP_RFFE_COMMAND                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 862
#define RSVNA_ATTR_SEQUENCER_UNSEGMENTED_SWEEP_GPIO_VOLTAGE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 863
#define RSVNA_ATTR_SEQUENCER_UNSEGMENTED_SWEEP_DELAY_TIME                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 864
#define RSVNA_ATTR_SEQUENCER_SEGMENTED_SWEEP_COUNT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 865
#define RSVNA_ATTR_SEQUENCER_SEGMENTED_SWEEP_DELETE_ALL                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 866
#define RSVNA_ATTR_SEQUENCER_SEGMENTED_SWEEP_RFFE_COMMAND                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 867
#define RSVNA_ATTR_SEQUENCER_SEGMENTED_SWEEP_GPIO_VOLTAGE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 868
#define RSVNA_ATTR_SEQUENCER_SEGMENTED_SWEEP_DELAY_TIME                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 869
#define RSVNA_ATTR_RFFE_TEST_START_ALL_MEASUREMENTS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 991
#define RSVNA_ATTR_RFFE_TEST_MEASUREMENT_TIME                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 992
#define RSVNA_ATTR_RFFE_TEST_DATA_VOLTAGE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 993
#define RSVNA_ATTR_RFFE_TEST_DATA_CURRENT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 994
#define RSVNA_ATTR_RFFE_TEST_CLOCK_VOLTAGE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 995
#define RSVNA_ATTR_RFFE_TEST_CLOCK_CURRENT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 996
#define RSVNA_ATTR_RFFE_TEST_VIO_VOLTAGE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 997
#define RSVNA_ATTR_RFFE_TEST_VIO_CURRENT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 998
#define RSVNA_ATTR_RFFE_TEST_APPLY                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 999
#define RSVNA_ATTR_RFFE_TEST_DATA                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1000
#define RSVNA_ATTR_RFFE_TEST_CLOCK                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1001
#define RSVNA_ATTR_RFFE_TEST_VIO                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1002
#define RSVNA_ATTR_RFFE_TEST_DATA_RANGE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1003
#define RSVNA_ATTR_RFFE_TEST_CLOCK_RANGE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1004
#define RSVNA_ATTR_RFFE_TEST_VIO_RANGE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1005
#define RSVNA_ATTR_RFFE_TEST_DATA_SHUNT                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1006
#define RSVNA_ATTR_RFFE_TEST_CLOCK_SHUNT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1007
#define RSVNA_ATTR_RFFE_TEST_VIO_SHUNT                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1008
#define RSVNA_ATTR_AGC_GLOBAL_MODE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 766
#define RSVNA_ATTR_AGC_SEGMENTED_ENABLE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 648
#define RSVNA_ATTR_START_AGC_LEARNING_SWEEP                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 817
#define RSVNA_ATTR_SAVE_AGC_LEARNING_RESULTS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 818
#define RSVNA_ATTR_AGC_ALL_A_AND_B_WAVES_MODE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 959
#define RSVNA_ATTR_AGC_SEGMENT_ALL_A_AND_B_WAVES_MODE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 960
#define RSVNA_ATTR_AGC_GLOBAL_METHOD                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1248
#define RSVNA_ATTR_MIXER_NUMBER_OF_STAGES                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 655
#define RSVNA_ATTR_MIXER_RF_PORT                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 656
#define RSVNA_ATTR_MIXER_IF_PORT                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 657
#define RSVNA_ATTR_MIXER_BASE_FREQUENCY_PORT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 658
#define RSVNA_ATTR_MIXER_FIXED_FREQUENCY_PORT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 659
#define RSVNA_ATTR_MIXER_INPUT_SIGNAL                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 807
#define RSVNA_ATTR_MIXER_CONVERSION_MODE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 660
#define RSVNA_ATTR_INTERMODULATION_STATE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 661
#define RSVNA_ATTR_INTERMODULATION_TWO_TONE_SIGNAL_SOURCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 662
#define RSVNA_ATTR_INTERMODULATION_RECEIVER_PORT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 663
#define RSVNA_ATTR_INTERMODULATION_CAL_SOURCE_PORT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 773
#define RSVNA_ATTR_INTERMODULATION_ORDER_STATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 664
#define RSVNA_ATTR_INTERMODULATION_MAXIMUM_ORDER                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 665
#define RSVNA_ATTR_INTERMODULATION_TONE_DISTANCE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 666
#define RSVNA_ATTR_INTERMODULATION_RESET                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 667
#define RSVNA_ATTR_INTERMODULATION_COMBINER                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1450
#define RSVNA_ATTR_INTERMODULATION_ADD_CW_MODE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1451
#define RSVNA_ATTR_PORT_CONFIG_USE_PRIMED_WAVES                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1452
#define RSVNA_ATTR_PORT_CONFIG_PRIMED_WAVES_FREQUENCY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1453
#define RSVNA_ATTR_PORT_CONFIG_WAVES_FREQUENCY_OFFSET                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1454
#define RSVNA_ATTR_PORT_CONFIG_RECEIVER_FREQUENCY_CONVERSION                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1455
#define RSVNA_ATTR_DIRECT_RECEIVER_INPUT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1456
#define RSVNA_ATTR_DIRECT_REAR_INPUT_PATH                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1457
#define RSVNA_ATTR_DIRECT_AUTO_DETERMINE_IF_OUT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1458
#define RSVNA_ATTR_DIRECT_REAR_OUT                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1460
#define RSVNA_ATTR_DIRECT_MEASUREMENT_INPUT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1461
#define RSVNA_ATTR_DIRECT_REFERENCE_INPUT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1462
#define RSVNA_ATTR_DIRECT_MEASUREMENT_REAR_INPUT_PATH                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1463
#define RSVNA_ATTR_DIRECT_REFERENCE_REAR_INPUT_PATH                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1464
#define RSVNA_ATTR_DIRECT_MEASUREMENT_AUTO_DETERMINE_IF_OUT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1465
#define RSVNA_ATTR_DIRECT_REFERENCE_AUTO_DETERMINE_IF_OUT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1466
#define RSVNA_ATTR_DIRECT_MEASUREMENT_REAR_OUT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1469
#define RSVNA_ATTR_DIRECT_REFERENCE_REAR_OUT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1470
#define RSVNA_ATTR_REAR_LO_OUTPUT_POWER_MODE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1472
#define RSVNA_ATTR_REAR_LO_PERMANENT_ENABLED                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1473
#define RSVNA_ATTR_REAR_LO_POWER_OFFSET                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1474
#define RSVNA_ATTR_TRIGGER_SOURCE                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 668
#define RSVNA_ATTR_TRIGGER_MEASUREMENT_SEQUENCE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 669
#define RSVNA_ATTR_TRIGGER_DELAY                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 670
#define RSVNA_ATTR_TRIGGER_SLOPE                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 671
#define RSVNA_ATTR_TRIGGER_EXTERNAL_INPUT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1475
#define RSVNA_ATTR_TRIGGER_STATUS_READY                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1747
#define RSVNA_ATTR_TRIGGER_AUX_OUT_ENABLE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 672
#define RSVNA_ATTR_TRIGGER_AUX_OUT_POLARITY                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 673
#define RSVNA_ATTR_TRIGGER_AUX_OUT_POSITION                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 674
#define RSVNA_ATTR_TRIGGER_AUX_OUT_INTERVAL                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 675
#define RSVNA_ATTR_TRIGGER_AUX_OUT_DURATION                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 676
#define RSVNA_ATTR_TRIGGER_AUX_OUT_DELAY                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1477
#define RSVNA_ATTR_TRIGGER_AUX_OUT_SIGNAL_TYPE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1478
#define RSVNA_ATTR_LOST_TRIGGER_ENABLED                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1479
#define RSVNA_ATTR_LOST_TRIGGER_COUNT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1480
#define RSVNA_ATTR_LOST_TRIGGER_RESET                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1481
#define RSVNA_ATTR_OFFSET_DELAY_TIME                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 677
#define RSVNA_ATTR_OFFSET_ELECTRICAL_LENGTH                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 678
#define RSVNA_ATTR_OFFSET_MECHANICAL_LENGTH                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 679
#define RSVNA_ATTR_OFFSET_PERMITTIVITY                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 680
#define RSVNA_ATTR_ADJUST_TIME_GATE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 777
#define RSVNA_ATTR_ONE_WAY_LOSS_AT_DC                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 681
#define RSVNA_ATTR_ONE_WAY_LOSS_AT_FREQUENCY                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 682
#define RSVNA_ATTR_ONE_WAY_LOSS_FREQUENCY_FOR_LOSS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 683
#define RSVNA_ATTR_ONE_WAY_LOSS_USE_SECOND_FREQUENCY                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1249
#define RSVNA_ATTR_ONE_WAY_LOSS_AT_SECOND_FREQUENCY                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1250
#define RSVNA_ATTR_ONE_WAY_LOSS_SECOND_FREQUENCY_FOR_LOSS                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1251
#define RSVNA_ATTR_AUTO_LENGTH                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 684
#define RSVNA_ATTR_AUTO_LENGTH_AND_LOSS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 685
#define RSVNA_ATTR_RESET_OFFSET                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 686
#define RSVNA_ATTR_OFFSET_COMPENSATION                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 870
#define RSVNA_ATTR_VIRTUAL_NETWORK_SINGLE_ENDED_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 687
#define RSVNA_ATTR_VIRTUAL_NETWORK_SINGLE_ENDED_CIRCUIT_MODEL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 688
#define RSVNA_ATTR_LOAD_SINGLE_ENDED_PORT_CIRCUIT_MODEL_DATA                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 689
#define RSVNA_ATTR_LOAD_SINGLE_ENDED_PORT_CIRCUIT_MODEL_DEEMBEDDING_DATA                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1009
#define RSVNA_ATTR_LOAD_SINGLE_ENDED_PORT_CIRCUIT_MODEL_EMBEDDING_DATA                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1010
#define RSVNA_ATTR_VIRTUAL_NETWORK_BALANCED_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 690
#define RSVNA_ATTR_VIRTUAL_NETWORK_BALANCED_CIRCUIT_MODEL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 691
#define RSVNA_ATTR_VIRTUAL_NETWORK_DELETE_PORT_PAIR_DEEMBEDDING                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 827
#define RSVNA_ATTR_VIRTUAL_NETWORK_DELETE_PORT_PAIR_EMBEDDING                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 828
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_PAIR_DEEMBEDDING_STATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 829
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_PAIR_EMBEDDING_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 830
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_PAIR_DEEMBEDDING_CIRCUIT_MODEL                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 831
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_PAIR_EMBEDDING_CIRCUIT_MODEL                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 832
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_STATE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 833
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_STATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 834
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_CIRCUIT_MODEL                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 835
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_CIRCUIT_MODEL                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 836
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_CAPACITANCE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1011
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_CAPACITANCE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1012
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_INDUCTANCE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1013
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_INDUCTANCE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1014
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_RESISTANCE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1015
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_RESISTANCE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1016
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_CONDUCTANCE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1017
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_CONDUCTANCE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1018
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_DEEMBEDDING_LOAD_CIRCUIT_MODEL            RS_SPECIFIC_PUBLIC_ATTR_BASE + 837
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_EMBEDDING_LOAD_CIRCUIT_MODEL              RS_SPECIFIC_PUBLIC_ATTR_BASE + 838
#define RSVNA_ATTR_VIRTUAL_NETWORK_DIFFERENTIAL_MATCH_EMBEDDING_STATE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 839
#define RSVNA_ATTR_VIRTUAL_NETWORK_DIFFERENTIAL_MATCH_EMBEDDING_CIRCUIT_MODEL            RS_SPECIFIC_PUBLIC_ATTR_BASE + 840
#define RSVNA_ATTR_VIRTUAL_NETWORK_WAVE_DE_EMBED_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1252
#define RSVNA_ATTR_FIXTURE_COMPENSATION_OFFSET_CORRECTION_TYPE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 692
#define RSVNA_ATTR_FIXTURE_COMPENSATION_OFFSET_CORRECTION_AUTO_TYPE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 693
#define RSVNA_ATTR_FIXTURE_COMPENSATION_RESET                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 694
#define RSVNA_ATTR_FIXTURE_COMPENSATION_APPLY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 695
#define RSVNA_ATTR_VIRTUAL_NETWORK_GROUND_LOOP_PER_PORT_GROUP_ENABLED                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1019
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_STATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1020
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_STATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1021
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_CIRCUIT_MODEL      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1022
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_CIRCUIT_MODEL        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1023
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_CAPACITANCE        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1024
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_CAPACITANCE          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1025
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_INDUCTANCE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1026
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_INDUCTANCE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1027
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_RESISTANCE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1028
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_RESISTANCE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1029
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_CONDUCTANCE        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1030
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_CONDUCTANCE          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1031
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_DEEMBEDDING_LOAD_CIRCUIT_MODEL RS_SPECIFIC_PUBLIC_ATTR_BASE + 1032
#define RSVNA_ATTR_VIRTUAL_NETWORK_PORT_GROUP_GROUND_LOOP_EMBEDDING_LOAD_CIRCUIT_MODEL   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1033
#define RSVNA_ATTR_FIXTURE_MODELING_TIMESTAMD_FILENAMES_ENABLED                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1034
#define RSVNA_ATTR_FIXTURE_MODELING_DIRECTORY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1627
#define RSVNA_ATTR_FIXTURE_MODELING_DIRECTORY_DEFAULT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1628
#define RSVNA_ATTR_FIXTURE_MODELING_CLEAR_DEFAULT_DIRECTORY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1629
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_COUPON_TYPE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1035
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_PORT_ACTIVE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1036
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_START_COUPON_MEASUREMENT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1037
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_LOAD_TOUCHSTONE_FILE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1038
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_COUPON_INVALIDATES_FILE_PATH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1630
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_DUT_PORT_ACTIVE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1039
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_MEASURE_DUT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1040
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_LOAD_DUT_FILE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1253
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_DUT_INVALIDATES_FILE_PATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1631
#define RSVNA_ATTR_FIXTURE_MODELING_RUN_SFD_SIMULATOR                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1041
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_APPLY_TO_PORT_ENABLED                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1042
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_USE_IMPEDANCE_CORRECTION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1632
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_RESET_TO_DEFAULT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1633
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_2X_THRU_PORT_ORDERING                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1043
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_TOTAL_PORT_ORDERING                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1044
#define RSVNA_ATTR_FIXTURE_MODELING_SFD_ADJUST_IMPEDANCE_MISMATCH_ENABLED                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1045
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_COUPON_TYPE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1046
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_PORT_ACTIVE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1047
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_START_COUPON_MEASUREMENT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1048
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_START_COUPON_MEASUREMENT_OPEN                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1049
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_START_COUPON_MEASUREMENT_SHORT                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1050
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LOAD_TOUCHSTONE_FILE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1051
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_COUPON_INVALIDATES_FILE_PATH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1634
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LOAD_TOUCHSTONE_FILE_OPEN                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1052
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LOAD_TOUCHSTONE_FILE_SHORT                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1053
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_PORT_ACTIVE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1054
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_MEASURE_DUT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1055
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LOAD_DUT_FILE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1254
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_INVALIDATES_FILE_PATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1635
#define RSVNA_ATTR_FIXTURE_MODELING_RUN_ISD_SIMULATOR                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1056
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_APPLY_TO_PORT_ENABLED                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1057
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_USE_IMPEDANCE_CORRECTION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1636
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_RESET_TO_DEFAULT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1637
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_TEST_COUPONS_INSERTION_LOSS                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1058
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LEAD_INS_ATTENUATION_SCALING                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1059
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LEAD_INS_FLIGHT_TIME_SCALING                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1060
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LEAD_INS_FLIGHT_TIME_OVERRIDE_ENABLED            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1061
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_LEAD_INS_FLIGHT_TIME_OVERRIDE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1062
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_WITH_FIXTURE_PORT_SEQUENCE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1063
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_WITH_FIXTURE_DUT_TYPE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1064
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_WITH_FIXTURE_PORTS_TO_SKIP_NONE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1065
#define RSVNA_ATTR_FIXTURE_MODELING_IDS_DUT_WITH_FIXTURE_PORTS_TO_SKIP_LEFT              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1066
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_WITH_FIXTURE_PORTS_TO_SKIP_RIGHT             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1067
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_DUT_WITH_FIXTURE_PORTS_TO_SKIP                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1068
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_CALCULATIONS_ENFORCE_PASSIVITY                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1069
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_CALCULATIONS_TRACE_COUPLING                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1070
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_CALCULATIONS_OPERATION_ACCURACY                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1071
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_CALCULATIONS_MAX_FREQUENCY                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1072
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_CALCULATIONS_DC_EXTRAPOLATION                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1073
#define RSVNA_ATTR_FIXTURE_MODELING_ISD_CALCULATIONS_SMALL_FIXTURE_ENABLED               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1255
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_PORT_ACTIVE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1256
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_START_COUPON_MEASUREMENT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1257
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_LOAD_TOUCHSTONE_FILE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1258
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_INVALIDATES_FILE_PATH                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1638
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_DUT_PORT_ACTIVE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1259
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_MEASURE_DUT                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1260
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_LOAD_DUT_FILE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1261
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_DUT_INVALIDATES_FILE_PATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1639
#define RSVNA_ATTR_FIXTURE_MODELING_RUN_EZD_SIMULATOR                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1262
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_APPLY_TO_PORT_ENABLED                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1263
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_PORT_ORDER                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1640
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_RESET_TO_DEFAULT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1641
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_GENERATE_SIDE_1_ENABLED                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1264
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_GENERATE_SIDE_2_ENABLED                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1265
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_DC_EXTRAPOLATION_ENABLED                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1266
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_IMPEDANCE_CORRECTION_ENABLED                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1267
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_REFERENCE_IMPEDANCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1268
#define RSVNA_ATTR_FIXTURE_MODELING_EZD_BANDWIDTH_LIMIT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1269
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_RESET                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1642
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_RUN                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1643
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_USE_SAME_COUPON                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1644
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_IMP_CORRECT                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1645
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_TOOL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1646
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_RIGHT_MODEL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1647
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_LEFT_MODEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1648
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_COUPON_LEFT_CLEAR                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1649
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_COUPON_RIGHT_CLEAR                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1650
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_COUPON_LEFT_MEASURE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1653
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_COUPON_RIGHT_MEASURE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1654
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_COUPON_LEFT_TYPE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1657
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_COUPON_RIGHT_TYPE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1658
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_DUT_MEASURE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1659
#define RSVNA_ATTR_FIXTURE_MODELING_DEEMBED_ASSISTANT_DUT_TYPE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1661
#define RSVNA_ATTR_DELTA_L_CHARACTERIZATION_METHOD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1270
#define RSVNA_ATTR_DELTA_L_CHARACTERIZATION_M1L_DIFFERENTIAL_ENABLED                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1271
#define RSVNA_ATTR_DELTA_L_CHARACTERIZATION_MEASUREMENT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1272
#define RSVNA_ATTR_DELTA_L_PORT_ORDER                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1273
#define RSVNA_ATTR_DELTA_L_FREQUENCIES_USE_DEFAULT_ENABLED                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1274
#define RSVNA_ATTR_DELTA_L_FREQUENCY_COUNT_CURRENT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1275
#define RSVNA_ATTR_DELTA_L_FREQUENCIES_CURRENT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1276
#define RSVNA_ATTR_DELTA_L_FREQUENCY_COUNT_DEFAULT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1277
#define RSVNA_ATTR_DELTA_L_FREQUENCIES_DEFAULT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1278
#define RSVNA_ATTR_DELTA_L_FREQUENCY_COUNT_USER                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1279
#define RSVNA_ATTR_DELTA_L_FREQUENCIES_USER                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1280
#define RSVNA_ATTR_DELTA_L_ADD_USER_FREQUENCY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1283
#define RSVNA_ATTR_DELTA_L_DELETE_USER_FREQUENCY                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1284
#define RSVNA_ATTR_DELTA_L_DELETE_ALL_USER_FREQUENCIES                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1285
#define RSVNA_ATTR_DELTA_L_SWEEP_CONTROL                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1286
#define RSVNA_ATTR_DELTA_L_SWEEP_CONTROL_STEP                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1287
#define RSVNA_ATTR_DELTA_L_SWEEP_CONTROL_IF_BANDWIDTH                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1288
#define RSVNA_ATTR_DELTA_L_SWEEP_CONTROL_DYNAMIC_BANDWIDTH                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1482
#define RSVNA_ATTR_DELTA_L_SWEEP_RANGE_MINIMUM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1289
#define RSVNA_ATTR_DELTA_L_SWEEP_RANGE_MAXIMUM                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1290
#define RSVNA_ATTR_DELTA_L_USE_RECOMMENDED_TRACE_CONFIGURATION_ENABLED                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1291
#define RSVNA_ATTR_DELTA_L_SHOW_TDR_IMPEDEDANCES_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1292
#define RSVNA_ATTR_DELTA_L_LOAD_1L_MEASUREMENT_FILE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1293
#define RSVNA_ATTR_DELTA_L_LOAD_2L_MEASUREMENT_FILE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1294
#define RSVNA_ATTR_DELTA_L_LOAD_3L_MEASUREMENT_FILE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1295
#define RSVNA_ATTR_DELTA_L_1L_PORT_ENABLED                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1297
#define RSVNA_ATTR_DELTA_L_2L_PORT_ENABLED                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1298
#define RSVNA_ATTR_DELTA_L_3L_PORT_ENABLED                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1299
#define RSVNA_ATTR_DELTA_L_1L_LENGTH                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1301
#define RSVNA_ATTR_DELTA_L_2L_LENGTH                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1302
#define RSVNA_ATTR_DELTA_L_3L_LENGTH                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1303
#define RSVNA_ATTR_DELTA_L_MEASURE_1L                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1305
#define RSVNA_ATTR_DELTA_L_MEASURE_2L                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1306
#define RSVNA_ATTR_DELTA_L_MEASURE_3L                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1307
#define RSVNA_ATTR_DELTA_L_RUN                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1309
#define RSVNA_ATTR_DELTA_L_CLEAR_ALL_1L_MEASUREMENTS                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1310
#define RSVNA_ATTR_DELTA_L_CLEAR_1L_MEASUREMENT_FILE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1311
#define RSVNA_ATTR_DELTA_L_DIRECTORY                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1662
#define RSVNA_ATTR_DELTA_L_DIRECTORY_DEFAULT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1663
#define RSVNA_ATTR_HARMONIC_AUTOMATIC_GRID_ENABLED                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1074
#define RSVNA_ATTR_HARMONIC_IS_CURRENT                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1075
#define RSVNA_ATTR_PIPELINING_STATE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1312
#define RSVNA_ATTR_PIPELINING_CHANNEL_RESULTS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1313
#define RSVNA_ATTR_MIXER_DELAY_CALIBRATION_START                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1483
#define RSVNA_ATTR_MIXER_DELAY_2_TONE_DELTA_FREQUENCY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1484
#define RSVNA_ATTR_MIXER_DELAY_CONSTANT_DELAY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1485
#define RSVNA_ATTR_MIXER_DELAY_DISABLE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1486
#define RSVNA_ATTR_MIXER_DELAY_CALIBRATION_ACTIVE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1487
#define RSVNA_ATTR_MIXER_DELAY_GROUP_DELAY_RECEIVER_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1488
#define RSVNA_ATTR_MIXER_DELAY_LOWER_TONE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1489
#define RSVNA_ATTR_MIXER_DELAY_USE_CONSTANT_DELAY                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1494
#define RSVNA_ATTR_PULSE_ACQUISITION_TIME                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1495
#define RSVNA_ATTR_PULSE_COUPLED_ENABLED                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1496
#define RSVNA_ATTR_PULSE_DUTY_CYCLE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1497
#define RSVNA_ATTR_PULSE_FIXED_DUTY_CYCLE_ENABLED                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1498
#define RSVNA_ATTR_PULSE_GENERATOR_PERIOD                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1499
#define RSVNA_ATTR_PULSE_GENERATOR_SOURCE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1500
#define RSVNA_ATTR_PULSE_GENERATOR_TRAIN_PERIOD                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1501
#define RSVNA_ATTR_PULSE_INTERNAL_ENABLED                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1502
#define RSVNA_ATTR_PULSE_MEASUREMENT_DELAY                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1503
#define RSVNA_ATTR_PULSE_PERIOD                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1504
#define RSVNA_ATTR_PULSE_SYNC_MEASUREMENT_TO_PULSES                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1505
#define RSVNA_ATTR_PULSE_MEASUREMENT_ENABLED                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1506
#define RSVNA_ATTR_PULSE_ALC_MEASUREMENT_DELAY                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1749
#define RSVNA_ATTR_PULSE_GENERATOR_ENABLED                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1507
#define RSVNA_ATTR_PULSE_GENERATOR_DELAY                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1508
#define RSVNA_ATTR_PULSE_GENERATOR_WIDTH                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1509
#define RSVNA_ATTR_PULSE_GENERATOR_TRAIN_DELETE_ALL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1511
#define RSVNA_ATTR_PULSE_GENERATOR_TRAIN_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1512
#define RSVNA_ATTR_PULSE_GENERATOR_SEGMENT_COUNT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1513
#define RSVNA_ATTR_PULSE_GENERATOR_SEGMENT_ENABLED                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1514
#define RSVNA_ATTR_PULSE_GENERATOR_SEGMENT_START_TIME                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1515
#define RSVNA_ATTR_PULSE_GENERATOR_SEGMENT_STOP_TIME                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1516
#define RSVNA_ATTR_PULSE_PORT_DELAY                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1517
#define RSVNA_ATTR_PULSE_PORT_INPUT_INVERTED                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1518
#define RSVNA_ATTR_PULSE_PORT_OUTPUT_INVERTED                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1519
#define RSVNA_ATTR_PULSE_PORT_EXTERNAL_OUTPUT_ENABLED                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1520
#define RSVNA_ATTR_PULSE_PORT_WIDTH                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1521
#define RSVNA_ATTR_PULSE_PORT_ENABLED                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1522
#define RSVNA_ATTR_SOURCE_COHERENCE_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1523
#define RSVNA_ATTR_SOURCE_COHERENCE_OPTIMIZED_ENABLED                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1524
#define RSVNA_ATTR_SOURCE_COHERENCE_PORT_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1525
#define RSVNA_ATTR_SOURCE_COHERENCE_PORT_PHASE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1526
#define RSVNA_ATTR_SOURCE_COHERENCE_PORT_PHASE_SPAN                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1527
#define RSVNA_ATTR_TDIF_MODE_ENABLED                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1528
#define RSVNA_ATTR_TDIF_MODE_COMPENSATE_IMBALANCE_ENABLED                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1529
#define RSVNA_ATTR_TDIF_MODE_IMBALANCE_PHASE_PORT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1530
#define RSVNA_ATTR_TDIF_MODE_IMBALANCE_PHASE_START                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1531
#define RSVNA_ATTR_TDIF_MODE_IMBALANCE_PHASE_STOP                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1532
#define RSVNA_ATTR_TDIF_MODE_IMBALANCE_AMPLITUDE_PORT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1533
#define RSVNA_ATTR_TDIF_MODE_IMBALANCE_AMPLITUDE_START                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1534
#define RSVNA_ATTR_TDIF_MODE_IMBALANCE_AMPLITUDE_STOP                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1535
#define RSVNA_ATTR_GAIN_COMPRESSION_VALUE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1536
#define RSVNA_ATTR_GAIN_COMPRESSION_POWER_POINTS                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1537
#define RSVNA_ATTR_GAIN_COMPRESSION_START_POWER_POINT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1538
#define RSVNA_ATTR_GAIN_COMPRESSION_STOP_POWER_POINT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1539
#define RSVNA_ATTR_GAIN_COMPRESSION_RECEIVING_PORT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1540
#define RSVNA_ATTR_GAIN_COMPRESSION_DRIVING_PORT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1541
#define RSVNA_ATTR_GAIN_COMPRESSION_SKIP                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1750
#define RSVNA_ATTR_GAIN_COMPRESSION_SKIP_OFFSET                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1751
#define RSVNA_ATTR_SPECTRUM_DETECTOR                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1542
#define RSVNA_ATTR_SPECTRUM_IMAGE_REJECTION                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1543
#define RSVNA_ATTR_SPECTRUM_DETECTOR_ZERO_PADDING                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1752
#define RSVNA_ATTR_DISPLAY_DIAGRAM_AREA_MAXIMIZE_ALL                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 696
#define RSVNA_ATTR_DISPLAY_DIAGRAM_AREA_SHOW_TITLE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 697
#define RSVNA_ATTR_DISPLAY_DIAGRAM_AREA_TITLE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 698
#define RSVNA_ATTR_DISPLAY_DIAGRAM_AREA_NAME                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 699
#define RSVNA_ATTR_DISPLAY_DIAGRAM_AREA_CATALOG                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 700
#define RSVNA_ATTR_DISPLAY_TRACE_DIAGRAM_AREA_CATALOG                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 701
#define RSVNA_ATTR_SELECT_COLOR_SCHEME                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 702
#define RSVNA_ATTR_DISPLAY_FREQUENCY_INFO_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 703
#define RSVNA_ATTR_DISPLAY_CHANNEL_INFO_STATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 704
#define RSVNA_ATTR_DISPLAY_TRACE_INFO_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 774
#define RSVNA_ATTR_DISPLAY_FONT_SIZE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 705
#define RSVNA_ATTR_DISPLAY_TRACE_COLOR_FOR_LIMIT_LINES                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 706
#define RSVNA_ATTR_DISPLAY_COLORIZE_FAILED_SEGMENTS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 707
#define RSVNA_ATTR_DISPLAY_SHOW_FAILED_SYMBOLS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 708
#define RSVNA_ATTR_DISPLAY_ALL_MARKERS_SAME_COLOR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 709
#define RSVNA_ATTR_DISPLAY_TRACE_COLORS_PER_DIAGRAM                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 710
#define RSVNA_ATTR_DISPLAY_COLOR_SCHEME_SAVE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 711
#define RSVNA_ATTR_DISPLAY_COLOR_SCHEME_HCOPY_SAVE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1664
#define RSVNA_ATTR_DISPLAY_COLOR_SCHEME_LOAD                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 712
#define RSVNA_ATTR_DISPLAY_COLOR_SCHEME_HCOPY_LOAD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1665
#define RSVNA_ATTR_DISPLAY_MENU_BAR_STATE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 713
#define RSVNA_ATTR_DISPLAY_STATUS_BAR_STATE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 714
#define RSVNA_ATTR_DISPLAY_HARDKEY_PANEL_STATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 715
#define RSVNA_ATTR_DISPLAY_TITLE_BAR_STATE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 716
#define RSVNA_ATTR_DISPLAY_TOOLBAR_STATE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 717
#define RSVNA_ATTR_DISPLAY_SOFTTOOL_PANEL_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 718
#define RSVNA_ATTR_DISPLAY_ERROR_TOOLTIP                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 719
#define RSVNA_ATTR_DISPLAY_SPLIT_TYPE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 720
#define RSVNA_ATTR_SELECT_DISPLAY_LAYOUT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 775
#define RSVNA_ATTR_ZOOM_OVERVIEW                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 776
#define RSVNA_ATTR_DISPLAY_CONDUCTANCES                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 871
#define RSVNA_ATTR_DISPLAY_TRACE_LABEL                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1077
#define RSVNA_ATTR_DISPLAY_MENU_KEY_DIALOG_CATALOG                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1078
#define RSVNA_ATTR_DISPLAY_MENU_KEY_DIALOG                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1079
#define RSVNA_ATTR_DISPLAY_MENU_KEY_TAB_CATALOG                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1080
#define RSVNA_ATTR_DISPLAY_MENU_KEY_TAB                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1081
#define RSVNA_ATTR_DISPLAY_INFO_WINDOW_ENABLED                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1082
#define RSVNA_ATTR_DISPLAY_INFO_WINDOW_BANDFILTER_SEARCH_RESULTS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1084
#define RSVNA_ATTR_DISPLAY_INFO_WINDOW_MARKER_SEARCH_RESULTS                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1086
#define RSVNA_ATTR_DISPLAY_OVERLAY_ALL                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1666
#define RSVNA_ATTR_DISPLAY_SPLIT_ALL                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1667
#define RSVNA_ATTR_DISPLAY_X_AXIS_GRID_LABELS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1668
#define RSVNA_ATTR_TDR_STIMULUS_MAX_DUT_DELAY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1087
#define RSVNA_ATTR_TDR_STIMULUS_MAX_ELECTRICAL_LENGTH                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1088
#define RSVNA_ATTR_TDR_STIMULUS_MAX_MECHANICAL_LENGTH                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1089
#define RSVNA_ATTR_TDR_STIMULUS_PERMITIVITY                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1090
#define RSVNA_ATTR_TDR_STIMULUS_VELOCITY_FACTOR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1091
#define RSVNA_ATTR_TDR_STIMULUS_RISE_TIME_DEFINITION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1092
#define RSVNA_ATTR_TDR_STIMULUS_RISE_TIME                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1093
#define RSVNA_ATTR_TDR_EYE_ENABLED                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1094
#define RSVNA_ATTR_TDR_EYE_VIEW                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1095
#define RSVNA_ATTR_TDR_EYE_BIT_STREAM                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1096
#define RSVNA_ATTR_TDR_EYE_BIT_STREAM_PRBS_LENGTH                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1097
#define RSVNA_ATTR_TDR_EYE_BIT_STREAM_BIT_LENGTH                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1098
#define RSVNA_ATTR_TDR_EYE_BIT_STREAM_LOAD                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1099
#define RSVNA_ATTR_TDR_EYE_SYMBOL_RATE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1100
#define RSVNA_ATTR_TDR_EYE_MODULATION                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1101
#define RSVNA_ATTR_TDR_EYE_HIGH_LEVEL                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1102
#define RSVNA_ATTR_TDR_EYE_LOW_LEVEL                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1103
#define RSVNA_ATTR_TDR_EYE_LOW_PASS                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1104
#define RSVNA_ATTR_TDR_EYE_RISE_TIME                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1105
#define RSVNA_ATTR_TDR_EYE_RISE_TIME_DEFINITION                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1106
#define RSVNA_ATTR_TDR_EYE_TRANSFER_RATE_UNITS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1753
#define RSVNA_ATTR_TDR_EYE_GENERATOR_ENCODER                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1107
#define RSVNA_ATTR_TDR_EYE_GENERATOR_SCRAMBLER                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1108
#define RSVNA_ATTR_TDR_EYE_EMPHASIS_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1109
#define RSVNA_ATTR_TDR_EYE_EMPHASIS_PRE_CURSOR                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1110
#define RSVNA_ATTR_TDR_EYE_EMPHASIS_POST_CURSOR                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1111
#define RSVNA_ATTR_TDR_EYE_JITTER_ENABLED                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1112
#define RSVNA_ATTR_TDR_EYE_JITTER_RANDOM_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1113
#define RSVNA_ATTR_TDR_EYE_JITTER_RANDOM_STANDARD_DEVIATION                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1114
#define RSVNA_ATTR_TDR_EYE_JITTER_PERIODIC_ENABLED                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1115
#define RSVNA_ATTR_TDR_EYE_JITTER_PERIODIC_MAGNITUDE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1116
#define RSVNA_ATTR_TDR_EYE_JITTER_PERIODIC_FREQUENCY                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1117
#define RSVNA_ATTR_TDR_EYE_JITTER_PERIODIC_PHASE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1118
#define RSVNA_ATTR_TDR_EYE_JITTER_DIRAC_ENABLED                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1119
#define RSVNA_ATTR_TDR_EYE_JITTER_DIRAC_DELTA                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1120
#define RSVNA_ATTR_TDR_EYE_JITTER_DIRAC_PROBABILITY                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1121
#define RSVNA_ATTR_TDR_EYE_JITTER_USER_ENABLED                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1122
#define RSVNA_ATTR_TDR_EYE_JITTER_LOAD_USER                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1123
#define RSVNA_ATTR_TDR_EYE_NOISE_ENABLED                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1124
#define RSVNA_ATTR_TDR_EYE_NOISE_RMS                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1125
#define RSVNA_ATTR_TDR_EYE_DUT_MODE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1126
#define RSVNA_ATTR_TDR_EYE_EQUALIZATION_ENABLED                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1127
#define RSVNA_ATTR_TDR_EYE_EQUALIZATION_CTLE_DC                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1128
#define RSVNA_ATTR_TDR_EYE_EQUALIZATION_CTLE_ZERO                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1129
#define RSVNA_ATTR_TDR_EYE_EQUALIZATION_CTLE_POLE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1130
#define RSVNA_ATTR_TDR_EYE_MASK_SHOW                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1131
#define RSVNA_ATTR_TDR_EYE_MASK_ENABLED                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1132
#define RSVNA_ATTR_TDR_EYE_MASK_AUTO                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1133
#define RSVNA_ATTR_TDR_EYE_MASK_FAIL_BEEP                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1134
#define RSVNA_ATTR_TDR_EYE_MASK_FAIL_STATUS                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1135
#define RSVNA_ATTR_TDR_EYE_MASK_POLYGON_ENABLED                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1136
#define RSVNA_ATTR_TDR_EYE_MASK_POLYGON_SHAPE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1137
#define RSVNA_ATTR_TDR_EYE_MASK_POLYGON_WIDTH_MAIN                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1138
#define RSVNA_ATTR_TDR_EYE_MASK_POLYGON_WIDTH                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1139
#define RSVNA_ATTR_TDR_EYE_MASK_POLYGON_HEIGHT_MAIN                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1140
#define RSVNA_ATTR_TDR_EYE_MASK_POLYGON_HEIGHT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1141
#define RSVNA_ATTR_TDR_EYE_MASK_TOP_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1142
#define RSVNA_ATTR_TDR_EYE_MASK_TOP_WIDTH                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1143
#define RSVNA_ATTR_TDR_EYE_MASK_TOP_HEIGHT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1144
#define RSVNA_ATTR_TDR_EYE_MASK_BOTTOM_ENABLED                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1145
#define RSVNA_ATTR_TDR_EYE_MASK_BOTTOM_WIDTH                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1146
#define RSVNA_ATTR_TDR_EYE_MASK_BOTTOM_HEIGHT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1147
#define RSVNA_ATTR_TDR_EYE_MASK_CENTER_WIDTH                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1148
#define RSVNA_ATTR_TDR_EYE_MASK_CENTER_HEIGHT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1149
#define RSVNA_ATTR_TDR_EYE_MASK_FAIL_CONDITION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1150
#define RSVNA_ATTR_TDR_EYE_MASK_VIOLATION_TOLERANCE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1151
#define RSVNA_ATTR_TDR_EYE_MASK_VIOLATION_RATE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1152
#define RSVNA_ATTR_TDR_EYE_MASK_DATA                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1153
#define RSVNA_ATTR_TDR_EYE_MASK_LOAD                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1154
#define RSVNA_ATTR_TDR_EYE_MASK_STORE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1155
#define RSVNA_ATTR_TDR_EYE_MASK_STORE_RESULTS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1156
#define RSVNA_ATTR_TDR_EYE_MEASUREMENT_ENABLED                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1157
#define RSVNA_ATTR_TDR_EYE_MEASUREMENT_DATA                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1158
#define RSVNA_ATTR_TDR_EYE_MEASUREMENT_THRESHOLD                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1159
#define RSVNA_ATTR_TDR_EYE_MEASUREMENT_THRESHOLD_CUSTOM                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1160
#define RSVNA_ATTR_TDR_EYE_MEASUREMENT_EXPORT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1161
#define RSVNA_ATTR_TDR_RISE_TIME_ENABLED                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1163
#define RSVNA_ATTR_TDR_RISE_TIME_DATA                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1165
#define RSVNA_ATTR_TDR_RISE_TIME_THRESHOLD                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1166
#define RSVNA_ATTR_TDR_RISE_TIME_THRESHOLD_CUSTOM                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1167
#define RSVNA_ATTR_SKEW_ENABLED                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1169
#define RSVNA_ATTR_SKEW_REFERENCE_TRACE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1170
#define RSVNA_ATTR_SKEW_POSITION                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1171
#define RSVNA_ATTR_SKEW_CHECK_ON                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1172
#define RSVNA_ATTR_SKEW_LIMIT                                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1173
#define RSVNA_ATTR_SKEW_FAIL_BEEP                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1174
#define RSVNA_ATTR_SKEW_FAIL                                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1175
#define RSVNA_ATTR_SKEW_RESULT                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1176
#define RSVNA_ATTR_SKEW_ALL_RESULTS                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1178
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1754
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_FREQUENCY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1755
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_POWER                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1756
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_RF_OFF                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1757
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_PORT_ENABLED                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1758
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_PORT_FREQUENCY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1759
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_PORT_POWER                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1760
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_PORT_RF_OFF                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1761
#define RSVNA_ATTR_INDEPENDENT_GENERATOR_PORT_STEP_ATTENUATOR                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1762
#define RSVNA_ATTR_PRESET_SCOPE                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 721
#define RSVNA_ATTR_REMOTE_PRESET                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 841
#define RSVNA_ATTR_FACTORY_PRESET                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 722
#define RSVNA_ATTR_USER_DEFINED_PRESET_STATE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 723
#define RSVNA_ATTR_USER_DEFINED_PRESET_NAME                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 724
#define RSVNA_ATTR_USER_DEFINED_PRESET_CAL                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 819
#define RSVNA_ATTR_CALIBRATION_AUTO_AVERAGING                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 820
#define RSVNA_ATTR_KEY_LOCK                                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 725
#define RSVNA_ATTR_TOUCHSCREEN_LOCK                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 726
#define RSVNA_ATTR_REMOTE_LANGUAGE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 727
#define RSVNA_ATTR_SAVE_REPORT                                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 728
#define RSVNA_ATTR_REMOTE_LOGGING                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 729
#define RSVNA_ATTR_REMOTE_ENCODING                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 872
#define RSVNA_ATTR_CALIBRATION_MANAGER_STATE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 808
#define RSVNA_ATTR_SYSTEM_SHUTDOWN                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 730
#define RSVNA_ATTR_SELF_TEST_QUERY                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 731
#define RSVNA_ATTR_QUESTIONABLE_REGISTER                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 732
#define RSVNA_ATTR_PRESET_STATUS_REGISTERS                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 733
#define RSVNA_ATTR_SOUND_ALARM_STATE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 734
#define RSVNA_ATTR_SOUND_STATUS_STATE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 735
#define RSVNA_ATTR_USER_PORT_CHANNEL_BITS                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 736
#define RSVNA_ATTR_USER_PORT_PINS_USAGE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 737
#define RSVNA_ATTR_USER_PORT_OPTIONAL_COLUMNS_ENABLED                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1179
#define RSVNA_ATTR_USER_PORT_OPTIONAL_COLUMNS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1180
#define RSVNA_ATTR_REFERENCE_OSCILATOR_SOURCE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 738
#define RSVNA_ATTR_REFERENCE_OSCILLATOR_SMA_INPUT_FREQUENCY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1544
#define RSVNA_ATTR_REFERENCE_OSCILLATOR_SMA_OUTPUT_FREQUENCY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1545
#define RSVNA_ATTR_EXTERNAL_REFERENCE_OSCILATOR_FREQUENCY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 739
#define RSVNA_ATTR_SYST_VERSION                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 740
#define RSVNA_ATTR_INIT_WAIT_STATE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 824
#define RSVNA_ATTR_LOGICAL_PORT_ALINGMENT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 825
#define RSVNA_ATTR_DISPLAY_TITLE                                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1181
#define RSVNA_ATTR_DISPLAY_NUMBER_OF_TRACE_COLOURS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1182
#define RSVNA_ATTR_DISPLAY_UPDATE                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 765
#define RSVNA_ATTR_DISPLAY_UPDATE_ONCE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 876
#define RSVNA_ATTR_START_IN_PRESET                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1183
#define RSVNA_ATTR_SET_ALL_TRACES_TO_0                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1184
#define RSVNA_ATTR_SYSTEM_SCPI_COMMANDS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1185
#define RSVNA_ATTR_QUERY_IMPLEMENTED_SCPI_COMMANDS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1188
#define RSVNA_ATTR_GPIB_SUPRESSES                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1189
#define RSVNA_ATTR_HOSTNAME                                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1190
#define RSVNA_ATTR_CORRECTION_WIZARD_DIALOG                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1191
#define RSVNA_ATTR_USER_DEFINED_LED_COLOR                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1546
#define RSVNA_ATTR_RF_OFF_BEHAVIOR                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1547
#define RSVNA_ATTR_SYSTEM_PRESET_MODE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1548
#define RSVNA_ATTR_OPC_CHECK_SELFTEST                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1669
#define RSVNA_ATTR_FORMAT_IDENTIFY                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1763
#define RSVNA_ATTR_EXTERNAL_GENERATOR_POWER_MODE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 741
#define RSVNA_ATTR_EXTERNAL_GENERATOR_POWER                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 742
#define RSVNA_ATTR_EXTERNAL_GENERATORS_COUNT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 743
#define RSVNA_ATTR_EXTERNAL_GENERATORS_NUMBERS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 744
#define RSVNA_ATTR_EXTERNAL_GENERATOR_DELETE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 745
#define RSVNA_ATTR_EXTERNAL_POWER_METER_S_PARAMETER_CORRECTION                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 809
#define RSVNA_ATTR_EXTERNAL_POWER_METER_AUTO_ZEROING                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 746
#define RSVNA_ATTR_EXTERNAL_POWER_METER_AUTO_CONFIG                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 747
#define RSVNA_ATTR_EXTERNAL_POWER_METER_COUNT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 748
#define RSVNA_ATTR_EXTERNAL_POWER_METER_NUMBERS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 749
#define RSVNA_ATTR_EXTERNAL_POWER_METER_DELETE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 750
#define RSVNA_ATTR_SWITCH_MATRIX_STATE                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 786
#define RSVNA_ATTR_DELETE_SWITCH_MATRIX                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 787
#define RSVNA_ATTR_SWITCH_MATRIX_COUNT                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 788
#define RSVNA_ATTR_SWITCH_MATRIX_NUMBERS                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 789
#define RSVNA_ATTR_SWITCH_MATRIX_TEST_PORTS_COUNT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 790
#define RSVNA_ATTR_SWITCH_MATRIX_START_CONFIGURATION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 791
#define RSVNA_ATTR_SWITCH_MATRIX_ABORT_CONFIGURATION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 792
#define RSVNA_ATTR_SWITCH_MATRIX_END_CONFIGURATION                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 793
#define RSVNA_ATTR_SWITCH_MATRIX_OPTIMIZATION                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 794
#define RSVNA_ATTR_SCAN_SWITCH_MATRIX                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1192
#define RSVNA_ATTR_SWITCH_MATRIX_SERIAL_NUMBER                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1193
#define RSVNA_ATTR_PORT_POWER_LIMIT_ENABLED                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1549
#define RSVNA_ATTR_PORT_POWER_LIMIT                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1550
#define RSVNA_ATTR_EXTERNAL_GENERATOR_LIMIT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1551
#define RSVNA_ATTR_EXTERNAL_GENERATOR_LIMIT_ENABLED                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1552
#define RSVNA_ATTR_SELECT_APPLICATION                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 751
#define RSVNA_ATTR_WAIT_TO_FINISH                                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 752
#define RSVNA_ATTR_QUERY_APPLICATION_RETURN_VALUE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 873
#define RSVNA_ATTR_UI_PORT_DATA                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 753
#define RSVNA_ATTR_UI_PORT_DATA_MODE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 810
#define RSVNA_ATTR_UI_PORT_WRITING_DATA                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 811
#define RSVNA_ATTR_UI_OUTPUT_PORT_DATA                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 812
#define RSVNA_ATTR_UI_OUTPUT_PORT_STATE                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 813
#define RSVNA_ATTR_UI_DIGITAL_SIGNAL                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 814
#define RSVNA_ATTR_UI_DATA_PORTS_LOGIC                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 754
#define RSVNA_ATTR_UI_PASS_FAIL_LOGIC                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 755
#define RSVNA_ATTR_UI_PASS_FAIL_DEFAULT_STATE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 756
#define RSVNA_ATTR_UI_PASS_FAIL_SCOPE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 757
#define RSVNA_ATTR_UI_PASS_FAIL_POLICY                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 758
#define RSVNA_ATTR_UI_SWEEP_END_EVENT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 759
#define RSVNA_ATTR_UI_PASS_FAIL_STATUS                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 760
#define RSVNA_ATTR_UI_TRANSITION_STATUS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 761
#define RSVNA_ATTR_USER_DEFINED_PHYSICAL_PORT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 822
#define RSVNA_ATTR_USER_DEFINED_PHYSICAL_PORT_STATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 823
#define RSVNA_ATTR_DATA_MEMORY_ADD_TRACE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1195
#define RSVNA_ATTR_DATA_MEMORY_NAMES_CATALOG                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1196
#define RSVNA_ATTR_DATA_MEMORY_INITIALIZE_NEW_SHARED_BUFFER                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1197
#define RSVNA_ATTR_DATA_MEMORY_COMMIT_NAME                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1198
#define RSVNA_ATTR_DATA_MEMORY_DELETE                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1199
#define RSVNA_ATTR_DATA_MEMORY_OFFSET                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1200
#define RSVNA_ATTR_DATA_MEMORY_DELETE_ALL_SHARED_BUFFERS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1201
#define RSVNA_ATTR_DATA_MEMORY_SIZE                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1202
#define RSVNA_ATTR_HUMS_ENABLED                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1315
#define RSVNA_ATTR_HUMS_EXPORT_HISTORY                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1316
#define RSVNA_ATTR_HUMS_DELETE_HISTORY                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1317
#define RSVNA_ATTR_HUMS_DELETE_DEVICE_HISTORY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1318
#define RSVNA_ATTR_HUMS_DELETE_UTILIZATION_HISTORY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1319
#define RSVNA_ATTR_HUMS_DATA_FORMAT                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1320
#define RSVNA_ATTR_HUMS_REST_ENABLED                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1335
#define RSVNA_ATTR_HUMS_SNMP_VERSION                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1336
#define RSVNA_ATTR_HUMS_SNMP_COMMUNITY_READ_ONLY                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1337
#define RSVNA_ATTR_HUMS_SNMP_COMMUNITY_READ_WRITE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1338
#define RSVNA_ATTR_HUMS_SNMP_SYSTEM_CONTACT                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1339
#define RSVNA_ATTR_HUMS_SNMP_SYSTEM_LOCATION                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1340
#define RSVNA_ATTR_HUMS_SNMP_USER_LIST                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1343
#define RSVNA_ATTR_HUMS_SNMP_DELETE_USER                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1344
#define RSVNA_ATTR_HUMS_SNMP_DELETE_ALL_USERS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1345
#define RSVNA_ATTR_HUMS_TAG_KEY_VALUE_PAIR                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1348
#define RSVNA_ATTR_HUMS_ALL_DEVICE_TAGS                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1349
#define RSVNA_ATTR_HUMS_TAG_DELETE                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1350
#define RSVNA_ATTR_HUMS_TAG_DELETE_ALL                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1351
#define RSVNA_ATTR_HUMS_DEVICE_STATUS_SUMMARY                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1354
#define RSVNA_ATTR_HUMS_UTILIZATION_DELETE_CUSTOM                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1674
#define RSVNA_ATTR_HUMS_UTILIZATION_DELETE_CUSTOM_ALL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1675
#define RSVNA_ATTR_HUMS_UTILIZATION_ALL                                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1676
#define RSVNA_ATTR_SERVICE_CALIBRATION_LAST_DATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1355
#define RSVNA_ATTR_SERVICE_CALIBRATION_DUE_DATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1356
#define RSVNA_ATTR_SERVICE_CALIBRATION_DUE_STATE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1357
#define RSVNA_ATTR_SERVICE_CALIBRATION_INTERVAL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1358
#define RSVNA_ATTR_SERVICE_LAST_SERVICE_DATE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1359
#define RSVNA_ATTR_SERVICE_DUE_STATE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1360
#define RSVNA_ATTR_GENERIC_DEVICE_CATALOG                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1553
#define RSVNA_ATTR_GENERIC_DEVICE_DELETE_ALL                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1554
#define RSVNA_ATTR_GENERIC_DEVICE_DELETE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1555
#define RSVNA_ATTR_GENERIC_DEVICES_ADDRESSES                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1557
#define RSVNA_ATTR_GENERIC_DEVICES_DELETE_ADDRESS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1558
#define RSVNA_ATTR_GENERIC_DEVICE_CHARACTERIZATION                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1678
#define RSVNA_ATTR_GENERIC_DEVICE_CONFIGURED_DEVICES                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1679
#define RSVNA_ATTR_CONVERTER_TYPES_CATALOG                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1560
#define RSVNA_ATTR_CONVERTER_COUNT                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1561
#define RSVNA_ATTR_CONVERTER_DEFINE_NEW_TYPE                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1562
#define RSVNA_ATTR_CONVERTER_DEFINITION_DELETE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1563
#define RSVNA_ATTR_FREQUENCY_CONVERSION_VNA_PORT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1582
#define RSVNA_ATTR_CONVERTER_SPLITTER_LOSS                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1584
#define RSVNA_ATTR_CONVERTER_DEFAULT_CABLE_LOSS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1585
#define RSVNA_ATTR_CONVERTER_DEFAULT_SLOPE_FACTOR                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1586
#define RSVNA_ATTR_CONVERTER_IF_INPUT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1587
#define RSVNA_ATTR_CONVERTER_IF_FREQUENCY                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1588
#define RSVNA_ATTR_CONVERTER_LO_LOSS                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1589
#define RSVNA_ATTR_CONVERTER_SOURCE_PORT_LOSS                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1591
#define RSVNA_ATTR_CONVERTER_PORT_ENABLED                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1593
#define RSVNA_ATTR_CONVERTER_PORT_RF_SLOPE                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1765
#define RSVNA_ATTR_LEVELING_DATASET_POWER_CONTROL_MODE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1594
#define RSVNA_ATTR_LEVELING_DATASET_MODE                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1595
#define RSVNA_ATTR_LEVELING_DATASET_PATH                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1597
#define RSVNA_ATTR_ID_QUERY_RESPONSE                                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 762
#define RSVNA_ATTR_SYSTEM_ERROR                                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 763
#define RSVNA_ATTR_SYSTEM_ERROR_COUNT                                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1361
#define RSVNA_ATTR_NUMBER_OF_PORTS                                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 764

// rsvna_rngFilterSelectivity
#define RSVNA_VAL_SELECTIVITY_NORMAL                                        0
#define RSVNA_VAL_SELECTIVITY_HIGH                                          1
#define RSVNA_VAL_SELECTIVITY_MEDIUM                                        2
#define RSVNA_VAL_SELECTIVITY_NFIG                                          3

// rsvna_rngSweepType
#define RSVNA_VAL_SWEEP_LIN                                                 0
#define RSVNA_VAL_SWEEP_LOG                                                 1
#define RSVNA_VAL_SWEEP_SEG                                                 2
#define RSVNA_VAL_SWEEP_POW                                                 3
#define RSVNA_VAL_SWEEP_TIM                                                 4
#define RSVNA_VAL_SWEEP_CW                                                  5
#define RSVNA_VAL_SWEEP_PULSE                                               6
#define RSVNA_VAL_SWEEP_IAMP                                                7
#define RSVNA_VAL_SWEEP_IPH                                                 8

// rsvna_rngTimeMode
#define RSVNA_VAL_SEG_TIME                                                  0
#define RSVNA_VAL_SEG_POINT                                                 1

// rsvna_rngTSweepScope
#define RSVNA_VAL_SWEEP_SINGLE_CHAN                                         0
#define RSVNA_VAL_SWEEP_ALL_CHAN                                            1

// rsvna_rngPrintFileFormat
#define RSVNA_VAL_BMP                                                       0
#define RSVNA_VAL_JPG                                                       1
#define RSVNA_VAL_PNG                                                       2
#define RSVNA_VAL_PDF                                                       3
#define RSVNA_VAL_SVG                                                       4

// rsvna_rngPrintDiagramAreas
#define RSVNA_VAL_HCOPY_ALL                                                 0
#define RSVNA_VAL_HCOPY_ACTIVE                                              1
#define RSVNA_VAL_HCOPY_SINGLE                                              2
#define RSVNA_VAL_HCOPY_HARD                                                3
#define RSVNA_VAL_HCOPY_NONE                                                4

// rsvna_rngPrintPageOrientation
#define RSVNA_VAL_PORTRAIT                                                  0
#define RSVNA_VAL_LANDSCAPE                                                 1

// rsvna_rngMarkerFormat
#define RSVNA_VAL_MARKER_FORMAT_DB_MAG                                      0
#define RSVNA_VAL_MARKER_FORMAT_LIN_MAG                                     1
#define RSVNA_VAL_MARKER_FORMAT_PHASE                                       2
#define RSVNA_VAL_MARKER_FORMAT_REAL                                        3
#define RSVNA_VAL_MARKER_FORMAT_IMAG                                        4
#define RSVNA_VAL_MARKER_FORMAT_SWR                                         5
#define RSVNA_VAL_MARKER_FORMAT_DELAY                                       6
#define RSVNA_VAL_MARKER_FORMAT_DB_MAG_PHASE                                7
#define RSVNA_VAL_MARKER_FORMAT_LIN_MAG_PHASE                               8
#define RSVNA_VAL_MARKER_FORMAT_REAL_IMAG                                   9
#define RSVNA_VAL_MARKER_FORMAT_DEFAULT                                     10
#define RSVNA_VAL_MARKER_FORMAT_R_JX                                        11
#define RSVNA_VAL_MARKER_FORMAT_G_JB                                        12
#define RSVNA_VAL_MARKER_FORMAT_MIMP                                        13
#define RSVNA_VAL_MARKER_FORMAT_MDB                                         14
#define RSVNA_VAL_MARKER_FORMAT_COMP                                        15
#define RSVNA_VAL_MARKER_FORMAT_MLPH                                        16
#define RSVNA_VAL_MARKER_FORMAT_MDPH                                        17
#define RSVNA_VAL_MARKER_FORMAT_PIMP                                        18
#define RSVNA_VAL_MARKER_FORMAT_PADM                                        19
#define RSVNA_VAL_MARKER_FORMAT_MADM                                        20
#define RSVNA_VAL_MARKER_FORMAT_MPIM                                        21
#define RSVNA_VAL_MARKER_FORMAT_MPAD                                        22
#define RSVNA_VAL_MARKER_FORMAT_INDX                                        23
#define RSVNA_VAL_MARKER_FORMAT_NOIS                                        24

// rsvna_rngMathFunctions
#define RSVNA_VAL_MATH_FUNC_NORM                                            0
#define RSVNA_VAL_MATH_FUNC_ADD                                             1
#define RSVNA_VAL_MATH_FUNC_SUB                                             2
#define RSVNA_VAL_MATH_FUNC_MULT                                            3
#define RSVNA_VAL_MATH_FUNC_DIV                                             4

// rsvna_rngTraceFormat
#define RSVNA_VAL_DB_MAG                                                    0
#define RSVNA_VAL_PHASE                                                     1
#define RSVNA_VAL_SMITH                                                     2
#define RSVNA_VAL_POLAR                                                     3
#define RSVNA_VAL_DELAY                                                     4
#define RSVNA_VAL_SWR                                                       5
#define RSVNA_VAL_LIN_MAG                                                   6
#define RSVNA_VAL_REAL                                                      7
#define RSVNA_VAL_IMAG                                                      8
#define RSVNA_VAL_ISMITH                                                    9
#define RSVNA_VAL_UPHASE                                                    10
#define RSVNA_VAL_ZVR_COMPLEX                                               11
#define RSVNA_VAL_ZVR_MAGNITUDE                                             12
#define RSVNA_VAL_LOGARITHMIC                                               13

// rsvna_rngCalibrationStatus
#define RSVNA_VAL_CALSTATE_CAL                                              0
#define RSVNA_VAL_CALSTATE_CAI                                              1
#define RSVNA_VAL_CALSTATE_CA                                               2
#define RSVNA_VAL_CALSTATE_CAV                                              3
#define RSVNA_VAL_CALSTATE_CALOFF                                           4
#define RSVNA_VAL_CALSTATE_NONE                                             5

// rsvna_rngCalibrationConnectorShort
#define RSVNA_VAL_CONNECTOR_N50FEMALE                                       0
#define RSVNA_VAL_CONNECTOR_N50MALE                                         1
#define RSVNA_VAL_CONNECTOR_N75FEMALE                                       2
#define RSVNA_VAL_CONNECTOR_N75MALE                                         3
#define RSVNA_VAL_CONNECTOR_PC7                                             4
#define RSVNA_VAL_CONNECTOR_PC35FEMALE                                      5
#define RSVNA_VAL_CONNECTOR_PC35MALE                                        6
#define RSVNA_VAL_CONNECTOR_PC292FEMALE                                     7
#define RSVNA_VAL_CONNECTOR_PC292MALE                                       8
#define RSVNA_VAL_CONNECTOR_UFEMALE1                                        9
#define RSVNA_VAL_CONNECTOR_UMALE1                                          10
#define RSVNA_VAL_CONNECTOR_UFEMALE2                                        11
#define RSVNA_VAL_CONNECTOR_UMALE2                                          12
#define RSVNA_VAL_CONNECTOR_SMAFEMALE                                       13
#define RSVNA_VAL_CONNECTOR_SMAMALE                                         14
#define RSVNA_VAL_CONNECTOR_BNC50FEMALE                                     15
#define RSVNA_VAL_CONNECTOR_BNC50MALE                                       16
#define RSVNA_VAL_CONNECTOR_BNC75FEMALE                                     17
#define RSVNA_VAL_CONNECTOR_BNC75MALE                                       18
#define RSVNA_VAL_CONNECTOR_PC24FEMALE                                      19
#define RSVNA_VAL_CONNECTOR_PC24MALE                                        20
#define RSVNA_VAL_CONNECTOR_PC185FEMALE                                     21
#define RSVNA_VAL_CONNECTOR_PC185MALE                                       22
#define RSVNA_VAL_CONNECTOR_S7_16FEMALE                                     23
#define RSVNA_VAL_CONNECTOR_S7_16MALE                                       24
#define RSVNA_VAL_CONNECTOR_WM864                                           25
#define RSVNA_VAL_CONNECTOR_WM1295                                          26
#define RSVNA_VAL_CONNECTOR_WM1651                                          27
#define RSVNA_VAL_CONNECTOR_WM2032                                          28
#define RSVNA_VAL_CONNECTOR_WM2540                                          29
#define RSVNA_VAL_CONNECTOR_WR02                                            30
#define RSVNA_VAL_CONNECTOR_WR03                                            31
#define RSVNA_VAL_CONNECTOR_WR05                                            32
#define RSVNA_VAL_CONNECTOR_WR06                                            33
#define RSVNA_VAL_CONNECTOR_WR08                                            34
#define RSVNA_VAL_CONNECTOR_WR10                                            35
#define RSVNA_VAL_CONNECTOR_WR12                                            36
#define RSVNA_VAL_CONNECTOR_WR15                                            37

// rsvna_rngCalibrationSameConnectorGender
#define RSVNA_VAL_CONNECTOR_TYPE_SINGLE                                     0
#define RSVNA_VAL_CONNECTOR_TYPE_ALL                                        1

// rsvna_rngSpuriousAvoidance
#define RSVNA_VAL_AVOID_AUTO                                                0
#define RSVNA_VAL_AVOID_POS                                                 1
#define RSVNA_VAL_AVOID_NEG                                                 2

// rsvna_rngAGCMode
#define RSVNA_VAL_AGC_MODE_AUTO                                             0
#define RSVNA_VAL_AGC_MODE_LOW_NOISE                                        1
#define RSVNA_VAL_AGC_MODE_LOW_DISTORTION                                   2

// rsvna_rngTriggerSource
#define RSVNA_VAL_TRIGGER_SOURCE_IMMEDIATE                                  0
#define RSVNA_VAL_TRIGGER_SOURCE_EXTERNAL                                   1
#define RSVNA_VAL_TRIGGER_SOURCE_MANUAL                                     2
#define RSVNA_VAL_TRIGGER_SOURCE_MULTIPLE                                   3

// rsvna_rngTriggerSequence
#define RSVNA_VAL_TRG_SEQ_SWEEP                                             0
#define RSVNA_VAL_TRG_SEQ_SEGMENT                                           1
#define RSVNA_VAL_TRG_SEQ_POINT                                             2
#define RSVNA_VAL_TRG_SEQ_PARTIAL_MEAS                                      3

// rsvna_rngTriggerSlope
#define RSVNA_VAL_NEG                                                       0
#define RSVNA_VAL_POS                                                       1
#define RSVNA_VAL_HIGH                                                      2
#define RSVNA_VAL_LOW                                                       3

// rsvna_rngRefOscillatorSource
#define RSVNA_VAL_INT                                                       0
#define RSVNA_VAL_EXT                                                       1
#define RSVNA_VAL_ESMA                                                      2

// rsvna_rngSweepSegmentSelectivity
#define RSVNA_VAL_SEGMENT_SELECTIVITY_NORMAL                                0
#define RSVNA_VAL_SEGMENT_SELECTIVITY_MEDIUM                                1
#define RSVNA_VAL_SEGMENT_SELECTIVITY_HIGH                                  2

// rsvna_rngParameterConversion
#define RSVNA_VAL_CONV_S                                                    0
#define RSVNA_VAL_CONV_Y                                                    1
#define RSVNA_VAL_CONV_Z                                                    2

// rsvna_rngWaveQuantitiesUnitType
#define RSVNA_VAL_WAVE_UNIT_TYPE_POWER                                      0
#define RSVNA_VAL_WAVE_UNIT_TYPE_VOLTAGE                                    1

// rsvna_rngDCRange
#define RSVNA_VAL_DC_RANGE_MV300                                            0
#define RSVNA_VAL_DC_RANGE_V3                                               1
#define RSVNA_VAL_DC_RANGE_V20                                              2

// rsvna_rngImpedanceRenormalization
#define RSVNA_VAL_RENORMALIZATION_TWAV                                      0
#define RSVNA_VAL_RENORMALIZATION_PWAV                                      1

// rsvna_rngTimeDomainType
#define RSVNA_VAL_TYPE_BAND_PASS                                            0
#define RSVNA_VAL_TYPE_LOW_PASS                                             1

// rsvna_rngTimeDomainStimulus
#define RSVNA_VAL_STIMULUS_IMPULSE                                          0
#define RSVNA_VAL_STIMULUS_STEP                                             1

// rsvna_rngTimeDomainImpulseResponse
#define RSVNA_VAL_FILTER_RECT                                               0
#define RSVNA_VAL_FILTER_HANN                                               1
#define RSVNA_VAL_FILTER_HAMM                                               2
#define RSVNA_VAL_FILTER_BOHM                                               3
#define RSVNA_VAL_FILTER_DCH                                                4

// rsvna_rngTimeDomainHarmonicGrid
#define RSVNA_VAL_GRID_KFST                                                 0
#define RSVNA_VAL_GRID_KDFR                                                 1
#define RSVNA_VAL_GRID_KSDFR                                                2

// rsvna_rngTimeGateType
#define RSVNA_VAL_TGATE_TYPE_BPAS                                           0
#define RSVNA_VAL_TGATE_TYPE_NOTCH                                          1

// rsvna_rngStatisticsEvalRange
#define RSVNA_VAL_FULL_SPAN                                                 0
#define RSVNA_VAL_RANGE_1                                                   1
#define RSVNA_VAL_RANGE_2                                                   2
#define RSVNA_VAL_RANGE_3                                                   3
#define RSVNA_VAL_RANGE_4                                                   4
#define RSVNA_VAL_RANGE_5                                                   5
#define RSVNA_VAL_RANGE_6                                                   6
#define RSVNA_VAL_RANGE_7                                                   7
#define RSVNA_VAL_RANGE_8                                                   8
#define RSVNA_VAL_RANGE_9                                                   9
#define RSVNA_VAL_RANGE_10                                                  10

// rsvna_rngHoldTrace
#define RSVNA_VAL_HOLD_OFF                                                  0
#define RSVNA_VAL_HOLD_MIN                                                  1
#define RSVNA_VAL_HOLD_MAX                                                  2

// rsvna_rngLimitSegmentType
#define RSVNA_VAL_LIMIT_OFF                                                 0
#define RSVNA_VAL_LIMIT_UPPER                                               1
#define RSVNA_VAL_LIMIT_LOWER                                               2

// rsvna_rngRipplePhysicalUnits
#define RSVNA_VAL_LIMIT_DOM_FLIN                                            0
#define RSVNA_VAL_LIMIT_DOM_FLOG                                            1
#define RSVNA_VAL_LIMIT_DOM_FSEG                                            2
#define RSVNA_VAL_LIMIT_DOM_FSIN                                            3
#define RSVNA_VAL_LIMIT_DOM_TLIN                                            4
#define RSVNA_VAL_LIMIT_DOM_TLOG                                            5
#define RSVNA_VAL_LIMIT_DOM_PLIN                                            6
#define RSVNA_VAL_LIMIT_DOM_PLOG                                            7
#define RSVNA_VAL_LIMIT_DOM_PSIN                                            8

// rsvna_rngRippleResponseDomainFormatUnits
#define RSVNA_VAL_LIMIT_RDOM_FORM_COMP                                      0
#define RSVNA_VAL_LIMIT_RDOM_FORM_MAGN                                      1
#define RSVNA_VAL_LIMIT_RDOM_FORM_PHAS                                      2
#define RSVNA_VAL_LIMIT_RDOM_FORM_REAL                                      3
#define RSVNA_VAL_LIMIT_RDOM_FORM_IMAG                                      4
#define RSVNA_VAL_LIMIT_RDOM_FORM_SWR                                       5
#define RSVNA_VAL_LIMIT_RDOM_FORM_GDEL                                      6
#define RSVNA_VAL_LIMIT_RDOM_FORM_L                                         7
#define RSVNA_VAL_LIMIT_RDOM_FORM_C                                         8

// rsvna_rngMarkerMode
#define RSVNA_VAL_CONTINUOUS                                                0
#define RSVNA_VAL_DISCRETE                                                  1

// rsvna_rngMarkerType
#define RSVNA_VAL_NORMAL                                                    0
#define RSVNA_VAL_FIXED                                                     1
#define RSVNA_VAL_ARBITRARY                                                 2

// rsvna_rngMarkerSearch
#define RSVNA_VAL_MARKER_MAX                                                0
#define RSVNA_VAL_MARKER_MIN                                                1
#define RSVNA_VAL_MARKER_NEXT                                               2
#define RSVNA_VAL_MARKER_RPEAK                                              3
#define RSVNA_VAL_MARKER_LPEAK                                              4
#define RSVNA_VAL_MARKER_TARGET                                             5
#define RSVNA_VAL_MARKER_RTARGET                                            6
#define RSVNA_VAL_MARKER_LTARGET                                            7
#define RSVNA_VAL_MARKER_BFILTER                                            8
#define RSVNA_VAL_MARKER_NMAXIMUM                                           9
#define RSVNA_VAL_MARKER_NMINIMUM                                           10
#define RSVNA_VAL_MARKER_SPROGRESS                                          11

// rsvna_rngMarkerBfilterSearchMode
#define RSVNA_VAL_BWID_MODE_NONE                                            -1
#define RSVNA_VAL_BWID_MODE_BPAS                                            0
#define RSVNA_VAL_BWID_MODE_BST                                             1
#define RSVNA_VAL_BWID_MODE_BPRM                                            2
#define RSVNA_VAL_BWID_MODE_BSRM                                            3
#define RSVNA_VAL_BWID_MODE_BPAB                                            4
#define RSVNA_VAL_BWID_MODE_BSAB                                            5

// rsvna_rngTimeDomainStimulusXAxisScaling
#define RSVNA_VAL_DISTANCE                                                  0
#define RSVNA_VAL_TIME                                                      1

// rsvna_rngDrivingMode
#define RSVNA_VAL_DRIVING_MODE_AUTO                                         0
#define RSVNA_VAL_DRIVING_MODE_ALTERNATE                                    1
#define RSVNA_VAL_DRIVING_MODE_CHOPPED                                      2

// rsvna_rngVNetworkSingleEndedCircuitModel
#define RSVNA_VAL_SENDED_CIRCUIT_FIMP                                       0
#define RSVNA_VAL_SENDED_CIRCUIT_CSL                                        1
#define RSVNA_VAL_SENDED_CIRCUIT_LSC                                        2
#define RSVNA_VAL_SENDED_CIRCUIT_CSC                                        3
#define RSVNA_VAL_SENDED_CIRCUIT_LSL                                        4
#define RSVNA_VAL_SENDED_CIRCUIT_SLC                                        5
#define RSVNA_VAL_SENDED_CIRCUIT_SCL                                        6
#define RSVNA_VAL_SENDED_CIRCUIT_SCC                                        7
#define RSVNA_VAL_SENDED_CIRCUIT_SLL                                        8
#define RSVNA_VAL_SENDED_CIRCUIT_SHLC                                       9
#define RSVNA_VAL_SENDED_CIRCUIT_GSL                                        10
#define RSVNA_VAL_SENDED_CIRCUIT_LSG                                        11
#define RSVNA_VAL_SENDED_CIRCUIT_GSG                                        12
#define RSVNA_VAL_SENDED_CIRCUIT_SGL                                        13
#define RSVNA_VAL_SENDED_CIRCUIT_SLG                                        14
#define RSVNA_VAL_SENDED_CIRCUIT_SGG                                        15

// rsvna_rngVNetworkBalancedCircuitModel
#define RSVNA_VAL_BALANCED_CIRCUIT_FIMP                                     0
#define RSVNA_VAL_BALANCED_CIRCUIT_STSL                                     1
#define RSVNA_VAL_BALANCED_CIRCUIT_STSC                                     2
#define RSVNA_VAL_BALANCED_CIRCUIT_SLST                                     3
#define RSVNA_VAL_BALANCED_CIRCUIT_SCST                                     4
#define RSVNA_VAL_BALANCED_CIRCUIT_CSSL                                     5
#define RSVNA_VAL_BALANCED_CIRCUIT_LSSC                                     6
#define RSVNA_VAL_BALANCED_CIRCUIT_CSSC                                     7
#define RSVNA_VAL_BALANCED_CIRCUIT_LSSL                                     8
#define RSVNA_VAL_BALANCED_CIRCUIT_SLCS                                     9
#define RSVNA_VAL_BALANCED_CIRCUIT_SCLS                                     10
#define RSVNA_VAL_BALANCED_CIRCUIT_SCCS                                     11
#define RSVNA_VAL_BALANCED_CIRCUIT_SLLS                                     12
#define RSVNA_VAL_BALANCED_CIRCUIT_STSG                                     13
#define RSVNA_VAL_BALANCED_CIRCUIT_SGST                                     14
#define RSVNA_VAL_BALANCED_CIRCUIT_GSSL                                     15
#define RSVNA_VAL_BALANCED_CIRCUIT_LSSG                                     16
#define RSVNA_VAL_BALANCED_CIRCUIT_GSSG                                     17
#define RSVNA_VAL_BALANCED_CIRCUIT_SLGS                                     18
#define RSVNA_VAL_BALANCED_CIRCUIT_SGLS                                     19
#define RSVNA_VAL_BALANCED_CIRCUIT_SGGS                                     20

// rsvna_rngColorScheme
#define RSVNA_VAL_DBACKGROUND                                               0
#define RSVNA_VAL_LBACKGROUND                                               1
#define RSVNA_VAL_BW_LSTYLES                                                2
#define RSVNA_VAL_BW_SOLID                                                  3
#define RSVNA_VAL_UDEF                                                      4

// rsvna_rngPresetScope
#define RSVNA_VAL_PRESET_SINGLE                                             0
#define RSVNA_VAL_PRESET_ALL                                                1

// rsvna_rngRemoteLanguage
#define RSVNA_VAL_LANG_SCPI                                                 0
#define RSVNA_VAL_LANG_ZVR                                                  1
#define RSVNA_VAL_LANG_ZVABT                                                2
#define RSVNA_VAL_LANG_PNA                                                  3
#define RSVNA_VAL_LANG_HP8510                                               4
#define RSVNA_VAL_LANG_HP8720                                               5
#define RSVNA_VAL_LANG_HP8753                                               6

// rsvna_rngSelectApplication
#define RSVNA_VAL_SELECT_APPL_PROG                                          0

// rsvna_rngPAEType
#define RSVNA_VAL_PAE_MEASUREMENT_TYPE_VOLTAGE                              0
#define RSVNA_VAL_PAE_MEASUREMENT_TYPE_CURRENT                              1
#define RSVNA_VAL_PAE_MEASUREMENT_TYPE_VCURRENT                             2

// rsvna_rngPAEDCInput
#define RSVNA_VAL_PAE_DC_INPUT_DC1                                          1
#define RSVNA_VAL_PAE_DC_INPUT_DC2                                          2
#define RSVNA_VAL_PAE_DC_INPUT_DC3                                          3
#define RSVNA_VAL_PAE_DC_INPUT_DC4                                          4

// rsvna_rngSegmentScaling
#define RSVNA_VAL_SEG_X_AXIS_POINT                                          0
#define RSVNA_VAL_SEG_X_AXIS_FREQ                                           1

// rsvna_rngConversionMode
#define RSVNA_VAL_CONVERSION_MODE_FUNDAMENTAL                               0
#define RSVNA_VAL_CONVERSION_MODE_ARBITRARY                                 1
#define RSVNA_VAL_CONVERSION_MODE_MIXER                                     2
#define RSVNA_VAL_CONVERSION_MODE_INTERMODULATION                           3
#define RSVNA_VAL_CONVERSION_MODE_VMIX                                      4

// rsvna_rngExtGenMode
#define RSVNA_VAL_EXTERNAL_GENERATOR_MODE_OFF                               0
#define RSVNA_VAL_EXTERNAL_GENERATOR_MODE_LOW                               1
#define RSVNA_VAL_EXTERNAL_GENERATOR_MODE_KEEP                              2
#define RSVNA_VAL_EXTERNAL_GENERATOR_MODE_USER                              3

// rsvna_rngPowerSelect
#define RSVNA_VAL_PWR_CAL_POWER_USER                                        0
#define RSVNA_VAL_PWR_CAL_POWER_PORT                                        1

// rsvna_rngTouchLock
#define RSVNA_VAL_TOUCHSCREEN_OFF                                           0
#define RSVNA_VAL_TOUCHSCREEN_DIAGRAM                                       1
#define RSVNA_VAL_TOUCHSCREEN_SCREEN                                        2

// rsvna_rngChannelSweepAverageMode
#define RSVNA_VAL_AVERAGING_AUTO                                            0
#define RSVNA_VAL_AVERAGING_REDUCE_NOISE                                    1
#define RSVNA_VAL_AVERAGING_FLATTEN_NOISE                                   2
#define RSVNA_VAL_AVERAGING_FLATTEN_MOVING_AVG                              3

// rsvna_rngActivatePowerCalibration
#define RSVNA_VAL_POWER_CAL_ACTIVE_NONE                                     0
#define RSVNA_VAL_POWER_CAL_ACTIVE_ALL                                      1

// rsvna_rngMixerBasePort
#define RSVNA_VAL_MIXER_PORT_RF                                             0
#define RSVNA_VAL_MIXER_PORT_LO                                             1
#define RSVNA_VAL_MIXER_PORT_LO1                                            2
#define RSVNA_VAL_MIXER_PORT_LO2                                            3
#define RSVNA_VAL_MIXER_PORT_IF                                             4

// rsvna_rngMixerConversionMode
#define RSVNA_VAL_CONVERSION_DCLOWER                                        0
#define RSVNA_VAL_CONVERSION_DCUPPER                                        1
#define RSVNA_VAL_CONVERSION_UCONVERSION                                    2

// rsvna_rngPolarity
#define RSVNA_VAL_NEGATIVE                                                  0
#define RSVNA_VAL_POSITIVE                                                  1

// rsvna_rngTriggerAUXOutPosition
#define RSVNA_VAL_TRIGGER_AUX_BEFORE                                        0
#define RSVNA_VAL_TRIGGER_AUX_AFTER                                         1

// rsvna_rngTriggerAUXOutInterval
#define RSVNA_VAL_TRIGGER_AUX_PER_SWEEP                                     0
#define RSVNA_VAL_TRIGGER_AUX_PER_POINT                                     1
#define RSVNA_VAL_TRIGGER_AUX_PER_ACQUISITION                               2
#define RSVNA_VAL_TRIGGER_AUX_PER_NONE                                      3
#define RSVNA_VAL_TRIGGER_AUX_PER_SEGM                                      4
#define RSVNA_VAL_TRIGGER_AUX_PER_PPO                                       5
#define RSVNA_VAL_TRIGGER_AUX_PER_PGEN                                      6

// rsvna_rngDisplaySplitType
#define RSVNA_VAL_SPLIT_TYPE_LINEUP                                         0
#define RSVNA_VAL_SPLIT_TYPE_STACK                                          1
#define RSVNA_VAL_SPLIT_TYPE_HORIZONTAL                                     2
#define RSVNA_VAL_SPLIT_TYPE_VERTICAL                                       3
#define RSVNA_VAL_SPLIT_TYPE_GRID                                           4

// rsvna_rngUIPassFailDefaultState
#define RSVNA_VAL_UI_FAIL                                                   0
#define RSVNA_VAL_UI_PASS                                                   1
#define RSVNA_VAL_UI_NOWAIT                                                 2

// rsvna_rngUIPassFailScope
#define RSVNA_VAL_UI_CHANNEL                                                0
#define RSVNA_VAL_UI_GLOBAL                                                 1

// rsvna_rngUIPassFailPolicy
#define RSVNA_VAL_UI_ALL_TESTS                                              0
#define RSVNA_VAL_UI_ALL_MEAS                                               1

// rsvna_rngUIPassFailStatus
#define RSVNA_VAL_UI_NONE                                                   2

// rsvna_rngUISweepEndEvent
#define RSVNA_VAL_UI_SWEEP_END_SWEEP                                        0
#define RSVNA_VAL_UI_SWEEP_END_CHANNEL                                      1
#define RSVNA_VAL_UI_SWEEP_END_GLOBAL                                       2

// rsvna_rngFixtureCompensationOffsetCorrectionType
#define RSVNA_VAL_FIXTURE_COMPENSATION_DIRECT                               0
#define RSVNA_VAL_FIXTURE_COMPENSATION_AUTO                                 1

// rsvna_rngFixtureCompensationOffsetCorrectionAutoType
#define RSVNA_VAL_FIXTURE_COMPEMSATION_AUTO                                 0
#define RSVNA_VAL_FIXTURE_COMPEMSATION_AUTO_LOSS                            1

// rsvna_rngIntermodulationTwoToneSignalSource
#define RSVNA_VAL_TTO_SOURCE_EXTERNAL                                       0
#define RSVNA_VAL_TTO_SOURCE_PORT1                                          1

// rsvna_rngAutomaticGainControlGlobalMode
#define RSVNA_VAL_AGC_MODE_MANUAL                                           3

// rsvna_rngSweepMeasurementDelayInsertionPoint
#define RSVNA_VAL_INSERT_BEFORE_ALL                                         0
#define RSVNA_VAL_INSERT_BEFORE_FIRST                                       1

// rsvna_rngPortConnectorType
#define RSVNA_VAL_PORT_CONNECTOR_TYPE_ALL                                   0
#define RSVNA_VAL_PORT_CONNECTOR_TYPE_SINGLE                                1

// rsvna_rngMixerInputSignal
#define RSVNA_VAL_INPUT_SIGNAL_RF                                           0
#define RSVNA_VAL_INPUT_SIGNAL_LO                                           1

// rsvna_rngUIPortDataDirectionMode
#define RSVNA_VAL_DATA_PORT_INPUT                                           0
#define RSVNA_VAL_DATA_PORT_OUTPUT                                          1

// rsvna_rngSwitchMatrixOptimization
#define RSVNA_VAL_SWITCH_MATRIX_OPTIMIZATION_SPEED                          0
#define RSVNA_VAL_SWITCH_MATRIX_OPTIMIZATION_PRECISION                      1

// rsvna_rngMarkerTraceOutputFormat
#define RSVNA_VAL_TARGET_FORMAT_MLIN                                        0
#define RSVNA_VAL_TARGET_FORMAT_MLOG                                        1
#define RSVNA_VAL_TARGET_FORMAT_PHASE                                       2
#define RSVNA_VAL_TARGET_FORMAT_UPHASE                                      3
#define RSVNA_VAL_TARGET_FORMAT_REAL                                        4
#define RSVNA_VAL_TARGET_FORMAT_IMAG                                        5
#define RSVNA_VAL_TARGET_FORMAT_SWR                                         6
#define RSVNA_VAL_TARGET_FORMAT_DEF                                         7

// rsvna_rngCalibrationAutoAveraging
#define RSVNA_VAL_AVERAGING_MAN                                             1

// rsvna_rngVirtualNetworkGroundLoopDeembeddingCircuitModel
#define RSVNA_VAL_GLOOP_CIRCUIT_FIMP                                        0
#define RSVNA_VAL_GLOOP_CIRCUIT_SL                                          1
#define RSVNA_VAL_GLOOP_CIRCUIT_SC                                          2
#define RSVNA_VAL_GLOOP_CIRCUIT_SG                                          3

// rsvna_rngVirtualNetworkDifferentialMatchEmbeddingCircuitModel
#define RSVNA_VAL_CIRCUIT_FIMP                                              0
#define RSVNA_VAL_CIRCUIT_SHLC                                              1

// rsvna_rngRemoteEncoding
#define RSVNA_VAL_ENCODING_ANSI                                             0
#define RSVNA_VAL_ENCODING_UTF8                                             1
#define RSVNA_VAL_ENCODING_SJIS                                             2

// rsvna_rngStatisticalFormat
#define RSVNA_VAL_STATISTICAL_FORMAT_ZVAB                                   0
#define RSVNA_VAL_STATISTICAL_FORMAT_IMPEDANCE                              1
#define RSVNA_VAL_STATISTICAL_FORMAT_ADMITTANCE                             2

// rsvna_rngStimulusMinMaxDef
#define RSVNA_VAL_STIMULUS_MIN                                              0
#define RSVNA_VAL_STIMULUS_MAX                                              1
#define RSVNA_VAL_STIMULUS_DEF                                              2

// rsvna_rngFixtureModelingISDCouponType
#define RSVNA_VAL_COUPON_SYMM_2X_THRU                                       0
#define RSVNA_VAL_COUPON_1X_OPEN                                            1
#define RSVNA_VAL_COUPON_1X_SHORT                                           2
#define RSVNA_VAL_COUPON_1XOPEN_1XSHORT                                     3

// rsvna_rngFixtureModelingISDDUTType
#define RSVNA_VAL_DUT_TYPE_PASSIVE                                          0
#define RSVNA_VAL_DUT_TYPE_ACTIVE                                           1

// rsvna_rngFixtureModelingISDInsertionLoss
#define RSVNA_VAL_INSERTION_LOSS_LINEAR                                     1
#define RSVNA_VAL_INSERTION_LOSS_NONLINEAR                                  2
#define RSVNA_VAL_INSERTION_LOSS_RESONANT                                   3

// rsvna_rngFixtureModelingISDPortSequence
#define RSVNA_VAL_ISD_PORT_SEQUENCE_1_TO_N_ON_LEFT                          0
#define RSVNA_VAL_ISD_PORT_SEQUENCE_ODD_ON_LEFT                             1
#define RSVNA_VAL_ISD_PORT_SEQUENCE_ALL_ON_LEFT                             2

// rsvna_rngFixtureModelingISDTraceCoupling
#define RSVNA_VAL_ISD_TRACE_COUPLING_NONE                                   0
#define RSVNA_VAL_ISD_TRACE_COUPLING_WEAK                                   1
#define RSVNA_VAL_ISD_TRACE_COUPLING_STRONG                                 2

// rsvna_rngFixtureModelingISDOperationAccuracy
#define RSVNA_VAL_ISD_CALCULATION_FAST                                      0
#define RSVNA_VAL_ISD_CALCULATION_ACCURATE                                  1

// rsvna_rngFixtureModelingSFDPortOrdering
#define RSVNA_VAL_SDF_PORT_ORDERING_1_2_ON_LEFT                             0
#define RSVNA_VAL_SDF_PORT_ORDERING_ODD_ON_LEFT                             1

// rsvna_rngErrorTermParameter
#define RSVNA_VAL_CAL_DATA_DIRECTIVITY                                      0
#define RSVNA_VAL_CAL_DATA_SRCMATCH                                         1
#define RSVNA_VAL_CAL_DATA_REFLTRACK                                        2
#define RSVNA_VAL_CAL_DATA_LOADMATCH                                        3
#define RSVNA_VAL_CAL_DATA_TRANSTRACK                                       4
#define RSVNA_VAL_CAL_DATA_G11                                              5
#define RSVNA_VAL_CAL_DATA_G12                                              6
#define RSVNA_VAL_CAL_DATA_G21                                              7
#define RSVNA_VAL_CAL_DATA_G22                                              8
#define RSVNA_VAL_CAL_DATA_H11                                              9
#define RSVNA_VAL_CAL_DATA_H12                                              10
#define RSVNA_VAL_CAL_DATA_H21                                              11
#define RSVNA_VAL_CAL_DATA_H22                                              12
#define RSVNA_VAL_CAL_DATA_Q11                                              13
#define RSVNA_VAL_CAL_DATA_Q12                                              14
#define RSVNA_VAL_CAL_DATA_Q21                                              15
#define RSVNA_VAL_CAL_DATA_Q22                                              16
#define RSVNA_VAL_CAL_DATA_PREL                                             17
#define RSVNA_VAL_CAL_DATA_PABS                                             18
#define RSVNA_VAL_CAL_DATA_L1                                               19
#define RSVNA_VAL_CAL_DATA_L2                                               20

// rsvna_rngSweepMeasurementArg
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_A1                                  0
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_A2                                  1
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_B1                                  2
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_B2                                  3
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_ABSA1                               4
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_ABSA2                               5
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_DCIN1                               6
#define RSVNA_VAL_SWEEP_MEASUREMENT_ARG_DCIN2                               7

// rsvna_rngCalibrationConnectorKitString
#define RSVNA_VAL_CONNECTOR_KIT_N50                                         0
#define RSVNA_VAL_CONNECTOR_KIT_N75                                         1
#define RSVNA_VAL_CONNECTOR_KIT_PC7                                         2
#define RSVNA_VAL_CONNECTOR_KIT_PC35                                        3
#define RSVNA_VAL_CONNECTOR_KIT_PC292                                       4
#define RSVNA_VAL_CONNECTOR_KIT_USER1                                       5
#define RSVNA_VAL_CONNECTOR_KIT_USER2                                       6
#define RSVNA_VAL_CONNECTOR_KIT_SMA                                         7
#define RSVNA_VAL_CONNECTOR_KIT_PC24                                        8
#define RSVNA_VAL_CONNECTOR_KIT_PC185                                       9
#define RSVNA_VAL_CONNECTOR_KIT_7_16                                        10
#define RSVNA_VAL_CONNECTOR_KIT_TYPE_F_75                                   11
#define RSVNA_VAL_CONNECTOR_KIT_BNC50                                       12
#define RSVNA_VAL_CONNECTOR_KIT_BNC75                                       13
#define RSVNA_VAL_CONNECTOR_KIT_43_10                                       14

// rsvna_rngCalibrationStandardOneOrTwoPorts
#define RSVNA_VAL_CALSTD_MMTH                                               0
#define RSVNA_VAL_CALSTD_FFTH                                               1
#define RSVNA_VAL_CALSTD_MFTH                                               2
#define RSVNA_VAL_CALSTD_MMLI                                               3
#define RSVNA_VAL_CALSTD_FFLI                                               4
#define RSVNA_VAL_CALSTD_MFLI                                               5
#define RSVNA_VAL_CALSTD_MMLI_2                                             6
#define RSVNA_VAL_CALSTD_FFLI_2                                             7
#define RSVNA_VAL_CALSTD_MFLI_2                                             8
#define RSVNA_VAL_CALSTD_MMLI_3                                             9
#define RSVNA_VAL_CALSTD_FFLI_3                                             10
#define RSVNA_VAL_CALSTD_MFLI_3                                             11
#define RSVNA_VAL_CALSTD_MOSH                                               12
#define RSVNA_VAL_CALSTD_FOSH                                               13
#define RSVNA_VAL_CALSTD_MOSH_2                                             14
#define RSVNA_VAL_CALSTD_FOSH_2                                             15
#define RSVNA_VAL_CALSTD_MOSH_3                                             16
#define RSVNA_VAL_CALSTD_FOSH_3                                             17
#define RSVNA_VAL_CALSTD_MMAT                                               18
#define RSVNA_VAL_CALSTD_FFAT                                               19
#define RSVNA_VAL_CALSTD_MFAT                                               20
#define RSVNA_VAL_CALSTD_MMSN                                               21
#define RSVNA_VAL_CALSTD_FFSN                                               22
#define RSVNA_VAL_CALSTD_MFSN                                               23
#define RSVNA_VAL_CALSTD_MOP                                                24
#define RSVNA_VAL_CALSTD_FOP                                                25
#define RSVNA_VAL_CALSTD_MSH                                                26
#define RSVNA_VAL_CALSTD_FSH                                                27
#define RSVNA_VAL_CALSTD_MREF                                               28
#define RSVNA_VAL_CALSTD_FREF                                               29
#define RSVNA_VAL_CALSTD_MMTC                                               30
#define RSVNA_VAL_CALSTD_FMTC                                               31
#define RSVNA_VAL_CALSTD_MSM                                                32
#define RSVNA_VAL_CALSTD_FSM                                                33

// rsvna_rngCalibrationLoadCircuitModel
#define RSVNA_VAL_CAL_CIRCUIT_OPEN                                          0
#define RSVNA_VAL_CAL_CIRCUIT_SHORT                                         1
#define RSVNA_VAL_CAL_CIRCUIT_MATCH                                         2
#define RSVNA_VAL_CAL_CIRCUIT_RESISTANCE                                    3
#define RSVNA_VAL_CAL_CIRCUIT_NONE                                          4

// rsvna_rngCalibrationTypeSimple
#define RSVNA_VAL_CALTYPE_SIMPLE_FRTRANS                                    0
#define RSVNA_VAL_CALTYPE_SIMPLE_PFRTRANS                                   1
#define RSVNA_VAL_CALTYPE_SIMPLE_FTRANS                                     2
#define RSVNA_VAL_CALTYPE_SIMPLE_PFTRANS                                    3
#define RSVNA_VAL_CALTYPE_SIMPLE_RTRANS                                     4
#define RSVNA_VAL_CALTYPE_SIMPLE_PRTRANS                                    5
#define RSVNA_VAL_CALTYPE_SIMPLE_TOM                                        6
#define RSVNA_VAL_CALTYPE_SIMPLE_TSM                                        7
#define RSVNA_VAL_CALTYPE_SIMPLE_TRM                                        8
#define RSVNA_VAL_CALTYPE_SIMPLE_TRL                                        9
#define RSVNA_VAL_CALTYPE_SIMPLE_TNA                                        10
#define RSVNA_VAL_CALTYPE_SIMPLE_TOSM                                       11
#define RSVNA_VAL_CALTYPE_SIMPLE_UOSM                                       12
#define RSVNA_VAL_CALTYPE_SIMPLE_ETOM                                       13
#define RSVNA_VAL_CALTYPE_SIMPLE_ETSM                                       14
#define RSVNA_VAL_CALTYPE_SIMPLE_FOPORT1                                    15
#define RSVNA_VAL_CALTYPE_SIMPLE_FOPORT2                                    16
#define RSVNA_VAL_CALTYPE_SIMPLE_FOPORT12                                   17
#define RSVNA_VAL_CALTYPE_SIMPLE_FOPTPORT                                   18
#define RSVNA_VAL_CALTYPE_SIMPLE_ROPTPORT                                   19
#define RSVNA_VAL_CALTYPE_SIMPLE_REFL1                                      20
#define RSVNA_VAL_CALTYPE_SIMPLE_REFL2                                      21
#define RSVNA_VAL_CALTYPE_SIMPLE_REFL12                                     22
#define RSVNA_VAL_CALTYPE_SIMPLE_TPORT                                      23

// rsvna_rngCalibrationDataParameter
#define RSVNA_VAL_CAL_PARAM_ACAL                                            0
#define RSVNA_VAL_CAL_PARAM_START                                           1
#define RSVNA_VAL_CAL_PARAM_STOP                                            2
#define RSVNA_VAL_CAL_PARAM_POINTS                                          3
#define RSVNA_VAL_CAL_PARAM_POWER                                           4
#define RSVNA_VAL_CAL_PARAM_SWEEP_TYPE                                      5
#define RSVNA_VAL_CAL_PARAM_BANDWIDTH                                       6
#define RSVNA_VAL_CAL_PARAM_POINT_DELAY                                     7
#define RSVNA_VAL_CAL_PARAM_ATTENUATION                                     8
#define RSVNA_VAL_CAL_PARAM_CAL_TYPE                                        9
#define RSVNA_VAL_CAL_PARAM_PORTS                                           10
#define RSVNA_VAL_CAL_PARAM_SPORT                                           11
#define RSVNA_VAL_CAL_PARAM_THROUGH                                         12
#define RSVNA_VAL_CAL_PARAM_TIMESTAMP                                       13
#define RSVNA_VAL_CAL_PARAM_TVNA                                            14
#define RSVNA_VAL_CAL_PARAM_MVNA                                            15
#define RSVNA_VAL_CAL_PARAM_MTEST                                           16
#define RSVNA_VAL_CAL_PARAM_TIMESTAMP_LOCAL                                 17
#define RSVNA_VAL_CAL_PARAM_CALKIT                                          18
#define RSVNA_VAL_CAL_PARAM_SWEEP_MODE                                      19

// rsvna_rngVnetDataInterchange
#define RSVNA_VAL_PORT_NUMBERS_NOT_INTERCHANGED                             0
#define RSVNA_VAL_PORT_NUMBERS_INTERCHANGED                                 1
#define RSVNA_VAL_PORT_NUMBERS_SWAPPED_GATES                                2
#define RSVNA_VAL_PORT_NUMBERS_SWAPPED_INCREASING                           3

// rsvna_rngFreqSweepMode
#define RSVNA_VAL_FREQ_SWEEP_MODE_STEPPED                                   0
#define RSVNA_VAL_FREQ_SWEEP_MODE_ANALOG                                    1

// rsvna_rngSourcePowerCalibrationData
#define RSVNA_VAL_SETTING_TYPE_WAVE                                         0
#define RSVNA_VAL_SETTING_TYPE_START                                        1
#define RSVNA_VAL_SETTING_TYPE_STOP                                         2
#define RSVNA_VAL_SETTING_TYPE_POINTS                                       3
#define RSVNA_VAL_SETTING_TYPE_STYPE                                        4
#define RSVNA_VAL_SETTING_TYPE_ATTENUATION                                  5
#define RSVNA_VAL_SETTING_TYPE_CPOWER                                       6
#define RSVNA_VAL_SETTING_TYPE_CFREQUENCY                                   7
#define RSVNA_VAL_SETTING_TYPE_TIMESTAMP                                    8
#define RSVNA_VAL_SETTING_TYPE_TVNA                                         9
#define RSVNA_VAL_SETTING_TYPE_MVNA                                         10
#define RSVNA_VAL_SETTING_TYPE_MTEST                                        11
#define RSVNA_VAL_SETTING_TYPE_TIMESTAMP_LOCAL                              12
#define RSVNA_VAL_SETTING_TYPE_SWEEP_MODE                                   13

// rsvna_rngDataFormat
#define RSVNA_VAL_DATA_FORMAT_SDATA                                         0
#define RSVNA_VAL_DATA_FORMAT_FDATA                                         1

// rsvna_rngInfiniteAveragingMode
#define RSVNA_VAL_INFINITE_AVERAGING_MAGPHASE                               0
#define RSVNA_VAL_INFINITE_AVERAGING_REIMAG                                 1

// rsvna_rngTDRStimulusThreshold
#define RSVNA_VAL_TDR_THRESHOLD_T1_9                                        0
#define RSVNA_VAL_TDR_THRESHOLD_T2_8                                        1

// rsvna_rngEyeView
#define RSVNA_VAL_EYE_VIEW_STIMULUS                                         0
#define RSVNA_VAL_EYE_VIEW_EMPHASIS                                         1
#define RSVNA_VAL_EYE_VIEW_JITTER                                           2
#define RSVNA_VAL_EYE_VIEW_NOISE                                            3
#define RSVNA_VAL_EYE_VIEW_DUT                                              4
#define RSVNA_VAL_EYE_VIEW_EQUALIZATION                                     5

// rsvna_rngTDREyeBitStream
#define RSVNA_VAL_TDR_BIT_STREAM_PRBS                                       0
#define RSVNA_VAL_TDR_BIT_STREAM_USER                                       1

// rsvna_rngTDRPRBSLength
#define RSVNA_VAL_TDR_PRBS_LENGTH_L5                                        0
#define RSVNA_VAL_TDR_PRBS_LENGTH_L7                                        1
#define RSVNA_VAL_TDR_PRBS_LENGTH_L9                                        2
#define RSVNA_VAL_TDR_PRBS_LENGTH_L10                                       3
#define RSVNA_VAL_TDR_PRBS_LENGTH_L11                                       4
#define RSVNA_VAL_TDR_PRBS_LENGTH_L13                                       5
#define RSVNA_VAL_TDR_PRBS_LENGTH_L15                                       6

// rsvna_rngTDREyeModulation
#define RSVNA_VAL_TDR_MODULATION_NRZ                                        0
#define RSVNA_VAL_TDR_MODULATION_PM4                                        1
#define RSVNA_VAL_TDR_MODULATION_PM8                                        2
#define RSVNA_VAL_TDR_MODULATION_PM16                                       3

// rsvna_rngTDRDUTMode
#define RSVNA_VAL_TDR_DUT_MODE_IDEAL                                        0
#define RSVNA_VAL_TDR_DUT_MODE_MEASURED                                     1

// rsvna_rngPassFail
#define RSVNA_VAL_FAIL                                                      0
#define RSVNA_VAL_PASS                                                      1

// rsvna_rngPolygonShape
#define RSVNA_VAL_POLYGON_SHAPE_OCTOGON                                     0
#define RSVNA_VAL_POLYGON_SHAPE_HEXAGON                                     1
#define RSVNA_VAL_POLYGON_SHAPE_RECTANGLE                                   2

// rsvna_rngFailCondition
#define RSVNA_VAL_FAIL_CONDITION_SAMPLES                                    0
#define RSVNA_VAL_FAIL_CONDITION_RATE                                       1

// rsvna_rngCorrectionWizardDialog
#define RSVNA_VAL_CORRECTION_WIZARD_DIALOG_MAIN                             0
#define RSVNA_VAL_CORRECTION_WIZARD_DIALOG_CKIT                             1

// rsvna_rngLimitLineSegmentInterpolation
#define RSVNA_VAL_LIMIT_LINE_INTERPOLATION_LOG                              0
#define RSVNA_VAL_LIMIT_LINE_INTERPOLATION_LIN                              1

// rsvna_rngMarkerCouplingType
#define RSVNA_VAL_COUPLING_TYPE_ALL                                         0
#define RSVNA_VAL_COUPLING_TYPE_CHANNEL                                     1
#define RSVNA_VAL_COUPLING_TYPE_DIAGRAM                                     2

// rsvna_rngMarkerReferenceFormat
#define RSVNA_VAL_MARKER_REF_FORMAT_DEFAULT                                 0
#define RSVNA_VAL_MARKER_REF_FORMAT_LIN_MAG                                 1
#define RSVNA_VAL_MARKER_REF_FORMAT_DB_MAG                                  2
#define RSVNA_VAL_MARKER_REF_FORMAT_PHASE                                   3
#define RSVNA_VAL_MARKER_REF_FORMAT_REAL_IMAG                               4
#define RSVNA_VAL_MARKER_REF_FORMAT_DELAY                                   5
#define RSVNA_VAL_MARKER_REF_FORMAT_REAL                                    6
#define RSVNA_VAL_MARKER_REF_FORMAT_IMAG                                    7
#define RSVNA_VAL_MARKER_REF_FORMAT_SWR                                     8
#define RSVNA_VAL_MARKER_REF_FORMAT_LIN_MAG_PHASE                           9
#define RSVNA_VAL_MARKER_REF_FORMAT_DB_MAG_PHASE                            10
#define RSVNA_VAL_MARKER_REF_FORMAT_R_JX                                    11
#define RSVNA_VAL_MARKER_REF_FORMAT_G_JB                                    12
#define RSVNA_VAL_MARKER_REF_FORMAT_MIMP                                    13
#define RSVNA_VAL_MARKER_REF_FORMAT_MDB                                     14
#define RSVNA_VAL_MARKER_REF_FORMAT_COMP                                    15
#define RSVNA_VAL_MARKER_REF_FORMAT_MLPH                                    16
#define RSVNA_VAL_MARKER_REF_FORMAT_MDPH                                    17
#define RSVNA_VAL_MARKER_REF_FORMAT_PIMP                                    18
#define RSVNA_VAL_MARKER_REF_FORMAT_PADM                                    19
#define RSVNA_VAL_MARKER_REF_FORMAT_MADM                                    20
#define RSVNA_VAL_MARKER_REF_FORMAT_MPIM                                    21
#define RSVNA_VAL_MARKER_REF_FORMAT_MPAD                                    22
#define RSVNA_VAL_MARKER_REF_FORMAT_INDX                                    23
#define RSVNA_VAL_MARKER_REF_FORMAT_NOIS                                    24

// rsvna_rngPortGroupsParallelMeasurementMinimumOffsetMode
#define RSVNA_VAL_PORT_MIN_OFFSET_MODE_DIRECT                               0
#define RSVNA_VAL_PORT_MIN_OFFSET_MODE_BANDWIDTH                            1

// rsvna_rngPowerCalMethod
#define RSVNA_VAL_POWER_CAL_METHOD_PMONLY                                   0
#define RSVNA_VAL_POWER_CAL_METHOD_RRAFTER                                  1
#define RSVNA_VAL_POWER_CAL_METHOD_RRONLY                                   2

// rsvna_rngPowerSweepEndMode
#define RSVNA_VAL_POWER_SWEEP_END_MODE_AUTO                                 0
#define RSVNA_VAL_POWER_SWEEP_END_MODE_REDUCE                               1
#define RSVNA_VAL_POWER_SWEEP_END_MODE_KEEP                                 2

// rsvna_rngStatisticalCompressionReferenceValue
#define RSVNA_VAL_STAT_COMP_REF_VALUE_FIRST_POINT                           0
#define RSVNA_VAL_STAT_COMP_REF_VALUE_MANUAL                                1
#define RSVNA_VAL_STAT_COMP_REF_VALUE_MARKER                                2
#define RSVNA_VAL_STAT_COMP_REF_VALUE_RANGE                                 3

// rsvna_rngDisplayInfoWindowBandfilter
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_CENTER                            0
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_BANDWIDTH                         1
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_UPPER                             2
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_LOWER                             3
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_LOSS                              4
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_QFAC                              5
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_QFA3                              6
#define RSVNA_VAL_DISPLAY_IWIN_BANDFILTER_HEADER                            7

// rsvna_rngLimitFailType
#define RSVNA_VAL_LIMIT_FAIL_TYPE_LIMIT                                     0
#define RSVNA_VAL_LIMIT_FAIL_TYPE_RIPPLE                                    1
#define RSVNA_VAL_LIMIT_FAIL_TYPE_CIRCLE                                    2

// rsvna_rngDecimalSeparator
#define RSVNA_VAL_DECIMAL_SEPARATOR_POINT                                   0
#define RSVNA_VAL_DECIMAL_SEPARATOR_COMMA                                   1

// rsvna_rngFieldSeparator
#define RSVNA_VAL_FIELD_SEPARATOR_SEMICOLON                                 0
#define RSVNA_VAL_FIELD_SEPARATOR_COMMA                                     1
#define RSVNA_VAL_FIELD_SEPARATOR_TABULATOR                                 2
#define RSVNA_VAL_FIELD_SEPARATOR_SPACE                                     3

// rsvna_rngPrintColorScheme
#define RSVNA_VAL_BW_LSTYLES_OPTIMIZED                                      4
#define RSVNA_VAL_LBACKGROUND_OPTIMIZED                                     5
#define RSVNA_VAL_LBACKGROUND_UDEF                                          6

// rsvna_rngCalDetectorType
#define RSVNA_VAL_CALIBRATION_DETECTOR_TYPE_NORMAL                          0
#define RSVNA_VAL_CALIBRATION_DETECTOR_TYPE_AVERAGE                         1

// rsvna_rngDeltaLMeasurement
#define RSVNA_VAL_DELTA_L_MEASUREMENT_SINGLE_ENDED                          0
#define RSVNA_VAL_DELTA_L_MEASUREMENT_BALANCED                              1

// rsvna_rngDeltaLCharacterisationMethod
#define RSVNA_VAL_DELTA_L_1L                                                0
#define RSVNA_VAL_DELTA_L_2L                                                1
#define RSVNA_VAL_DELTA_L_3L                                                2

// rsvna_rngDeltaLPortOrder
#define RSVNA_VAL_DELTA_L_PORT_ORDER_SEQUENTIAL                             0
#define RSVNA_VAL_DELTA_L_PORT_ORDER_ODD_EVEN                               1

// rsvna_rngDeltaLSweepControl
#define RSVNA_VAL_DELTA_L_SWEEP_DEFAULT                                     0
#define RSVNA_VAL_DELTA_L_SWEEP_CHANNEL                                     1

// rsvna_rngHUMSDataFormat
#define RSVNA_VAL_HUMS_FORMAT_JSON                                          0
#define RSVNA_VAL_HUMS_FORMAT_XML                                           1

// rsvna_rngTouchstoneExportFormat
#define RSVNA_VAL_TOUCHSTONE_FORMAT_L11                                     0
#define RSVNA_VAL_TOUCHSTONE_FORMAT_F11                                     1
#define RSVNA_VAL_TOUCHSTONE_FORMAT_F20                                     2

// rsvna_rngPowerDeEmbType
#define RSVNA_VAL_POWER_EMBEDDING                                           0
#define RSVNA_VAL_POWER_DEEMBEDDING                                         1

// rsvna_rngAGCMethod
#define RSVNA_VAL_AGC_METHOD_MEMORY                                         0
#define RSVNA_VAL_AGC_METHOD_POINT                                          1

// rsvna_rngSNMPAccess
#define RSVNA_VAL_SNMP_ACCESS_READ_ONLY                                     0
#define RSVNA_VAL_SNMP_ACCESS_READ_WRITE                                    1

// rsvna_rngSNMPSecurityLevel
#define RSVNA_VAL_SNMP_SECURITY_NONE                                        0
#define RSVNA_VAL_SNMP_SECURITY_AUTH                                        1
#define RSVNA_VAL_SNMP_SECURITY_PRIVACY                                     2

// rsvna_rngCalibrationGeneralStandard
#define RSVNA_VAL_CALCOLLSTDGEN_THR                                         0
#define RSVNA_VAL_CALCOLLSTDGEN_OPEN                                        1
#define RSVNA_VAL_CALCOLLSTDGEN_SHORT                                       2
#define RSVNA_VAL_CALCOLLSTDGEN_MATCH                                       3
#define RSVNA_VAL_CALCOLLSTDGEN_NET                                         4
#define RSVNA_VAL_CALCOLLSTDGEN_ATT                                         5
#define RSVNA_VAL_CALCOLLSTDGEN_REFL                                        6
#define RSVNA_VAL_CALCOLLSTDGEN_LINE                                        7
#define RSVNA_VAL_CALCOLLSTDGEN_LINE_2                                      8
#define RSVNA_VAL_CALCOLLSTDGEN_LINE_3                                      9
#define RSVNA_VAL_CALCOLLSTDGEN_SLID                                        10
#define RSVNA_VAL_CALCOLLSTDGEN_OSH                                         11
#define RSVNA_VAL_CALCOLLSTDGEN_OSH_2                                       12
#define RSVNA_VAL_CALCOLLSTDGEN_OSH_3                                       13
#define RSVNA_VAL_CALCOLLSTDGEN_POWER                                       14
#define RSVNA_VAL_CALCOLLSTDGEN_ISOLATION                                   15
#define RSVNA_VAL_CALCOLLSTDGEN_UTHROUGH                                    16
#define RSVNA_VAL_CALCOLLSTDGEN_MIX                                         17
#define RSVNA_VAL_CALCOLLSTDGEN_TERM                                        18
#define RSVNA_VAL_CALCOLLSTDGEN_EATT                                        19
#define RSVNA_VAL_CALCOLLSTDGEN_MDEL                                        20

// rsvna_rngPowerDeEmbNwType
#define RSVNA_VAL_POWER_VNET_OFF                                            0
#define RSVNA_VAL_POWER_VNET_ALL                                            1
#define RSVNA_VAL_POWER_VNET_SINGLE_ENDED                                   2
#define RSVNA_VAL_POWER_VNET_GROUND_LOOP                                    3
#define RSVNA_VAL_POWER_VNET_DIFFERENTIAL_MATCH                             4
#define RSVNA_VAL_POWER_VNET_PORT_SET                                       5
#define RSVNA_VAL_POWER_VNET_RENORMALIZATION                                6
#define RSVNA_VAL_POWER_VNET_OFFSET                                         7
#define RSVNA_VAL_POWER_VNET_ADJUSTED_OFFSET                                8

// rsvna_rngPipeliningState
#define RSVNA_VAL_PIPELINING_OFF                                            0
#define RSVNA_VAL_PIPELINING_CHANNEL                                        1

// rsvna_rngHUMSSNMPVersion
#define RSVNA_VAL_HUMPS_SNMP_OFF                                            0
#define RSVNA_VAL_HUMPS_SNMP_V12                                            1
#define RSVNA_VAL_HUMPS_SNMP_V123                                           2
#define RSVNA_VAL_HUMPS_SNMP_V3                                             3

// rsvna_rngHUMSDeviceData
#define RSVNA_VAL_HUMS_DATA_HISTORY                                         0
#define RSVNA_VAL_HUMS_DATA_UTILIZATION                                     1
#define RSVNA_VAL_HUMS_DATA_UTILIZATION_HISTORY                             2
#define RSVNA_VAL_HUMS_DATA_FOOTPRINT                                       3
#define RSVNA_VAL_HUMS_DATA_SECURITY                                        4
#define RSVNA_VAL_HUMS_DATA_SERVICE                                         5
#define RSVNA_VAL_HUMS_DATA_BIOS                                            6
#define RSVNA_VAL_HUMS_DATA_STORAGE                                         7
#define RSVNA_VAL_HUMS_DATA_SOFTWARE                                        8
#define RSVNA_VAL_HUMS_DATA_SYSTEM_INFO                                     9
#define RSVNA_VAL_HUMS_DATA_SYSTEM_STATUS                                   10
#define RSVNA_VAL_HUMS_DATA_REST_API_ENDPOINTS                              11

// rsvna_rngHUMSDeviceStatus
#define RSVNA_VAL_HUMS_DEVICE_STATUS_ERROR                                  -1
#define RSVNA_VAL_HUMS_DEVICE_STATUS_OK                                     0
#define RSVNA_VAL_HUMS_DEVICE_STATUS_WARNING                                1

// rsvna_rngServiceCalibrationDueState
#define RSVNA_VAL_CALIBRATION_DUE_OK                                        0
#define RSVNA_VAL_CALIBRATION_DUE_NAN                                       1
#define RSVNA_VAL_CALIBRATION_DUE_EXPIRED                                   2

// rsvna_rngServiceDueState
#define RSVNA_VAL_SERVICE_STATE_OK                                          0
#define RSVNA_VAL_SERVICE_STATE_NAN                                         1
#define RSVNA_VAL_SERVICE_STATE_DEVIATIONS                                  2
#define RSVNA_VAL_SERVICE_STATE_REQUIRED                                    3

// rsvna_rngDeltaLFrequencyList
#define RSVNA_VAL_DELTA_L_FREQUENCIES_CURRENT                               0
#define RSVNA_VAL_DELTA_L_FREQUENCIES_DEFAULT                               1
#define RSVNA_VAL_DELTA_L_FREQUENCIES_USER                                  2

// rsvna_rngInterpolationMethod
#define RSVNA_VAL_INTERPOLATION_METHOD_LINEAR                               0
#define RSVNA_VAL_INTERPOLATION_METHOD_HORDER                               1

// rsvna_rngSMAInputFreq
#define RSVNA_VAL_SMA_INPUT_FREQ_F100MHZ                                    0
#define RSVNA_VAL_SMA_INPUT_FREQ_F1GHZ                                      1

// rsvna_rngSMAOutputFreq
#define RSVNA_VAL_SMA_OUTPUT_FREQ_OFF                                       0
#define RSVNA_VAL_SMA_OUTPUT_FREQ_F100MHZ                                   1

// rsvna_rngIFSignalPath
#define RSVNA_VAL_IF_SIGNAL_PATH_WIDEBAND                                   0
#define RSVNA_VAL_IF_SIGNAL_PATH_NORMAL                                     1
#define RSVNA_VAL_IF_SIGNAL_PATH_NARROWBAND                                 2

// rsvna_rngPhaseMode
#define RSVNA_VAL_PHASE_MODE_NCOHERENT                                      0
#define RSVNA_VAL_PHASE_MODE_COHERENT                                       1
#define RSVNA_VAL_PHASE_MODE_LNNCOHERENT                                    2
#define RSVNA_VAL_PHASE_MODE_LNCOHERENT                                     3

// rsvna_rngFastPowerSweep
#define RSVNA_VAL_FAST_POWER_SWEEP_OFF                                      0
#define RSVNA_VAL_FAST_POWER_SWEEP_ON                                       1
#define RSVNA_VAL_FAST_POWER_SWEEP_FAST                                     2

// rsvna_rngSweepTimerMode
#define RSVNA_VAL_SWEEP_TIMER_MODE_AUTO                                     0
#define RSVNA_VAL_SWEEP_TIMER_MODE_MONO                                     1
#define RSVNA_VAL_SWEEP_TIMER_MODE_CONV                                     2
#define RSVNA_VAL_SWEEP_TIMER_MODE_SLOW                                     3

// rsvna_rngLedColor
#define RSVNA_VAL_LED_COLOR_OFF                                             0
#define RSVNA_VAL_LED_COLOR_GREEN                                           1
#define RSVNA_VAL_LED_COLOR_RED                                             2
#define RSVNA_VAL_LED_COLOR_FGREEN                                          3
#define RSVNA_VAL_LED_COLOR_FRED                                            4

// rsvna_rngRFPowerOff
#define RSVNA_VAL_RF_POWER_OFF_FAST                                         0
#define RSVNA_VAL_RF_POWER_OFF_LBNOISE                                      1

// rsvna_rngReceiverInput
#define RSVNA_VAL_RECEIVER_INPUT_NONE                                       0
#define RSVNA_VAL_RECEIVER_INPUT_B16                                        1
#define RSVNA_VAL_RECEIVER_INPUT_REV                                        2
#define RSVNA_VAL_RECEIVER_INPUT_SRC                                        3

// rsvna_rngALCOptimize
#define RSVNA_VAL_ALC_OPTIMIZE_AUTO                                         0
#define RSVNA_VAL_ALC_OPTIMIZE_ON                                           1
#define RSVNA_VAL_ALC_OPTIMIZE_OFF                                          2

// rsvna_rngALCPI
#define RSVNA_VAL_ALC_PI_AUTO                                               0
#define RSVNA_VAL_ALC_PI_MANUAL                                             1

// rsvna_rngALCNetworkusage
#define RSVNA_VAL_ALC_NETWORK_USAGE_OFFSET                                  0
#define RSVNA_VAL_ALC_NETWORK_USAGE_DEEMBED                                 1
#define RSVNA_VAL_ALC_NETWORK_USAGE_OFF                                     2

// rsvna_rngAttMode
#define RSVNA_VAL_ATT_MODE_AUTO                                             0
#define RSVNA_VAL_ATT_MODE_MANUAL                                           1

// rsvna_rngRearLOSweepType
#define RSVNA_VAL_REAR_LO_SWEEP_TYPE_FB                                     0
#define RSVNA_VAL_REAR_LO_SWEEP_TYPE_CW                                     1

// rsvna_rngPresetMode
#define RSVNA_VAL_PRESET_MODE_NORMAL                                        0
#define RSVNA_VAL_PRESET_MODE_FILE                                          1
#define RSVNA_VAL_PRESET_MODE_GLOBAL                                        2
#define RSVNA_VAL_PRESET_MODE_RFOFF                                         3

// rsvna_rngTriggerAUXSignalType
#define RSVNA_VAL_TRIGGER_AUX_SIGNAL_TYPE_POSITIVE                          0
#define RSVNA_VAL_TRIGGER_AUX_SIGNAL_TYPE_NEGATIVE                          1
#define RSVNA_VAL_TRIGGER_AUX_SIGNAL_TYPE_AHIGH                             2
#define RSVNA_VAL_TRIGGER_AUX_SIGNAL_TYPE_ALOW                              3

// rsvna_rngTriggerExtInput
#define RSVNA_VAL_TRIGGER_EXT_INPUT_EXTA                                    0
#define RSVNA_VAL_TRIGGER_EXT_INPUT_EXTB                                    1
#define RSVNA_VAL_TRIGGER_EXT_INPUT_EXTC                                    2
#define RSVNA_VAL_TRIGGER_EXT_INPUT_EXTD                                    3
#define RSVNA_VAL_TRIGGER_EXT_INPUT_UPORT2                                  4

// rsvna_rngTriggerLogicSeq
#define RSVNA_VAL_TRIGGER_LOGIC_SEQ_SWEEP                                   0
#define RSVNA_VAL_TRIGGER_LOGIC_SEQ_SEGMENT                                 1
#define RSVNA_VAL_TRIGGER_LOGIC_SEQ_POINT                                   2
#define RSVNA_VAL_TRIGGER_LOGIC_SEQ_PPOINT                                  3

// rsvna_rngTriggerLogic
#define RSVNA_VAL_TRIGGER_LOGIC_NONE                                        0
#define RSVNA_VAL_TRIGGER_LOGIC_AND                                         1
#define RSVNA_VAL_TRIGGER_LOGIC_OR                                          2

// rsvna_rngLinearityDeviationMode
#define RSVNA_VAL_LDEV_MODE_DISCARD                                         0
#define RSVNA_VAL_LDEV_MODE_APPLY                                           1
#define RSVNA_VAL_LDEV_MODE_REPEAT                                          2

// rsvna_rngGroupDelayReceiverMode
#define RSVNA_VAL_GROUP_DELAY_RECEIVER_MODE_INT                             0
#define RSVNA_VAL_GROUP_DELAY_RECEIVER_MODE_EXT                             1

// rsvna_rngUpperToneSource
#define RSVNA_VAL_UPPER_TONE_SOURCE_NONE                                    0
#define RSVNA_VAL_UPPER_TONE_SOURCE_PORT                                    1
#define RSVNA_VAL_UPPER_TONE_SOURCE_GENERATOR                               2

// rsvna_rngSplitterLOPort
#define RSVNA_VAL_SPLITTER_LO_PORT_NONE                                     0
#define RSVNA_VAL_SPLITTER_LO_PORT_RLO                                      1
#define RSVNA_VAL_SPLITTER_LO_PORT_PORT                                     2
#define RSVNA_VAL_SPLITTER_LO_PORT_GEN                                      3

// rsvna_rngConverterIFInput
#define RSVNA_VAL_CONVERTER_IF_INPUT_DRAC                                   0
#define RSVNA_VAL_CONVERTER_IF_INPUT_REAR                                   1

// rsvna_rngPulseGeneratorSource
#define RSVNA_VAL_PULSE_GENERATOR_SOURCE_EXTA                               0
#define RSVNA_VAL_PULSE_GENERATOR_SOURCE_EXTB                               1
#define RSVNA_VAL_PULSE_GENERATOR_SOURCE_EXTC                               2
#define RSVNA_VAL_PULSE_GENERATOR_SOURCE_EXTD                               3
#define RSVNA_VAL_PULSE_GENERATOR_SOURCE_INT                                4

// rsvna_rngNoiseFigureSweeps
#define RSVNA_VAL_NOISE_FIGURE_SWEEPS_RECEIVER                              0
#define RSVNA_VAL_NOISE_FIGURE_SWEEPS_SOURCE                                1
#define RSVNA_VAL_NOISE_FIGURE_SWEEPS_ATTENUATOR                            2

// rsvna_rngTransferPowerControlMode
#define RSVNA_VAL_TRANSFER_POWER_CONTROL_MODE_NONE                          0
#define RSVNA_VAL_TRANSFER_POWER_CONTROL_MODE_DSET                          1

// rsvna_rngLevelingDatasetMode
#define RSVNA_VAL_LEVELING_DATASET_MODE_FACT                                0
#define RSVNA_VAL_LEVELING_DATASET_MODE_USER                                1
#define RSVNA_VAL_LEVELING_DATASET_MODE_GENERIC                             2

// rsvna_rngCombiner
#define RSVNA_VAL_COMBINER_INT                                              0
#define RSVNA_VAL_COMBINER_EXT                                              1
#define RSVNA_VAL_COMBINER_B16                                              2
#define RSVNA_VAL_COMBINER_COUP                                             3

// rsvna_rngLOSequenceMode
#define RSVNA_VAL_LO_SEQUENCE_MODE_ALTERNATED                               0
#define RSVNA_VAL_LO_SEQUENCE_MODE_CHOPPED                                  1

// rsvna_rngLOUsage
#define RSVNA_VAL_LO_USAGE_OPTIMIZED                                        0
#define RSVNA_VAL_LO_USAGE_DEFAULT                                          1

// rsvna_rngSpectrumDetector
#define RSVNA_VAL_SPECTRUM_DETECTOR_MAXP                                    0
#define RSVNA_VAL_SPECTRUM_DETECTOR_MINP                                    1
#define RSVNA_VAL_SPECTRUM_DETECTOR_RMS                                     2
#define RSVNA_VAL_SPECTRUM_DETECTOR_AVER                                    3

// rsvna_rngSpectrumImageRejection
#define RSVNA_VAL_SPECTRUM_IMAGE_REJECTION_OFF                              0
#define RSVNA_VAL_SPECTRUM_IMAGE_REJECTION_LOW                              1
#define RSVNA_VAL_SPECTRUM_IMAGE_REJECTION_NORMAL                           2
#define RSVNA_VAL_SPECTRUM_IMAGE_REJECTION_HIGH                             3

// rsvna_rngFrequencyConversionReceiverFrequency
#define RSVNA_VAL_FREQ_CONVERSION_RECEIVER_FREQ_OFF                         0
#define RSVNA_VAL_FREQ_CONVERSION_RECEIVER_FREQ_ON                          1
#define RSVNA_VAL_FREQ_CONVERSION_RECEIVER_FREQ_SRC                         2

// rsvna_rngDirectReceiverInput
#define RSVNA_VAL_DIRECT_RECEIVER_INPUT_NONE                                0
#define RSVNA_VAL_DIRECT_RECEIVER_INPUT_B16                                 1
#define RSVNA_VAL_DIRECT_RECEIVER_INPUT_REV                                 2
#define RSVNA_VAL_DIRECT_RECEIVER_INPUT_SRC                                 3

// rsvna_rngDirectRearInputPath
#define RSVNA_VAL_DIRECT_REAR_INPUT_PATH_DIRECT                             0
#define RSVNA_VAL_DIRECT_REAR_INPUT_PATH_LP60                               1
#define RSVNA_VAL_DIRECT_REAR_INPUT_PATH_BP279                              2

// rsvna_rngDirectRearOut
#define RSVNA_VAL_DIRECT_REAR_OUT_OFF                                       0
#define RSVNA_VAL_DIRECT_REAR_OUT_IN                                        1
#define RSVNA_VAL_DIRECT_REAR_OUT_OUT                                       2

// rsvna_rngLOPort
#define RSVNA_VAL_LO_PORT_NONE                                              0
#define RSVNA_VAL_LO_PORT_RLO                                               1
#define RSVNA_VAL_LO_PORT_PORT                                              2
#define RSVNA_VAL_LO_PORT_GEN                                               3
#define RSVNA_VAL_LO_PORT_SPL                                               4

// rsvna_rngDeembedAssistantTool
#define RSVNA_VAL_DEEMBED_ASSISTANT_TOOL_ISD                                0
#define RSVNA_VAL_DEEMBED_ASSISTANT_TOOL_SFD                                1
#define RSVNA_VAL_DEEMBED_ASSISTANT_TOOL_EZD                                2

// rsvna_rngDeembedAssistantCouponType
#define RSVNA_VAL_DEEMBED_ASSISTANT_COUPON_TYPE_SYMMETRIC2X                 0
#define RSVNA_VAL_DEEMBED_ASSISTANT_COUPON_TYPE_OPEN1X                      1
#define RSVNA_VAL_DEEMBED_ASSISTANT_COUPON_TYPE_SHORT1X                     2
#define RSVNA_VAL_DEEMBED_ASSISTANT_COUPON_TYPE_OPENSHORT                   3

// rsvna_rngDeembedAssistantDUTType
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_SEX                            0
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_SEXS                           1
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_BX                             2
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_BXB                            3
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_BXSE                           4
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_SE21                           5
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_SE22                           6
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_NXSE                           7
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_NXMS                           8
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_NXB                            9
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_TYPE_NXMB                           10

// rsvna_rngDeembedAssistantPosition
#define RSVNA_VAL_DEEMBED_ASSISTANT_POSITION_LEFT                           0
#define RSVNA_VAL_DEEMBED_ASSISTANT_POSITION_RIGHT                          1

// rsvna_rngDeembedAssistantDUTPosition
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_POSITION_UPLEFT                     0
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_POSITION_UPRIGHT                    1
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_POSITION_LOWLEFT                    2
#define RSVNA_VAL_DEEMBED_ASSISTANT_DUT_POSITION_LOWRIGHT                   3

// rsvna_rngEZDPortOrder
#define RSVNA_VAL_EZD_PORT_ORDER_ODD                                        0
#define RSVNA_VAL_EZD_PORT_ORDER_NON                                        1

// rsvna_rngHUMSUtilizationUpdateValue
#define RSVNA_VAL_HUMS_UTILIZATION_UPDATE_VALUE_ABS                         0
#define RSVNA_VAL_HUMS_UTILIZATION_UPDATE_VALUE_INC                         1

// rsvna_rngHUMSSeverity
#define RSVNA_VAL_HUMS_SEVERITY_INFO                                        0
#define RSVNA_VAL_HUMS_SEVERITY_WARNING                                     1
#define RSVNA_VAL_HUMS_SEVERITY_ERROR                                       2

// rsvna_rngDeembedModel
#define RSVNA_VAL_DEEMBED_MODEL_NONE                                        0
#define RSVNA_VAL_DEEMBED_MODEL_SINGLE_ENDED                                1
#define RSVNA_VAL_DEEMBED_MODEL_BALANCED                                    2
#define RSVNA_VAL_DEEMBED_MODEL_COUPLED                                     3
#define RSVNA_VAL_DEEMBED_MODEL_UNCOUPLED                                   4

// rsvna_rngCalLocation
#define RSVNA_VAL_CAL_LOCATION_CHANNEL                                      0
#define RSVNA_VAL_CAL_LOCATION_POOL                                         1

// rsvna_rngDeEmbeddingPosition
#define RSVNA_VAL_DE_EMBEDDING_POSITION_DUT                                 0
#define RSVNA_VAL_DE_EMBEDDING_POSITION_PM                                  1
#define RSVNA_VAL_DE_EMBEDDING_POSITION_BOTH                                2

// rsvna_rngSweepTimeMode
#define RSVNA_VAL_SWEEP_TIME_MODE_AUTO                                      0
#define RSVNA_VAL_SWEEP_TIME_MODE_USER                                      1
#define RSVNA_VAL_SWEEP_TIME_MODE_LIMITED                                   2

// rsvna_rngCalibrationValidationFormat
#define RSVNA_VAL_CALIBRATION_VALIDATION_FORMAT_MLIN                        0
#define RSVNA_VAL_CALIBRATION_VALIDATION_FORMAT_REAL                        1

// rsvna_rngTDREyeTransferRateUnits
#define RSVNA_VAL_TDR_EYE_TRANSFER_RATE_UNITS_XPS                           0
#define RSVNA_VAL_TDR_EYE_TRANSFER_RATE_UNITS_KXPS                          1
#define RSVNA_VAL_TDR_EYE_TRANSFER_RATE_UNITS_MXPS                          2
#define RSVNA_VAL_TDR_EYE_TRANSFER_RATE_UNITS_GXPS                          3

// rsvna_rngPremeasurementMode
#define RSVNA_VAL_PREMEASUREMENT_MODE_ONCE                                  0
#define RSVNA_VAL_PREMEASUREMENT_MODE_PSW                                   1

// rsvna_rngPremeasurementDetector
#define RSVNA_VAL_PREMEASUREMENT_DETECTOR_SAMPLE                            0
#define RSVNA_VAL_PREMEASUREMENT_DETECTOR_AVER                              1
#define RSVNA_VAL_PREMEASUREMENT_DETECTOR_RMS                               2
#define RSVNA_VAL_PREMEASUREMENT_DETECTOR_PEAK                              3

// rsvna_rngSweepGenerationMode
#define RSVNA_VAL_SWEEP_GENERATION_MODE_STEPPED                             0

// rsvna_rngFormatIdentify
#define RSVNA_VAL_FORMAT_IDENTIFY_LEGACY                                    0
#define RSVNA_VAL_FORMAT_IDENTIFY_NEW                                       1

// rsvna_rngCalibrationConverterOffsetType
#define RSVNA_VAL_CALIBRATION_CONVERTER_OFFSET_TYPE_ONLY                    0
#define RSVNA_VAL_CALIBRATION_CONVERTER_OFFSET_TYPE_CPAD                    1

//repcap Channel
#define RSVNA_REPCAP_CHANNEL_CHANNEL1                                       1
#define RSVNA_REPCAP_CHANNEL_CHANNEL2                                       2
#define RSVNA_REPCAP_CHANNEL_CHANNEL3                                       3
#define RSVNA_REPCAP_CHANNEL_CHANNEL4                                       4

//repcap Port
#define RSVNA_REPCAP_PORT_PT1                                               1
#define RSVNA_REPCAP_PORT_PT2                                               2
#define RSVNA_REPCAP_PORT_PT3                                               3
#define RSVNA_REPCAP_PORT_PT4                                               4
#define RSVNA_REPCAP_PORT_PT5                                               5
#define RSVNA_REPCAP_PORT_PT6                                               6
#define RSVNA_REPCAP_PORT_PT7                                               7
#define RSVNA_REPCAP_PORT_PT8                                               8
#define RSVNA_REPCAP_PORT_PT9                                               9
#define RSVNA_REPCAP_PORT_PT10                                              10
#define RSVNA_REPCAP_PORT_PT11                                              11
#define RSVNA_REPCAP_PORT_PT12                                              12
#define RSVNA_REPCAP_PORT_PT13                                              13
#define RSVNA_REPCAP_PORT_PT14                                              14
#define RSVNA_REPCAP_PORT_PT15                                              15
#define RSVNA_REPCAP_PORT_PT16                                              16
#define RSVNA_REPCAP_PORT_PT17                                              17
#define RSVNA_REPCAP_PORT_PT18                                              18
#define RSVNA_REPCAP_PORT_PT19                                              19
#define RSVNA_REPCAP_PORT_PT20                                              20
#define RSVNA_REPCAP_PORT_PT21                                              21
#define RSVNA_REPCAP_PORT_PT22                                              22
#define RSVNA_REPCAP_PORT_PT23                                              23
#define RSVNA_REPCAP_PORT_PT24                                              24
#define RSVNA_REPCAP_PORT_PT25                                              25
#define RSVNA_REPCAP_PORT_PT26                                              26
#define RSVNA_REPCAP_PORT_PT27                                              27
#define RSVNA_REPCAP_PORT_PT28                                              28
#define RSVNA_REPCAP_PORT_PT29                                              29
#define RSVNA_REPCAP_PORT_PT30                                              30
#define RSVNA_REPCAP_PORT_PT31                                              31
#define RSVNA_REPCAP_PORT_PT32                                              32
#define RSVNA_REPCAP_PORT_PT33                                              33
#define RSVNA_REPCAP_PORT_PT34                                              34
#define RSVNA_REPCAP_PORT_PT35                                              35
#define RSVNA_REPCAP_PORT_PT36                                              36
#define RSVNA_REPCAP_PORT_PT37                                              37
#define RSVNA_REPCAP_PORT_PT38                                              38
#define RSVNA_REPCAP_PORT_PT39                                              39
#define RSVNA_REPCAP_PORT_PT40                                              40
#define RSVNA_REPCAP_PORT_PT41                                              41
#define RSVNA_REPCAP_PORT_PT42                                              42
#define RSVNA_REPCAP_PORT_PT43                                              43
#define RSVNA_REPCAP_PORT_PT44                                              44
#define RSVNA_REPCAP_PORT_PT45                                              45
#define RSVNA_REPCAP_PORT_PT46                                              46
#define RSVNA_REPCAP_PORT_PT47                                              47
#define RSVNA_REPCAP_PORT_PT48                                              48
#define RSVNA_REPCAP_PORT_PT49                                              49
#define RSVNA_REPCAP_PORT_PT50                                              50
#define RSVNA_REPCAP_PORT_PT51                                              51
#define RSVNA_REPCAP_PORT_PT52                                              52
#define RSVNA_REPCAP_PORT_PT53                                              53
#define RSVNA_REPCAP_PORT_PT54                                              54
#define RSVNA_REPCAP_PORT_PT55                                              55
#define RSVNA_REPCAP_PORT_PT56                                              56
#define RSVNA_REPCAP_PORT_PT57                                              57
#define RSVNA_REPCAP_PORT_PT58                                              58
#define RSVNA_REPCAP_PORT_PT59                                              59
#define RSVNA_REPCAP_PORT_PT60                                              60
#define RSVNA_REPCAP_PORT_PT61                                              61
#define RSVNA_REPCAP_PORT_PT62                                              62
#define RSVNA_REPCAP_PORT_PT63                                              63
#define RSVNA_REPCAP_PORT_PT64                                              64
#define RSVNA_REPCAP_PORT_PT65                                              65
#define RSVNA_REPCAP_PORT_PT66                                              66
#define RSVNA_REPCAP_PORT_PT67                                              67
#define RSVNA_REPCAP_PORT_PT68                                              68
#define RSVNA_REPCAP_PORT_PT69                                              69
#define RSVNA_REPCAP_PORT_PT70                                              70
#define RSVNA_REPCAP_PORT_PT71                                              71
#define RSVNA_REPCAP_PORT_PT72                                              72
#define RSVNA_REPCAP_PORT_PT73                                              73
#define RSVNA_REPCAP_PORT_PT74                                              74
#define RSVNA_REPCAP_PORT_PT75                                              75
#define RSVNA_REPCAP_PORT_PT76                                              76
#define RSVNA_REPCAP_PORT_PT77                                              77
#define RSVNA_REPCAP_PORT_PT78                                              78
#define RSVNA_REPCAP_PORT_PT79                                              79
#define RSVNA_REPCAP_PORT_PT80                                              80
#define RSVNA_REPCAP_PORT_PT81                                              81
#define RSVNA_REPCAP_PORT_PT82                                              82
#define RSVNA_REPCAP_PORT_PT83                                              83
#define RSVNA_REPCAP_PORT_PT84                                              84
#define RSVNA_REPCAP_PORT_PT85                                              85
#define RSVNA_REPCAP_PORT_PT86                                              86
#define RSVNA_REPCAP_PORT_PT87                                              87
#define RSVNA_REPCAP_PORT_PT88                                              88
#define RSVNA_REPCAP_PORT_PT89                                              89
#define RSVNA_REPCAP_PORT_PT90                                              90
#define RSVNA_REPCAP_PORT_PT91                                              91
#define RSVNA_REPCAP_PORT_PT92                                              92
#define RSVNA_REPCAP_PORT_PT93                                              93
#define RSVNA_REPCAP_PORT_PT94                                              94
#define RSVNA_REPCAP_PORT_PT95                                              95
#define RSVNA_REPCAP_PORT_PT96                                              96
#define RSVNA_REPCAP_PORT_PT97                                              97
#define RSVNA_REPCAP_PORT_PT98                                              98
#define RSVNA_REPCAP_PORT_PT99                                              99
#define RSVNA_REPCAP_PORT_PT100                                             100
#define RSVNA_REPCAP_PORT_PT101                                             101
#define RSVNA_REPCAP_PORT_PT102                                             102
#define RSVNA_REPCAP_PORT_PT103                                             103
#define RSVNA_REPCAP_PORT_PT104                                             104
#define RSVNA_REPCAP_PORT_PT105                                             105
#define RSVNA_REPCAP_PORT_PT106                                             106
#define RSVNA_REPCAP_PORT_PT107                                             107
#define RSVNA_REPCAP_PORT_PT108                                             108
#define RSVNA_REPCAP_PORT_PT109                                             109
#define RSVNA_REPCAP_PORT_PT110                                             110
#define RSVNA_REPCAP_PORT_PT111                                             111
#define RSVNA_REPCAP_PORT_PT112                                             112
#define RSVNA_REPCAP_PORT_PT113                                             113
#define RSVNA_REPCAP_PORT_PT114                                             114
#define RSVNA_REPCAP_PORT_PT115                                             115
#define RSVNA_REPCAP_PORT_PT116                                             116
#define RSVNA_REPCAP_PORT_PT117                                             117
#define RSVNA_REPCAP_PORT_PT118                                             118
#define RSVNA_REPCAP_PORT_PT119                                             119
#define RSVNA_REPCAP_PORT_PT120                                             120
#define RSVNA_REPCAP_PORT_PT121                                             121
#define RSVNA_REPCAP_PORT_PT122                                             122
#define RSVNA_REPCAP_PORT_PT123                                             123
#define RSVNA_REPCAP_PORT_PT124                                             124
#define RSVNA_REPCAP_PORT_PT125                                             125
#define RSVNA_REPCAP_PORT_PT126                                             126
#define RSVNA_REPCAP_PORT_PT127                                             127
#define RSVNA_REPCAP_PORT_PT128                                             128
#define RSVNA_REPCAP_PORT_PT129                                             129
#define RSVNA_REPCAP_PORT_PT130                                             130
#define RSVNA_REPCAP_PORT_PT131                                             131
#define RSVNA_REPCAP_PORT_PT132                                             132
#define RSVNA_REPCAP_PORT_PT133                                             133
#define RSVNA_REPCAP_PORT_PT134                                             134
#define RSVNA_REPCAP_PORT_PT135                                             135
#define RSVNA_REPCAP_PORT_PT136                                             136
#define RSVNA_REPCAP_PORT_PT137                                             137
#define RSVNA_REPCAP_PORT_PT138                                             138
#define RSVNA_REPCAP_PORT_PT139                                             139
#define RSVNA_REPCAP_PORT_PT140                                             140
#define RSVNA_REPCAP_PORT_PT141                                             141
#define RSVNA_REPCAP_PORT_PT142                                             142
#define RSVNA_REPCAP_PORT_PT143                                             143
#define RSVNA_REPCAP_PORT_PT144                                             144
#define RSVNA_REPCAP_PORT_PT145                                             145
#define RSVNA_REPCAP_PORT_PT146                                             146
#define RSVNA_REPCAP_PORT_PT147                                             147
#define RSVNA_REPCAP_PORT_PT148                                             148
#define RSVNA_REPCAP_PORT_PT149                                             149
#define RSVNA_REPCAP_PORT_PT150                                             150
#define RSVNA_REPCAP_PORT_PT151                                             151
#define RSVNA_REPCAP_PORT_PT152                                             152
#define RSVNA_REPCAP_PORT_PT153                                             153
#define RSVNA_REPCAP_PORT_PT154                                             154
#define RSVNA_REPCAP_PORT_PT155                                             155
#define RSVNA_REPCAP_PORT_PT156                                             156
#define RSVNA_REPCAP_PORT_PT157                                             157
#define RSVNA_REPCAP_PORT_PT158                                             158
#define RSVNA_REPCAP_PORT_PT159                                             159
#define RSVNA_REPCAP_PORT_PT160                                             160
#define RSVNA_REPCAP_PORT_PT161                                             161
#define RSVNA_REPCAP_PORT_PT162                                             162
#define RSVNA_REPCAP_PORT_PT163                                             163
#define RSVNA_REPCAP_PORT_PT164                                             164
#define RSVNA_REPCAP_PORT_PT165                                             165
#define RSVNA_REPCAP_PORT_PT166                                             166
#define RSVNA_REPCAP_PORT_PT167                                             167
#define RSVNA_REPCAP_PORT_PT168                                             168
#define RSVNA_REPCAP_PORT_PT169                                             169
#define RSVNA_REPCAP_PORT_PT170                                             170
#define RSVNA_REPCAP_PORT_PT171                                             171
#define RSVNA_REPCAP_PORT_PT172                                             172
#define RSVNA_REPCAP_PORT_PT173                                             173
#define RSVNA_REPCAP_PORT_PT174                                             174
#define RSVNA_REPCAP_PORT_PT175                                             175
#define RSVNA_REPCAP_PORT_PT176                                             176
#define RSVNA_REPCAP_PORT_PT177                                             177
#define RSVNA_REPCAP_PORT_PT178                                             178
#define RSVNA_REPCAP_PORT_PT179                                             179
#define RSVNA_REPCAP_PORT_PT180                                             180
#define RSVNA_REPCAP_PORT_PT181                                             181
#define RSVNA_REPCAP_PORT_PT182                                             182
#define RSVNA_REPCAP_PORT_PT183                                             183
#define RSVNA_REPCAP_PORT_PT184                                             184
#define RSVNA_REPCAP_PORT_PT185                                             185
#define RSVNA_REPCAP_PORT_PT186                                             186
#define RSVNA_REPCAP_PORT_PT187                                             187
#define RSVNA_REPCAP_PORT_PT188                                             188
#define RSVNA_REPCAP_PORT_PT189                                             189
#define RSVNA_REPCAP_PORT_PT190                                             190
#define RSVNA_REPCAP_PORT_PT191                                             191
#define RSVNA_REPCAP_PORT_PT192                                             192
#define RSVNA_REPCAP_PORT_PT193                                             193
#define RSVNA_REPCAP_PORT_PT194                                             194
#define RSVNA_REPCAP_PORT_PT195                                             195
#define RSVNA_REPCAP_PORT_PT196                                             196
#define RSVNA_REPCAP_PORT_PT197                                             197
#define RSVNA_REPCAP_PORT_PT198                                             198
#define RSVNA_REPCAP_PORT_PT199                                             199
#define RSVNA_REPCAP_PORT_PT200                                             200
#define RSVNA_REPCAP_PORT_PT201                                             201
#define RSVNA_REPCAP_PORT_PT202                                             202
#define RSVNA_REPCAP_PORT_PT203                                             203
#define RSVNA_REPCAP_PORT_PT204                                             204
#define RSVNA_REPCAP_PORT_PT205                                             205
#define RSVNA_REPCAP_PORT_PT206                                             206
#define RSVNA_REPCAP_PORT_PT207                                             207
#define RSVNA_REPCAP_PORT_PT208                                             208
#define RSVNA_REPCAP_PORT_PT209                                             209
#define RSVNA_REPCAP_PORT_PT210                                             210
#define RSVNA_REPCAP_PORT_PT211                                             211
#define RSVNA_REPCAP_PORT_PT212                                             212
#define RSVNA_REPCAP_PORT_PT213                                             213
#define RSVNA_REPCAP_PORT_PT214                                             214
#define RSVNA_REPCAP_PORT_PT215                                             215
#define RSVNA_REPCAP_PORT_PT216                                             216
#define RSVNA_REPCAP_PORT_PT217                                             217
#define RSVNA_REPCAP_PORT_PT218                                             218
#define RSVNA_REPCAP_PORT_PT219                                             219
#define RSVNA_REPCAP_PORT_PT220                                             220
#define RSVNA_REPCAP_PORT_PT221                                             221
#define RSVNA_REPCAP_PORT_PT222                                             222
#define RSVNA_REPCAP_PORT_PT223                                             223
#define RSVNA_REPCAP_PORT_PT224                                             224
#define RSVNA_REPCAP_PORT_PT225                                             225
#define RSVNA_REPCAP_PORT_PT226                                             226
#define RSVNA_REPCAP_PORT_PT227                                             227
#define RSVNA_REPCAP_PORT_PT228                                             228
#define RSVNA_REPCAP_PORT_PT229                                             229
#define RSVNA_REPCAP_PORT_PT230                                             230
#define RSVNA_REPCAP_PORT_PT231                                             231
#define RSVNA_REPCAP_PORT_PT232                                             232
#define RSVNA_REPCAP_PORT_PT233                                             233
#define RSVNA_REPCAP_PORT_PT234                                             234
#define RSVNA_REPCAP_PORT_PT235                                             235
#define RSVNA_REPCAP_PORT_PT236                                             236
#define RSVNA_REPCAP_PORT_PT237                                             237
#define RSVNA_REPCAP_PORT_PT238                                             238
#define RSVNA_REPCAP_PORT_PT239                                             239
#define RSVNA_REPCAP_PORT_PT240                                             240
#define RSVNA_REPCAP_PORT_PT241                                             241
#define RSVNA_REPCAP_PORT_PT242                                             242
#define RSVNA_REPCAP_PORT_PT243                                             243
#define RSVNA_REPCAP_PORT_PT244                                             244
#define RSVNA_REPCAP_PORT_PT245                                             245
#define RSVNA_REPCAP_PORT_PT246                                             246
#define RSVNA_REPCAP_PORT_PT247                                             247
#define RSVNA_REPCAP_PORT_PT248                                             248
#define RSVNA_REPCAP_PORT_PT249                                             249
#define RSVNA_REPCAP_PORT_PT250                                             250
#define RSVNA_REPCAP_PORT_PT251                                             251
#define RSVNA_REPCAP_PORT_PT252                                             252
#define RSVNA_REPCAP_PORT_PT253                                             253
#define RSVNA_REPCAP_PORT_PT254                                             254
#define RSVNA_REPCAP_PORT_PT255                                             255
#define RSVNA_REPCAP_PORT_PT256                                             256
#define RSVNA_REPCAP_PORT_PT257                                             257
#define RSVNA_REPCAP_PORT_PT258                                             258
#define RSVNA_REPCAP_PORT_PT259                                             259
#define RSVNA_REPCAP_PORT_PT260                                             260
#define RSVNA_REPCAP_PORT_PT261                                             261
#define RSVNA_REPCAP_PORT_PT262                                             262
#define RSVNA_REPCAP_PORT_PT263                                             263
#define RSVNA_REPCAP_PORT_PT264                                             264
#define RSVNA_REPCAP_PORT_PT265                                             265
#define RSVNA_REPCAP_PORT_PT266                                             266
#define RSVNA_REPCAP_PORT_PT267                                             267
#define RSVNA_REPCAP_PORT_PT268                                             268
#define RSVNA_REPCAP_PORT_PT269                                             269
#define RSVNA_REPCAP_PORT_PT270                                             270
#define RSVNA_REPCAP_PORT_PT271                                             271
#define RSVNA_REPCAP_PORT_PT272                                             272
#define RSVNA_REPCAP_PORT_PT273                                             273
#define RSVNA_REPCAP_PORT_PT274                                             274
#define RSVNA_REPCAP_PORT_PT275                                             275
#define RSVNA_REPCAP_PORT_PT276                                             276
#define RSVNA_REPCAP_PORT_PT277                                             277
#define RSVNA_REPCAP_PORT_PT278                                             278
#define RSVNA_REPCAP_PORT_PT279                                             279
#define RSVNA_REPCAP_PORT_PT280                                             280
#define RSVNA_REPCAP_PORT_PT281                                             281
#define RSVNA_REPCAP_PORT_PT282                                             282
#define RSVNA_REPCAP_PORT_PT283                                             283
#define RSVNA_REPCAP_PORT_PT284                                             284
#define RSVNA_REPCAP_PORT_PT285                                             285
#define RSVNA_REPCAP_PORT_PT286                                             286
#define RSVNA_REPCAP_PORT_PT287                                             287
#define RSVNA_REPCAP_PORT_PT288                                             288
#define RSVNA_REPCAP_PORT_PT289                                             289
#define RSVNA_REPCAP_PORT_PT290                                             290
#define RSVNA_REPCAP_PORT_PT291                                             291
#define RSVNA_REPCAP_PORT_PT292                                             292
#define RSVNA_REPCAP_PORT_PT293                                             293
#define RSVNA_REPCAP_PORT_PT294                                             294
#define RSVNA_REPCAP_PORT_PT295                                             295
#define RSVNA_REPCAP_PORT_PT296                                             296
#define RSVNA_REPCAP_PORT_PT297                                             297
#define RSVNA_REPCAP_PORT_PT298                                             298
#define RSVNA_REPCAP_PORT_PT299                                             299
#define RSVNA_REPCAP_PORT_PT300                                             300
#define RSVNA_REPCAP_PORT_PT301                                             301
#define RSVNA_REPCAP_PORT_PT302                                             302
#define RSVNA_REPCAP_PORT_PT303                                             303
#define RSVNA_REPCAP_PORT_PT304                                             304
#define RSVNA_REPCAP_PORT_PT305                                             305
#define RSVNA_REPCAP_PORT_PT306                                             306
#define RSVNA_REPCAP_PORT_PT307                                             307
#define RSVNA_REPCAP_PORT_PT308                                             308
#define RSVNA_REPCAP_PORT_PT309                                             309
#define RSVNA_REPCAP_PORT_PT310                                             310
#define RSVNA_REPCAP_PORT_PT311                                             311
#define RSVNA_REPCAP_PORT_PT312                                             312

//repcap PageMargin
#define RSVNA_REPCAP_PAGEMARGIN_LEFT                                        0
#define RSVNA_REPCAP_PAGEMARGIN_RIGHT                                       1
#define RSVNA_REPCAP_PAGEMARGIN_TOP                                         2
#define RSVNA_REPCAP_PAGEMARGIN_BOTTOM                                      3

//repcap Marker
#define RSVNA_REPCAP_MARKER_MK1                                             1
#define RSVNA_REPCAP_MARKER_MK2                                             2
#define RSVNA_REPCAP_MARKER_MK3                                             3
#define RSVNA_REPCAP_MARKER_MK4                                             4
#define RSVNA_REPCAP_MARKER_MK5                                             5
#define RSVNA_REPCAP_MARKER_MK6                                             6
#define RSVNA_REPCAP_MARKER_MK7                                             7
#define RSVNA_REPCAP_MARKER_MK8                                             8
#define RSVNA_REPCAP_MARKER_MK9                                             9
#define RSVNA_REPCAP_MARKER_MK10                                            10

//repcap Window
#define RSVNA_REPCAP_WINDOW_WIN1                                            1

//repcap QRegister
#define RSVNA_REPCAP_QREGISTER_MAIN                                         0
#define RSVNA_REPCAP_QREGISTER_REGLIMITA                                    1
#define RSVNA_REPCAP_QREGISTER_REGLIMITB                                    2
#define RSVNA_REPCAP_QREGISTER_REGINTEGRITY                                 3
#define RSVNA_REPCAP_QREGISTER_REGINTEGRITYHW                               4

//repcap RegisterPart
#define RSVNA_REPCAP_REGISTERPART_REGENABLE                                 0
#define RSVNA_REPCAP_REGISTERPART_REGNTR                                    1
#define RSVNA_REPCAP_REGISTERPART_REGPTR                                    2
#define RSVNA_REPCAP_REGISTERPART_REGCOND                                   3
#define RSVNA_REPCAP_REGISTERPART_REGEVENT                                  4

//repcap ResultType
#define RSVNA_REPCAP_RESULTTYPE_PHASEDELAY                                  0
#define RSVNA_REPCAP_RESULTTYPE_MAXMINPEAK                                  1
#define RSVNA_REPCAP_RESULTTYPE_MEANSTDDEV                                  2
#define RSVNA_REPCAP_RESULTTYPE_RMS                                         3
#define RSVNA_REPCAP_RESULTTYPE_GAINSLOPEFLATNESS                           4
#define RSVNA_REPCAP_RESULTTYPE_COMPPOINT                                   5

//repcap StatParam
#define RSVNA_REPCAP_STATPARAM_STATMEAN                                     0
#define RSVNA_REPCAP_STATPARAM_STATSDEV                                     1
#define RSVNA_REPCAP_STATPARAM_STATMAX                                      2
#define RSVNA_REPCAP_STATPARAM_STATMIN                                      3
#define RSVNA_REPCAP_STATPARAM_STATRMS                                      4
#define RSVNA_REPCAP_STATPARAM_STATPTP                                      5
#define RSVNA_REPCAP_STATPARAM_STATELEN                                     6
#define RSVNA_REPCAP_STATPARAM_STATPDEL                                     7
#define RSVNA_REPCAP_STATPARAM_STATGAIN                                     8
#define RSVNA_REPCAP_STATPARAM_STATSLOPE                                    9
#define RSVNA_REPCAP_STATPARAM_STATFLAT                                     10

//repcap LimitCheck
#define RSVNA_REPCAP_LIMITCHECK_LIMITCHECKUPPER                             0
#define RSVNA_REPCAP_LIMITCHECK_LIMITCHECKLOWER                             1
#define RSVNA_REPCAP_LIMITCHECK_LIMITCHECKALL                               2

//repcap StimAmpl
#define RSVNA_REPCAP_STIMAMPL_STIMULUS                                      0
#define RSVNA_REPCAP_STIMAMPL_AMPLITUDE                                     1

//repcap StartStop
#define RSVNA_REPCAP_STARTSTOP_START                                        0
#define RSVNA_REPCAP_STARTSTOP_STOP                                         1

//repcap OutputNumber
#define RSVNA_REPCAP_OUTPUTNUMBER_TTL1                                      1
#define RSVNA_REPCAP_OUTPUTNUMBER_TTL2                                      2

//repcap MarkerType
#define RSVNA_REPCAP_MARKERTYPE_NORMALMARKER                                0
#define RSVNA_REPCAP_MARKERTYPE_REFERENCEMARKER                             1

//repcap StartStopCenterSpan
#define RSVNA_REPCAP_STARTSTOPCENTERSPAN_START                              0
#define RSVNA_REPCAP_STARTSTOPCENTERSPAN_STOP                               1
#define RSVNA_REPCAP_STARTSTOPCENTERSPAN_CENTER                             2
#define RSVNA_REPCAP_STARTSTOPCENTERSPAN_SPAN                               3

//repcap ConnType
#define RSVNA_REPCAP_CONNTYPE_CONNN50                                       0
#define RSVNA_REPCAP_CONNTYPE_CONNN75                                       1
#define RSVNA_REPCAP_CONNTYPE_CONNPC7                                       2
#define RSVNA_REPCAP_CONNTYPE_CONNPC35                                      3
#define RSVNA_REPCAP_CONNTYPE_CONNPC292                                     4
#define RSVNA_REPCAP_CONNTYPE_CONNUSER1                                     5
#define RSVNA_REPCAP_CONNTYPE_CONNUSER2                                     6
#define RSVNA_REPCAP_CONNTYPE_CONNSMA                                       7

//repcap EmbedDeembed
#define RSVNA_REPCAP_EMBEDDEEMBED_DEEMBEDDING                               0
#define RSVNA_REPCAP_EMBEDDEEMBED_EMBEDDING                                 1

//repcap LogicalPort
#define RSVNA_REPCAP_LOGICALPORT_LOGPORT1                                   1
#define RSVNA_REPCAP_LOGICALPORT_LOGPORT2                                   2
#define RSVNA_REPCAP_LOGICALPORT_LOGPORT3                                   3
#define RSVNA_REPCAP_LOGICALPORT_LOGPORT4                                   4

//repcap DCInput
#define RSVNA_REPCAP_DCINPUT_DCINPUT1                                       1
#define RSVNA_REPCAP_DCINPUT_DCINPUT2                                       2
#define RSVNA_REPCAP_DCINPUT_DCINPUT3                                       3
#define RSVNA_REPCAP_DCINPUT_DCINPUT4                                       4

//repcap PAEType
#define RSVNA_REPCAP_PAETYPE_MAIN                                           0
#define RSVNA_REPCAP_PAETYPE_SECONDARY                                      1

//repcap PAEParam
#define RSVNA_REPCAP_PAEPARAM_I                                             0
#define RSVNA_REPCAP_PAEPARAM_R                                             1
#define RSVNA_REPCAP_PAEPARAM_U                                             2

//repcap ListNo
#define RSVNA_REPCAP_LISTNO_LN1                                             1
#define RSVNA_REPCAP_LISTNO_LN2                                             2
#define RSVNA_REPCAP_LISTNO_LN3                                             3
#define RSVNA_REPCAP_LISTNO_LN4                                             4
#define RSVNA_REPCAP_LISTNO_LN5                                             5
#define RSVNA_REPCAP_LISTNO_LN6                                             6
#define RSVNA_REPCAP_LISTNO_LN7                                             7
#define RSVNA_REPCAP_LISTNO_LN8                                             8

//repcap Generator
#define RSVNA_REPCAP_GENERATOR_GEN1                                         1
#define RSVNA_REPCAP_GENERATOR_GEN2                                         2
#define RSVNA_REPCAP_GENERATOR_GEN3                                         3
#define RSVNA_REPCAP_GENERATOR_GEN4                                         4
#define RSVNA_REPCAP_GENERATOR_GEN5                                         5
#define RSVNA_REPCAP_GENERATOR_GEN6                                         6
#define RSVNA_REPCAP_GENERATOR_GEN7                                         7
#define RSVNA_REPCAP_GENERATOR_GEN8                                         8
#define RSVNA_REPCAP_GENERATOR_GEN9                                         9
#define RSVNA_REPCAP_GENERATOR_GEN10                                        10
#define RSVNA_REPCAP_GENERATOR_GEN11                                        11
#define RSVNA_REPCAP_GENERATOR_GEN12                                        12
#define RSVNA_REPCAP_GENERATOR_GEN13                                        13
#define RSVNA_REPCAP_GENERATOR_GEN14                                        14
#define RSVNA_REPCAP_GENERATOR_GEN15                                        15
#define RSVNA_REPCAP_GENERATOR_GEN16                                        16
#define RSVNA_REPCAP_GENERATOR_GEN17                                        17
#define RSVNA_REPCAP_GENERATOR_GEN18                                        18
#define RSVNA_REPCAP_GENERATOR_GEN19                                        19
#define RSVNA_REPCAP_GENERATOR_GEN20                                        20
#define RSVNA_REPCAP_GENERATOR_GEN21                                        21
#define RSVNA_REPCAP_GENERATOR_GEN22                                        22
#define RSVNA_REPCAP_GENERATOR_GEN23                                        23
#define RSVNA_REPCAP_GENERATOR_GEN24                                        24
#define RSVNA_REPCAP_GENERATOR_GEN25                                        25
#define RSVNA_REPCAP_GENERATOR_GEN26                                        26
#define RSVNA_REPCAP_GENERATOR_GEN27                                        27
#define RSVNA_REPCAP_GENERATOR_GEN28                                        28
#define RSVNA_REPCAP_GENERATOR_GEN29                                        29
#define RSVNA_REPCAP_GENERATOR_GEN30                                        30
#define RSVNA_REPCAP_GENERATOR_GEN31                                        31
#define RSVNA_REPCAP_GENERATOR_GEN32                                        32
#define RSVNA_REPCAP_GENERATOR_GEN33                                        33
#define RSVNA_REPCAP_GENERATOR_GEN34                                        34
#define RSVNA_REPCAP_GENERATOR_GEN35                                        35
#define RSVNA_REPCAP_GENERATOR_GEN36                                        36
#define RSVNA_REPCAP_GENERATOR_GEN37                                        37
#define RSVNA_REPCAP_GENERATOR_GEN38                                        38
#define RSVNA_REPCAP_GENERATOR_GEN39                                        39
#define RSVNA_REPCAP_GENERATOR_GEN40                                        40
#define RSVNA_REPCAP_GENERATOR_GEN41                                        41
#define RSVNA_REPCAP_GENERATOR_GEN42                                        42
#define RSVNA_REPCAP_GENERATOR_GEN43                                        43
#define RSVNA_REPCAP_GENERATOR_GEN44                                        44
#define RSVNA_REPCAP_GENERATOR_GEN45                                        45
#define RSVNA_REPCAP_GENERATOR_GEN46                                        46
#define RSVNA_REPCAP_GENERATOR_GEN47                                        47
#define RSVNA_REPCAP_GENERATOR_GEN48                                        48
#define RSVNA_REPCAP_GENERATOR_GEN49                                        49
#define RSVNA_REPCAP_GENERATOR_GEN50                                        50
#define RSVNA_REPCAP_GENERATOR_GEN51                                        51
#define RSVNA_REPCAP_GENERATOR_GEN52                                        52
#define RSVNA_REPCAP_GENERATOR_GEN53                                        53
#define RSVNA_REPCAP_GENERATOR_GEN54                                        54
#define RSVNA_REPCAP_GENERATOR_GEN55                                        55
#define RSVNA_REPCAP_GENERATOR_GEN56                                        56
#define RSVNA_REPCAP_GENERATOR_GEN57                                        57
#define RSVNA_REPCAP_GENERATOR_GEN58                                        58
#define RSVNA_REPCAP_GENERATOR_GEN59                                        59
#define RSVNA_REPCAP_GENERATOR_GEN60                                        60
#define RSVNA_REPCAP_GENERATOR_GEN61                                        61
#define RSVNA_REPCAP_GENERATOR_GEN62                                        62
#define RSVNA_REPCAP_GENERATOR_GEN63                                        63
#define RSVNA_REPCAP_GENERATOR_GEN64                                        64
#define RSVNA_REPCAP_GENERATOR_GEN65                                        65
#define RSVNA_REPCAP_GENERATOR_GEN66                                        66
#define RSVNA_REPCAP_GENERATOR_GEN67                                        67
#define RSVNA_REPCAP_GENERATOR_GEN68                                        68
#define RSVNA_REPCAP_GENERATOR_GEN69                                        69
#define RSVNA_REPCAP_GENERATOR_GEN70                                        70
#define RSVNA_REPCAP_GENERATOR_GEN71                                        71
#define RSVNA_REPCAP_GENERATOR_GEN72                                        72
#define RSVNA_REPCAP_GENERATOR_GEN73                                        73
#define RSVNA_REPCAP_GENERATOR_GEN74                                        74
#define RSVNA_REPCAP_GENERATOR_GEN75                                        75
#define RSVNA_REPCAP_GENERATOR_GEN76                                        76
#define RSVNA_REPCAP_GENERATOR_GEN77                                        77
#define RSVNA_REPCAP_GENERATOR_GEN78                                        78
#define RSVNA_REPCAP_GENERATOR_GEN79                                        79
#define RSVNA_REPCAP_GENERATOR_GEN80                                        80
#define RSVNA_REPCAP_GENERATOR_GEN81                                        81
#define RSVNA_REPCAP_GENERATOR_GEN82                                        82
#define RSVNA_REPCAP_GENERATOR_GEN83                                        83
#define RSVNA_REPCAP_GENERATOR_GEN84                                        84
#define RSVNA_REPCAP_GENERATOR_GEN85                                        85
#define RSVNA_REPCAP_GENERATOR_GEN86                                        86
#define RSVNA_REPCAP_GENERATOR_GEN87                                        87
#define RSVNA_REPCAP_GENERATOR_GEN88                                        88
#define RSVNA_REPCAP_GENERATOR_GEN89                                        89
#define RSVNA_REPCAP_GENERATOR_GEN90                                        90
#define RSVNA_REPCAP_GENERATOR_GEN91                                        91
#define RSVNA_REPCAP_GENERATOR_GEN92                                        92
#define RSVNA_REPCAP_GENERATOR_GEN93                                        93
#define RSVNA_REPCAP_GENERATOR_GEN94                                        94
#define RSVNA_REPCAP_GENERATOR_GEN95                                        95
#define RSVNA_REPCAP_GENERATOR_GEN96                                        96
#define RSVNA_REPCAP_GENERATOR_GEN97                                        97
#define RSVNA_REPCAP_GENERATOR_GEN98                                        98
#define RSVNA_REPCAP_GENERATOR_GEN99                                        99
#define RSVNA_REPCAP_GENERATOR_GEN100                                       100
#define RSVNA_REPCAP_GENERATOR_GEN101                                       101
#define RSVNA_REPCAP_GENERATOR_GEN102                                       102
#define RSVNA_REPCAP_GENERATOR_GEN103                                       103
#define RSVNA_REPCAP_GENERATOR_GEN104                                       104
#define RSVNA_REPCAP_GENERATOR_GEN105                                       105
#define RSVNA_REPCAP_GENERATOR_GEN106                                       106
#define RSVNA_REPCAP_GENERATOR_GEN107                                       107
#define RSVNA_REPCAP_GENERATOR_GEN108                                       108
#define RSVNA_REPCAP_GENERATOR_GEN109                                       109
#define RSVNA_REPCAP_GENERATOR_GEN110                                       110
#define RSVNA_REPCAP_GENERATOR_GEN111                                       111
#define RSVNA_REPCAP_GENERATOR_GEN112                                       112
#define RSVNA_REPCAP_GENERATOR_GEN113                                       113
#define RSVNA_REPCAP_GENERATOR_GEN114                                       114
#define RSVNA_REPCAP_GENERATOR_GEN115                                       115
#define RSVNA_REPCAP_GENERATOR_GEN116                                       116
#define RSVNA_REPCAP_GENERATOR_GEN117                                       117
#define RSVNA_REPCAP_GENERATOR_GEN118                                       118
#define RSVNA_REPCAP_GENERATOR_GEN119                                       119
#define RSVNA_REPCAP_GENERATOR_GEN120                                       120
#define RSVNA_REPCAP_GENERATOR_GEN121                                       121
#define RSVNA_REPCAP_GENERATOR_GEN122                                       122
#define RSVNA_REPCAP_GENERATOR_GEN123                                       123
#define RSVNA_REPCAP_GENERATOR_GEN124                                       124
#define RSVNA_REPCAP_GENERATOR_GEN125                                       125
#define RSVNA_REPCAP_GENERATOR_GEN126                                       126
#define RSVNA_REPCAP_GENERATOR_GEN127                                       127
#define RSVNA_REPCAP_GENERATOR_GEN128                                       128

//repcap PowMeter
#define RSVNA_REPCAP_POWMETER_PM1                                           1
#define RSVNA_REPCAP_POWMETER_PM2                                           2
#define RSVNA_REPCAP_POWMETER_PM3                                           3
#define RSVNA_REPCAP_POWMETER_PM4                                           4
#define RSVNA_REPCAP_POWMETER_PM5                                           5
#define RSVNA_REPCAP_POWMETER_PM6                                           6
#define RSVNA_REPCAP_POWMETER_PM7                                           7
#define RSVNA_REPCAP_POWMETER_PM8                                           8

//repcap FreqPower
#define RSVNA_REPCAP_FREQPOWER_FREQUENCY                                    0
#define RSVNA_REPCAP_FREQPOWER_POWER                                        1

//repcap MinMax
#define RSVNA_REPCAP_MINMAX_MIN                                             0
#define RSVNA_REPCAP_MINMAX_MAX                                             1

//repcap MixerStage
#define RSVNA_REPCAP_MIXERSTAGE_STAGE1                                      1
#define RSVNA_REPCAP_MIXERSTAGE_STAGE2                                      2

//repcap AUXPort
#define RSVNA_REPCAP_AUXPORT_AUX1                                           1
#define RSVNA_REPCAP_AUXPORT_AUX2                                           2

//repcap UIPort
#define RSVNA_REPCAP_UIPORT_E                                               0
#define RSVNA_REPCAP_UIPORT_F                                               1
#define RSVNA_REPCAP_UIPORT_G                                               2
#define RSVNA_REPCAP_UIPORT_H                                               3

//repcap IMSource
#define RSVNA_REPCAP_IMSOURCE_LOWER                                         0
#define RSVNA_REPCAP_IMSOURCE_UPPER                                         1

//repcap IMOrder
#define RSVNA_REPCAP_IMORDER_ORDER3                                         0
#define RSVNA_REPCAP_IMORDER_ORDER5                                         1
#define RSVNA_REPCAP_IMORDER_ORDER7                                         2
#define RSVNA_REPCAP_IMORDER_ORDER9                                         3

//repcap UIPortInit
#define RSVNA_REPCAP_UIPORTINIT_A                                           0
#define RSVNA_REPCAP_UIPORTINIT_B                                           1
#define RSVNA_REPCAP_UIPORTINIT_C                                           2
#define RSVNA_REPCAP_UIPORTINIT_D                                           3

//repcap OutputPort
#define RSVNA_REPCAP_OUTPUTPORT_OUTPORT1                                    1
#define RSVNA_REPCAP_OUTPUTPORT_OUTPORT2                                    2

//repcap SignalRouting
#define RSVNA_REPCAP_SIGNALROUTING_INDEX                                    0
#define RSVNA_REPCAP_SIGNALROUTING_RTRIGGER                                 1

//repcap Wave
#define RSVNA_REPCAP_WAVE_WAVE1                                             1
#define RSVNA_REPCAP_WAVE_WAVE2                                             2
#define RSVNA_REPCAP_WAVE_WAVE3                                             3
#define RSVNA_REPCAP_WAVE_WAVE4                                             4

//repcap RFFEBus
#define RSVNA_REPCAP_RFFEBUS_BUS1                                           1
#define RSVNA_REPCAP_RFFEBUS_BUS2                                           2

//repcap GPIOPort
#define RSVNA_REPCAP_GPIOPORT_GPIO1                                         1
#define RSVNA_REPCAP_GPIOPORT_GPIO2                                         2
#define RSVNA_REPCAP_GPIOPORT_GPIO3                                         3
#define RSVNA_REPCAP_GPIOPORT_GPIO4                                         4
#define RSVNA_REPCAP_GPIOPORT_GPIO5                                         5
#define RSVNA_REPCAP_GPIOPORT_GPIO6                                         6
#define RSVNA_REPCAP_GPIOPORT_GPIO7                                         7
#define RSVNA_REPCAP_GPIOPORT_GPIO8                                         8
#define RSVNA_REPCAP_GPIOPORT_GPIO9                                         9
#define RSVNA_REPCAP_GPIOPORT_GPIO10                                        10

//repcap StandardType
#define RSVNA_REPCAP_STANDARDTYPE_MOPEN                                     0
#define RSVNA_REPCAP_STANDARDTYPE_FOPEN                                     1
#define RSVNA_REPCAP_STANDARDTYPE_MSHORT                                    2
#define RSVNA_REPCAP_STANDARDTYPE_FSHORT                                    3
#define RSVNA_REPCAP_STANDARDTYPE_OFFSHORT1                                 4
#define RSVNA_REPCAP_STANDARDTYPE_MOFFSHORT1                                5
#define RSVNA_REPCAP_STANDARDTYPE_FOFFSHORT1                                6
#define RSVNA_REPCAP_STANDARDTYPE_OFFSHORT2                                 7
#define RSVNA_REPCAP_STANDARDTYPE_MOFFSHORT2                                8
#define RSVNA_REPCAP_STANDARDTYPE_FOFFSHORT2                                9
#define RSVNA_REPCAP_STANDARDTYPE_OFFSHORT3                                 10
#define RSVNA_REPCAP_STANDARDTYPE_MOFFSHORT3                                11
#define RSVNA_REPCAP_STANDARDTYPE_FOFFSHORT3                                12
#define RSVNA_REPCAP_STANDARDTYPE_MMATCH                                    13
#define RSVNA_REPCAP_STANDARDTYPE_FMATCH                                    14
#define RSVNA_REPCAP_STANDARDTYPE_MSLIDINGMATCH                             15
#define RSVNA_REPCAP_STANDARDTYPE_FSLIDINGMATCH                             16
#define RSVNA_REPCAP_STANDARDTYPE_MREFLECT                                  17
#define RSVNA_REPCAP_STANDARDTYPE_FREFLECT                                  18
#define RSVNA_REPCAP_STANDARDTYPE_MMTHROUGH                                 19
#define RSVNA_REPCAP_STANDARDTYPE_MFTHROUGH                                 20
#define RSVNA_REPCAP_STANDARDTYPE_FFTHROUGH                                 21
#define RSVNA_REPCAP_STANDARDTYPE_MMLINE1                                   22
#define RSVNA_REPCAP_STANDARDTYPE_MFLINE1                                   23
#define RSVNA_REPCAP_STANDARDTYPE_FFLINE1                                   24
#define RSVNA_REPCAP_STANDARDTYPE_MMLINE2                                   25
#define RSVNA_REPCAP_STANDARDTYPE_MFLINE2                                   26
#define RSVNA_REPCAP_STANDARDTYPE_FFLINE2                                   27
#define RSVNA_REPCAP_STANDARDTYPE_MMLINE3                                   28
#define RSVNA_REPCAP_STANDARDTYPE_MFLINE3                                   29
#define RSVNA_REPCAP_STANDARDTYPE_FFLINE3                                   30
#define RSVNA_REPCAP_STANDARDTYPE_MMATTEN                                   31
#define RSVNA_REPCAP_STANDARDTYPE_MFATTEN                                   32
#define RSVNA_REPCAP_STANDARDTYPE_FFATTEN                                   33
#define RSVNA_REPCAP_STANDARDTYPE_MMSYMMNETWORK                             34
#define RSVNA_REPCAP_STANDARDTYPE_MFSYMMNETWORK                             35
#define RSVNA_REPCAP_STANDARDTYPE_FFSYMMNETWORK                             36

//repcap Cursor
#define RSVNA_REPCAP_CURSOR_CURSOR1                                         1
#define RSVNA_REPCAP_CURSOR_CURSOR2                                         2

//repcap Pole
#define RSVNA_REPCAP_POLE_POLE1                                             1
#define RSVNA_REPCAP_POLE_POLE2                                             2

//repcap FavId
#define RSVNA_REPCAP_FAVID_FAVID1                                           1
#define RSVNA_REPCAP_FAVID_FAVID2                                           2
#define RSVNA_REPCAP_FAVID_FAVID3                                           3
#define RSVNA_REPCAP_FAVID_FAVID4                                           4
#define RSVNA_REPCAP_FAVID_FAVID5                                           5
#define RSVNA_REPCAP_FAVID_FAVID6                                           6
#define RSVNA_REPCAP_FAVID_FAVID7                                           7
#define RSVNA_REPCAP_FAVID_FAVID8                                           8
#define RSVNA_REPCAP_FAVID_FAVID9                                           9
#define RSVNA_REPCAP_FAVID_FAVID10                                          10
#define RSVNA_REPCAP_FAVID_FAVID11                                          11
#define RSVNA_REPCAP_FAVID_FAVID12                                          12
#define RSVNA_REPCAP_FAVID_FAVID13                                          13
#define RSVNA_REPCAP_FAVID_FAVID14                                          14
#define RSVNA_REPCAP_FAVID_FAVID15                                          15
#define RSVNA_REPCAP_FAVID_FAVID16                                          16
#define RSVNA_REPCAP_FAVID_FAVID17                                          17
#define RSVNA_REPCAP_FAVID_FAVID18                                          18
#define RSVNA_REPCAP_FAVID_FAVID19                                          19
#define RSVNA_REPCAP_FAVID_FAVID20                                          20
#define RSVNA_REPCAP_FAVID_FAVID21                                          21
#define RSVNA_REPCAP_FAVID_FAVID22                                          22
#define RSVNA_REPCAP_FAVID_FAVID23                                          23
#define RSVNA_REPCAP_FAVID_FAVID24                                          24
#define RSVNA_REPCAP_FAVID_FAVID25                                          25
#define RSVNA_REPCAP_FAVID_FAVID26                                          26
#define RSVNA_REPCAP_FAVID_FAVID27                                          27
#define RSVNA_REPCAP_FAVID_FAVID28                                          28
#define RSVNA_REPCAP_FAVID_FAVID29                                          29
#define RSVNA_REPCAP_FAVID_FAVID30                                          30
#define RSVNA_REPCAP_FAVID_FAVID31                                          31
#define RSVNA_REPCAP_FAVID_FAVID32                                          32
#define RSVNA_REPCAP_FAVID_FAVID33                                          33
#define RSVNA_REPCAP_FAVID_FAVID34                                          34
#define RSVNA_REPCAP_FAVID_FAVID35                                          35
#define RSVNA_REPCAP_FAVID_FAVID36                                          36
#define RSVNA_REPCAP_FAVID_FAVID37                                          37
#define RSVNA_REPCAP_FAVID_FAVID38                                          38
#define RSVNA_REPCAP_FAVID_FAVID39                                          39
#define RSVNA_REPCAP_FAVID_FAVID40                                          40
#define RSVNA_REPCAP_FAVID_FAVID41                                          41
#define RSVNA_REPCAP_FAVID_FAVID42                                          42
#define RSVNA_REPCAP_FAVID_FAVID43                                          43
#define RSVNA_REPCAP_FAVID_FAVID44                                          44
#define RSVNA_REPCAP_FAVID_FAVID45                                          45
#define RSVNA_REPCAP_FAVID_FAVID46                                          46
#define RSVNA_REPCAP_FAVID_FAVID47                                          47
#define RSVNA_REPCAP_FAVID_FAVID48                                          48
#define RSVNA_REPCAP_FAVID_FAVID49                                          49
#define RSVNA_REPCAP_FAVID_FAVID50                                          50
#define RSVNA_REPCAP_FAVID_FAVID51                                          51
#define RSVNA_REPCAP_FAVID_FAVID52                                          52
#define RSVNA_REPCAP_FAVID_FAVID53                                          53
#define RSVNA_REPCAP_FAVID_FAVID54                                          54
#define RSVNA_REPCAP_FAVID_FAVID55                                          55
#define RSVNA_REPCAP_FAVID_FAVID56                                          56
#define RSVNA_REPCAP_FAVID_FAVID57                                          57
#define RSVNA_REPCAP_FAVID_FAVID58                                          58
#define RSVNA_REPCAP_FAVID_FAVID59                                          59
#define RSVNA_REPCAP_FAVID_FAVID60                                          60
#define RSVNA_REPCAP_FAVID_FAVID61                                          61
#define RSVNA_REPCAP_FAVID_FAVID62                                          62
#define RSVNA_REPCAP_FAVID_FAVID63                                          63
#define RSVNA_REPCAP_FAVID_FAVID64                                          64
#define RSVNA_REPCAP_FAVID_FAVID65                                          65
#define RSVNA_REPCAP_FAVID_FAVID66                                          66
#define RSVNA_REPCAP_FAVID_FAVID67                                          67
#define RSVNA_REPCAP_FAVID_FAVID68                                          68
#define RSVNA_REPCAP_FAVID_FAVID69                                          69
#define RSVNA_REPCAP_FAVID_FAVID70                                          70
#define RSVNA_REPCAP_FAVID_FAVID71                                          71
#define RSVNA_REPCAP_FAVID_FAVID72                                          72
#define RSVNA_REPCAP_FAVID_FAVID73                                          73
#define RSVNA_REPCAP_FAVID_FAVID74                                          74
#define RSVNA_REPCAP_FAVID_FAVID75                                          75
#define RSVNA_REPCAP_FAVID_FAVID76                                          76
#define RSVNA_REPCAP_FAVID_FAVID77                                          77
#define RSVNA_REPCAP_FAVID_FAVID78                                          78
#define RSVNA_REPCAP_FAVID_FAVID79                                          79
#define RSVNA_REPCAP_FAVID_FAVID80                                          80
#define RSVNA_REPCAP_FAVID_FAVID81                                          81
#define RSVNA_REPCAP_FAVID_FAVID82                                          82
#define RSVNA_REPCAP_FAVID_FAVID83                                          83
#define RSVNA_REPCAP_FAVID_FAVID84                                          84
#define RSVNA_REPCAP_FAVID_FAVID85                                          85
#define RSVNA_REPCAP_FAVID_FAVID86                                          86
#define RSVNA_REPCAP_FAVID_FAVID87                                          87
#define RSVNA_REPCAP_FAVID_FAVID88                                          88
#define RSVNA_REPCAP_FAVID_FAVID89                                          89
#define RSVNA_REPCAP_FAVID_FAVID90                                          90
#define RSVNA_REPCAP_FAVID_FAVID91                                          91
#define RSVNA_REPCAP_FAVID_FAVID92                                          92
#define RSVNA_REPCAP_FAVID_FAVID93                                          93
#define RSVNA_REPCAP_FAVID_FAVID94                                          94
#define RSVNA_REPCAP_FAVID_FAVID95                                          95
#define RSVNA_REPCAP_FAVID_FAVID96                                          96
#define RSVNA_REPCAP_FAVID_FAVID97                                          97
#define RSVNA_REPCAP_FAVID_FAVID98                                          98
#define RSVNA_REPCAP_FAVID_FAVID99                                          99
#define RSVNA_REPCAP_FAVID_FAVID100                                         100

//repcap HUMSTag
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG0                                       0
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG1                                       1
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG2                                       2
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG3                                       3
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG4                                       4
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG5                                       5
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG6                                       6
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG7                                       7
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG8                                       8
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG9                                       9
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG10                                      10
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG11                                      11
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG12                                      12
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG13                                      13
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG14                                      14
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG15                                      15
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG16                                      16
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG17                                      17
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG18                                      18
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG19                                      19
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG20                                      20
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG21                                      21
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG22                                      22
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG23                                      23
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG24                                      24
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG25                                      25
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG26                                      26
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG27                                      27
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG28                                      28
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG29                                      29
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG30                                      30
#define RSVNA_REPCAP_HUMSTAG_HUMSTAG31                                      31

//repcap GenericDevice
#define RSVNA_REPCAP_GENERICDEVICE_GDEV1                                    1
#define RSVNA_REPCAP_GENERICDEVICE_GDEV2                                    2
#define RSVNA_REPCAP_GENERICDEVICE_GDEV3                                    3
#define RSVNA_REPCAP_GENERICDEVICE_GDEV4                                    4
#define RSVNA_REPCAP_GENERICDEVICE_GDEV5                                    5
#define RSVNA_REPCAP_GENERICDEVICE_GDEV6                                    6
#define RSVNA_REPCAP_GENERICDEVICE_GDEV7                                    7
#define RSVNA_REPCAP_GENERICDEVICE_GDEV8                                    8
#define RSVNA_REPCAP_GENERICDEVICE_GDEV9                                    9
#define RSVNA_REPCAP_GENERICDEVICE_GDEV10                                   10
#define RSVNA_REPCAP_GENERICDEVICE_GDEV11                                   11
#define RSVNA_REPCAP_GENERICDEVICE_GDEV12                                   12
#define RSVNA_REPCAP_GENERICDEVICE_GDEV13                                   13
#define RSVNA_REPCAP_GENERICDEVICE_GDEV14                                   14
#define RSVNA_REPCAP_GENERICDEVICE_GDEV15                                   15
#define RSVNA_REPCAP_GENERICDEVICE_GDEV16                                   16
#define RSVNA_REPCAP_GENERICDEVICE_GDEV17                                   17
#define RSVNA_REPCAP_GENERICDEVICE_GDEV18                                   18
#define RSVNA_REPCAP_GENERICDEVICE_GDEV19                                   19
#define RSVNA_REPCAP_GENERICDEVICE_GDEV20                                   20
#define RSVNA_REPCAP_GENERICDEVICE_GDEV21                                   21
#define RSVNA_REPCAP_GENERICDEVICE_GDEV22                                   22
#define RSVNA_REPCAP_GENERICDEVICE_GDEV23                                   23
#define RSVNA_REPCAP_GENERICDEVICE_GDEV24                                   24
#define RSVNA_REPCAP_GENERICDEVICE_GDEV25                                   25
#define RSVNA_REPCAP_GENERICDEVICE_GDEV26                                   26
#define RSVNA_REPCAP_GENERICDEVICE_GDEV27                                   27
#define RSVNA_REPCAP_GENERICDEVICE_GDEV28                                   28
#define RSVNA_REPCAP_GENERICDEVICE_GDEV29                                   29
#define RSVNA_REPCAP_GENERICDEVICE_GDEV30                                   30
#define RSVNA_REPCAP_GENERICDEVICE_GDEV31                                   31
#define RSVNA_REPCAP_GENERICDEVICE_GDEV32                                   32
#define RSVNA_REPCAP_GENERICDEVICE_GDEV33                                   33
#define RSVNA_REPCAP_GENERICDEVICE_GDEV34                                   34
#define RSVNA_REPCAP_GENERICDEVICE_GDEV35                                   35
#define RSVNA_REPCAP_GENERICDEVICE_GDEV36                                   36
#define RSVNA_REPCAP_GENERICDEVICE_GDEV37                                   37
#define RSVNA_REPCAP_GENERICDEVICE_GDEV38                                   38
#define RSVNA_REPCAP_GENERICDEVICE_GDEV39                                   39
#define RSVNA_REPCAP_GENERICDEVICE_GDEV40                                   40
#define RSVNA_REPCAP_GENERICDEVICE_GDEV41                                   41
#define RSVNA_REPCAP_GENERICDEVICE_GDEV42                                   42
#define RSVNA_REPCAP_GENERICDEVICE_GDEV43                                   43
#define RSVNA_REPCAP_GENERICDEVICE_GDEV44                                   44
#define RSVNA_REPCAP_GENERICDEVICE_GDEV45                                   45
#define RSVNA_REPCAP_GENERICDEVICE_GDEV46                                   46
#define RSVNA_REPCAP_GENERICDEVICE_GDEV47                                   47
#define RSVNA_REPCAP_GENERICDEVICE_GDEV48                                   48
#define RSVNA_REPCAP_GENERICDEVICE_GDEV49                                   49
#define RSVNA_REPCAP_GENERICDEVICE_GDEV50                                   50
#define RSVNA_REPCAP_GENERICDEVICE_GDEV51                                   51
#define RSVNA_REPCAP_GENERICDEVICE_GDEV52                                   52
#define RSVNA_REPCAP_GENERICDEVICE_GDEV53                                   53
#define RSVNA_REPCAP_GENERICDEVICE_GDEV54                                   54
#define RSVNA_REPCAP_GENERICDEVICE_GDEV55                                   55
#define RSVNA_REPCAP_GENERICDEVICE_GDEV56                                   56
#define RSVNA_REPCAP_GENERICDEVICE_GDEV57                                   57
#define RSVNA_REPCAP_GENERICDEVICE_GDEV58                                   58
#define RSVNA_REPCAP_GENERICDEVICE_GDEV59                                   59
#define RSVNA_REPCAP_GENERICDEVICE_GDEV60                                   60
#define RSVNA_REPCAP_GENERICDEVICE_GDEV61                                   61
#define RSVNA_REPCAP_GENERICDEVICE_GDEV62                                   62
#define RSVNA_REPCAP_GENERICDEVICE_GDEV63                                   63
#define RSVNA_REPCAP_GENERICDEVICE_GDEV64                                   64
#define RSVNA_REPCAP_GENERICDEVICE_GDEV65                                   65
#define RSVNA_REPCAP_GENERICDEVICE_GDEV66                                   66
#define RSVNA_REPCAP_GENERICDEVICE_GDEV67                                   67
#define RSVNA_REPCAP_GENERICDEVICE_GDEV68                                   68
#define RSVNA_REPCAP_GENERICDEVICE_GDEV69                                   69
#define RSVNA_REPCAP_GENERICDEVICE_GDEV70                                   70
#define RSVNA_REPCAP_GENERICDEVICE_GDEV71                                   71
#define RSVNA_REPCAP_GENERICDEVICE_GDEV72                                   72
#define RSVNA_REPCAP_GENERICDEVICE_GDEV73                                   73
#define RSVNA_REPCAP_GENERICDEVICE_GDEV74                                   74
#define RSVNA_REPCAP_GENERICDEVICE_GDEV75                                   75
#define RSVNA_REPCAP_GENERICDEVICE_GDEV76                                   76
#define RSVNA_REPCAP_GENERICDEVICE_GDEV77                                   77
#define RSVNA_REPCAP_GENERICDEVICE_GDEV78                                   78
#define RSVNA_REPCAP_GENERICDEVICE_GDEV79                                   79
#define RSVNA_REPCAP_GENERICDEVICE_GDEV80                                   80
#define RSVNA_REPCAP_GENERICDEVICE_GDEV81                                   81
#define RSVNA_REPCAP_GENERICDEVICE_GDEV82                                   82
#define RSVNA_REPCAP_GENERICDEVICE_GDEV83                                   83
#define RSVNA_REPCAP_GENERICDEVICE_GDEV84                                   84
#define RSVNA_REPCAP_GENERICDEVICE_GDEV85                                   85
#define RSVNA_REPCAP_GENERICDEVICE_GDEV86                                   86
#define RSVNA_REPCAP_GENERICDEVICE_GDEV87                                   87
#define RSVNA_REPCAP_GENERICDEVICE_GDEV88                                   88
#define RSVNA_REPCAP_GENERICDEVICE_GDEV89                                   89
#define RSVNA_REPCAP_GENERICDEVICE_GDEV90                                   90
#define RSVNA_REPCAP_GENERICDEVICE_GDEV91                                   91
#define RSVNA_REPCAP_GENERICDEVICE_GDEV92                                   92
#define RSVNA_REPCAP_GENERICDEVICE_GDEV93                                   93
#define RSVNA_REPCAP_GENERICDEVICE_GDEV94                                   94
#define RSVNA_REPCAP_GENERICDEVICE_GDEV95                                   95
#define RSVNA_REPCAP_GENERICDEVICE_GDEV96                                   96
#define RSVNA_REPCAP_GENERICDEVICE_GDEV97                                   97
#define RSVNA_REPCAP_GENERICDEVICE_GDEV98                                   98
#define RSVNA_REPCAP_GENERICDEVICE_GDEV99                                   99
#define RSVNA_REPCAP_GENERICDEVICE_GDEV100                                  100
#define RSVNA_REPCAP_GENERICDEVICE_GDEV101                                  101
#define RSVNA_REPCAP_GENERICDEVICE_GDEV102                                  102
#define RSVNA_REPCAP_GENERICDEVICE_GDEV103                                  103
#define RSVNA_REPCAP_GENERICDEVICE_GDEV104                                  104
#define RSVNA_REPCAP_GENERICDEVICE_GDEV105                                  105
#define RSVNA_REPCAP_GENERICDEVICE_GDEV106                                  106
#define RSVNA_REPCAP_GENERICDEVICE_GDEV107                                  107
#define RSVNA_REPCAP_GENERICDEVICE_GDEV108                                  108
#define RSVNA_REPCAP_GENERICDEVICE_GDEV109                                  109
#define RSVNA_REPCAP_GENERICDEVICE_GDEV110                                  110
#define RSVNA_REPCAP_GENERICDEVICE_GDEV111                                  111
#define RSVNA_REPCAP_GENERICDEVICE_GDEV112                                  112
#define RSVNA_REPCAP_GENERICDEVICE_GDEV113                                  113
#define RSVNA_REPCAP_GENERICDEVICE_GDEV114                                  114
#define RSVNA_REPCAP_GENERICDEVICE_GDEV115                                  115
#define RSVNA_REPCAP_GENERICDEVICE_GDEV116                                  116
#define RSVNA_REPCAP_GENERICDEVICE_GDEV117                                  117
#define RSVNA_REPCAP_GENERICDEVICE_GDEV118                                  118
#define RSVNA_REPCAP_GENERICDEVICE_GDEV119                                  119
#define RSVNA_REPCAP_GENERICDEVICE_GDEV120                                  120
#define RSVNA_REPCAP_GENERICDEVICE_GDEV121                                  121
#define RSVNA_REPCAP_GENERICDEVICE_GDEV122                                  122
#define RSVNA_REPCAP_GENERICDEVICE_GDEV123                                  123
#define RSVNA_REPCAP_GENERICDEVICE_GDEV124                                  124
#define RSVNA_REPCAP_GENERICDEVICE_GDEV125                                  125
#define RSVNA_REPCAP_GENERICDEVICE_GDEV126                                  126
#define RSVNA_REPCAP_GENERICDEVICE_GDEV127                                  127
#define RSVNA_REPCAP_GENERICDEVICE_GDEV128                                  128

//repcap Splitter
#define RSVNA_REPCAP_SPLITTER_SPLITTER1                                     1
#define RSVNA_REPCAP_SPLITTER_SPLITTER2                                     2

//repcap TrainSegment
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG1                                 1
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG2                                 2
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG3                                 3
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG4                                 4
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG5                                 5
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG6                                 6
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG7                                 7
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG8                                 8
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG9                                 9
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG10                                10
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG11                                11
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG12                                12
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG13                                13
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG14                                14
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG15                                15
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG16                                16
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG17                                17
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG18                                18
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG19                                19
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG20                                20
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG21                                21
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG22                                22
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG23                                23
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG24                                24
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG25                                25
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG26                                26
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG27                                27
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG28                                28
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG29                                29
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG30                                30
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG31                                31
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG32                                32
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG33                                33
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG34                                34
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG35                                35
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG36                                36
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG37                                37
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG38                                38
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG39                                39
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG40                                40
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG41                                41
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG42                                42
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG43                                43
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG44                                44
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG45                                45
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG46                                46
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG47                                47
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG48                                48
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG49                                49
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG50                                50
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG51                                51
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG52                                52
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG53                                53
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG54                                54
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG55                                55
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG56                                56
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG57                                57
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG58                                58
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG59                                59
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG60                                60
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG61                                61
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG62                                62
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG63                                63
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG64                                64
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG65                                65
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG66                                66
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG67                                67
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG68                                68
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG69                                69
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG70                                70
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG71                                71
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG72                                72
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG73                                73
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG74                                74
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG75                                75
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG76                                76
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG77                                77
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG78                                78
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG79                                79
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG80                                80
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG81                                81
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG82                                82
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG83                                83
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG84                                84
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG85                                85
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG86                                86
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG87                                87
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG88                                88
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG89                                89
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG90                                90
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG91                                91
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG92                                92
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG93                                93
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG94                                94
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG95                                95
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG96                                96
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG97                                97
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG98                                98
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG99                                99
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG100                               100
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG101                               101
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG102                               102
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG103                               103
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG104                               104
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG105                               105
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG106                               106
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG107                               107
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG108                               108
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG109                               109
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG110                               110
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG111                               111
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG112                               112
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG113                               113
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG114                               114
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG115                               115
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG116                               116
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG117                               117
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG118                               118
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG119                               119
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG120                               120
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG121                               121
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG122                               122
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG123                               123
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG124                               124
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG125                               125
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG126                               126
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG127                               127
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG128                               128
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG129                               129
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG130                               130
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG131                               131
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG132                               132
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG133                               133
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG134                               134
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG135                               135
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG136                               136
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG137                               137
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG138                               138
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG139                               139
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG140                               140
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG141                               141
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG142                               142
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG143                               143
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG144                               144
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG145                               145
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG146                               146
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG147                               147
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG148                               148
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG149                               149
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG150                               150
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG151                               151
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG152                               152
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG153                               153
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG154                               154
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG155                               155
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG156                               156
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG157                               157
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG158                               158
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG159                               159
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG160                               160
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG161                               161
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG162                               162
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG163                               163
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG164                               164
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG165                               165
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG166                               166
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG167                               167
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG168                               168
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG169                               169
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG170                               170
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG171                               171
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG172                               172
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG173                               173
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG174                               174
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG175                               175
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG176                               176
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG177                               177
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG178                               178
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG179                               179
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG180                               180
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG181                               181
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG182                               182
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG183                               183
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG184                               184
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG185                               185
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG186                               186
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG187                               187
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG188                               188
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG189                               189
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG190                               190
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG191                               191
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG192                               192
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG193                               193
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG194                               194
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG195                               195
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG196                               196
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG197                               197
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG198                               198
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG199                               199
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG200                               200
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG201                               201
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG202                               202
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG203                               203
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG204                               204
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG205                               205
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG206                               206
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG207                               207
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG208                               208
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG209                               209
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG210                               210
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG211                               211
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG212                               212
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG213                               213
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG214                               214
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG215                               215
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG216                               216
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG217                               217
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG218                               218
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG219                               219
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG220                               220
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG221                               221
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG222                               222
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG223                               223
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG224                               224
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG225                               225
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG226                               226
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG227                               227
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG228                               228
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG229                               229
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG230                               230
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG231                               231
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG232                               232
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG233                               233
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG234                               234
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG235                               235
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG236                               236
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG237                               237
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG238                               238
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG239                               239
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG240                               240
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG241                               241
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG242                               242
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG243                               243
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG244                               244
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG245                               245
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG246                               246
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG247                               247
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG248                               248
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG249                               249
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG250                               250
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG251                               251
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG252                               252
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG253                               253
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG254                               254
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG255                               255
#define RSVNA_REPCAP_TRAINSEGMENT_TRAINSEG256                               256
/* END GENERATE */

    /*****************************************************************************
     *- Hidden Attribute Declarations -------------------------------------------*
     *****************************************************************************/

#define RSVNA_ATTR_OPC_TIMEOUT               (RS_ATTR_OPC_TIMEOUT)           /* ViInt32 */
#define RSVNA_ATTR_IO_SESSION                (RS_ATTR_IO_SESSION)            /* ViSession */
#define RSVNA_ATTR_OPC_CALLBACK              (RS_ATTR_OPC_CALLBACK)          /* ViAddr */
#define RSVNA_ATTR_CHECK_STATUS_CALLBACK     (RS_ATTR_CHECK_STATUS_CALLBACK) /* ViAddr */

    /****************************************************************************
     *---------------- Constant definition  ------------------------------------*
     ****************************************************************************/
#define RSVNA_VAL_RECALL_SET_NEW    0
#define RSVNA_VAL_RECALL_SET_OPEN   1
#define RSVNA_VAL_RECALL_SET_SAVE   2
#define RSVNA_VAL_RECALL_SET_ACTIVE 3
#define RSVNA_VAL_RECALL_SET_CLOSE  4

#define RSVNA_VAL_UNFORMATTED      0
#define RSVNA_VAL_FORMATTED        1
#define RSVNA_VAL_UNFORMATTED_MATH 2
#define RSVNA_VAL_PULSE_PROFILE    3
#define RSVNA_VAL_FACTORY_CAL      4
#define RSVNA_VAL_UNCAL            5

#define RSVNA_VAL_FILE_MAN_CDRIVE    0
#define RSVNA_VAL_FILE_MAN_CDIR      1
#define RSVNA_VAL_FILE_MAN_MDIR      2
#define RSVNA_VAL_FILE_MAN_RDIR      3
#define RSVNA_VAL_FILE_MAN_COPY      4
#define RSVNA_VAL_FILE_MAN_MOVE      5
#define RSVNA_VAL_FILE_MAN_DELETE    6
#define RSVNA_VAL_FILE_MAN_CDIR_DEF  7

#define RSVNA_VAL_SINGLE_ENDED 0
#define RSVNA_VAL_DIFFERENTIAL 1
#define RSVNA_VAL_COMMON       2
#define RSVNA_VAL_SINGLE_ENDED_RAW 3

#define RSVNA_VAL_RATIO_B1_A1_SRC_PORT_1 0
#define RSVNA_VAL_RATIO_B2_A1_SRC_PORT_1 1
#define RSVNA_VAL_RATIO_B2_A2_SRC_PORT_2 2
#define RSVNA_VAL_RATIO_B1_A2_SRC_PORT_2 3

#define RSVNA_VAL_CALCOLLSTDGEN_THR    0
#define RSVNA_VAL_CALCOLLSTDGEN_OPEN   1
#define RSVNA_VAL_CALCOLLSTDGEN_SHORT  2
#define RSVNA_VAL_CALCOLLSTDGEN_MATCH  3
#define RSVNA_VAL_CALCOLLSTDGEN_NET    4
#define RSVNA_VAL_CALCOLLSTDGEN_ATT    5
#define RSVNA_VAL_CALCOLLSTDGEN_REFL   6
#define RSVNA_VAL_CALCOLLSTDGEN_LINE   7
#define RSVNA_VAL_CALCOLLSTDGEN_LINE_2 8
#define RSVNA_VAL_CALCOLLSTDGEN_LINE_3 9
#define RSVNA_VAL_CALCOLLSTDGEN_SLID   10
#define RSVNA_VAL_CALCOLLSTDGEN_OSH    11
#define RSVNA_VAL_CALCOLLSTDGEN_OSH_2  12
#define RSVNA_VAL_CALCOLLSTDGEN_OSH_3  13
#define RSVNA_VAL_CALCOLLSTDGEN_POWER  14

#define RSVNA_VAL_CALCOLLSTDGEN_UTHR 0

#define RSVNA_VAL_UTHR_AUTO     0
#define RSVNA_VAL_UTHR_MAN      1

#define RSVNA_VAL_CORR_MAN_COPY     0
#define RSVNA_VAL_CORR_MAN_APPLY    1
#define RSVNA_VAL_CORR_MAN_RESOLVE  2
#define RSVNA_VAL_CORR_MAN_MERGE    3
#define RSVNA_VAL_CORR_MAN_DELETE   4

#define RSVNA_VAL_CALTYPE_FNP   0
#define RSVNA_VAL_CALTYPE_SFTP  1
#define RSVNA_VAL_CALTYPE_FOP   2
#define RSVNA_VAL_CALTYPE_OPTP  3
#define RSVNA_VAL_CALTYPE_REFL 4
#define RSVNA_VAL_CALTYPE_RSH  5
#define RSVNA_VAL_CALTYPE_FRTR 6
#define RSVNA_VAL_CALTYPE_TOSM 7
#define RSVNA_VAL_CALTYPE_TOM  8
#define RSVNA_VAL_CALTYPE_TSM  9
#define RSVNA_VAL_CALTYPE_TRM  10
#define RSVNA_VAL_CALTYPE_TRL  11
#define RSVNA_VAL_CALTYPE_TNA  12
#define RSVNA_VAL_CALTYPE_UOSM 13
#define RSVNA_VAL_CALTYPE_FTR  14
#define RSVNA_VAL_CALTYPE_RTR  15
#define RSVNA_VAL_CALTYPE_PTOS  16
#define RSVNA_VAL_CALTYPE_PTOM  17
#define RSVNA_VAL_CALTYPE_PTSM  18
#define RSVNA_VAL_CALTYPE_PTRM  19
#define RSVNA_VAL_CALTYPE_PTRL  20
#define RSVNA_VAL_CALTYPE_PTNA  21
#define RSVNA_VAL_CALTYPE_PUOS  22
#define RSVNA_VAL_CALTYPE_ART  23
#define RSVNA_VAL_CALTYPE_ARPT 24

#define RSVNA_VAL_CALTYPE_PFNP 25
#define RSVNA_VAL_CALTYPE_VUOS 26
#define RSVNA_VAL_CALTYPE_VPU  27
#define RSVNA_VAL_CALTYPE_CPU  28
#define RSVNA_VAL_CALTYPE_MTOS 29
#define RSVNA_VAL_CALTYPE_MUOS 30
#define RSVNA_VAL_CALTYPE_MOSM 31
#define RSVNA_VAL_CALTYPE_MTRL 32
#define RSVNA_VAL_CALTYPE_UMD  33
#define RSVNA_VAL_CALTYPE_PUMD 34

#define RSVNA_VAL_CALSTD_MMTH   0
#define RSVNA_VAL_CALSTD_FFTH   1
#define RSVNA_VAL_CALSTD_MFTH   2
#define RSVNA_VAL_CALSTD_MMLI   3
#define RSVNA_VAL_CALSTD_FFLI   4
#define RSVNA_VAL_CALSTD_MFLI   5
#define RSVNA_VAL_CALSTD_MMLI_2 6
#define RSVNA_VAL_CALSTD_FFLI_2 7
#define RSVNA_VAL_CALSTD_MFLI_2 8
#define RSVNA_VAL_CALSTD_MMLI_3 9
#define RSVNA_VAL_CALSTD_FFLI_3 10
#define RSVNA_VAL_CALSTD_MFLI_3 11
#define RSVNA_VAL_CALSTD_MOSH   12
#define RSVNA_VAL_CALSTD_FOSH   13
#define RSVNA_VAL_CALSTD_MOSH_2 14
#define RSVNA_VAL_CALSTD_FOSH_2 15
#define RSVNA_VAL_CALSTD_MOSH_3 16
#define RSVNA_VAL_CALSTD_FOSH_3 17
#define RSVNA_VAL_CALSTD_MMAT   18
#define RSVNA_VAL_CALSTD_FFAT   19
#define RSVNA_VAL_CALSTD_MFAT   20
#define RSVNA_VAL_CALSTD_MMSN   21
#define RSVNA_VAL_CALSTD_FFSN   22
#define RSVNA_VAL_CALSTD_MFSN   23
#define RSVNA_VAL_CALSTD_MOP    24
#define RSVNA_VAL_CALSTD_FOP    25
#define RSVNA_VAL_CALSTD_MSH    26
#define RSVNA_VAL_CALSTD_FSH    27
#define RSVNA_VAL_CALSTD_MREF   28
#define RSVNA_VAL_CALSTD_FREF   29
#define RSVNA_VAL_CALSTD_MMTC   30
#define RSVNA_VAL_CALSTD_FMTC   31
#define RSVNA_VAL_CALSTD_MSM    32
#define RSVNA_VAL_CALSTD_FSM    33

#define RSVNA_VAL_A 0
#define RSVNA_VAL_B 1
#define RSVNA_VAL_A_PRIMED 2
#define RSVNA_VAL_B_PRIMED 3

#define RSVNA_VAL_DET_RMS  0
#define RSVNA_VAL_DET_PEAK 1
#define RSVNA_VAL_DET_AVG  2

#define RSVNA_VAL_WQUANTITY_A1_SRC_PORT_1 0
#define RSVNA_VAL_WQUANTITY_B1_SRC_PORT_1 1
#define RSVNA_VAL_WQUANTITY_B2_SRC_PORT_1 2
#define RSVNA_VAL_WQUANTITY_A2_SRC_PORT_2 3
#define RSVNA_VAL_WQUANTITY_B1_SRC_PORT_2 4
#define RSVNA_VAL_WQUANTITY_B2_SRC_PORT_2 5

#define RSVNA_VAL_SFACTOR_K  0
#define RSVNA_VAL_SFACTOR_U1 1
#define RSVNA_VAL_SFACTOR_U2 2

#define RSVNA_VAL_DC_MEAS_1V  0
#define RSVNA_VAL_DC_MEAS_10V 1

#define RSVNA_UNFORMATTED                           0
#define RSVNA_FORMATTED                             1
#define RSVNA_UNFORMATTED_MATH                      2
#define RSVNA_PULSE_PROFILE                         3

#define RSVNA_VAL_COMPLEX 0
#define RSVNA_VAL_LINP    1
#define RSVNA_VAL_LOGP    2

#define RSVNA_VAL_POIN 0
#define RSVNA_VAL_COMM 1

#define RSVNA_VAL_SEM  0
#define RSVNA_VAL_TAB  2
#define RSVNA_VAL_SPAC 3

#define RSVNA_VAL_CHANNEL_TRACE_SINGLE 0
#define RSVNA_VAL_CHANNEL_TRACE_ALL    1

#define RSVNA_VAL_DISP_TRAC_DATA 0
#define RSVNA_VAL_DISP_TRAC_MEM  1
#define RSVNA_VAL_DISP_TRAC_SING 2

#define RSVNA_VAL_CONNECTION_MODE_TEM      0
#define RSVNA_VAL_CONNECTION_MODE_WGUIDE   1

#define RSVNA_VAL_CONNECTION_CONNECTOR_GENDER    0
#define RSVNA_VAL_CONNECTION_CONNECTOR_NGENDER   1

#define RSVNA_VAL_CONNECTOR_GENDER_MALE    0
#define RSVNA_VAL_CONNECTOR_GENDER_FEMALE  1

#define RSVNA_VAL_CAL_DATA_DIRECTIVITY  0
#define RSVNA_VAL_CAL_DATA_SRCMATCH     1
#define RSVNA_VAL_CAL_DATA_REFLTRACK    2
#define RSVNA_VAL_CAL_DATA_LOADMATCH    3
#define RSVNA_VAL_CAL_DATA_TRANSTRACK   4
#define RSVNA_VAL_CAL_DATA_G11          5
#define RSVNA_VAL_CAL_DATA_G12          6
#define RSVNA_VAL_CAL_DATA_G21          7
#define RSVNA_VAL_CAL_DATA_G22          8
#define RSVNA_VAL_CAL_DATA_H11          9
#define RSVNA_VAL_CAL_DATA_H12          10
#define RSVNA_VAL_CAL_DATA_H21          11
#define RSVNA_VAL_CAL_DATA_H22          12
#define RSVNA_VAL_CAL_DATA_Q11          13
#define RSVNA_VAL_CAL_DATA_Q12          14
#define RSVNA_VAL_CAL_DATA_Q21          15
#define RSVNA_VAL_CAL_DATA_Q22          16
#define RSVNA_VAL_CAL_DATA_PREL         17
#define RSVNA_VAL_CAL_DATA_PABS         18
#define RSVNA_VAL_CAL_DATA_L1           19
#define RSVNA_VAL_CAL_DATA_L2           20

#define RSVNA_VAL_LIMIT_ON    1
#define RSVNA_VAL_LIMIT_ALL   3

#define RSVNA_VAL_LOCAL_OSCILLATOR_NEGATIVE 0
#define RSVNA_VAL_LOCAL_OSCILLATOR_POSITIVE 1
#define RSVNA_VAL_LOCAL_OSCILLATOR_AUTO     2

#define RSVNA_VAL_DRIVE_PORTS  0
#define RSVNA_VAL_CHANNEL_BITS 1

#define RSVNA_VAL_RESTORE  0
#define RSVNA_VAL_MAXIMIZE 1

#define RSVNA_VAL_REG_COND   3
#define RSVNA_VAL_REG_EVENT  4

#define RSVNA_VAL_TYPE_BPAS_IMP  0
#define RSVNA_VAL_TYPE_LPAS_IMP  1
#define RSVNA_VAL_TYPE_LPAS_STEP 2

#define RSVNA_VAL_HORIZONTAL_LEFT  0
#define RSVNA_VAL_HORIZONTAL_MID   1
#define RSVNA_VAL_HORIZONTAL_RIGHT 2

#define RSVNA_VAL_VERTICAL_TOP    0
#define RSVNA_VAL_VERTICAL_MID    1
#define RSVNA_VAL_VERTICAL_BOTTOM 2

#define RSVNA_VAL_SWITCH_MATRIX_START  0
#define RSVNA_VAL_SWITCH_MATRIX_ABORT  1
#define RSVNA_VAL_SWITCH_MATRIX_END    2

#define RSVNA_VAL_MEAN     0
#define RSVNA_VAL_STDDEV   1
#define RSVNA_VAL_MAX      2
#define RSVNA_VAL_MIN      3
#define RSVNA_VAL_RMS      4
#define RSVNA_VAL_PTPEAK   5
#define RSVNA_VAL_ELENGTH  6
#define RSVNA_VAL_PDELAY   7
#define RSVNA_VAL_GAIN     8
#define RSVNA_VAL_SLOPE    9
#define RSVNA_VAL_FLATNESS 10

#define RSVNA_VAL_SAVE   0
#define RSVNA_VAL_RECALL 1

#define RSVNA_VAL_MARKER           0
#define RSVNA_VAL_REFERENCE_MARKER 1

#define RSVNA_VAL_SET_BY_MARKER_START  0
#define RSVNA_VAL_SET_BY_MARKER_STOP   1
#define RSVNA_VAL_SET_BY_MARKER_CENTER 2
#define RSVNA_VAL_SET_BY_MARKER_SPAN   3

#define RSVNA_VAL_CALUNIT_STD_THR  0
#define RSVNA_VAL_CALUNIT_STD_OPEN 1
#define RSVNA_VAL_CALUNIT_STD_SHOR 2
#define RSVNA_VAL_CALUNIT_STD_MATC 3

#define RSVNA_VAL_S11 0
#define RSVNA_VAL_S12 1
#define RSVNA_VAL_S21 2
#define RSVNA_VAL_S22 3

#define RSVNA_VAL_CONNECTOR_KIT_N50   0
#define RSVNA_VAL_CONNECTOR_KIT_N75   1
#define RSVNA_VAL_CONNECTOR_KIT_PC7   2
#define RSVNA_VAL_CONNECTOR_KIT_PC35  3
#define RSVNA_VAL_CONNECTOR_KIT_PC292 4
#define RSVNA_VAL_CONNECTOR_KIT_USER1 5
#define RSVNA_VAL_CONNECTOR_KIT_USER2 6
#define RSVNA_VAL_CONNECTOR_KIT_SMA   7

#define RSVNA_VAL_OPEN  0
#define RSVNA_VAL_SHORT 1
#define RSVNA_VAL_MATCH 2

#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_IMM         0
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_1       1
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_2       2
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_1_AND_2 3
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_1_OR_2  4
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_MANUAL      5
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_A       6
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_B       7
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_C       8
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_EXT_D       9
#define RSVNA_VAL_MULTIPLE_TRIGGER_SOURCE_UPORT       10

#define RSVNA_VAL_VNET_FUNC_DEEMBED  0
#define RSVNA_VAL_VNET_FUNC_EMBED    1

#define RSVNA_VAL_VNET_PARAM_C  0
#define RSVNA_VAL_VNET_PARAM_L  1
#define RSVNA_VAL_VNET_PARAM_R  2
#define RSVNA_VAL_VNET_PARAM_G  3

#define RSVNA_VAL_VNET_PARAM_PMAIN     0
#define RSVNA_VAL_VNET_PARAM_PSECOND   1

#define RSVNA_VAL_ELEMENT_BACKGROUND      1
#define RSVNA_VAL_ELEMENT_TEXT            2
#define RSVNA_VAL_ELEMENT_SELTEXT         3
#define RSVNA_VAL_ELEMENT_GRID            4
#define RSVNA_VAL_ELEMENT_REFLINE         5
#define RSVNA_VAL_ELEMENT_ALLMAKERS       6
#define RSVNA_VAL_ELEMENT_HLINE           7
#define RSVNA_VAL_ELEMENT_DTITLE          8
#define RSVNA_VAL_ELEMENT_LIMITFAILTRACE  9
#define RSVNA_VAL_ELEMENT_LIMITLINEOFF    10
#define RSVNA_VAL_ELEMENT_LIMITLINEUPPER  11
#define RSVNA_VAL_ELEMENT_LIMITLINELOWER  12
#define RSVNA_VAL_ELEMENT_TRACE1          13
#define RSVNA_VAL_ELEMENT_TRACE2          14
#define RSVNA_VAL_ELEMENT_TRACE3          15
#define RSVNA_VAL_ELEMENT_TRACE4          16
#define RSVNA_VAL_ELEMENT_TRACE5          17
#define RSVNA_VAL_ELEMENT_TRACE6          18
#define RSVNA_VAL_ELEMENT_TRACE7          19
#define RSVNA_VAL_ELEMENT_TRACE8          20
#define RSVNA_VAL_ELEMENT_TRACE9          21
#define RSVNA_VAL_ELEMENT_TRACE10         22
#define RSVNA_VAL_ELEMENT_TRACE11         23
#define RSVNA_VAL_ELEMENT_TRACE12         24
#define RSVNA_VAL_ELEMENT_TRACE13         25
#define RSVNA_VAL_ELEMENT_TRACE14         26
#define RSVNA_VAL_ELEMENT_TRACE15         27
#define RSVNA_VAL_ELEMENT_TRACE16         28

#define RSVNA_VAL_TRACE_STYLE_SOLID     0
#define RSVNA_VAL_TRACE_STYLE_DASHED    1
#define RSVNA_VAL_TRACE_STYLE_DOTTED    2
#define RSVNA_VAL_TRACE_STYLE_DDOTTED   3
#define RSVNA_VAL_TRACE_STYLE_DDDOTTED  4

#define RSVNA_VAL_PAE_DC_INPUT_MAIN      1
#define RSVNA_VAL_PAE_DC_INPUT_SECONDARY 2

#define RSVNA_VAL_PAE_PARAMETER_TYPE_I 0
#define RSVNA_VAL_PAE_PARAMETER_TYPE_R 1
#define RSVNA_VAL_PAE_PARAMETER_TYPE_U 2

#define RSVNA_VAL_SWEEP_AXIS_FREQUENCY 0
#define RSVNA_VAL_SWEEP_AXIS_POWER     1

#define RSVNA_VAL_PWR_CAL_PORT  0
#define RSVNA_VAL_PWR_CAL_GEN   1
#define RSVNA_VAL_PWR_CAL_A1    2
#define RSVNA_VAL_PWR_CAL_A2    3
#define RSVNA_VAL_PWR_CAL_A3    4
#define RSVNA_VAL_PWR_CAL_A4    5
#define RSVNA_VAL_PWR_CAL_ESRC1 6
#define RSVNA_VAL_PWR_CAL_ESRC2 7

#define RSVNA_VAL_SETTING_TYPE_WAVE         0
#define RSVNA_VAL_SETTING_TYPE_START        1
#define RSVNA_VAL_SETTING_TYPE_STOP         2
#define RSVNA_VAL_SETTING_TYPE_POINTS       3
#define RSVNA_VAL_SETTING_TYPE_STYPE        4
#define RSVNA_VAL_SETTING_TYPE_ATTENUATION  5
#define RSVNA_VAL_SETTING_TYPE_CPOWER       6
#define RSVNA_VAL_SETTING_TYPE_CFREQUENCY   7
#define RSVNA_VAL_SETTING_TYPE_TIMESTAMP    8
#define RSVNA_VAL_SETTING_TYPE_TVNA         9
#define RSVNA_VAL_SETTING_TYPE_MVNA         10
#define RSVNA_VAL_SETTING_TYPE_MTEST        11

#define RSVNA_VAL_PWR_CAL_AWAVE 0
#define RSVNA_VAL_PWR_CAL_BWAVE 1
#define RSVNA_VAL_PWR_CAL_B1    2
#define RSVNA_VAL_PWR_CAL_B2    3
#define RSVNA_VAL_PWR_CAL_B3    4
#define RSVNA_VAL_PWR_CAL_B4    5

#define RSVNA_VAL_POWER_METER_NO_COEFFICIENTS      0
#define RSVNA_VAL_POWER_METER_TWO_PORT_DUT         1
#define RSVNA_VAL_POWER_METER_TWO_PORT_POWER_METER 2

#define RSVNA_VAL_SWEEP_TYPE_SWEEP 0
#define RSVNA_VAL_SWEEP_TYPE_FIXED 1

#define RSVNA_VAL_PWR_CAL_OFFSET_ONLY   0
#define RSVNA_VAL_PWR_CAL_OFFSET_CPADD  1

#define RSVNA_VAL_MIXER_STAGE1 1
#define RSVNA_VAL_MIXER_STAGE2 2

#define RSVNA_VAL_MIXER_LO_NONE      0
#define RSVNA_VAL_MIXER_LO_PORT      1
#define RSVNA_VAL_MIXER_LO_GENERATOR 2

#define RSVNA_VAL_MIXER_FIXED 0
#define RSVNA_VAL_MIXER_BASE  1

#define RSVNA_VAL_MIXER_FUND  1

#define RSVNA_VAL_TRIGGER_AUX1 1
#define RSVNA_VAL_TRIGGER_AUX2 2

#define RSVNA_VAL_INTERMODULATION_MAJOR 0
#define RSVNA_VAL_INTERMODULATION_UPPER 1
#define RSVNA_VAL_INTERMODULATION_LOWER 2

#define RSVNA_VAL_INTERMODULATION_OUT 0
#define RSVNA_VAL_INTERMODULATION_IN 1

#define RSVNA_VAL_INTERMODULATION_LOWER_TONE 0
#define RSVNA_VAL_INTERMODULATION_UPPER_TONE 1
#define RSVNA_VAL_INTERMODULATION_RECEIVER 2

#define RSVNA_VAL_SOURCE_NONE 0
#define RSVNA_VAL_SOURCE_PORT 1
#define RSVNA_VAL_SOURCE_GENERATOR 2

#define RSVNA_VAL_OFFSET_COMPENSATION_AUTO_LENGTH 0
#define RSVNA_VAL_OFFSET_COMPENSATION_AUTO_LENGTH_LOSS 1
#define RSVNA_VAL_OFFSET_COMPENSATION_DIRECT 2

#define RSVNA_VAL_FIXTURE_OPEN  0
#define RSVNA_VAL_FIXTURE_SHORT 1

#define RSVNA_VAL_UI_PORT_A 0
#define RSVNA_VAL_UI_PORT_B 1
#define RSVNA_VAL_UI_PORT_C 2
#define RSVNA_VAL_UI_PORT_D 3
#define RSVNA_VAL_UI_PORT_E 4
#define RSVNA_VAL_UI_PORT_F 5
#define RSVNA_VAL_UI_PORT_G 6
#define RSVNA_VAL_UI_PORT_H 7

#define RSVNA_VAL_INDEX     0
#define RSVNA_VAL_RTRIGGER  1

#define RSVNA_VAL_MIXER_MULTIPLICATION_PORT_RF  0
#define RSVNA_VAL_MIXER_MULTIPLICATION_PORT_LO1 1
#define RSVNA_VAL_MIXER_MULTIPLICATION_PORT_LO2 2

#define RSVNA_VAL_FREQ_MIN 0
#define RSVNA_VAL_FREQ_MAX 1

#define RSVNA_VAL_LAYOUT_FORMAT_HORIZONTAL 0
#define RSVNA_VAL_LAYOUT_FORMAT_VERTICAL   1

#define RSVNA_VAL_REVERT_DISABLING 0
#define RSVNA_VAL_DISABLE          1

#define RSVNA_VAL_GENDER_NGENDER 0
#define RSVNA_VAL_GENDER_MALE    1
#define RSVNA_VAL_GENDER_FEMALE  2

#define RSVNA_VAL_ALL_PORTS      0
#define RSVNA_VAL_SELECTED_PORTS 1

#define RSVNA_VAL_CONNECTOR 0
#define RSVNA_VAL_LABEL     1
#define RSVNA_VAL_NAME      2
#define RSVNA_VAL_GENDER    3

#define RSVNA_VAL_PORT_NUMBERS_NOT_INTERCHANGED 0
#define RSVNA_VAL_PORT_NUMBERS_INTERCHANGED     1

#define RSVNA_VAL_CAL_PARAM_ACAL            0
#define RSVNA_VAL_CAL_PARAM_START           1
#define RSVNA_VAL_CAL_PARAM_STOP            2
#define RSVNA_VAL_CAL_PARAM_POINTS          3
#define RSVNA_VAL_CAL_PARAM_POWER           4
#define RSVNA_VAL_CAL_PARAM_SWEEP_TYPE      5
#define RSVNA_VAL_CAL_PARAM_BANDWIDTH       6
#define RSVNA_VAL_CAL_PARAM_POINT_DELAY     7
#define RSVNA_VAL_CAL_PARAM_ATTENUATION     8
#define RSVNA_VAL_CAL_PARAM_CAL_TYPE        9
#define RSVNA_VAL_CAL_PARAM_PORTS           10
#define RSVNA_VAL_CAL_PARAM_SPORT           11
#define RSVNA_VAL_CAL_PARAM_THROUGH         12
#define RSVNA_VAL_CAL_PARAM_TIMESTAMP       13
#define RSVNA_VAL_CAL_PARAM_TVNA            14
#define RSVNA_VAL_CAL_PARAM_MVNA            15
#define RSVNA_VAL_CAL_PARAM_MTEST           16
#define RSVNA_VAL_CAL_PARAM_TIMESTAMP_LOCAL 17
#define RSVNA_VAL_CAL_PARAM_CALKIT          18
#define RSVNA_VAL_CAL_PARAM_SWEEP_MODE      19

#define RSVNA_VAL_COMMON_IMPEDANCE 0
#define RSVNA_VAL_PORT_IMPEDANCE   1

#define RSVNA_VAL_UNSEGMENTED_SWEEP 0
#define RSVNA_VAL_SEGMENTED_SWEEP   1

#define RSVNA_VAL_AGILENT 0
#define RSVNA_VAL_ZVR     1

#define RSVNA_VAL_FREQ_START 0
#define RSVNA_VAL_FREQ_STOP     1

#define RSVNA_VAL_PMET_EXT_GEN  0
#define RSVNA_VAL_PMET_ANALYZER 1

#define RSVNA_VAL_GAIN_COMPRESSION_POWER_IN    0
#define RSVNA_VAL_GAIN_COMPRESSION_POWER_OUT   1
#define RSVNA_VAL_GAIN_COMPRESSION_S_PARAMETER 2

#define RSVNA_VAL_LTO 0
#define RSVNA_VAL_LTI 1
#define RSVNA_VAL_UTO 2
#define RSVNA_VAL_UTI 3

    /* Fast Mode */
#define RSVNA_VAL_FAST_SWEEP_NORMAL 0
#define RSVNA_VAL_FAST_SWEEP_FAST   1

#define RSVNA_VAL_ALL            0
#define RSVNA_VAL_CHANNEL        1
#define RSVNA_VAL_WINDOW         2
#define RSVNA_VAL_SEGMENT        3
#define RSVNA_VAL_LIMIT_SEGMENT  4
#define RSVNA_VAL_RIPPLE_SEGMENT 5
#define RSVNA_VAL_GROUP          6

    /* Distance to Fault */
#define RSVNA_VAL_FIXTURE_TOOL_SFD 0
#define RSVNA_VAL_FIXTURE_TOOL_ISD 1

#define RSVNA_VAL_FIXTURE_PORTS_TO_SKIP_NONE   0
#define RSVNA_VAL_FIXTURE_PORTS_TO_SKIP_LEFT   1
#define RSVNA_VAL_FIXTURE_PORTS_TO_SKIP_RIGHT  2
#define RSVNA_VAL_FIXTURE_PORTS_TO_SKIP_MANUAL 3

#define RSVNA_RFFE_PIN_DATA  0
#define RSVNA_RFFE_PIN_CLOCK 1
#define RSVNA_RFFE_PIN_VIO   2

    /* Set Status Register */
#define RSVNA_VAL_STATUS_REGISTER_PRESET 0
#define RSVNA_VAL_STATUS_REGISTER_SET    1

    /* Backwards compatibility */
#define RSVNA_ATTR_SWEEP_SEGMET_INSERT_TIME_MODE RSVNA_ATTR_SWEEP_SEGMENT_INSERT_TIME_MODE

#define RSVNA_VAL_HARMONIC_MEASUREMENT_THD_F 0
#define RSVNA_VAL_HARMONIC_MEASUREMENT_THD_R 1

#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_DELAY        0
#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_PHASE        1
#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_DERIVATIVE   2
#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_INPUT_LOWER  3
#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_INPUT_UPPER  4
#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_OUTPUT_LOWER 5
#define RSZNX_VAL_TWO_TONE_GROUP_DELAY_MEASUREMENT_OUTPUT_UPPER 6

    /****************************************************************************
    *---------------- Instrument Driver Function Declarations -----------------*
     ****************************************************************************/

ViStatus _VI_FUNC rsvna_init(ViRsrc     resourceName,
                             ViBoolean  IDQuery,
                             ViBoolean  resetDevice,
                             ViSession* instrSession);

ViStatus _VI_FUNC rsvna_InitWithOptions(ViRsrc     resourceName,
                                        ViBoolean  IDQuery,
                                        ViBoolean  resetDevice,
                                        ViString   optionString,
                                        ViSession* instrSession);

ViStatus _VI_FUNC rsvna_FileListofFavoriteRecallSet(ViSession instrSession,
                                                    ViInt32   favId,
                                                    ViString  fileName);

ViStatus _VI_FUNC rsvna_RecallSetManager(ViSession instrSession,
                                         ViInt32   operationToBePerformed,
                                         ViString  recallSetName);

ViStatus _VI_FUNC rsvna_Print(ViSession instrSession,
                              ViString  printerName);

ViStatus _VI_FUNC rsvna_PrintToFile(ViSession instrSession,
                                    ViString  fileName,
                                    ViInt32   fileFormat);

ViStatus _VI_FUNC rsvna_ConfigurePrintPageSetup(ViSession instrSession,
                                                ViInt32   printDiagramArea,
                                                ViBoolean useColor,
                                                ViInt32   pageOrientation,
                                                ViReal64  leftMargin,
                                                ViReal64  rightMargin,
                                                ViReal64  topMargin,
                                                ViReal64  bottomMargin);

ViStatus _VI_FUNC rsvna_ConfigurePrintColorScheme(ViSession instrSession,
                                                  ViInt32   colorScheme);

ViStatus _VI_FUNC rsvna_FileManager(ViSession instrSession,
                                    ViInt32   operationToBePerformed,
                                    ViString  source,
                                    ViString  destination);

ViStatus _VI_FUNC rsvna_GetCurrentDirectory(ViSession instrSession,
                                            ViChar    currentDirectory[]);

ViStatus _VI_FUNC rsvna_FileCatalog(ViSession instrSession,
                                    ViString  directory,
                                    ViChar    catalog[],
                                    ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_ReadToFileFromInstrument(ViSession instrSession,
                                                 ViString  source,
                                                 ViString  destination);

ViStatus _VI_FUNC rsvna_WriteFromFileToInstrument(ViSession instrSession,
                                                  ViString  source,
                                                  ViString  destination);

ViStatus _VI_FUNC rsvna_FileStoreDistancetoFaultCableTypes(ViSession instrSession,
                                                           ViString  directory);

ViStatus _VI_FUNC rsvna_FileLoadDistancetoFaultCableTypes(ViSession instrSession,
                                                          ViString  directory);

ViStatus _VI_FUNC rsvna_FileAddedComment(ViSession instrSession,
                                         ViString  comment);

ViStatus _VI_FUNC rsvna_ConfigureFileSymmetricParams(ViSession instrSession,
                                                     ViBoolean symmetricParams);

ViStatus _VI_FUNC rsvna_ConfigureFileDecimalPlaces(ViSession instrSession,
                                                   ViInt32   dataDecimalPlaces,
                                                   ViInt32   stimulusDecimalPlaces);

ViStatus _VI_FUNC rsvna_ConfigureParametersConversion(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   conversion);

ViStatus _VI_FUNC rsvna_DefineSweepandMeasurement(ViSession instrSession,
                                                  ViString  sweepType);

ViStatus _VI_FUNC rsvna_DefineSweepandMeasurementWithArguments(
                            ViSession instrSession,
                            ViString  sweepType,
                            ViInt32   arg1,
                            ViInt32   arg2);

ViStatus _VI_FUNC rsvna_QuerySweepandMeasurementDefinition(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   bufferSize,
                                                           ViChar    sweepType[]);

ViStatus _VI_FUNC rsvna_SelectSParameters(ViSession instrSession,
                                          ViInt32   channel,
                                          ViString  traceName,
                                          ViInt32   outPort,
                                          ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectMoreSParameters(ViSession instrSession,
                                              ViInt32   channel,
                                              ViString  traceName,
                                              ViInt32   outMode,
                                              ViInt32   outPort,
                                              ViInt32   inMode,
                                              ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectMoreSParametersWithDetector(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViString  traceName,
                                                          ViInt32   outMode,
                                                          ViInt32   outPort,
                                                          ViInt32   inMode,
                                                          ViInt32   inPort,
                                                          ViInt32   detector);

ViStatus _VI_FUNC rsvna_SelectBalancedPortsSParameters(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViString  traceName,
                                                       ViInt32   outMode,
                                                       ViInt32   outPort,
                                                       ViInt32   inMode,
                                                       ViInt32   inPort);

ViStatus _VI_FUNC rsvna_CreateTraceSParametersGroup(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   numberOfLogicalPorts,
                                                    ViInt32   logicalPorts[]);

ViStatus _VI_FUNC rsvna_SelectRatios(ViSession instrSession,
                                     ViInt32   channel,
                                     ViString  traceName,
                                     ViInt32   ratios);

ViStatus _VI_FUNC rsvna_SelectMoreRatios(ViSession instrSession,
                                         ViInt32   channel,
                                         ViString  traceName,
                                         ViInt32   sourcePort,
                                         ViInt32   numeratorType,
                                         ViInt32   numeratorPortNumber,
                                         ViInt32   denominatorType,
                                         ViInt32   denominatorPortNumber);

ViStatus _VI_FUNC rsvna_SelectMoreRatiosWithDetector(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViString  traceName,
                                                     ViInt32   sourcePort,
                                                     ViInt32   numeratorType,
                                                     ViInt32   numeratorPortNumber,
                                                     ViInt32   denominatorType,
                                                     ViInt32   denominatorPortNumber,
                                                     ViInt32   detector,
                                                     ViReal64  measurementTime);

ViStatus _VI_FUNC rsvna_SelectWaveQuantities(ViSession instrSession,
                                             ViInt32   channel,
                                             ViString  traceName,
                                             ViInt32   waveQuantities);

ViStatus _VI_FUNC rsvna_SelectMoreWaveQuantities(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViString  traceName,
                                                 ViInt32   waveQuantityType,
                                                 ViInt32   waveQuantityPortNumber,
                                                 ViInt32   sourcePort,
                                                 ViInt32   showAs);

ViStatus _VI_FUNC rsvna_SelectMoreWaveQuantitiesWithDetector(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViString  traceName,
                                                             ViInt32   waveQuantityType,
                                                             ViInt32   waveQuantityPortNumber,
                                                             ViInt32   sourcePort,
                                                             ViInt32   showAs,
                                                             ViInt32   detector,
                                                             ViReal64  measurementTime);

ViStatus _VI_FUNC rsvna_ConfigureWaveQuantitiesPhaseNormalizationEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_SelectImpedances(ViSession instrSession,
                                         ViInt32   channel,
                                         ViString  traceName,
                                         ViInt32   outPort,
                                         ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectMoreImpedances(ViSession instrSession,
                                             ViInt32   channel,
                                             ViString  traceName,
                                             ViInt32   outMode,
                                             ViInt32   outPort,
                                             ViInt32   inMode,
                                             ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectAdmittances(ViSession instrSession,
                                          ViInt32   channel,
                                          ViString  traceName,
                                          ViInt32   outPort,
                                          ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectMoreAdmittances(ViSession instrSession,
                                              ViInt32   channel,
                                              ViString  traceName,
                                              ViInt32   outMode,
                                              ViInt32   outPort,
                                              ViInt32   inMode,
                                              ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectZParameters(ViSession instrSession,
                                          ViInt32   channel,
                                          ViString  traceName,
                                          ViInt32   outMode,
                                          ViInt32   outPort,
                                          ViInt32   inMode,
                                          ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectYParameters(ViSession instrSession,
                                          ViInt32   channel,
                                          ViString  traceName,
                                          ViInt32   outMode,
                                          ViInt32   outPort,
                                          ViInt32   inMode,
                                          ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectImbalance(ViSession instrSession,
                                        ViInt32   channel,
                                        ViString  traceName,
                                        ViInt32   outPort,
                                        ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectStabilityFactors(ViSession instrSession,
                                               ViInt32   channel,
                                               ViString  traceName,
                                               ViInt32   DUTOut,
                                               ViInt32   DUTIn,
                                               ViInt32   stabilityFactor);

ViStatus _VI_FUNC rsvna_SelectDCMeasurement(ViSession instrSession,
                                            ViInt32   channel,
                                            ViString  traceName,
                                            ViInt32   DCMeas);

ViStatus _VI_FUNC rsvna_ConfigureDCRange(ViSession instrSession,
                                         ViInt32   channel,
                                         ViInt32   DCInput,
                                         ViInt32   range);

ViStatus _VI_FUNC rsvna_ConfigurePAEMeasurementType(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   measurementType);

ViStatus _VI_FUNC rsvna_ConfigurePAEDCInput(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   type,
                                            ViInt32   DCInput);

ViStatus _VI_FUNC rsvna_ConfigurePAEParameters(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   parameterType,
                                               ViReal64  parameterValue);

ViStatus _VI_FUNC rsvna_DefineBalancedPort(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   logicalPort,
                                           ViInt32   physicalPort1,
                                           ViInt32   physicalPort2);

ViStatus _VI_FUNC rsvna_DeleteBalancedPort(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   logicalPort);

ViStatus _VI_FUNC rsvna_DeleteAllBalancedPorts(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureLogicalPortReferenceImpedanceCommonMode(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   logicalPort,
                            ViReal64  real,
                            ViReal64  imaginary);

ViStatus _VI_FUNC rsvna_ConfigureLogicalPortReferenceImpedanceDifferentialMode(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   logicalPort,
                            ViReal64  real,
                            ViReal64  imaginary);

ViStatus _VI_FUNC rsvna_ConfigureLogicalPortReferenceImpedanceRenormalization(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   theory);

ViStatus _VI_FUNC rsvna_ConfigureLogicalPortReferenceImpedanceDefault(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   logicalPort,
                            ViBoolean useDefault);

ViStatus _VI_FUNC rsvna_SelectIntermodulationProducts(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViString  traceName,
                                                      ViInt32   order,
                                                      ViInt32   side,
                                                      ViInt32   DUTPort,
                                                      ViBoolean relative_dBc);

ViStatus _VI_FUNC rsvna_SelectInterceptPoints(ViSession instrSession,
                                              ViInt32   channel,
                                              ViString  traceName,
                                              ViInt32   order,
                                              ViInt32   side,
                                              ViInt32   DUTPort);

ViStatus _VI_FUNC rsvna_SelectNoiseLevel(ViSession instrSession,
                                         ViInt32   channel,
                                         ViString  traceName);

ViStatus _VI_FUNC rsvna_SelectMainTone(ViSession instrSession,
                                       ViInt32   channel,
                                       ViString  traceName,
                                       ViInt32   tone);

ViStatus _VI_FUNC rsvna_SelectPowerSensorMeasurement(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViString  traceName,
                                                     ViInt32   powerSensor,
                                                     ViInt32   source,
                                                     ViInt32   sourcePort);

ViStatus _VI_FUNC rsvna_ConfigureInfiniteAveraging(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean enabled,
                                                   ViInt32   mode);

ViStatus _VI_FUNC rsvna_SelectNoiseFigure(ViSession instrSession,
                                          ViInt32   channel,
                                          ViString  traceName,
                                          ViInt32   outPort,
                                          ViInt32   inPort);

ViStatus _VI_FUNC rsvna_SelectNoiseFigureDUTGain(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViString  traceName,
                                                 ViInt32   bWave,
                                                 ViInt32   aWave,
                                                 ViInt32   sourcePort);

ViStatus _VI_FUNC rsvna_SelectReceiverNoiseFigure(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViString  traceName,
                                                  ViInt32   port);

ViStatus _VI_FUNC rsvna_SelectNoiseDensity(ViSession instrSession,
                                           ViInt32   channel,
                                           ViString  traceName,
                                           ViInt32   port);

ViStatus _VI_FUNC rsvna_SelectSpectrum(ViSession instrSession,
                                       ViInt32   channel,
                                       ViString  traceName,
                                       ViInt32   port);

ViStatus _VI_FUNC rsvna_SelectSpectrumReference(ViSession instrSession,
                                                ViInt32   channel,
                                                ViString  traceName,
                                                ViInt32   port);

ViStatus _VI_FUNC rsvna_SelectGainCompression(ViSession instrSession,
                                              ViInt32   channel,
                                              ViString  traceName,
                                              ViInt32   measurement,
                                              ViReal64  compressionValue);

ViStatus _VI_FUNC rsvna_SelectHarmonicMeasurement(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViString  traceName,
                                                  ViInt32   order,
                                                  ViInt32   receiverPort,
                                                  ViInt32   sourcePort,
                                                  ViBoolean relative);

ViStatus _VI_FUNC rsvna_SelectHarmonicMeasurementTHD(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViString  traceName,
                                                     ViInt32   order,
                                                     ViInt32   receiverPort,
                                                     ViInt32   sourcePort,
                                                     ViInt32   mode);

ViStatus _VI_FUNC rsvna_SelectTwoToneGroupDelayMeasurement(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViString  traceName,
                                                           ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigureTraceFormat(ViSession instrSession,
                                             ViInt32   channel_Trace,
                                             ViInt32   format,
                                             ViInt32   aperturePoints);

ViStatus _VI_FUNC rsvna_AutoScaleTrace(ViSession instrSession,
                                       ViString  traceName);

ViStatus _VI_FUNC rsvna_ConfigureTrace(ViSession instrSession,
                                       ViString  traceName,
                                       ViReal64  scaleDivisions,
                                       ViReal64  referenceLevel,
                                       ViReal64  referencePosition);

ViStatus _VI_FUNC rsvna_ConfigureTraceMaxMin(ViSession instrSession,
                                             ViString  traceName,
                                             ViReal64  top,
                                             ViReal64  bottom);

ViStatus _VI_FUNC rsvna_TraceAdd(ViSession instrSession,
                                 ViInt32   channel,
                                 ViString  traceName);

ViStatus _VI_FUNC rsvna_ConfigureTraceDisplayState(ViSession instrSession,
                                                   ViInt32   traceType,
                                                   ViString  singleTraceName,
                                                   ViBoolean showTrace);

ViStatus _VI_FUNC rsvna_ConfigureTraceDisplayLabel(ViSession instrSession,
                                                   ViString  singleTraceName,
                                                   ViBoolean displayed);

ViStatus _VI_FUNC rsvna_TraceAddDiagramArea(ViSession instrSession,
                                            ViInt32   window,
                                            ViInt32   channel,
                                            ViString  traceName);

ViStatus _VI_FUNC rsvna_TraceAssignDiagramArea(ViSession instrSession,
                                               ViInt32   window,
                                               ViString  traceName);

ViStatus _VI_FUNC rsvna_TraceUnassignDiagramArea(ViSession instrSession,
                                                 ViInt32   window,
                                                 ViInt32   window_Trace);

ViStatus _VI_FUNC rsvna_TraceSelect(ViSession instrSession,
                                    ViInt32   channel,
                                    ViString  traceName);

ViStatus _VI_FUNC rsvna_TraceDelete(ViSession instrSession,
                                    ViInt32   channel,
                                    ViString  traceName);

ViStatus _VI_FUNC rsvna_TraceDeleteAll(ViSession instrSession,
                                       ViInt32   channel);

ViStatus _VI_FUNC rsvna_TraceDeleteAllChannels(ViSession instrSession);

ViStatus _VI_FUNC rsvna_TraceList(ViSession instrSession,
                                  ViInt32   channel,
                                  ViChar    catalog[],
                                  ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_TraceListBalancedPorts(ViSession instrSession,
                                               ViInt32   channel,
                                               ViChar    catalog[],
                                               ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_TraceRename(ViSession instrSession,
                                    ViString  oldTraceName,
                                    ViString  newTraceName);

ViStatus _VI_FUNC rsvna_ChannelTraceRename(ViSession instrSession,
                                           ViInt32   channel,
                                           ViString  traceName);

ViStatus _VI_FUNC rsvna_TraceListCatalog(ViSession instrSession,
                                         ViChar    catalog[],
                                         ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_TraceListChannelCatalog(ViSession instrSession,
                                                ViInt32   channel,
                                                ViChar    catalog[],
                                                ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_QueryTraceName(ViSession instrSession,
                                       ViInt32   traceNumber,
                                       ViChar    traceName[]);

ViStatus _VI_FUNC rsvna_QueryTraceNumber(ViSession instrSession,
                                         ViString  traceName,
                                         ViInt32*  traceNumber);

ViStatus _VI_FUNC rsvna_QueryTraceDiagramNumber(ViSession instrSession,
                                                ViString  traceName,
                                                ViInt32*  diagramNumber);

ViStatus _VI_FUNC rsvna_QueryTraceChannelName(ViSession instrSession,
                                              ViString  traceName,
                                              ViChar    channelName[]);

ViStatus _VI_FUNC rsvna_QueryTraceChannelNumber(ViSession instrSession,
                                                ViString  traceName,
                                                ViInt32*  channelNumber);

ViStatus _VI_FUNC rsvna_TraceDataToMemory(ViSession instrSession,
                                          ViInt32   channel_Trace);

ViStatus _VI_FUNC rsvna_TraceDataToMemoryTrace(ViSession instrSession,
                                               ViString  memoryTrace,
                                               ViString  dataTrace);

ViStatus _VI_FUNC rsvna_TraceMathToMemoryTrace(ViSession instrSession,
                                               ViString  memoryTrace,
                                               ViString  dataTrace);

ViStatus _VI_FUNC rsvna_DeleteMemoryTraceAll(ViSession instrSession,
                                             ViInt32   channel);

ViStatus _VI_FUNC rsvna_DeleteMemoryTraceAllChannel(ViSession instrSession);

ViStatus _VI_FUNC rsvna_TraceUserDefinedMath(ViSession instrSession,
                                             ViInt32   channel_Trace,
                                             ViString  mathematicalExpression);

ViStatus _VI_FUNC rsvna_ConfigureTraceMathState(ViSession instrSession,
                                                ViInt32   channel_Trace,
                                                ViBoolean mathState);

ViStatus _VI_FUNC rsvna_ConfigureTraceMathFunction(ViSession instrSession,
                                                   ViInt32   channel_Trace,
                                                   ViInt32   mathematicalFunction);

ViStatus _VI_FUNC rsvna_TraceUserDefinedMathFormatted(ViSession instrSession,
                                                      ViInt32   channel_Trace,
                                                      ViString  mathematicalExpression);

ViStatus _VI_FUNC rsvna_ConfigureTraceMathFormattedState(ViSession instrSession,
                                                         ViInt32   channel_Trace,
                                                         ViBoolean mathState);

ViStatus _VI_FUNC rsvna_ConfigureTraceMathFormattedFunction(ViSession instrSession,
                                                            ViInt32   channel_Trace,
                                                            ViInt32   function);

ViStatus _VI_FUNC rsvna_ConfigureTraceMathWaveQuantityState(ViSession instrSession,
                                                            ViInt32   channel_Trace,
                                                            ViBoolean mathWaveQuantityState);

ViStatus _VI_FUNC rsvna_ConfigureTimeDomain(ViSession instrSession,
                                            ViInt32   channel,
                                            ViBoolean timeDomain,
                                            ViInt32   type,
                                            ViInt32   impulseResponse,
                                            ViReal64  sideLobeLevel,
                                            ViReal64  resolutionEnhancement);

ViStatus _VI_FUNC rsvna_ConfigureTimeDomainVSWREnabled(ViSession instrSession,
                                                       ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureHarmonicGridAndKeep(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   calculationMethod);

ViStatus _VI_FUNC rsvna_ExtrapolateDCValue(ViSession instrSession,
                                           ViInt32   channel);

ViStatus _VI_FUNC rsvna_CalculateHarmonicGrid(ViSession instrSession,
                                              ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureTimeGate(ViSession instrSession,
                                          ViInt32   channel,
                                          ViBoolean timeGate,
                                          ViInt32   timeGateType,
                                          ViBoolean rangeLines,
                                          ViInt32   filterType,
                                          ViReal64  sidebandSuppression);

ViStatus _VI_FUNC rsvna_ConfigureTimeGateStartStop(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64  startTime,
                                                   ViReal64  stopTime);

ViStatus _VI_FUNC rsvna_ConfigureTimeGateCenterSpan(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViReal64  centerTime,
                                                    ViReal64  span);

ViStatus _VI_FUNC rsvna_ConfigureTimeGateOffsetActivation(ViSession instrSession,
                                                          ViBoolean adjustTimeGate);

ViStatus _VI_FUNC rsvna_ConfigureTimeGateLinking(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViString  timeGateLinking);

ViStatus _VI_FUNC rsvna_ConfigureDistanceToFault(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean DTFEnabled,
                                                 ViString  cableType);

ViStatus _VI_FUNC rsvna_ConfigureDistanceToFaultStartStop(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  startDistance,
                                                          ViReal64  stopDistance);

ViStatus _VI_FUNC rsvna_ConfigureDistanceToFaultCenterSpan(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViReal64  centerDistance,
                                                           ViReal64  span);

ViStatus _VI_FUNC rsvna_DistanceToFaultAutoNumberOfPoints(ViSession instrSession,
                                                          ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureDistanceToFaultLimitCheck(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean limitCheckEnabled,
                                                           ViReal64  limit);

ViStatus _VI_FUNC rsvna_DefineDistanceToFaultCableType(ViSession instrSession,
                                                       ViString  cableName,
                                                       ViReal64  permittivity,
                                                       ViInt32   arrayLengths,
                                                       ViReal64  attenuationValues[],
                                                       ViReal64  frequencyValues[]);

ViStatus _VI_FUNC rsvna_DeleteDistanceToFaultCableType(ViSession instrSession,
                                                       ViString  cableType);

ViStatus _VI_FUNC rsvna_QueryDistanceToFaultPeakListCount(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32*  peakCount);

ViStatus _VI_FUNC rsvna_DistanceToFaultPeakListResults(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   arrayLengths,
                                                       ViReal64  peakDistances[],
                                                       ViReal64  responseValues[],
                                                       ViInt32*  peakCount);

ViStatus _VI_FUNC rsvna_ConfigureStatisticsInfoField(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean infoField,
                                                     ViInt32   horizontalPosition,
                                                     ViInt32   verticalPosition);

ViStatus _VI_FUNC rsvna_ConfigureTraceEvaluationRange(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   evaluationRange,
                                                      ViReal64  start,
                                                      ViReal64  stop);

ViStatus _VI_FUNC rsvna_ConfigureTraceStatisticalReference(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   referenceValue);

ViStatus _VI_FUNC rsvna_ConfigureTraceStatisticalManualReferenceLevel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  referenceLevel);

ViStatus _VI_FUNC rsvna_ConfigureTraceStatisticalManualReferencePhase(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  referencePhase);

ViStatus _VI_FUNC rsvna_ConfigureTraceStatisticalReferenceMarker(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   selectedMarker);

ViStatus _VI_FUNC rsvna_ConfigureTraceStatisticalReferenceRange(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean show,
                            ViInt32   referenceRange,
                            ViReal64  start,
                            ViReal64  stop);

ViStatus _VI_FUNC rsvna_ConfigureTraceStatisticalCompressionPhase(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  phase);

ViStatus _VI_FUNC rsvna_QueryTraceStatisticalParameter(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   statisticalParameter,
                                                       ViReal64* responseValue);

ViStatus _VI_FUNC rsvna_QueryTraceStatisticalAllParameters(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   bufferSize,
                                                           ViReal64  responseValues[]);

ViStatus _VI_FUNC rsvna_QueryTraceCompressionPoint(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64* compressionPointIn,
                                                   ViReal64* compressionPointOut);

ViStatus _VI_FUNC rsvna_ConfigureTraceSmoothing(ViSession instrSession,
                                                ViInt32   channel,
                                                ViBoolean smoothing,
                                                ViReal64  aperture);

ViStatus _VI_FUNC rsvna_TraceShiftStimulusValue(ViSession instrSession,
                                                ViInt32   window,
                                                ViReal64  shiftStimulusValue);

ViStatus _VI_FUNC rsvna_TraceShiftResponseValue(ViSession instrSession,
                                                ViInt32   window,
                                                ViReal64  magnitude,
                                                ViReal64  phase,
                                                ViReal64  real,
                                                ViReal64  imaginary);

ViStatus _VI_FUNC rsvna_TraceResponseData(ViSession instrSession,
                                          ViInt32   channel_Trace,
                                          ViInt32   arraySize,
                                          ViInt32   dataFormat,
                                          ViInt32*  noOfValues,
                                          ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseDataError(ViSession instrSession,
                                               ViInt32   channel_Trace,
                                               ViInt32   errorTerm,
                                               ViInt32   arraySize,
                                               ViInt32*  noOfValues,
                                               ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseDataAll(ViSession instrSession,
                                             ViInt32   arraySize,
                                             ViInt32   dataFormat,
                                             ViInt32*  noOfValues,
                                             ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseDataAllData(ViSession instrSession,
                                                 ViInt32   arraySize,
                                                 ViInt32   dataFormat,
                                                 ViInt32*  noOfValues,
                                                 ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseDataAllSData(ViSession instrSession,
                                                  ViInt32   channel_Trace,
                                                  ViInt32   arraySize,
                                                  ViInt32   dataFormat,
                                                  ViInt32*  noOfValues,
                                                  ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseSDataCatalog(ViSession instrSession,
                                                  ViInt32   channel_Trace,
                                                  ViInt32   bufferSize,
                                                  ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_TraceResponseSingleSweepData(ViSession instrSession,
                                                     ViInt32   channel_Trace,
                                                     ViInt32   sweepNumber,
                                                     ViInt32   arraySize,
                                                     ViInt32*  noOfValues,
                                                     ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseSingleSweepDataCount(ViSession instrSession,
                                                          ViInt32   channel_Trace,
                                                          ViInt32*  sweepCount);

ViStatus _VI_FUNC rsvna_TraceResponseSingleSweepDataForward(ViSession instrSession,
                                                            ViInt32   channel_Trace,
                                                            ViInt32   sweepNumber,
                                                            ViInt32   arraySize,
                                                            ViInt32*  noOfValues,
                                                            ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceResponseSingleSweepDataFirstLast(
                            ViSession instrSession,
                            ViInt32   channel_Trace,
                            ViInt32   firstSweep,
                            ViInt32   lastSweep,
                            ViInt32   arraySize,
                            ViInt32*  noOfValues,
                            ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_ArbitraryTraceResponseData(ViSession instrSession,
                                                   ViString  traceName,
                                                   ViInt32   arraySize,
                                                   ViInt32   dataFormat,
                                                   ViInt32*  noOfValues,
                                                   ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceStimulusData(ViSession instrSession,
                                          ViInt32   channel_Trace,
                                          ViInt32   arraySize,
                                          ViInt32*  noOfValues,
                                          ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_TraceDataExtrapolation(ViSession instrSession,
                                               ViInt32   channel_Trace);

ViStatus _VI_FUNC rsvna_WriteMemoryTraceData(ViSession instrSession,
                                             ViInt32   channel_Trace,
                                             ViInt32   dataFormat,
                                             ViInt32   noOfPoints,
                                             ViReal64  traceData[]);

ViStatus _VI_FUNC rsvna_ConfigureTraceLOSupression(ViSession instrSession,
                                                   ViInt32   channel_Trace,
                                                   ViBoolean LOSupression);

ViStatus _VI_FUNC rsvna_TraceImportData(ViSession instrSession,
                                        ViString  traceName,
                                        ViString  fileName);

ViStatus _VI_FUNC rsvna_TraceImportDataAutoDistribute(ViSession instrSession,
                                                      ViString  traceName);

ViStatus _VI_FUNC rsvna_TraceExportData(ViSession instrSession,
                                        ViString  traceName,
                                        ViString  fileName,
                                        ViInt32   exportFormat,
                                        ViInt32   exportData,
                                        ViInt32   decimalSeparator,
                                        ViInt32   fieldSeparator);

ViStatus _VI_FUNC rsvna_TraceExportDataPorts(ViSession instrSession,
                                             ViInt32   channelNumber,
                                             ViString  fileName,
                                             ViInt32   exportData,
                                             ViInt32   impedanceMode,
                                             ViInt32   port1,
                                             ViInt32   port2,
                                             ViInt32   port3,
                                             ViInt32   port4);

ViStatus _VI_FUNC rsvna_TraceExportDataPortsVariable(ViSession instrSession,
                                                     ViInt32   channelNumber,
                                                     ViString  fileName,
                                                     ViInt32   exportData,
                                                     ViInt32   impedanceMode,
                                                     ViInt32   arraySize,
                                                     ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_ChannelTraceExportData(ViSession instrSession,
                                               ViInt32   selectChannel,
                                               ViInt32   channelNumber,
                                               ViString  fileName,
                                               ViInt32   exportFormat,
                                               ViInt32   exportData,
                                               ViInt32   decimalSeparator,
                                               ViInt32   fieldSeparator);

ViStatus _VI_FUNC rsvna_ShowLimitLine(ViSession instrSession,
                                      ViInt32   channel,
                                      ViBoolean showLimitLine);

ViStatus _VI_FUNC rsvna_ConfigureLimitCheck(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   limitLine,
                                            ViBoolean limitCheck,
                                            ViInt32   horizontalPosition,
                                            ViInt32   verticalPosition);

ViStatus _VI_FUNC rsvna_QueryLimitCheckResult(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32*  limitCheckResult);

ViStatus _VI_FUNC rsvna_QueryGlobalLimitCheckResult(ViSession instrSession,
                                                    ViInt32*  globalLimitCheck);

ViStatus _VI_FUNC rsvna_ClearLimitCheckResults(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_AddLimitLineSegment(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   reserved,
                                            ViInt32   type,
                                            ViReal64  startStimulus,
                                            ViReal64  stopStimulus,
                                            ViReal64  startResponse,
                                            ViReal64  stopResponse);

ViStatus _VI_FUNC rsvna_EditLimitLineSegment(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   segment,
                                             ViInt32   type,
                                             ViReal64  startStimulus,
                                             ViReal64  stopStimulus,
                                             ViReal64  startResponse,
                                             ViReal64  stopResponse);

ViStatus _VI_FUNC rsvna_QueryLimitLineSegmentCount(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32*  segmentCount);

ViStatus _VI_FUNC rsvna_ReadLimitLineSegmentList(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   listSize,
                                                 ViInt32*  segmentsCount,
                                                 ViInt32   type[],
                                                 ViReal64  startStimulus[],
                                                 ViReal64  stopStimulus[],
                                                 ViReal64  startResponse[],
                                                 ViReal64  stopResponse[]);

ViStatus _VI_FUNC rsvna_WriteLimitLineSegmentList(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   listSize,
                                                  ViInt32   type,
                                                  ViReal64  startStimulus[],
                                                  ViReal64  stopStimulus[],
                                                  ViReal64  startResponse[],
                                                  ViReal64  stopResponse[]);

ViStatus _VI_FUNC rsvna_DeleteLimitLineSegments(ViSession instrSession,
                                                ViInt32   channel);

ViStatus _VI_FUNC rsvna_RecallLimitLine(ViSession instrSession,
                                        ViString  traceName,
                                        ViString  fileName,
                                        ViString  sParameter,
                                        ViReal64  xOffset,
                                        ViReal64  yOffset,
                                        ViInt32   type);

ViStatus _VI_FUNC rsvna_SaveLimitLine(ViSession instrSession,
                                      ViString  traceName,
                                      ViString  fileName);

ViStatus _VI_FUNC rsvna_ImportTraceAsLimitLine(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   limitLineType,
                                               ViReal64  stimulusOffset,
                                               ViReal64  responseOffset,
                                               ViString  traceName);

ViStatus _VI_FUNC rsvna_ConfigureLimitLineSegmentFormula(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   segment,
                                                         ViBoolean enabled,
                                                         ViString  formula);

ViStatus _VI_FUNC rsvna_ConfigureLimitLineSegmentInterpolation(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   segment,
                            ViInt32   interpolation);

ViStatus _VI_FUNC rsvna_QueryLimitLineFailData(ViSession instrSession,
                                               ViString  traceName,
                                               ViInt32   limitFailType,
                                               ViInt32   arraySize,
                                               ViReal64  results[],
                                               ViInt32*  returnedValues);

ViStatus _VI_FUNC rsvna_ConfigureLimitLineTTLPass(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   outputNo,
                                                  ViBoolean TTLOutput);

ViStatus _VI_FUNC rsvna_ConfigureLimitLineTTLPassDefault(ViSession instrSession,
                                                         ViInt32   outputNo,
                                                         ViBoolean TTLPassDefault);

ViStatus _VI_FUNC rsvna_ConfigureHorizontalLine(ViSession instrSession,
                                                ViInt32   channel,
                                                ViBoolean displayLine,
                                                ViReal64  position);

ViStatus _VI_FUNC rsvna_ConfigureLimitCheckPoints(ViSession instrSession,
                                                  ViInt32   type,
                                                  ViString  traceName,
                                                  ViChar    data[]);

ViStatus _VI_FUNC rsvna_ShowRippleLimits(ViSession instrSession,
                                         ViInt32   channel,
                                         ViBoolean showRippleLimits);

ViStatus _VI_FUNC rsvna_ConfigureRippleCheck(ViSession instrSession,
                                             ViInt32   channel,
                                             ViBoolean rippleCheck,
                                             ViInt32   horizontalPosition,
                                             ViInt32   verticalPosition);

ViStatus _VI_FUNC rsvna_QueryRippleLimitGlobalCheckResult(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32*  rippleLimitCheckResult);

ViStatus _VI_FUNC rsvna_ClearRippleLimitGlobalCheckResult(ViSession instrSession,
                                                          ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureRippleLimitRangeSegmentCheck(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   segment,
                            ViBoolean limitCheck);

ViStatus _VI_FUNC rsvna_QueryRippleLimitCheckSegmentResult(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   segment,
                                                           ViInt32*  fail,
                                                           ViReal64* limitCheckResult);

ViStatus _VI_FUNC rsvna_AddRippleLimitLineRangesSegment(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   noOfValues,
                                                        ViInt32   type[],
                                                        ViReal64  startStimulus[],
                                                        ViReal64  stopStimulus[],
                                                        ViReal64  limit[]);

ViStatus _VI_FUNC rsvna_EditRippleLimitLineSegment(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   segment,
                                                   ViReal64  startStimulus,
                                                   ViReal64  stopStimulus,
                                                   ViReal64  limit);

ViStatus _VI_FUNC rsvna_QueryRippleLimitRangesCount(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32*  rippleLimitRangesCount);

ViStatus _VI_FUNC rsvna_DeleteAllRippleLimitRanges(ViSession instrSession,
                                                   ViInt32   channel);

ViStatus _VI_FUNC rsvna_SaveRecallRippleLimit(ViSession instrSession,
                                              ViInt32   operationToBePerformed,
                                              ViString  traceName,
                                              ViString  fileName);

ViStatus _VI_FUNC rsvna_ShowCircleLimits(ViSession instrSession,
                                         ViInt32   channel,
                                         ViBoolean showCircleLimitLine);

ViStatus _VI_FUNC rsvna_ConfigureCircleLimitCheck(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViBoolean circleCheck);

ViStatus _VI_FUNC rsvna_ConfigureCircleLimitLine(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViReal64  centerX,
                                                 ViReal64  centerY,
                                                 ViReal64  radius);

ViStatus _VI_FUNC rsvna_QueryCircleLimitCheckResult(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32*  circleLimitCheckResult);

ViStatus _VI_FUNC rsvna_ClearCircleLimitCheckResult(ViSession instrSession,
                                                    ViInt32   channel);

ViStatus _VI_FUNC rsvna_CircleClear(ViSession instrSession,
                                    ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureCircleDisplay(ViSession instrSession,
                                               ViInt32   channel,
                                               ViBoolean limitToCircle,
                                               ViBoolean showBorder);

ViStatus _VI_FUNC rsvna_ConfigureCircleData(ViSession instrSession,
                                            ViInt32   channel,
                                            ViReal64  centerX,
                                            ViReal64  centerY,
                                            ViReal64  radius);

ViStatus _VI_FUNC rsvna_ConfigureMarkerState(ViSession instrSession,
                                             ViInt32   channel_Trace,
                                             ViInt32   marker,
                                             ViBoolean markerState);

ViStatus _VI_FUNC rsvna_ConfigureMarker(ViSession instrSession,
                                        ViInt32   channel_Trace,
                                        ViInt32   marker,
                                        ViBoolean markerState,
                                        ViReal64  markerStimulus);

ViStatus _VI_FUNC rsvna_ConfigureMarkerFormat(ViSession instrSession,
                                              ViInt32   channel_Trace,
                                              ViInt32   marker,
                                              ViInt32   markerFormat);

ViStatus _VI_FUNC rsvna_ConfigureMarkerDefaultFormat(ViSession instrSession,
                                                     ViInt32   channel_Trace,
                                                     ViInt32   markerDefaultFormat);

ViStatus _VI_FUNC rsvna_QueryMarkerResponse(ViSession instrSession,
                                            ViInt32   channel_Trace,
                                            ViInt32   marker,
                                            ViReal64  markerResponse[]);

ViStatus _VI_FUNC rsvna_ConfigureReferenceMarker(ViSession instrSession,
                                                 ViInt32   channel_Trace,
                                                 ViInt32   marker,
                                                 ViBoolean referenceMarkerState,
                                                 ViReal64  referenceMarkerStimulus);

ViStatus _VI_FUNC rsvna_QueryReferenceMarkerResponse(ViSession instrSession,
                                                     ViInt32   channel_Trace,
                                                     ViInt32   marker,
                                                     ViReal64* referenceMarkerResponse);

ViStatus _VI_FUNC rsvna_ConfigureMarkerDeltaMode(ViSession instrSession,
                                                 ViInt32   channel_Trace,
                                                 ViInt32   marker,
                                                 ViBoolean deltaMarkerState);

ViStatus _VI_FUNC rsvna_ConfigureMarkerReferenceFormat(ViSession instrSession,
                                                       ViInt32   channel_Trace,
                                                       ViInt32   marker,
                                                       ViInt32   markerReferenceFormat);

ViStatus _VI_FUNC rsvna_SetAllMarkersOff(ViSession instrSession,
                                         ViInt32   channel_Trace);

ViStatus _VI_FUNC rsvna_ConfigureMarkerInfoFieldPosition(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   horizontalPosition,
                                                         ViInt32   verticalPosition);

ViStatus _VI_FUNC rsvna_ConfigureMarkerProperties(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   marker,
                                                  ViInt32   markerType,
                                                  ViString  markerName,
                                                  ViInt32   discreteMode,
                                                  ViInt32   fixedMarker);

ViStatus _VI_FUNC rsvna_ConfigureMarkerThreshold(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   marker,
                                                 ViBoolean enabled,
                                                 ViReal64  threshold);

ViStatus _VI_FUNC rsvna_ConfigureMarkerCoupledMarkers(ViSession instrSession,
                                                      ViBoolean coupledMarkers);

ViStatus _VI_FUNC rsvna_ConfigureMarkerCouplingType(ViSession instrSession,
                                                    ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureMarkerExcursion(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   marker,
                                                 ViBoolean enabled,
                                                 ViReal64  excursionMinimum);

ViStatus _VI_FUNC rsvna_MarkerSpectrumEqualsMarker(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   marker);

ViStatus _VI_FUNC rsvna_ConfigureMarkerMultiplePeakExcursion(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViBoolean enabled,
                                                             ViReal64  excursionMinimum);

ViStatus _VI_FUNC rsvna_ConfigureMarkerMultiplePeakThreshold(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViBoolean enabled,
                                                             ViReal64  threshold);

ViStatus _VI_FUNC rsvna_ConfigureMarkerReferenceExcursion(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   marker,
                                                          ViBoolean enabled,
                                                          ViReal64  reference);

ViStatus _VI_FUNC rsvna_MarkerReferenceMaximum(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_MarkerReferenceMinimum(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMarkerReferenceThreshold(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean enabled,
                                                          ViReal64  reference);

ViStatus _VI_FUNC rsvna_MarkerSearch(ViSession instrSession,
                                     ViInt32   channel,
                                     ViInt32   marker,
                                     ViInt32   search);

ViStatus _VI_FUNC rsvna_MarkerSearchReference(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   marker,
                                              ViInt32   search);

ViStatus _VI_FUNC rsvna_MarkerTargetSearch(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   marker,
                                           ViInt32   search);

ViStatus _VI_FUNC rsvna_ConfigureMarkerTargetValue(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   marker,
                                                   ViReal64  targetValue);

ViStatus _VI_FUNC rsvna_ConfigureMarkerTargetValueReference(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViReal64  targetValue);

ViStatus _VI_FUNC rsvna_ConfigureMarkerTargetFormat(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   marker,
                                                    ViInt32   targetFormat);

ViStatus _VI_FUNC rsvna_ConfigureMarkerTargetFormatReference(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   targetFormat);

ViStatus _VI_FUNC rsvna_MarkerTargetFormatReferenceSearchLeft(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_MarkerTargetFormatReferenceSearchNext(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_MarkerTargetFormatReferenceSearchRight(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_MarkerTargetFormatReferenceSearchImmediate(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMarkerTrackingReference(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViBoolean enabled);

ViStatus _VI_FUNC rsvna_MarkerBandfilterSearch(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMarkerBandfilterSearchMode(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   searchMode);

ViStatus _VI_FUNC rsvna_ConfigureMarkerxdBBandwidth(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViReal64  xDBBandwidth);

ViStatus _VI_FUNC rsvna_QueryMarkerBandfilterResults(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViReal64* bandwidth,
                                                     ViReal64* centerStimulus,
                                                     ViReal64* qualityFactor,
                                                     ViReal64* loss,
                                                     ViReal64* lowerEdge,
                                                     ViReal64* upperEdge);

ViStatus _VI_FUNC rsvna_ConfigureMarkerBandfilterSearchResultState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean searchResults,
                            ViInt32   horizontalPosition,
                            ViInt32   verticalPosition);

ViStatus _VI_FUNC rsvna_ConfigureMarkerTracking(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   marker,
                                                ViBoolean markerTracking);

ViStatus _VI_FUNC rsvna_ConfigureMarkerSearchRange(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   marker,
                                                   ViInt32   searchRange,
                                                   ViReal64  start,
                                                   ViReal64  stop);

ViStatus _VI_FUNC rsvna_ConfigureMarkerSearchRangeReference(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   searchRange,
                                                            ViReal64  start,
                                                            ViReal64  stop);

ViStatus _VI_FUNC rsvna_QueryMarkerSearchResults(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   marker,
                                                 ViReal64* markerStimulus,
                                                 ViReal64  markerResponse[]);

ViStatus _VI_FUNC rsvna_ConfigureSweepRangeToMarker(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   marker,
                                                    ViInt32   setByMarker);

ViStatus _VI_FUNC rsvna_ConfigureSweepRangeToMarkerReference(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   setByMarker);

ViStatus _VI_FUNC rsvna_MarkerCenterSweepRangeToMarkerReference(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMarkerSearchRangeShowLimitLinesReference(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ExternalDLLList(ViSession instrSession,
                                        ViInt32   bufferSize,
                                        ViChar    externalDLLList[],
                                        ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_ExternalDLLTaskList(ViSession instrSession,
                                            ViString  dllName,
                                            ViInt32   bufferSize,
                                            ViChar    tasks[],
                                            ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_ConfigureExternalDLL(ViSession instrSession,
                                             ViInt32   channel,
                                             ViString  dllName,
                                             ViBoolean active,
                                             ViBoolean permanent,
                                             ViString  task,
                                             ViString  additional,
                                             ViString  address);

ViStatus _VI_FUNC rsvna_QueryExternalDLLConfiguration(ViSession  instrSession,
                                                      ViInt32    channel,
                                                      ViString   dllName,
                                                      ViBoolean* active,
                                                      ViBoolean* permanent,
                                                      ViInt32    buffersSize,
                                                      ViChar     task[],
                                                      ViChar     additional[],
                                                      ViChar     address[]);

ViStatus _VI_FUNC rsvna_AddExternalDLL(ViSession instrSession,
                                       ViString  dllPath);

ViStatus _VI_FUNC rsvna_RemoveExternalDLL(ViSession instrSession,
                                          ViString  dllName);

ViStatus _VI_FUNC rsvna_SelectExternalDLL(ViSession instrSession,
                                          ViInt32   channel,
                                          ViString  traceName,
                                          ViInt32   port,
                                          ViString  dllName,
                                          ViString  taskType,
                                          ViString  additionalInput);

ViStatus _VI_FUNC rsvna_ConfigureLinearityDeviation(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   mode,
                                                    ViReal64  constant,
                                                    ViReal64  slope);

ViStatus _VI_FUNC rsvna_ConfigureLinearityDeviationElectricalLength(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  electricalLength);

ViStatus _VI_FUNC rsvna_RecalculateLinearityDeviation(ViSession instrSession,
                                                      ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyStartStop(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViReal64  startFrequency,
                                                    ViReal64  stopFrequency);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyCenterSpan(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViReal64  centerFrequency,
                                                     ViReal64  span);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyMinMax(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   frequency,
                                                 ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigurePower(ViSession instrSession,
                                       ViInt32   channel,
                                       ViReal64  power);

ViStatus _VI_FUNC rsvna_ConfigurePowerStartStop(ViSession instrSession,
                                                ViInt32   channel,
                                                ViReal64  startPower,
                                                ViReal64  stopPower);

ViStatus _VI_FUNC rsvna_ConfigureCWFrequency(ViSession instrSession,
                                             ViInt32   channel,
                                             ViReal64  CWFrequency);

ViStatus _VI_FUNC rsvna_ConfigureSourcePort(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   sourcePort);

ViStatus _VI_FUNC rsvna_QueryFrequencyRange(ViSession instrSession,
                                            ViInt32   range,
                                            ViReal64* frequency);

ViStatus _VI_FUNC rsvna_ConfigurePreamplifierEnabled(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   physicalPort,
                                                     ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePreamplifierGain(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViReal64  gain);

ViStatus _VI_FUNC rsvna_ConfigureTimeDomainStartStopTime(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViReal64  startTime,
                                                         ViReal64  stopTime);

ViStatus _VI_FUNC rsvna_ConfigureTimeDomainCenterSpanTime(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  centerTime,
                                                          ViReal64  timeSpan);

ViStatus _VI_FUNC rsvna_ConfigureTimeDomainXAxisScaling(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   xAxisScaling);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCAutomaticIFBandwidth(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean automaticIFBandwidth);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCBandwidth(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64  bandwidth);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCEnabled(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortCouple(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean portCouple);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCSettling(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViBoolean autoSettling,
                                                  ViReal64  controlTime);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCSettlingTolerance(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViReal64  settlingTolerance);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortOptimize(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   physicalPort,
                                                      ViInt32   portOptimize);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortPI(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   physicalPort,
                                                ViInt32   portPI,
                                                ViReal64  PIGain,
                                                ViReal64  PIIntegrationTime);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortControlRange(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   physicalPort,
                                                          ViReal64  portControlRange);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortReceiver(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   physicalPort,
                                                      ViString  receiver);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortStartOffset(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort,
                                                         ViReal64  startOffset);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortTableAreaOn(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort,
                                                         ViBoolean tableAreaOn);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortTargetPower(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort,
                                                         ViBoolean targetPowerDefault,
                                                         ViReal64  targetPower);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCPortDeembedingNetworkUsage(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   deembedingNetworkUsage);

ViStatus _VI_FUNC rsvna_ConfigurePowerALCClamp(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   physicalPort,
                                               ViBoolean clamp);

ViStatus _VI_FUNC rsvna_ConfigureRFState(ViSession instrSession,
                                         ViBoolean RFState);

ViStatus _VI_FUNC rsvna_ConfigureBandwidth(ViSession instrSession,
                                           ViInt32   channel,
                                           ViReal64  bandwidth,
                                           ViInt32   filterSelectivity);

ViStatus _VI_FUNC rsvna_ConfigureAverageMode(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigureAverage(ViSession instrSession,
                                         ViInt32   channel,
                                         ViBoolean averageState,
                                         ViInt32   averageFactor);

ViStatus _VI_FUNC rsvna_RestartAverage(ViSession instrSession,
                                       ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureChannelDynamicBwatLowFrequencies(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean bwAtLowFrequencies);

ViStatus _VI_FUNC rsvna_ConfigureChannelAnalogIFFilter(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   IFFilter);

ViStatus _VI_FUNC rsvna_ConfigureSweepNumberOfPoints(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   numberOfPoints);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyStepSize(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64  stepSize);

ViStatus _VI_FUNC rsvna_ConfigureSweepTime(ViSession instrSession,
                                           ViInt32   channel,
                                           ViBoolean autoSweepTime,
                                           ViReal64  sweepTime);

ViStatus _VI_FUNC rsvna_ConfigureSweepMeasDelay(ViSession instrSession,
                                                ViInt32   channel,
                                                ViReal64  measDelay);

ViStatus _VI_FUNC rsvna_ConfigureSweepMeasDelayInsertionPoint(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   insertionPoint);

ViStatus _VI_FUNC rsvna_ConfigureSweepType(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   sweepType);

ViStatus _VI_FUNC rsvna_ConfigureSweepStimulusAxis(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   sweepAxisType,
                                                   ViString  scale);

ViStatus _VI_FUNC rsvna_ConfigureAcquisition(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   continuousSweepScope,
                                             ViBoolean sweepModeContinuous,
                                             ViInt32   sweepCount,
                                             ViInt32   singleSweepScope);

ViStatus _VI_FUNC rsvna_RestartSweep(ViSession instrSession,
                                     ViInt32   channel,
                                     ViInt32   sweepScope);

ViStatus _VI_FUNC rsvna_ConfigureMeasurementInChannel(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   channelScope,
                                                      ViBoolean enableSweep);

ViStatus _VI_FUNC rsvna_SweepOptimizeChannelSwitchingAuto(ViSession instrSession);

ViStatus _VI_FUNC rsvna_SweepOptimizeChannelSwitchingManual(ViSession instrSession,
                                                            ViInt32   arraySize,
                                                            ViInt32   channelList[]);

ViStatus _VI_FUNC rsvna_SendTrigger(ViSession instrSession);

ViStatus _VI_FUNC rsvna_SendTriggerWaitOPC(ViSession instrSession,
                                           ViInt32   timeout);

ViStatus _VI_FUNC rsvna_ConfigureSweepWaitTimeperPoint(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64  waitTimePerPoint);

ViStatus _VI_FUNC rsvna_ConfigureSweepModeControlledTiming(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean modeControlledTiming);

ViStatus _VI_FUNC rsvna_ConfigureSweepTimeMode(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   sweepTimeMode);

ViStatus _VI_FUNC rsvna_ConfigureSweepFastPowerSweep(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   fastPowerSweep);

ViStatus _VI_FUNC rsvna_ConfigureSweepReverse(ViSession instrSession,
                                              ViInt32   channel,
                                              ViBoolean reverse);

ViStatus _VI_FUNC rsvna_ConfigureSweepTimerMode(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   timerMode);

ViStatus _VI_FUNC rsvna_ConfigureSweepOnlyActiveChannelOn(ViSession instrSession,
                                                          ViBoolean sweepOnlyActiveChannelOn);

ViStatus _VI_FUNC rsvna_ConfigureSweepMinimumWaitTime(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViBoolean enabled,
                                                      ViReal64  minimumWaitTime);

ViStatus _VI_FUNC rsvna_ConfigureSweepFixedWaitTime(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean enabled,
                                                    ViReal64  fixedWaitTime);

ViStatus _VI_FUNC rsvna_ConfigureSweepManualWaitTimeEnabled(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureSweepMinimumWaitTimeInitializationChannel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled,
                            ViReal64  timeInitializationChannel);

ViStatus _VI_FUNC rsvna_ConfigureSweepMaximumWaitTimePort(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean enabled,
                                                          ViReal64  maximumWaitTimePort);

ViStatus _VI_FUNC rsvna_ConfigureSweepMinimumWaitTimePort(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean enabled,
                                                          ViReal64  minimumWaitTimePort);

ViStatus _VI_FUNC rsvna_InsertNewSegment(ViSession instrSession,
                                         ViInt32   channel,
                                         ViInt32   segment,
                                         ViReal64  startFrequency,
                                         ViReal64  stopFrequency,
                                         ViInt32   numberOfPoints,
                                         ViReal64  power,
                                         ViInt32   sweepTimeSelect,
                                         ViReal64  time,
                                         ViReal64  pointDelay,
                                         ViReal64  measBandwidth,
                                         ViInt32   localOscillatorFrequency,
                                         ViInt32   selectivity,
                                         ViInt32   freqSweepMode);

ViStatus _VI_FUNC rsvna_RedefineSegment(ViSession instrSession,
                                        ViInt32   channel,
                                        ViInt32   segment,
                                        ViReal64  startFrequency,
                                        ViReal64  stopFrequency,
                                        ViInt32   numberOfPoints,
                                        ViReal64  power,
                                        ViInt32   sweepTimeSelect,
                                        ViReal64  time,
                                        ViReal64  pointDelay,
                                        ViReal64  measBandwidth,
                                        ViInt32   localOscillatorFrequency,
                                        ViInt32   selectivity,
                                        ViInt32   freqSweepMode);

ViStatus _VI_FUNC rsvna_AddNewSegment(ViSession instrSession,
                                      ViInt32   channel,
                                      ViInt32   segment);

ViStatus _VI_FUNC rsvna_QuerySweepSegmentNumber(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32*  numberOfSweepSegments);

ViStatus _VI_FUNC rsvna_DeleteSelectedSegment(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   segment);

ViStatus _VI_FUNC rsvna_DeleteAllSegments(ViSession instrSession,
                                          ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureSweepSegmentState(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   segment,
                                                   ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureSweepSegmentXAxis(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   xAxis);

ViStatus _VI_FUNC rsvna_SaveSegment(ViSession instrSession,
                                    ViInt32   channel,
                                    ViString  fileName);

ViStatus _VI_FUNC rsvna_LoadSegment(ViSession instrSession,
                                    ViInt32   channel,
                                    ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureSweepSegmentPortLevel(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   segment,
                                                       ViInt32   port,
                                                       ViBoolean enabled,
                                                       ViReal64  powerLevel);

ViStatus _VI_FUNC rsvna_ConfigureSweepSegmentWaitTimeFixed(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean enabled,
                                                           ViReal64  waitTimeFixed);

ViStatus _VI_FUNC rsvna_ConfigureSweepSegmentWaitTimeFixedMaximum(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  waitTimeFixedMaximum);

ViStatus _VI_FUNC rsvna_ConfigureSweepSegmentWaitTimeFixedMinimum(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  waitTimeFixedMinimum);

ViStatus _VI_FUNC rsvna_ConfigureSweepTimingInfoBoxEnabled(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureSweepGenerationMode(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   mode);

ViStatus _VI_FUNC rsvna_SelectCalibrationType(ViSession instrSession,
                                              ViInt32   channel,
                                              ViString  calibrationName,
                                              ViInt32   parameters,
                                              ViInt32   port1,
                                              ViInt32   port2,
                                              ViInt32   port3,
                                              ViInt32   port4);

ViStatus _VI_FUNC rsvna_SelectCalibrationTypeSimple(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   calibrationType);

ViStatus _VI_FUNC rsvna_SelectCalibrationTypeExtended(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViString  calibrationName,
                                                      ViInt32   parameters,
                                                      ViInt32   ports[],
                                                      ViInt32   arraySize,
                                                      ViString  adapterCalKit);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationDetectorType(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViString  calibrationName,
                                                         ViInt32   detectorType);

ViStatus _VI_FUNC rsvna_StartCalibration(ViSession instrSession,
                                         ViInt32   channel,
                                         ViInt32   standard,
                                         ViInt32   port1,
                                         ViInt32   port2);

ViStatus _VI_FUNC rsvna_StartCalibrationWithOptions(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   standard,
                                                    ViInt32   port1,
                                                    ViInt32   port2,
                                                    ViBoolean dispersion,
                                                    ViInt32   delayPhase,
                                                    ViReal64  delayPhaseValue);

ViStatus _VI_FUNC rsvna_ConfigureSystemErrorCorrectionState(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViBoolean correctionState);

ViStatus _VI_FUNC rsvna_LoadCalibrationDataFromCalGroupFile(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViString  calPoolFile,
                                                            ViInt32   standard,
                                                            ViInt32   port1,
                                                            ViInt32   port2);

ViStatus _VI_FUNC rsvna_ActivatePowerCalibrations(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   activate);

ViStatus _VI_FUNC rsvna_CalibrationManager(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   operationToBePerformed,
                                           ViString  fileName,
                                           ViString  secondFileName);

ViStatus _VI_FUNC rsvna_CalibrationAuto(ViSession instrSession,
                                        ViInt32   channel,
                                        ViString  calibrationKitName,
                                        ViInt32   analyzerPort1,
                                        ViInt32   analyzerPort2,
                                        ViInt32   analyzerPort3,
                                        ViInt32   analyzerPort4,
                                        ViInt32   calUnitPort1,
                                        ViInt32   calUnitPort2,
                                        ViInt32   calUnitPort3,
                                        ViInt32   calUnitPort4,
                                        ViUInt32  timeout);

ViStatus _VI_FUNC rsvna_QueryCalibrationPortsState(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   bufferSize,
                                                   ViChar    state[]);

ViStatus _VI_FUNC rsvna_CalibrationAutoSimplified(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViString  calibrationKitName,
                                                  ViInt32   analyzerPort1,
                                                  ViInt32   analyzerPort2,
                                                  ViInt32   analyzerPort3,
                                                  ViInt32   analyzerPort4,
                                                  ViUInt32  timeout);

ViStatus _VI_FUNC rsvna_CalibrationAutoType(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   parameters,
                                            ViString  calibrationKitName,
                                            ViInt32   analyzerPort1,
                                            ViInt32   analyzerPort2,
                                            ViInt32   analyzerPort3,
                                            ViInt32   analyzerPort4,
                                            ViInt32   calUnitPort1,
                                            ViInt32   calUnitPort2,
                                            ViInt32   calUnitPort3,
                                            ViInt32   calUnitPort4,
                                            ViUInt32  timeout);

ViStatus _VI_FUNC rsvna_CalibrationAutoTypeSimplified(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   parameters,
                                                      ViString  calibrationKitName,
                                                      ViInt32   analyzerPort1,
                                                      ViInt32   analyzerPort2,
                                                      ViInt32   analyzerPort3,
                                                      ViInt32   analyzerPort4,
                                                      ViUInt32  timeout);

ViStatus _VI_FUNC rsvna_CalibrationAutoPower(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   analyzerPort);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutoAssignment(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   assignment,
                                                           ViInt32   arraySize,
                                                           ViInt32   testPorts[],
                                                           ViInt32   calUnitPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutoAssignmentWithDetection(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   assignment,
                            ViInt32   arraySize,
                            ViInt32   testPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutoAssignmentType(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   calibrationType,
                            ViString  calibrationKitName);

ViStatus _VI_FUNC rsvna_SaveCalibrationAutoAssignment(ViSession instrSession,
                                                      ViInt32   channel);

ViStatus _VI_FUNC rsvna_InitiateCalibrationAutoAssignment(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   assignment);

ViStatus _VI_FUNC rsvna_DeleteAllCalibrationAutoAssignments(ViSession instrSession,
                                                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_GenerateDefaultCalibrationAutoAssignment(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   arraySize,
                            ViInt32   testPorts[]);

ViStatus _VI_FUNC rsvna_GenerateDefaultCalibrationAutoAssignmentWithDetection(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   arraySize,
                            ViInt32   testPorts[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationAutoAssignmentCount(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32*  count);

ViStatus _VI_FUNC rsvna_QueryAllCalibrationAutoAssignmentCount(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32*  count);

ViStatus _VI_FUNC rsvna_QueryCalibrationConnection(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32*  analyzerPort1,
                                                   ViInt32*  analyzerPort2,
                                                   ViInt32*  analyzerPort3,
                                                   ViInt32*  analyzerPort4);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationExternalPowerMeter(
                            ViSession instrSession,
                            ViInt32   powerMeterNumber,
                            ViInt32   VNAPort);

ViStatus _VI_FUNC rsvna_ConfigureMulticalibrationType(ViSession instrSession,
                                                      ViBoolean type);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutoMinimumConnect(
                            ViSession instrSession,
                            ViBoolean autoMinimumConnect);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutoCommonPort(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   commonPort);

ViStatus _VI_FUNC rsvna_CalibrationTerminateSystemErrorCorrection(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_CalibrationClean(ViSession instrSession,
                                         ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationWaveCorrectionEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationCalculateAfterDeembed(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryCalibrationDate(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   bufferSize,
                                             ViChar    calibrationDate[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationState(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32*  calibrationState);

ViStatus _VI_FUNC rsvna_ConfigureActiveCalibrationUnit(ViSession instrSession,
                                                       ViString  calibrationUnit);

ViStatus _VI_FUNC rsvna_QueryAllCalibrationUnits(ViSession instrSession,
                                                 ViInt32   bufferSize,
                                                 ViChar    calibrationUnits[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationUnitStandard(ViSession instrSession,
                                                         ViInt32   standard,
                                                         ViInt32   port1,
                                                         ViInt32   port2);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitCharacterizations(
                            ViSession instrSession,
                            ViInt32   bufferSize,
                            ViChar    calUnitCharacterizations[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitCharacterizationsStandards(
                            ViSession instrSession,
                            ViString  calUnitCharacterization,
                            ViInt32   bufferSize,
                            ViChar    calUnitCharacterizations[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitCharacterizationDate(
                            ViSession instrSession,
                            ViString  calUnitCharacterization,
                            ViInt32   bufferSize,
                            ViChar    calUnitCharacterizationDate[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitCharacterizationFrequencyRange(
                            ViSession instrSession,
                            ViString  calUnitCharacterization,
                            ViReal64* startFrequency,
                            ViReal64* stopFrequency);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitCharacterizationPorts(
                            ViSession instrSession,
                            ViString  calUnitCharacterization,
                            ViInt32   bufferSize,
                            ViChar    calUnitCharacterizationPorts[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitCharacterizationData(
                            ViSession instrSession,
                            ViString  calUnitCharacterization,
                            ViInt32   standard,
                            ViInt32   sParameter,
                            ViInt32   firstPortNumber,
                            ViInt32   secondPortNumber,
                            ViInt32   bufferSize,
                            ViReal64  calibrationData[],
                            ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_QueryCalibrationUnitWarmupState(ViSession  instrSession,
                                                        ViBoolean* warmupState);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutomaticVectorMixer(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   port1,
                            ViInt32   port2);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationNoiseDataReceivingPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   receivingPort);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationSameSweepSetup(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean sameSweepSetup);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationSweepSetup(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean sweepSetup);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationReferenceReceiverAttenuation(
                            ViSession instrSession,
                            ViReal64  attenuation);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationIMDMethod(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViBoolean IMDMethod);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationFastMultiportCorrection(
                            ViSession instrSession,
                            ViBoolean fastMultiportCorrection);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationInterpolationMethod(
                            ViSession instrSession,
                            ViInt32   interpolationMethod);

ViStatus _VI_FUNC rsvna_StartHarmonicReceiverCalibration(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort);

ViStatus _VI_FUNC rsvna_StartHarmonicSourceCalibration(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   physicalPort);

ViStatus _VI_FUNC rsvna_CalibrationICSRepeat(ViSession instrSession,
                                             ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationICSTemperatureCompensation(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean temperatureCompensation);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationICSUnknownThrough(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   port1,
                            ViInt32   port2);

ViStatus _VI_FUNC rsvna_JoinCalibrations(ViSession instrSession,
                                         ViString  firstCalibration,
                                         ViInt32   locationFirstCalibration,
                                         ViString  secondCalibration,
                                         ViInt32   locationSecondCalibration,
                                         ViString  outputCalibration,
                                         ViInt32   locationOutputCalibration);

ViStatus _VI_FUNC rsvna_CalibrationDelayMixer(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   RFPort,
                                              ViInt32   IFPort);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationInternalCombinerEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_CalibrationConverterOffset(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64  value,
                                                   ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureConnector(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   port,
                                           ViInt32   calibrationConnector);

ViStatus _VI_FUNC rsvna_ConfigureUserConnectorType(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViString  connectorName,
                                                   ViInt32   propagationMode,
                                                   ViInt32   connectorType,
                                                   ViReal64  relativePermittivity,
                                                   ViReal64  impedance);

ViStatus _VI_FUNC rsvna_ConfigureUserConnector(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   port,
                                               ViString  connector,
                                               ViInt32   connectorGender);

ViStatus _VI_FUNC rsvna_DeleteUserConnectorType(ViSession instrSession,
                                                ViInt32   channel,
                                                ViString  connectorName);

ViStatus _VI_FUNC rsvna_ConfigurePortConnectorTypes(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   connectorType);

ViStatus _VI_FUNC rsvna_ConnectorTypesCatalog(ViSession instrSession,
                                              ViChar    catalog[],
                                              ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_SaveCalibrationData(ViSession instrSession,
                                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_GenerateDefaultCalibrationData(ViSession instrSession,
                                                       ViInt32   channel);

ViStatus _VI_FUNC rsvna_DeleteCalibrationData(ViSession instrSession,
                                              ViInt32   channel,
                                              ViString  calibrationName);

ViStatus _VI_FUNC rsvna_QueryAvailableCalibrationDataCatalog(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   catalog,
                                                             ViInt32   port1,
                                                             ViInt32   port2,
                                                             ViInt32   arraySize,
                                                             ViChar    availableBuffers[],
                                                             ViInt32*  numberOfValues);

ViStatus _VI_FUNC rsvna_ReadCalibrationData(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   errorTermParameters,
                                            ViInt32   sourcePort,
                                            ViInt32   loadPort,
                                            ViInt32   bufferSize,
                                            ViReal64  calibrationData[],
                                            ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_WriteCalibrationData(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   errorTermParameters,
                                             ViInt32   sourcePort,
                                             ViInt32   loadPort,
                                             ViReal64  calibrationData[]);

ViStatus _VI_FUNC rsvna_QueryAvailableCalibrationDataCatalogPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   catalog,
                            ViInt32   port1,
                            ViInt32   port2,
                            ViInt32   arraySize,
                            ViChar    availableBuffers[],
                            ViInt32*  numberOfValues);

ViStatus _VI_FUNC rsvna_ReadCalibrationDataPort(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   physicalPort,
                                                ViInt32   errorTermParameters,
                                                ViInt32   sourcePort,
                                                ViInt32   loadPort,
                                                ViInt32   bufferSize,
                                                ViReal64  calibrationData[],
                                                ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_WriteCalibrationDataPort(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   physicalPort,
                                                 ViInt32   errorTermParameters,
                                                 ViInt32   sourcePort,
                                                 ViInt32   loadPort,
                                                 ViReal64  calibrationData[]);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixAvailableCalibrationDataCatalog(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   catalog,
                            ViInt32   sourceTestPort,
                            ViInt32   loadTestPort,
                            ViInt32   arraySize,
                            ViChar    availableBuffers[],
                            ViInt32*  numberOfValues);

ViStatus _VI_FUNC rsvna_ReadSwitchMatrixCalibrationData(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   errorTermParameters,
                                                        ViInt32   sourceTestPort,
                                                        ViInt32   loadTestPort,
                                                        ViInt32   sourceVNAPort,
                                                        ViInt32   loadVNAPort,
                                                        ViInt32   bufferSize,
                                                        ViReal64  calibrationData[],
                                                        ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_WriteSwitchMatrixCalibrationData(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   errorTermParameters,
                                                         ViInt32   sourceTestPort,
                                                         ViInt32   loadTestPort,
                                                         ViInt32   sourceVNAPort,
                                                         ViInt32   loadVNAPort,
                                                         ViReal64  calibrationData[]);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixAvailableCalibrationDataCatalogPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   catalog,
                            ViInt32   sourceTestPort,
                            ViInt32   loadTestPort,
                            ViInt32   arraySize,
                            ViChar    availableBuffers[],
                            ViInt32*  numberOfValues);

ViStatus _VI_FUNC rsvna_ReadSwitchMatrixCalibrationDataPort(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   physicalPort,
                                                            ViInt32   errorTermParameters,
                                                            ViInt32   sourceTestPort,
                                                            ViInt32   loadTestPort,
                                                            ViInt32   sourceVNAPort,
                                                            ViInt32   loadVNAPort,
                                                            ViInt32   bufferSize,
                                                            ViReal64  calibrationData[],
                                                            ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_WriteSwitchMatrixCalibrationDataPort(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   physicalPort,
                                                             ViInt32   errorTermParameters,
                                                             ViInt32   sourceTestPort,
                                                             ViInt32   loadTestPort,
                                                             ViInt32   sourceVNAPort,
                                                             ViInt32   loadVNAPort,
                                                             ViReal64  calibrationData[]);

ViStatus _VI_FUNC rsvna_ExportFactoryCalibrationDataToTouchstoneFile(
                            ViSession instrSession,
                            ViString  directory);

ViStatus _VI_FUNC rsvna_ExportUserDefinedCalibrationDataToTouchstoneFile(
                            ViSession instrSession,
                            ViString  directory,
                            ViString  calibrationKitName);

ViStatus _VI_FUNC rsvna_QueryCalibrationStimulusFrequencies(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   bufferSize,
                                                            ViReal64  stimulusFrequencies[],
                                                            ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_QueryCalibrationStimulusFrequenciesPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   bufferSize,
                            ViReal64  stimulusFrequencies[],
                            ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_QueryCalibrationDataParameters(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64* frequencyStart,
                                                       ViReal64* frequencyStop,
                                                       ViInt32*  numberOfPoints,
                                                       ViReal64* internalSignalSourcePower,
                                                       ViInt32*  sweepType);

ViStatus _VI_FUNC rsvna_QueryCalibrationDataParametersSpecific(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   systemErrorCorrection,
                            ViInt32   type,
                            ViInt32   index,
                            ViInt32   bufferSize,
                            ViChar    info[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationDataParametersCount(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32*  parametersCount);

ViStatus _VI_FUNC rsvna_QueryCalibrationDataParametersPort(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   physicalPort,
                                                           ViReal64* frequencyStart,
                                                           ViReal64* frequencyStop,
                                                           ViInt32*  numberOfPoints,
                                                           ViReal64* internalSignalSourcePower,
                                                           ViInt32*  sweepType);

ViStatus _VI_FUNC rsvna_QueryCalibrationDataParametersPortSpecific(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   systemErrorCorrection,
                            ViInt32   physicalPort,
                            ViInt32   type,
                            ViInt32   index,
                            ViInt32   bufferSize,
                            ViChar    info[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKit(ViSession instrSession,
                                                ViInt32   connector,
                                                ViString  calibrationKitName);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitAuto(ViSession instrSession,
                                                    ViString  calibrationKitName,
                                                    ViInt32   analyzerPort1,
                                                    ViInt32   analyzerPort2,
                                                    ViInt32   analyzerPort3,
                                                    ViInt32   analyzerPort4,
                                                    ViInt32   calUnitPort1,
                                                    ViInt32   calUnitPort2,
                                                    ViInt32   calUnitPort3,
                                                    ViInt32   calUnitPort4,
                                                    ViUInt32  timeout);

ViStatus _VI_FUNC rsvna_AddCalibrationKitAuto(ViSession instrSession,
                                              ViString  characterization,
                                              ViInt32   analyzerPort1,
                                              ViInt32   analyzerPort2,
                                              ViInt32   analyzerPort3,
                                              ViInt32   analyzerPort4,
                                              ViInt32   calUnitPort1,
                                              ViInt32   calUnitPort2,
                                              ViInt32   calUnitPort3,
                                              ViInt32   calUnitPort4,
                                              ViUInt32  timeout);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitUserConnectorType(
                            ViSession instrSession,
                            ViString  connector,
                            ViString  calibrationKitName);

ViStatus _VI_FUNC rsvna_CalibrationKitCatalog(ViSession instrSession,
                                              ViString  connectorName,
                                              ViInt32   bufferSize,
                                              ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationStandard(ViSession instrSession,
                                                     ViInt32   standardType,
                                                     ViInt32   connector,
                                                     ViString  calKitName,
                                                     ViString  standardLabel,
                                                     ViReal64  minFreqHz,
                                                     ViReal64  maxFreqHz,
                                                     ViReal64  delayParameter,
                                                     ViReal64  loss,
                                                     ViReal64  z0,
                                                     ViReal64  capacitances[],
                                                     ViReal64  residualInductances[],
                                                     ViInt32   loadCircuitModel,
                                                     ViReal64  resistance,
                                                     ViInt32   port1,
                                                     ViInt32   port2);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationStandardExt(ViSession instrSession,
                                                        ViInt32   connector,
                                                        ViInt32   standard,
                                                        ViString  kit,
                                                        ViString  serialNumber,
                                                        ViReal64  minFreqHz,
                                                        ViReal64  maxFreqHz,
                                                        ViReal64  lengthmm,
                                                        ViReal64  loss,
                                                        ViReal64  z0,
                                                        ViReal64  capacitances[],
                                                        ViReal64  residualInductances[],
                                                        ViInt32   approximation);

ViStatus _VI_FUNC rsvna_CalibrationStandardsCatalog(ViSession instrSession,
                                                    ViString  calibrationKitName,
                                                    ViInt32   bufferSize,
                                                    ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitLabel(ViSession instrSession,
                                                     ViString  calibrationKitName,
                                                     ViString  label);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitPassword(ViSession instrSession,
                                                        ViString  password);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationDelayMode(ViSession instrSession,
                                                      ViString  connectorName,
                                                      ViString  calibrationKitName,
                                                      ViString  calibrationKitLabel,
                                                      ViInt32   delayMode);

ViStatus _VI_FUNC rsvna_SaveCalibrationKit(ViSession instrSession,
                                           ViString  fileName);

ViStatus _VI_FUNC rsvna_DeleteCalibrationKit(ViSession instrSession,
                                             ViString  calibrationKitName);

ViStatus _VI_FUNC rsvna_LoadCalibrationKit(ViSession instrSession,
                                           ViString  connectorName,
                                           ViString  calibrationKitName,
                                           ViInt32   standard,
                                           ViString  calibrationKitLabel,
                                           ViString  fileName,
                                           ViInt32   portNumber1,
                                           ViInt32   portNumber2);

ViStatus _VI_FUNC rsvna_LoadCalibrationKitByGenderFromFile(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViString  calibrationKitFile,
                                                           ViInt32   gender,
                                                           ViInt32   portSelector,
                                                           ViInt32   arraySize,
                                                           ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_LoadCalibrationKitByGenderFromPool(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViString  calibrationKitName,
                                                           ViString  label,
                                                           ViInt32   gender,
                                                           ViInt32   portSelector,
                                                           ViInt32   arraySize,
                                                           ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_QueryCalibrationKitData(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   port,
                                                ViInt32   detail,
                                                ViInt32   bufferSize,
                                                ViChar    result[]);

ViStatus _VI_FUNC rsvna_ImportKit(ViSession instrSession,
                                  ViString  fileName);

ViStatus _VI_FUNC rsvna_ExportKit(ViSession instrSession,
                                  ViString  kitName,
                                  ViString  fileName);

ViStatus _VI_FUNC rsvna_QueryCalibrationKitInfo(ViSession instrSession,
                                                ViString  calKitFile,
                                                ViInt32   bufferSize,
                                                ViChar    connectorType[],
                                                ViChar    calKitName[],
                                                ViChar    calKitLabel[],
                                                ViInt32*  calKitGender);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitWithLabel(ViSession instrSession,
                                                         ViInt32   connector,
                                                         ViString  calibrationKitName,
                                                         ViString  calibrationKitLabel);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitUserConnectorTypeWithLabel(
                            ViSession instrSession,
                            ViString  connector,
                            ViString  calibrationKitName,
                            ViString  calibrationKitLabel);

ViStatus _VI_FUNC rsvna_CalibrationKitCatalogWithLabel(ViSession instrSession,
                                                       ViString  connectorName,
                                                       ViInt32   bufferSize,
                                                       ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationStandardWithLabel(
                            ViSession instrSession,
                            ViInt32   standardType,
                            ViInt32   connector,
                            ViString  calKitName,
                            ViString  calKitLabel,
                            ViString  standardLabel,
                            ViReal64  minFreqHz,
                            ViReal64  maxFreqHz,
                            ViReal64  delayParameter,
                            ViReal64  loss,
                            ViReal64  z0,
                            ViReal64  capacitances[],
                            ViReal64  residualInductances[],
                            ViInt32   loadCircuitModel,
                            ViReal64  resistance,
                            ViInt32   port1,
                            ViInt32   port2);

ViStatus _VI_FUNC rsvna_CalibrationStandardsCatalogWithLabel(ViSession instrSession,
                                                             ViString  calibrationKitName,
                                                             ViString  calibrationKitLabel,
                                                             ViInt32   bufferSize,
                                                             ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationKitLabelWithLabel(
                            ViSession instrSession,
                            ViString  calibrationKitName,
                            ViString  calibrationKitLabel,
                            ViString  newLabel);

ViStatus _VI_FUNC rsvna_DeleteCalibrationKitWithLabel(ViSession instrSession,
                                                      ViString  calibrationKitName,
                                                      ViString  calibrationKitLabel);

ViStatus _VI_FUNC rsvna_LoadCalibrationKitWithLabel(ViSession instrSession,
                                                    ViString  connectorName,
                                                    ViString  calibrationKitName,
                                                    ViString  calibrationKitLabel,
                                                    ViInt32   standard,
                                                    ViString  standardLabel,
                                                    ViString  fileName,
                                                    ViInt32   portNumber1,
                                                    ViInt32   portNumber2);

ViStatus _VI_FUNC rsvna_ReadCalibrationKitWithLabelSParameterData(
                            ViSession instrSession,
                            ViInt32   standard,
                            ViString  connectorName,
                            ViString  calibrationKitName,
                            ViString  calibrationKitLabel,
                            ViInt32   sParameter,
                            ViInt32   portNumber1,
                            ViInt32   portNumber2,
                            ViInt32   arraySize,
                            ViReal64  traceData[],
                            ViInt32*  noOfValues);

ViStatus _VI_FUNC rsvna_ExportKitWithLabel(ViSession instrSession,
                                           ViString  kitName,
                                           ViString  calkitLabel,
                                           ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibration(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   sourceType,
                                                        ViInt32   sourcePort);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerFlatnessCalibrationState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerFlatnessCalibration(
                            ViSession instrSession,
                            ViInt32   totalReadings,
                            ViReal64  tolerance,
                            ViReal64  convergence);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerFlatnessCalibrationMethod(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   method);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationPowerMeter(
                            ViSession instrSession,
                            ViInt32   externalPowerMeterNumber,
                            ViInt32   reserved);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationPowerDeembedding(
                            ViSession instrSession,
                            ViInt32   networkType,
                            ViInt32   embedding);

ViStatus _VI_FUNC rsvna_QuerySourcePowerCalibrationDataParameters(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   waveNumber,
                            ViInt32   settingType,
                            ViInt32   bufferSize,
                            ViChar    info[]);

ViStatus _VI_FUNC rsvna_QuerySourcePowerCalibrationDataParametersSwitchingMatrix(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   waveNumber,
                            ViInt32   settingType,
                            ViInt32   index,
                            ViInt32   bufferSize,
                            ViChar    info[]);

ViStatus _VI_FUNC rsvna_QuerySourcePowerCalibrationDataParametersCount(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   waveNumber,
                            ViInt32*  parametersCount);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationSource(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   source,
                            ViInt32   portNumber);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationOffset(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   portNumber,
                            ViReal64  portPowerOffset,
                            ViInt32   offsetParameter,
                            ViReal64  calibrationPowerOffset);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationGeneratorOffset(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   generator,
                            ViReal64  portPowerOffset,
                            ViInt32   offsetParameter,
                            ViReal64  calibrationPowerOffset);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationPower(ViSession instrSession,
                                                             ViInt32   powerSelect,
                                                             ViReal64  referencePower);

ViStatus _VI_FUNC rsvna_ConfigureElectronicSourceAttenuation(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   physicalPort,
                                                             ViInt32   mode,
                                                             ViInt32   attenuation);

ViStatus _VI_FUNC rsvna_ConfigureDummySourcePowerCalibrationEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureFastSourcePowerCalibrationEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationGeneratorMinimumSettlingDelay(
                            ViSession instrSession,
                            ViReal64  minimumSettlingDelay);

ViStatus _VI_FUNC rsvna_QuerySourcePowerCalibrationVerificationResults(
                            ViSession  instrSession,
                            ViChar     calibratedChannel[],
                            ViBoolean* calibrationPassed,
                            ViReal64*  maxOffset);

ViStatus _VI_FUNC rsvna_WriteSourcePowerCalibrationCorrectionData(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViString  calibratedWave,
                            ViInt32   numberOfValues,
                            ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_ReadSourcePowerCalibrationCorrectionData(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   arraySize,
                            ViString  calibratedWave,
                            ViInt32*  numberOfValues,
                            ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_WriteSourcePowerCalibrationCorrectionDataPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViString  calibratedWave,
                            ViInt32   numberOfValues,
                            ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_ReadSourcePowerCalibrationCorrectionDataPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   arraySize,
                            ViString  calibratedWave,
                            ViInt32*  numberOfValues,
                            ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerSweepEndMode(ViSession instrSession,
                                                         ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerSweepEndSettlingDelay(
                            ViSession instrSession,
                            ViReal64  delay);

ViStatus _VI_FUNC rsvna_ConfigureReceiverPowerCalibration(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   wave,
                                                          ViInt32   portNumber,
                                                          ViInt32   source,
                                                          ViInt32   sourceNumber);

ViStatus _VI_FUNC rsvna_ConfigureReceiverPowerAttenuation(ViSession instrSession,
                                                          ViInt32   physicalPort,
                                                          ViReal64  attenuation);

ViStatus _VI_FUNC rsvna_ConfigureReferenceReceiverEnabled(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean enabled);

ViStatus _VI_FUNC rsvna_WriteReceiverPowerCorrectionData(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViString  calibratedWave,
                                                         ViInt32   numberOfValues,
                                                         ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_ReadReceiverPowerCorrectionData(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   arraySize,
                                                        ViString  calibratedWave,
                                                        ViInt32*  numberOfValues,
                                                        ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_WriteReceiverPowerCorrectionDataPort(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   physicalPort,
                                                             ViString  calibratedWave,
                                                             ViInt32   numberOfValues,
                                                             ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_ReadReceiverPowerCorrectionDataPort(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   physicalPort,
                                                            ViInt32   arraySize,
                                                            ViString  calibratedWave,
                                                            ViInt32*  numberOfValues,
                                                            ViReal64  powerCorrectionValues[]);

ViStatus _VI_FUNC rsvna_ConfigurePowerMeterTestSetup(ViSession instrSession,
                                                     ViInt32   testSetup,
                                                     ViInt32   deEmbeddingPosition);

ViStatus _VI_FUNC rsvna_PowerMeterDefineCoefficient(ViSession instrSession,
                                                    ViReal64  frequency,
                                                    ViReal64  transmissionCoeficient);

ViStatus _VI_FUNC rsvna_PowerMeterInsertCoefficient(ViSession instrSession,
                                                    ViInt32   listNumber,
                                                    ViReal64  frequency,
                                                    ViReal64  transmissionCoeficient,
                                                    ViInt32   deEmbeddingPosition);

ViStatus _VI_FUNC rsvna_PowerMeterDeleteCoefficient(ViSession instrSession,
                                                    ViInt32   listNumber,
                                                    ViInt32   deEmbeddingPosition);

ViStatus _VI_FUNC rsvna_PowerMeterDeleteAllCoefficients(ViSession instrSession);

ViStatus _VI_FUNC rsvna_PowerMeterTraceToList(ViSession instrSession,
                                              ViString  trace,
                                              ViInt32   deEmbeddingPosition);

ViStatus _VI_FUNC rsvna_QueryPowerMeterNumberOfPoints(ViSession instrSession,
                                                      ViInt32   deEmbeddingPosition,
                                                      ViInt32*  numberOfPoints);

ViStatus _VI_FUNC rsvna_QueryPowerMeterPointValues(ViSession instrSession,
                                                   ViInt32   listNumber,
                                                   ViInt32   deEmbeddingPosition,
                                                   ViReal64* frequency,
                                                   ViReal64* transmissionCoefficient);

ViStatus _VI_FUNC rsvna_PowerMeterSavePowerLossList(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViString  traceFile);

ViStatus _VI_FUNC rsvna_PowerMeterLoadPowerLossList(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViString  traceFile,
                                                    ViString  trace);

ViStatus _VI_FUNC rsvna_StartCalibrationIntermodulation(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   timeout,
                                                        ViInt32   calibrate);

ViStatus _VI_FUNC rsvna_RunCalibrationValidation(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationValidationStandard(
                            ViSession instrSession,
                            ViInt32   standard);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationValidationCharacterization(
                            ViSession instrSession,
                            ViString  characterization);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationValidationMagnitudeDeviation(
                            ViSession instrSession,
                            ViInt32   port,
                            ViReal64  lowerDeviation,
                            ViReal64  upperDeviation);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationValidationPhaseDeviation(
                            ViSession instrSession,
                            ViInt32   port,
                            ViInt32   lowerDeviation,
                            ViInt32   upperDeviation);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationValidationAverageEnabled(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationValidationFormat(ViSession instrSession,
                                                             ViInt32   calibrationValidationFormat);

ViStatus _VI_FUNC rsvna_CalibrationValidationPreset(ViSession instrSession);

ViStatus _VI_FUNC rsvna_QueryCalibrationValidationResult(ViSession instrSession,
                                                         ViInt32   bufferSize,
                                                         ViChar    result[]);

ViStatus _VI_FUNC rsvna_CalibrationValidationImaginary(ViSession instrSession,
                                                       ViInt32   port,
                                                       ViReal64  lower,
                                                       ViReal64  upper);

ViStatus _VI_FUNC rsvna_CalibrationValidationReal(ViSession instrSession,
                                                  ViInt32   port,
                                                  ViReal64  lower,
                                                  ViReal64  upper);

ViStatus _VI_FUNC rsvna_ConfigureMETASCalibrationActive(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViBoolean METASCalibrationActive);

ViStatus _VI_FUNC rsvna_QueryMETASCableCatalog(ViSession instrSession,
                                               ViInt32   bufferSize,
                                               ViChar    catalog[],
                                               ViInt32*  returnedBytes);

ViStatus _VI_FUNC rsvna_ConfigureMETASCableToPorts(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViString  cableNameFile,
                                                   ViInt32   arrayLength,
                                                   ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_METASCableReconnection(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   arrayLength,
                                               ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_METASClearCableReconnectionHistory(ViSession instrSession,
                                                           ViInt32   channel);

ViStatus _VI_FUNC rsvna_METASCableMovement(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   arrayLength,
                                           ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_METASClearCableMovementHistory(ViSession instrSession,
                                                       ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMETASUncertainlyBand(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViReal64  uncertainlyBand);

ViStatus _VI_FUNC rsvna_QueryAdvancedHarmonicOrders(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   bufferSize,
                                                    ViChar    orders[]);

ViStatus _VI_FUNC rsvna_AddAdvancedHarmonicOrder(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   add);

ViStatus _VI_FUNC rsvna_RemoveAdvancedHarmonicOrder(ViSession instrSession,
                                                    ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureAdvancedPower(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   physicalPort,
                                               ViBoolean enabled,
                                               ViReal64  power);

ViStatus _VI_FUNC rsvna_ConfigureAdvancedPowerAttenuator(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort,
                                                         ViBoolean enabled,
                                                         ViReal64  attenuator);

ViStatus _VI_FUNC rsvna_ConfigureAdvancedPowerSourceAttenuator(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViBoolean enabled,
                            ViReal64  attenuator);

ViStatus _VI_FUNC rsvna_ConfigureAdvancedPowerPreamplifier(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   physicalPort,
                                                           ViBoolean enabled,
                                                           ViReal64  preamplifier);

ViStatus _VI_FUNC rsvna_NoiseFigureCalibrationTerminate(ViSession instrSession,
                                                        ViInt32   channel);

ViStatus _VI_FUNC rsvna_SaveNoiseFigureCalibration(ViSession instrSession,
                                                   ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationStart(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   DUTOutput,
                                                             ViInt32   DUTInput,
                                                             ViBoolean extAtt,
                                                             ViReal64  attenuationNoise,
                                                             ViReal64  attenuationMeas);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationSweep(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   sweep);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationTime(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViReal64  noiseTime,
                                                            ViReal64  gainTime);

ViStatus _VI_FUNC rsvna_InitiateNoiseFigureCalibrationMeasurement(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   receivePort,
                            ViInt32   drivePort);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationAVGDetectorTime(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  AVGDetectorTime);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationSidebandCorrection(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled,
                            ViReal64  minimumFrequency,
                            ViReal64  frequencyLimit,
                            ViInt32   maximumOrder);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationIFFrequency(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled,
                            ViReal64  IFFrequency);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationRMSDetectorTime(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  RMSDetectorTime);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationDUTCorrectionEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationAmbientTemperature(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   ambientTemperature);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationVirtualImageRejection(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean virtualImageRejection);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationSegmentVirtualImageRejection(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   segment,
                            ViBoolean virtualImageRejection);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureCalibrationSegmentVirtualImageControlEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureQuicksetExternalAttenuatorReceiver(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  externalAttenuatorReceiver);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureQuicksetExternalAttenuatorSource(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  externalAttenuatorSource);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureQuicksetTraceNoise(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  traceNoise);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureQuicksetDUTExpectedNoiseFigure(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  DUTExpectedNoiseFigure);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureQuicksetDUTExpectedGain(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  DUTExpectedGain);

ViStatus _VI_FUNC rsvna_ConfigureNoiseFigureQuicksetDUTPower(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViReal64  DUTPower);

ViStatus _VI_FUNC rsvna_NoiseFigureQuicksetExecute(ViSession instrSession,
                                                   ViInt32   channel);

ViStatus _VI_FUNC rsvna_StartMixerPhaseIFCalibration(ViSession instrSession,
                                                     ViInt32   channel);

ViStatus _VI_FUNC rsvna_StartCalibrationMixerPhaseIFSecondStep(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_StartCalibrationMixerPhaseRFCalibration(
                            ViSession instrSession,
                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_StartCalibrationMixerPhaseLOCalibration(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   mixerStage);

ViStatus _VI_FUNC rsvna_SourceLevelingClearAll(ViSession instrSession);

ViStatus _VI_FUNC rsvna_SourceLevelingClear(ViSession instrSession,
                                            ViInt32   physicalPort);

ViStatus _VI_FUNC rsvna_SourceLevelingRemove(ViSession instrSession,
                                             ViInt32   physicalPort);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingLoad(ViSession instrSession,
                                                    ViInt32   physicalPort,
                                                    ViString  sourceLevelingLoad);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingStore(ViSession instrSession,
                                                     ViInt32   physicalPort,
                                                     ViString  sourceLevelingStore);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingNumberofPoints(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViInt32   numberOfPoints);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingStartFrequency(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViReal64  startFrequency);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingStopFrequency(ViSession instrSession,
                                                             ViInt32   physicalPort,
                                                             ViReal64  stopFrequency);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingStartPower(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   physicalPort,
                                                          ViReal64  startPower);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingStopPower(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort,
                                                         ViReal64  stopPower);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingALCStepSize(ViSession instrSession,
                                                           ViInt32   physicalPort,
                                                           ViReal64  ALCStepSize);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingPowerStepSize(ViSession instrSession,
                                                             ViInt32   physicalPort,
                                                             ViReal64  powerStepSize);

ViStatus _VI_FUNC rsvna_QuerySourceLevelingState(ViSession instrSession,
                                                 ViInt32   physicalPort,
                                                 ViInt32   bufferSize,
                                                 ViChar    state[]);

ViStatus _VI_FUNC rsvna_SourceLevelingStartCalSweep(ViSession instrSession,
                                                    ViInt32   physicalPort);

ViStatus _VI_FUNC rsvna_ConfigureSourceLevelingPortEnabled(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   physicalPort,
                                                           ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementEnabled(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementMode(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementPortEnabled(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   physicalPort,
                                                           ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementConvergenceFactor(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViReal64  convergenceFactor);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementDetector(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   physicalPort,
                                                        ViInt32   detector);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementMaxPrecalSweeps(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   maxPrecalSweeps);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementTolerance(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   physicalPort,
                                                         ViReal64  tolerance);

ViStatus _VI_FUNC rsvna_ConfigurePremeasurementReceiver(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   physicalPort,
                                                        ViString  premeasurementReceiver);

ViStatus _VI_FUNC rsvna_ChannelAdd(ViSession instrSession,
                                   ViInt32   channel,
                                   ViString  channelName);

ViStatus _VI_FUNC rsvna_ChannelAddTrace(ViSession instrSession,
                                        ViInt32   channel,
                                        ViString  channelName,
                                        ViString  traceName);

ViStatus _VI_FUNC rsvna_ChannelAddTraceDiagramArea(ViSession instrSession,
                                                   ViInt32   window,
                                                   ViInt32   channel,
                                                   ViString  channelName,
                                                   ViString  traceName);

ViStatus _VI_FUNC rsvna_ChannelDelete(ViSession instrSession,
                                      ViInt32   channel);

ViStatus _VI_FUNC rsvna_ChannelRename(ViSession instrSession,
                                      ViInt32   channel,
                                      ViString  channelName);

ViStatus _VI_FUNC rsvna_ChannelList(ViSession instrSession,
                                    ViChar    catalog[],
                                    ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_QueryChannelNumber(ViSession instrSession,
                                           ViString  channelName,
                                           ViInt32*  channelNumber);

ViStatus _VI_FUNC rsvna_ConfigureActiveChannel(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureDrivingMode(ViSession instrSession,
                                             ViInt32   drivingMode);

ViStatus _VI_FUNC rsvna_ConfigureDrivingModeAlternatedBlocks(ViSession instrSession,
                                                             ViInt32   alternatedBlocks);

ViStatus _VI_FUNC rsvna_ConfigureTrackLO(ViSession instrSession,
                                         ViInt32   channel,
                                         ViBoolean trackLO);

ViStatus _VI_FUNC rsvna_QueryTrackLOFrequencyDifference(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViReal64* frequencyDifference);

ViStatus _VI_FUNC rsvna_ConfigureLOUsage(ViSession instrSession,
                                         ViInt32   channel,
                                         ViInt32   LOUsage);

ViStatus _VI_FUNC rsvna_ConfigurePhaseMode(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   phaseMode);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerThruMeas(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean enabled,
                                                     ViReal64  powerThruMeas);

ViStatus _VI_FUNC rsvna_DefineLogicalPort(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   logicalPort,
                                          ViInt32   physicalPort1,
                                          ViInt32   physicalPort2);

ViStatus _VI_FUNC rsvna_DeleteLogicalPort(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   logicalPort);

ViStatus _VI_FUNC rsvna_DeleteAllLogicalPorts(ViSession instrSession,
                                              ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigurePhysicalPortReferenceImpedance(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViReal64  real,
                            ViReal64  imaginary);

ViStatus _VI_FUNC rsvna_ConfigureRFFESettings(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   RFFENumber,
                                              ViBoolean RFFEState,
                                              ViReal64  frequency,
                                              ViReal64  IOVoltage,
                                              ViReal64  lowVoltage,
                                              ViReal64  highVoltage);

ViStatus _VI_FUNC rsvna_ConfigureRFFECommand(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   RFFENumber,
                                             ViString  command);

ViStatus _VI_FUNC rsvna_ConfigureGPIO(ViSession instrSession,
                                      ViInt32   channel,
                                      ViInt32   GPIOPortNumber,
                                      ViBoolean GPIOState,
                                      ViReal64  voltage);

ViStatus _VI_FUNC rsvna_ConfigureGPIORange(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   GPIOPortNumber,
                                           ViReal64  range);

ViStatus _VI_FUNC rsvna_QueryGPIOShunt(ViSession instrSession,
                                       ViInt32   channel,
                                       ViInt32   GPIOPortNumber,
                                       ViReal64* shunt);

ViStatus _VI_FUNC rsvna_GPIOStartAllMeasurements(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureGPIOMeasurementTime(ViSession instrSession,
                                                     ViReal64  time);

ViStatus _VI_FUNC rsvna_QueryGPIOSum(ViSession instrSession,
                                     ViInt32   arraySize,
                                     ViInt32   ports[],
                                     ViReal64* result);

ViStatus _VI_FUNC rsvna_GPIOCurrentMeasurement(ViSession instrSession,
                                               ViInt32   GPIOPortNumber,
                                               ViReal64* measurement);

ViStatus _VI_FUNC rsvna_GPIOCurrentMeasurementAll(ViSession instrSession,
                                                  ViInt32   arraySize,
                                                  ViReal64  results[],
                                                  ViInt32*  returnedValues);

ViStatus _VI_FUNC rsvna_GPIOVoltageMeasurement(ViSession instrSession,
                                               ViInt32   GPIOPortNumber,
                                               ViReal64* measurement);

ViStatus _VI_FUNC rsvna_GPIOVoltageMeasurementAll(ViSession instrSession,
                                                  ViInt32   arraySize,
                                                  ViReal64  results[],
                                                  ViInt32*  returnedValues);

ViStatus _VI_FUNC rsvna_QuerySweepSequencerCount(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   sweepType,
                                                 ViInt32   sweepSegment,
                                                 ViInt32*  count);

ViStatus _VI_FUNC rsvna_DeleteAllSweepSequencerItems(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   sweepType,
                                                     ViInt32   sweepSegment);

ViStatus _VI_FUNC rsvna_ConfigureSweepSequencerRFFECommand(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   sweepType,
                                                           ViInt32   sweepSegment,
                                                           ViInt32   sequenceNumber,
                                                           ViInt32   RFFENumber,
                                                           ViString  command);

ViStatus _VI_FUNC rsvna_ConfigureSweepSequencerGPIOVoltage(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   sweepType,
                                                           ViInt32   sweepSegment,
                                                           ViInt32   sequenceNumber,
                                                           ViInt32   GPIOPortNumber,
                                                           ViReal64  GPIOVoltage);

ViStatus _VI_FUNC rsvna_ConfigureSweepSequencerDelayTime(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   sweepType,
                                                         ViInt32   sweepSegment,
                                                         ViInt32   sequenceNumber,
                                                         ViReal64  delayTime);

ViStatus _VI_FUNC rsvna_RFFETestStartAllMeasurements(ViSession instrSession);

ViStatus _VI_FUNC rsvna_RFFETestApply(ViSession instrSession,
                                      ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureRFFETestMeasurementTime(ViSession instrSession,
                                                         ViReal64  time);

ViStatus _VI_FUNC rsvna_ConfigureRFFETestData(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   RFFENumber,
                                              ViInt32   pins,
                                              ViReal64  voltage,
                                              ViReal64  range);

ViStatus _VI_FUNC rsvna_QueryRFFETestDataShunt(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   RFFENumber,
                                               ViInt32   pins,
                                               ViReal64* shunt);

ViStatus _VI_FUNC rsvna_QueryRFFETestData(ViSession instrSession,
                                          ViInt32   RFFENumber,
                                          ViInt32   pins,
                                          ViReal64* voltage,
                                          ViReal64* current);

ViStatus _VI_FUNC rsvna_DefinePortGroupContinuous(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   group,
                                                  ViInt32   logicalPort1,
                                                  ViInt32   logicalPort2);

ViStatus _VI_FUNC rsvna_DefinePortGroupArbitrary(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   group,
                                                 ViInt32   arraySize,
                                                 ViInt32   logicalPorts[]);

ViStatus _VI_FUNC rsvna_DeletePortGroup(ViSession instrSession,
                                        ViInt32   channel,
                                        ViInt32   group);

ViStatus _VI_FUNC rsvna_DeleteAllPortGroups(ViSession instrSession,
                                            ViInt32   channel);

ViStatus _VI_FUNC rsvna_QueryPortGroupsCount(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32*  count);

ViStatus _VI_FUNC rsvna_DefineParallelMeasurementPortGroup(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   group,
                                                           ViInt32   arraySize,
                                                           ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_ConfigureParallelMeasurementPortGroupName(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViString  groupName);

ViStatus _VI_FUNC rsvna_ConfigureParallelMeasurementNumberOfPorts(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViInt32   numberOfPorts);

ViStatus _VI_FUNC rsvna_ConfigureParallelMeasurementDUTPort(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   group,
                                                            ViInt32   physicalPort,
                                                            ViInt32   DUTPort);

ViStatus _VI_FUNC rsvna_ConfigureParallelMeasurementPortGroupsPortOrder(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViString  portOrder);

ViStatus _VI_FUNC rsvna_ConfigureParallelMeasurementPortGroupsWithFrequencyOffsetEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryParallelMeasurementPortGroupsStateIndicationWarning(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViInt32   bufferSize,
                            ViChar    stateIndicationWarning[]);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyConversionMeasurementMode(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyConversionMixer(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  IFFrequency,
                                                          ViReal64  LOFrequency,
                                                          ViReal64  RFFrequency);

ViStatus _VI_FUNC rsvna_ConfigureSourcePowerCalibrationSlope(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   portNumber,
                                                             ViReal64  slope);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyConversion(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   portNumber,
                                                     ViInt32   numerator,
                                                     ViInt32   denominator,
                                                     ViReal64  offset,
                                                     ViInt32   sweepType);

ViStatus _VI_FUNC rsvna_ConfigureReceiverFrequencyConversion(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   powerMeterNumber,
                                                             ViInt32   numerator,
                                                             ViInt32   denominator,
                                                             ViReal64  offset,
                                                             ViInt32   sweepType);

ViStatus _VI_FUNC rsvna_ConfigureGeneratorFrequencyConversion(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   generatorNumber,
                            ViBoolean state,
                            ViInt32   numerator,
                            ViInt32   denominator,
                            ViReal64  offset,
                            ViInt32   sweepType);

ViStatus _VI_FUNC rsvna_ConfigurePortConfigUsePrimedWaves(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean usePrimedWaves);

ViStatus _VI_FUNC rsvna_ConfigurePortConfigPrimedWavesFrequency(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViReal64  frequency);

ViStatus _VI_FUNC rsvna_ConfigurePortConfigWavesFrequencyOffset(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViReal64  frequencyOffset);

ViStatus _VI_FUNC rsvna_ConfigurePortConfigReceiverFrequencyConversion(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViInt32   conversion);

ViStatus _VI_FUNC rsvna_ConfigureDirectReceiverInput(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   physicalPort,
                                                     ViInt32   directReceiverInput);

ViStatus _VI_FUNC rsvna_ConfigureDirectRearInputPath(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   physicalPort,
                                                     ViInt32   inputPath);

ViStatus _VI_FUNC rsvna_ConfigureDirectAutoDetermineIFOut(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   physicalPort,
                                                          ViBoolean autoDetermineIFOut);

ViStatus _VI_FUNC rsvna_ConfigureDirectRearIFFrequency(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   physicalPort,
                                                       ViInt32   numerator,
                                                       ViInt32   denominator,
                                                       ViReal64  offset,
                                                       ViInt32   sweepType);

ViStatus _VI_FUNC rsvna_ConfigureDirectRearOut(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   physicalPort,
                                               ViInt32   rearOut);

ViStatus _VI_FUNC rsvna_QueryDirectMeasurementInput(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   physicalPort,
                                                    ViInt32*  measurementInput);

ViStatus _VI_FUNC rsvna_QueryDirectReferenceInput(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   physicalPort,
                                                  ViInt32*  referenceInput);

ViStatus _VI_FUNC rsvna_QueryDirectMeasurementRearInputPath(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   physicalPort,
                                                            ViInt32*  inputPath);

ViStatus _VI_FUNC rsvna_QueryDirectReferenceRearInputPath(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   physicalPort,
                                                          ViInt32*  inputPath);

ViStatus _VI_FUNC rsvna_QueryDirectMeasurementAutoDetermineIFOut(
                            ViSession  instrSession,
                            ViInt32    channel,
                            ViInt32    physicalPort,
                            ViBoolean* autoDetermineIFOut);

ViStatus _VI_FUNC rsvna_QueryDirectReferenceAutoDetermineIFOut(
                            ViSession  instrSession,
                            ViInt32    channel,
                            ViInt32    physicalPort,
                            ViBoolean* autoDetermineIFOut);

ViStatus _VI_FUNC rsvna_QueryDirectMeasurementIFFrequency(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   physicalPort,
                                                          ViInt32*  numerator,
                                                          ViInt32*  denominator,
                                                          ViReal64* offset,
                                                          ViInt32*  sweepType);

ViStatus _VI_FUNC rsvna_QueryDirectReferenceIFFrequency(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   physicalPort,
                                                        ViInt32*  numerator,
                                                        ViInt32*  denominator,
                                                        ViReal64* offset,
                                                        ViInt32*  sweepType);

ViStatus _VI_FUNC rsvna_QueryDirectMeasurementRearOut(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   physicalPort,
                                                      ViInt32*  rearOut);

ViStatus _VI_FUNC rsvna_QueryDirectReferenceRearOut(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   physicalPort,
                                                    ViInt32*  rearOut);

ViStatus _VI_FUNC rsvna_ConfigureRearLOFrequency(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViReal64  numerator,
                                                 ViReal64  denominator,
                                                 ViReal64  offset,
                                                 ViInt32   sweepType);

ViStatus _VI_FUNC rsvna_ConfigureRearLOOutputPowerMode(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean powerMode);

ViStatus _VI_FUNC rsvna_ConfigureRearLOPermanentEnabled(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureRearLOPowerOffset(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64  powerOffset);

ViStatus _VI_FUNC rsvna_ConfigureAGCMode(ViSession instrSession,
                                         ViInt32   measurementChannel,
                                         ViInt32   reserved);

ViStatus _VI_FUNC rsvna_ConfigureAGCGlobalMode(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigureAGC(ViSession instrSession,
                                     ViInt32   channel,
                                     ViInt32   drivePort,
                                     ViInt32   wave,
                                     ViInt32   waveNumber,
                                     ViInt32   gainMode);

ViStatus _VI_FUNC rsvna_ConfigureAGCAllAAndBWaves(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   gainControl);

ViStatus _VI_FUNC rsvna_ConfigureAGCSegmentedEnable(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean enable);

ViStatus _VI_FUNC rsvna_ConfigureAGCSegment(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   segment,
                                            ViInt32   drivePort,
                                            ViInt32   wave,
                                            ViInt32   waveNumber,
                                            ViInt32   gainMode);

ViStatus _VI_FUNC rsvna_ConfigureAGCSegmentAllAAndBWaves(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   segment,
                                                         ViInt32   gainControl);

ViStatus _VI_FUNC rsvna_ConfigureAGCGlobalMethod(ViSession instrSession,
                                                 ViInt32   method);

ViStatus _VI_FUNC rsvna_StartAGCLearningSweep(ViSession instrSession,
                                              ViInt32   channel);

ViStatus _VI_FUNC rsvna_SaveAGCLearningResults(ViSession instrSession,
                                               ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMixerNumberOfStages(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   mixerStages);

ViStatus _VI_FUNC rsvna_ConfigureMixerPorts(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   RFPort,
                                            ViInt32   IFPort);

ViStatus _VI_FUNC rsvna_ConfigureMixerLO(ViSession instrSession,
                                         ViInt32   channel,
                                         ViInt32   stage,
                                         ViInt32   source,
                                         ViInt32   port);

ViStatus _VI_FUNC rsvna_ConfigureMixerFrequency(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   type,
                                                ViInt32   port,
                                                ViReal64  fixedFrequency);

ViStatus _VI_FUNC rsvna_ConfigureMixerPower(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   port,
                                            ViInt32   type,
                                            ViReal64  fixedPower);

ViStatus _VI_FUNC rsvna_ConfigureMixerInputSignal(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   inputSignal);

ViStatus _VI_FUNC rsvna_ConfigureMixerMultiplication(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   port,
                                                     ViInt32   numerator,
                                                     ViInt32   denominator);

ViStatus _VI_FUNC rsvna_ConfigureMixerConversion(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   stage,
                                                 ViInt32   frequencyConversionMode);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationState(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViBoolean enable);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationSourcePort(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   configure,
                                                           ViInt32   source,
                                                           ViInt32   port);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationReceivePort(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   port);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationCalSourcePort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViString  sourcePort);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationTwoToneSignalSource(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   source);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationOrderState(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   IMOrder,
                                                           ViBoolean enable);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationMaximumOrder(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   maximumOrder);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationToneDistance(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViReal64  toneDistance);

ViStatus _VI_FUNC rsvna_ResetIntermodulation(ViSession instrSession,
                                             ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureIntermodulationCombiner(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   combiner);

ViStatus _VI_FUNC rsvna_IntermodulationAddCWMode(ViSession instrSession,
                                                 ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureTriggerSource(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   triggerSource);

ViStatus _VI_FUNC rsvna_ConfigureTriggerSettings(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   triggerMeasSequence,
                                                 ViReal64  triggerDelay,
                                                 ViInt32   slopeLevel);

ViStatus _VI_FUNC rsvna_ConfigureMultipleTriggerSource(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   triggerMeasSequence,
                                                       ViInt32   source,
                                                       ViInt32   triggerEvent);

ViStatus _VI_FUNC rsvna_ConfigureMultipleTriggerSlope(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   triggerSignal,
                                                      ViInt32   triggerMeasSequence,
                                                      ViInt32   slopeLevel);

ViStatus _VI_FUNC rsvna_ConfigureMultipleTriggerDelay(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   triggerMeasSequence,
                                                      ViReal64  triggerDelay);

ViStatus _VI_FUNC rsvna_ConfigureTriggerExternalInput(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   externalInput);

ViStatus _VI_FUNC rsvna_ConfigureAUXTriggerOutEnable(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   connector,
                                                     ViBoolean enable);

ViStatus _VI_FUNC rsvna_ConfigureAUXTriggerOut(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   connector,
                                               ViInt32   polarity,
                                               ViInt32   position,
                                               ViInt32   interval,
                                               ViReal64  duration);

ViStatus _VI_FUNC rsvna_ConfigureAUXTriggerOutDelay(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   connector,
                                                    ViReal64  delay);

ViStatus _VI_FUNC rsvna_ConfigureAUXTriggerOutSignalType(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   connector,
                                                         ViInt32   signalType);

ViStatus _VI_FUNC rsvna_ConfigureTriggerLogic(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   sequence,
                                              ViInt32   logic);

ViStatus _VI_FUNC rsvna_QueryTriggerStatusReady(ViSession instrSession,
                                                ViInt32   bufferSize,
                                                ViChar    triggerStatusReady[]);

ViStatus _VI_FUNC rsvna_TriggerReadyTimeout(ViSession  instrSession,
                                            ViReal64   timeout,
                                            ViBoolean* output);

ViStatus _VI_FUNC rsvna_ConfigureLostTriggerEnabled(ViSession instrSession,
                                                    ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryLostTriggerCount(ViSession instrSession,
                                              ViInt32*  count);

ViStatus _VI_FUNC rsvna_LostTriggerReset(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureOffsetDelay(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   port,
                                             ViReal64  delay);

ViStatus _VI_FUNC rsvna_ConfigureOffsetElectricalLength(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   port,
                                                        ViReal64  electricalLength);

ViStatus _VI_FUNC rsvna_ConfigureOffsetMechanicalLength(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   port,
                                                        ViReal64  mechanicalLength);

ViStatus _VI_FUNC rsvna_ConfigureOffsetPermittivity(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   port,
                                                    ViReal64  permittivity);

ViStatus _VI_FUNC rsvna_ConfigureOneWayLoss(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   port,
                                            ViReal64  lossAtDC,
                                            ViReal64  lossAtFrequency,
                                            ViReal64  frequencyForLoss);

ViStatus _VI_FUNC rsvna_ConfigureOneWayLossSecondFrequency(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   port,
                                                           ViBoolean useSecondFrequency,
                                                           ViReal64  loss,
                                                           ViReal64  frequency);

ViStatus _VI_FUNC rsvna_AutoLength(ViSession instrSession,
                                   ViInt32   channel,
                                   ViInt32   port);

ViStatus _VI_FUNC rsvna_AutoLengthAndLoss(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   port);

ViStatus _VI_FUNC rsvna_ResetOffsets(ViSession instrSession,
                                     ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureOffsetCompensation(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   port,
                                                    ViBoolean compensation);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkSingleEndedState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   physicalPortNumber,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkSingleEndedPort(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   physicalPortNumber,
                            ViInt32   parameterType,
                            ViInt32   parameterNumber,
                            ViInt32   circuitModel,
                            ViReal64  parameterValue);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkSingleEndedData(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   physicalPortNumber,
                            ViInt32   interchange,
                            ViString  touchstoneFile);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkSingleEndedCircuitModel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   physicalPortNumber,
                            ViInt32   circuitModel);

ViStatus _VI_FUNC rsvna_LoadSingleEndedPortCircuitModelData(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   functionType,
                                                            ViInt32   physicalPortNumber,
                                                            ViString  fileName);

ViStatus _VI_FUNC rsvna_LoadSingleEndedPortCircuitModelDataInverted(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   physicalPortNumber,
                            ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkABalancedState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkBalancedPort(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   functionType,
                                                            ViInt32   logicalPortNumber,
                                                            ViInt32   parameterType,
                                                            ViInt32   parameterNumber,
                                                            ViInt32   circuitModel,
                                                            ViReal64  parameterValue);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkBalancedData(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   functionType,
                                                            ViInt32   logicalPortNumber,
                                                            ViInt32   portAssignment,
                                                            ViInt32   interchange,
                                                            ViString  touchstoneFile);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkBalancedCircuitModel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViInt32   circuitModel);

ViStatus _VI_FUNC rsvna_LoadBalancedPortCircuitModelData(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   functionType,
                                                         ViInt32   logicalPortNumber,
                                                         ViString  fileName,
                                                         ViInt32   parameter);

ViStatus _VI_FUNC rsvna_DefineVirtualNetworkPortSet(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   functionType,
                                                    ViInt32   portSet,
                                                    ViInt32   arraySize,
                                                    ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_DefineVirtualNetworkPortPair(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   functionType,
                                                     ViInt32   portPair,
                                                     ViInt32   arraySize,
                                                     ViInt32   ports[]);

ViStatus _VI_FUNC rsvna_DeleteVirtualNetworkPortPair(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   functionType);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkPortPairState(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   functionType,
                                                             ViInt32   portPair,
                                                             ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkPortPair(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   functionType,
                                                        ViInt32   portPair,
                                                        ViInt32   parameterType,
                                                        ViInt32   parameterNumber,
                                                        ViInt32   circuitModel,
                                                        ViReal64  parameterValue);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkPortPairCircuitModel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   portPair,
                            ViInt32   circuitModel);

ViStatus _VI_FUNC rsvna_LoadPortPairCircuitModelData(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   functionType,
                                                     ViInt32   portPair,
                                                     ViString  fileName,
                                                     ViInt32   parameter,
                                                     ViBoolean interchangePortNumbers);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkGroundLoopState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkGroundLoop(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   functionType,
                                                          ViInt32   parameterType,
                                                          ViInt32   circuitModel,
                                                          ViReal64  groundLoopValue);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkGroundLoopCircuitModel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   circuitModel);

ViStatus _VI_FUNC rsvna_LoadGroundLoopCircuitModelData(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   functionType,
                                                       ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkDifferentialMatchState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkDifferentialMatch(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViInt32   parameterType,
                            ViInt32   parameterNumber,
                            ViInt32   circuitModel,
                            ViReal64  parameterValue);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkDifferentialMatchData(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViInt32   interchange,
                            ViString  touchstoneFile);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkDifferentialMatchCircuitModel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViInt32   circuitModel);

ViStatus _VI_FUNC rsvna_LoadDifferentialMatchCircuitModelData(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   functionType,
                            ViInt32   logicalPortNumber,
                            ViString  fileName,
                            ViBoolean interchangePortNumbers);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkWaveDeEmbedEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkGroundLoopPerPortGroup(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean portGroup);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkPortGroupGroundLoopState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViInt32   functionType,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkPortGroupGroundLoop(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViInt32   functionType,
                            ViInt32   parameterType,
                            ViInt32   circuitModel,
                            ViReal64  groundLoopValue);

ViStatus _VI_FUNC rsvna_ConfigureVirtualNetworkPortGroupGroundLoopCircuitModel(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViInt32   functionType,
                            ViInt32   circuitModel);

ViStatus _VI_FUNC rsvna_LoadPortGroupGroundLoopCircuitModelData(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   group,
                            ViInt32   functionType,
                            ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingTimestampFilenames(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingCouponType(ViSession instrSession,
                                                           ViInt32   tool,
                                                           ViInt32   couponType);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingCouponPortActive(
                            ViSession instrSession,
                            ViInt32   tool,
                            ViInt32   port,
                            ViBoolean active);

ViStatus _VI_FUNC rsvna_MeasureFixtureModelingCoupon(ViSession instrSession,
                                                     ViInt32   tool);

ViStatus _VI_FUNC rsvna_MeasureFixtureModelingISDCouponOpenStep(ViSession instrSession);

ViStatus _VI_FUNC rsvna_MeasureFixtureModelingISDCouponShortStep(ViSession instrSession);

ViStatus _VI_FUNC rsvna_LoadFixtureModelingTouchstoneFile(ViSession instrSession,
                                                          ViInt32   tool,
                                                          ViString  fileName);

ViStatus _VI_FUNC rsvna_LoadFixtureModelingISDOpenShortTouchstoneFiles(
                            ViSession instrSession,
                            ViString  openCouponFileName,
                            ViString  shortCouponFileName);

ViStatus _VI_FUNC rsvna_FixtureModelingISDCouponInvalidatesFilePath(ViSession instrSession);

ViStatus _VI_FUNC rsvna_FixtureModelingISDDUTInvalidatesFilePath(ViSession instrSession);

ViStatus _VI_FUNC rsvna_FixtureModelingISDResetToDefault(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingISDUseImpedanceCorrection(
                            ViSession instrSession,
                            ViBoolean useImpedanceCorrection);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDUTPortActive(
                            ViSession instrSession,
                            ViInt32   tool,
                            ViInt32   port,
                            ViBoolean active);

ViStatus _VI_FUNC rsvna_MeasureFixtureModelingDUTAndTestFixture(
                            ViSession instrSession,
                            ViInt32   tool);

ViStatus _VI_FUNC rsvna_RunFixtureModelingSimulator(ViSession instrSession,
                                                    ViInt32   tool);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingApplyToPort(ViSession instrSession,
                                                            ViInt32   tool,
                                                            ViInt32   port,
                                                            ViBoolean apply);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDirectory(ViSession instrSession,
                                                          ViString  directory);

ViStatus _VI_FUNC rsvna_FixtureModelingClearDefaultDirectory(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingSFDAdvancedOptions(
                            ViSession instrSession,
                            ViInt32   _2xThruPortOrdering,
                            ViInt32   totalPortOrdering,
                            ViBoolean adjustImpedanceMismatch);

ViStatus _VI_FUNC rsvna_FixtureModelingSFDCouponInvalidatesFilePath(ViSession instrSession);

ViStatus _VI_FUNC rsvna_FixtureModelingSFDDUTInvalidatesFilePath(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingSFDUseImpedanceCorrection(
                            ViSession instrSession,
                            ViBoolean useImpedanceCorrection);

ViStatus _VI_FUNC rsvna_FixtureModelingSFDResetToDefault(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingISDTestCouponsInsertionLoss(
                            ViSession instrSession,
                            ViInt32   insertionLoss);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingISDLeadIns(ViSession instrSession,
                                                           ViReal64  scalingForAttenuation,
                                                           ViReal64  scalingForFlightTime,
                                                           ViBoolean flightTimeOverride,
                                                           ViReal64  flightTime);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingISDDUTWithFixture(
                            ViSession instrSession,
                            ViInt32   portSequence,
                            ViInt32   DUTType,
                            ViInt32   portsToSkip,
                            ViString  manualSkipPorts);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingISDCalculations(
                            ViSession instrSession,
                            ViBoolean enforcePassivity,
                            ViInt32   traceCoupling,
                            ViInt32   operationAccuracy,
                            ViReal64  maxFreqToDeembed);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingISDCalculationsDCExtrapolation(
                            ViSession instrSession,
                            ViBoolean DCExtrapolation);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDPortActive(
                            ViSession instrSession,
                            ViInt32   port,
                            ViBoolean active);

ViStatus _VI_FUNC rsvna_FixtureModelingEZDStartCouponMeasurement(ViSession instrSession);

ViStatus _VI_FUNC rsvna_FixtureModelingEZDLoadTouchstoneFile(ViSession instrSession,
                                                             ViString  file);

ViStatus _VI_FUNC rsvna_FixtureModelingEZDDUTInvalidatesFilePath(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDDUTPortActive(
                            ViSession instrSession,
                            ViInt32   port,
                            ViBoolean active);

ViStatus _VI_FUNC rsvna_FixtureModelingEZDMeasureDUT(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDLoadDUTFile(
                            ViSession instrSession,
                            ViString  file);

ViStatus _VI_FUNC rsvna_FixtureModelingRunEZDSimulator(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDApplyToPortEnabled(
                            ViSession instrSession,
                            ViInt32   port,
                            ViBoolean applyToPort);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDPortOrder(ViSession instrSession,
                                                             ViInt32   order);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDResetToDefault(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDGenerateSides(
                            ViSession instrSession,
                            ViBoolean side1,
                            ViBoolean side2);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDDCExtrapolationEnabled(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingEZDImpedanceCorrection(
                            ViSession instrSession,
                            ViBoolean enabled,
                            ViReal64  referenceImpedance,
                            ViReal64  bandwidthLimit);

ViStatus _VI_FUNC rsvna_FixtureModelingDeembedAssistantReset(ViSession instrSession);

ViStatus _VI_FUNC rsvna_FixtureModelingDeembedAssistantRun(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantUseSameCoupon(
                            ViSession instrSession,
                            ViBoolean useSameCoupon);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantIMPCorrect(
                            ViSession instrSession,
                            ViBoolean IMPCorrect);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantTool(
                            ViSession instrSession,
                            ViInt32   tool);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantRightModel(
                            ViSession instrSession,
                            ViInt32   deembedModel);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantLeftModel(
                            ViSession instrSession,
                            ViInt32   deembedModel);

ViStatus _VI_FUNC rsvna_FixtureModelingDeembedAssistantCouponLeftClear(ViSession instrSession);

ViStatus _VI_FUNC rsvna_FixtureModelingDeembedAssistantCouponRightClear(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponLeftFile(
                            ViSession instrSession,
                            ViInt32   type,
                            ViString  path);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponRightFile(
                            ViSession instrSession,
                            ViInt32   type,
                            ViString  path);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponLeftMeasure(
                            ViSession instrSession,
                            ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponRightMeasure(
                            ViSession instrSession,
                            ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponLeftPort(
                            ViSession instrSession,
                            ViInt32   position,
                            ViInt32   portNumber);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponRightPort(
                            ViSession instrSession,
                            ViInt32   position,
                            ViInt32   portNumber);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponLeftType(
                            ViSession instrSession,
                            ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantCouponRightType(
                            ViSession instrSession,
                            ViInt32   type);

ViStatus _VI_FUNC rsvna_FixtureModelingDeembedAssistantDUTMeasure(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantDUTPort(
                            ViSession instrSession,
                            ViInt32   position,
                            ViInt32   port);

ViStatus _VI_FUNC rsvna_ConfigureFixtureModelingDeembedAssistantDUTType(
                            ViSession instrSession,
                            ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLCharacterization(ViSession instrSession,
                                                        ViInt32   characterizationMethod,
                                                        ViBoolean oneLineDifferential,
                                                        ViInt32   measurement);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLPortOrder(ViSession instrSession,
                                                 ViInt32   portOrder);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLFrequenciesUseDefaultEnabled(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryDeltaLFequencyListCount(ViSession instrSession,
                                                     ViInt32   frequencyList,
                                                     ViInt32*  count);

ViStatus _VI_FUNC rsvna_QueryDeltaLFrequencyList(ViSession instrSession,
                                                 ViInt32   frequencyList,
                                                 ViInt32   arraySize,
                                                 ViReal64  frequencies[],
                                                 ViInt32*  returnedValues);

ViStatus _VI_FUNC rsvna_DeltaLAddUserFrequency(ViSession instrSession,
                                               ViReal64  frequency);

ViStatus _VI_FUNC rsvna_DeltaLDeleteUserFrequency(ViSession instrSession,
                                                  ViReal64  frequency);

ViStatus _VI_FUNC rsvna_DeltaLDeleteAllUserFrequencies(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLSweepControl(ViSession instrSession,
                                                    ViInt32   sweepControl);

ViStatus _VI_FUNC rsvna_QueryDeltaLSweepControlConfiguration(ViSession instrSession,
                                                             ViReal64* stepSize,
                                                             ViReal64* IFBandwidth);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLSweepControlDynamicBandwidth(
                            ViSession instrSession,
                            ViBoolean controlDynamicBandwidth);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLSweepRange(ViSession instrSession,
                                                  ViReal64  minimum,
                                                  ViReal64  maximum);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLUseRecommendedTraceConfigurationEnabled(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLShowTDRImpededancesEnabled(
                            ViSession instrSession,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_DeltaLLoadMeasurementFile(ViSession instrSession,
                                                  ViInt32   line,
                                                  ViString  filename);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLPortEnabled(ViSession instrSession,
                                                   ViInt32   line,
                                                   ViInt32   port,
                                                   ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLLineLength(ViSession instrSession,
                                                  ViInt32   line,
                                                  ViReal64  length);

ViStatus _VI_FUNC rsvna_DeltaLMeasureLine(ViSession instrSession,
                                          ViInt32   line);

ViStatus _VI_FUNC rsvna_DeltaLRun(ViSession instrSession);

ViStatus _VI_FUNC rsvna_DeltaLClearAll1LMeasurements(ViSession instrSession);

ViStatus _VI_FUNC rsvna_DeltaLClear1LMeasurementFile(ViSession instrSession,
                                                     ViString  filename);

ViStatus _VI_FUNC rsvna_ConfigureDeltaLDirectory(ViSession instrSession,
                                                 ViString  directory);

ViStatus _VI_FUNC rsvna_DeltaLDirectoryDefault(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureFixtureCompensationOffsetCorrection(
                            ViSession instrSession,
                            ViInt32   offsetCompensation);

ViStatus _VI_FUNC rsvna_ResetFixtureCompensation(ViSession instrSession,
                                                 ViInt32   channel);

ViStatus _VI_FUNC rsvna_StartFixtureCompensation(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   arraySize,
                                                 ViInt32   terminatedType,
                                                 ViInt32   portNumbers[]);

ViStatus _VI_FUNC rsvna_ApplyFixtureCompensation(ViSession instrSession,
                                                 ViInt32   channel);

ViStatus _VI_FUNC rsvna_ImportFixtureCompensation(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViString  fixtureFile,
                                                  ViInt32   terminatedType,
                                                  ViInt32   arraySize,
                                                  ViInt32   portNumbers[]);

ViStatus _VI_FUNC rsvna_ExportFixtureCompensation(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViString  fixtureFile,
                                                  ViInt32   terminatedType,
                                                  ViInt32   arraySize,
                                                  ViInt32   portNumbers[]);

ViStatus _VI_FUNC rsvna_ConfigureHarmonicAutomaticGridEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryIsCurrentFrequencyHarmonic(ViSession  instrSession,
                                                        ViInt32    channel,
                                                        ViBoolean* isHarmonic);

ViStatus _VI_FUNC rsvna_ConfigurePipeliningState(ViSession instrSession,
                                                 ViInt32   pipelining);

ViStatus _VI_FUNC rsvna_QueryPipeliningChannelResults(ViSession instrSession,
                                                      ViInt32   bufferSize,
                                                      ViChar    result[],
                                                      ViInt32*  returnedBytes);

ViStatus _VI_FUNC rsvna_DisableMixerDelay(ViSession instrSession,
                                          ViInt32   channel);

ViStatus _VI_FUNC rsvna_MixerDelayCalibrationStart(ViSession instrSession,
                                                   ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureMixerDelay2ToneDeltaFrequency(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  deltaFrequency);

ViStatus _VI_FUNC rsvna_ConfigureMixerDelayConstantDelay(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViReal64  constantDelay);

ViStatus _VI_FUNC rsvna_MixerDelayUseConstantDelay(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureMixerDelayCalibrationActive(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViBoolean active);

ViStatus _VI_FUNC rsvna_ConfigureMixerDelayLowerTone(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   lowerTone);

ViStatus _VI_FUNC rsvna_ConfigureMixerDelayGroupDelayReceiverMode(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   receiverMode);

ViStatus _VI_FUNC rsvna_ConfigureMixerDelayUpperTone(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   signal,
                                                     ViInt32   portNumber);

ViStatus _VI_FUNC rsvna_LoadMixerDelayKnownDelayValues(ViSession instrSession,
                                                       ViInt32   channelID,
                                                       ViString  filename);

ViStatus _VI_FUNC rsvna_LoadMixerDelayCorrectionData(ViSession instrSession,
                                                     ViInt32   channelID,
                                                     ViString  filename);

ViStatus _VI_FUNC rsvna_StoreMixerDelayCorrectionData(ViSession instrSession,
                                                      ViInt32   channelID,
                                                      ViString  storeCorrectionData);

ViStatus _VI_FUNC rsvna_ConfigurePulseAcquisitionTime(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViReal64  acquisitionTime);

ViStatus _VI_FUNC rsvna_ConfigurePulseCoupledEnabled(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePulseDutyCycle(ViSession instrSession,
                                                ViInt32   channel,
                                                ViBoolean fixedDutyCycleEnabled,
                                                ViReal64  dutyCycle);

ViStatus _VI_FUNC rsvna_ConfigurePulseGenerator(ViSession instrSession,
                                                ViInt32   channel,
                                                ViReal64  period,
                                                ViInt32   source,
                                                ViReal64  trainPeriod);

ViStatus _VI_FUNC rsvna_ConfigurePulseInternalEnabled(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePulseMeasurementDelay(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64  measurementDelay);

ViStatus _VI_FUNC rsvna_ConfigurePulsePeriod(ViSession instrSession,
                                             ViInt32   channel,
                                             ViReal64  pulsePeriod);

ViStatus _VI_FUNC rsvna_ConfigurePulseSyncMeasurementtoPulses(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean sync);

ViStatus _VI_FUNC rsvna_ConfigurePulseMeasurementEnabled(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePulseALCMeasurementDelay(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  ALCMeasurementDelay);

ViStatus _VI_FUNC rsvna_ConfigurePulseMeasurementGenerator(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   generatorNumber,
                                                           ViBoolean enabled,
                                                           ViReal64  delay,
                                                           ViReal64  width);

ViStatus _VI_FUNC rsvna_ConfigurePulseGeneratorTrainEnabled(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   generatorNumber,
                                                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigurePulseGeneratorTrainData(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   generatorNumber,
                                                         ViInt32   arraySize,
                                                         ViBoolean element[],
                                                         ViReal64  startTime[],
                                                         ViReal64  stopTime[]);

ViStatus _VI_FUNC rsvna_PulseGeneratorTrainDeleteAll(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   generatorNumber);

ViStatus _VI_FUNC rsvna_ConfigurePulseGeneratorTrainCount(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt32   generatorNumber,
                                                          ViInt32   segment,
                                                          ViInt32   segmentCount);

ViStatus _VI_FUNC rsvna_ConfigurePulseGeneratorTrain(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   generatorNumber,
                                                     ViInt32   segment,
                                                     ViBoolean enabled,
                                                     ViReal64  startTime,
                                                     ViReal64  stopTime);

ViStatus _VI_FUNC rsvna_ConfigurePulsePort(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   physicalPort,
                                           ViBoolean enabled,
                                           ViReal64  delay,
                                           ViReal64  width);

ViStatus _VI_FUNC rsvna_ConfigurePulsePortInverted(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   physicalPort,
                                                   ViBoolean input,
                                                   ViBoolean output);

ViStatus _VI_FUNC rsvna_ConfigurePulsePortExternalOutputEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   physicalPort,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureSourceCoherence(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean enabled,
                                                 ViBoolean optimizedEnabled);

ViStatus _VI_FUNC rsvna_ConfigureSourceCoherencePort(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   physicalPort,
                                                     ViBoolean enabled,
                                                     ViReal64  phase,
                                                     ViReal64  phaseSpan);

ViStatus _VI_FUNC rsvna_ConfigureTDIFMode(ViSession instrSession,
                                          ViInt32   channel,
                                          ViBoolean enabled,
                                          ViBoolean compensateImbalanceEnabled);

ViStatus _VI_FUNC rsvna_ConfigureTDIFModeImbalancePhase(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   port,
                                                        ViReal64  start,
                                                        ViReal64  stop);

ViStatus _VI_FUNC rsvna_ConfigureTDIFModeImbalanceAmplitude(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   port,
                                                            ViReal64  start,
                                                            ViReal64  stop);

ViStatus _VI_FUNC rsvna_ConfigureGainCompressionValue(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViReal64  gainCompression);

ViStatus _VI_FUNC rsvna_ConfigureGainCompressionPowerPoints(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   powerPoints);

ViStatus _VI_FUNC rsvna_ConfigureGainCompressionPowerPointPosition(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  start,
                            ViReal64  stop);

ViStatus _VI_FUNC rsvna_ConfigureGainCompressionPort(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   receivingPort,
                                                     ViInt32   drivingPort);

ViStatus _VI_FUNC rsvna_ConfigureGainCompressionSkip(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean skip);

ViStatus _VI_FUNC rsvna_ConfigureGainCompressionSkipOffset(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViReal64  skipOffset);

ViStatus _VI_FUNC rsvna_ConfigureSpectrumDetector(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   spectrumDetector);

ViStatus _VI_FUNC rsvna_ConfigureSpectrumImageRejection(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   imageRejection);

ViStatus _VI_FUNC rsvna_ConfigureSpectrumDetectorZeroPadding(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViBoolean detectorZeroPadding);

ViStatus _VI_FUNC rsvna_DiagramAreaAdd(ViSession instrSession,
                                       ViInt32   window);

ViStatus _VI_FUNC rsvna_DiagramAreaDelete(ViSession instrSession,
                                          ViInt32   window);

ViStatus _VI_FUNC rsvna_DiagramAreaMaximize(ViSession instrSession,
                                            ViInt32   window,
                                            ViInt32   diagramArea);

ViStatus _VI_FUNC rsvna_DiagramAreaTitle(ViSession instrSession,
                                         ViInt32   window,
                                         ViBoolean title,
                                         ViString  titleString);

ViStatus _VI_FUNC rsvna_DiagramAreaName(ViSession instrSession,
                                        ViInt32   window,
                                        ViString  areaName);

ViStatus _VI_FUNC rsvna_DiagramAreaCatalog(ViSession instrSession,
                                           ViInt32   reserved,
                                           ViChar    catalog[],
                                           ViInt32   bufferSize);

ViStatus _VI_FUNC rsvna_TraceDiagramAreaCatalog(ViSession instrSession,
                                                ViInt32   window,
                                                ViInt32   bufferSize,
                                                ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_ConfigureTraceZoom(ViSession instrSession,
                                           ViInt32   window,
                                           ViInt32   window_Trace,
                                           ViBoolean zoomState,
                                           ViReal64  topEdge,
                                           ViReal64  bottomEdge,
                                           ViReal64  leftEdge,
                                           ViReal64  rightEdge);

ViStatus _VI_FUNC rsvna_ConfigureZoomOverview(ViSession instrSession,
                                              ViInt32   window,
                                              ViBoolean zoomOverviewState);

ViStatus _VI_FUNC rsvna_ConfigureSplitType(ViSession instrSession,
                                           ViInt32   layoutMode);

ViStatus _VI_FUNC rsvna_ConfigureSplitGrid(ViSession instrSession,
                                           ViInt32   rows,
                                           ViInt32   columns);

ViStatus _VI_FUNC rsvna_ConfigureLayoutSimplified(ViSession instrSession,
                                                  ViInt32   layoutID,
                                                  ViInt32   layoutFormatMode,
                                                  ViString  layoutData);

ViStatus _VI_FUNC rsvna_ConfigureLayout(ViSession instrSession,
                                        ViString  layoutData);

ViStatus _VI_FUNC rsvna_JoinLayouts(ViSession instrSession,
                                    ViInt32   mainLayoutID,
                                    ViInt32   diagramNumber,
                                    ViInt32   subLayoutID);

ViStatus _VI_FUNC rsvna_DisplaySplitAll(ViSession instrSession);

ViStatus _VI_FUNC rsvna_DisplayOverlayAll(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureColorScheme(ViSession instrSession,
                                             ViInt32   colorScheme);

ViStatus _VI_FUNC rsvna_ConfigureRGBColor(ViSession instrSession,
                                          ViInt32   element,
                                          ViReal64  red,
                                          ViReal64  green,
                                          ViReal64  blue,
                                          ViInt32   traceStyle,
                                          ViInt32   traceWidth);

ViStatus _VI_FUNC rsvna_ConfigureTraceRGBColor(ViSession instrSession,
                                               ViString  traceName,
                                               ViReal64  red,
                                               ViReal64  green,
                                               ViReal64  blue,
                                               ViInt32   traceStyle,
                                               ViInt32   traceWidth);

ViStatus _VI_FUNC rsvna_SaveColorScheme(ViSession instrSession,
                                        ViString  fileName);

ViStatus _VI_FUNC rsvna_SaveColorSchemeHCOPY(ViSession instrSession,
                                             ViString  fileName);

ViStatus _VI_FUNC rsvna_RecallColorScheme(ViSession instrSession,
                                          ViString  fileName);

ViStatus _VI_FUNC rsvna_RecallColorSchemeHCOPY(ViSession instrSession,
                                               ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureDisplayXAxisGridLabels(ViSession instrSession,
                                                        ViBoolean xAxisGridLabels);

ViStatus _VI_FUNC rsvna_ConfigureDisplayInfoWindowEnabled(ViSession instrSession,
                                                          ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureDisplayInfoWindowBandfilterSearchResults(
                            ViSession instrSession,
                            ViString  traceName,
                            ViInt32   bandfilterItem,
                            ViBoolean displayed);

ViStatus _VI_FUNC rsvna_ConfigureDisplayInfoWindowMarkerSearchResults(
                            ViSession instrSession,
                            ViInt32   marker,
                            ViString  traceName,
                            ViBoolean displayed);

ViStatus _VI_FUNC rsvna_QueryMenuKeyDialogCatalog(ViSession instrSession,
                                                  ViInt32   bufferSize,
                                                  ViChar    catalog[],
                                                  ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_DisplayMenuKeyDialog(ViSession instrSession,
                                             ViString  menuKey);

ViStatus _VI_FUNC rsvna_QueryMenuKeyTabCatalog(ViSession instrSession,
                                               ViInt32   bufferSize,
                                               ViChar    catalog[],
                                               ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_DisplayMenuKeyTab(ViSession instrSession,
                                          ViString  menuKey);

ViStatus _VI_FUNC rsvna_ConfigureTDRStimulusLength(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64  maxDUTDelay,
                                                   ViReal64  maxElectricalLength,
                                                   ViReal64  maxMechanicalLength);

ViStatus _VI_FUNC rsvna_ConfigureTDRStimulusPermitivity(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViReal64  permitivity);

ViStatus _VI_FUNC rsvna_ConfigureTDRStimulusVelocityFactor(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViReal64  velocityFactor);

ViStatus _VI_FUNC rsvna_ConfigureTDRStimulusRiseTime(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViReal64  riseTime,
                                                     ViInt32   definition);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeEnabled(ViSession instrSession,
                                               ViInt32   channel,
                                               ViBoolean enabled);

ViStatus _VI_FUNC rsvna_TDREyeLoadBitStream(ViSession instrSession,
                                            ViString  traceName,
                                            ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeBitStream(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   bitStream,
                                                 ViInt64   symbolRate);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeBitStreamPRBSLength(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   length);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeBitStreamBitLength(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViInt64   length);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeSimulatedLevels(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   modulation,
                                                       ViReal64  highLevel,
                                                       ViReal64  lowLevel);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeSimulatedRiseTime(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViBoolean lowPass,
                                                         ViReal64  riseTime,
                                                         ViInt32   definition);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeView(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   view);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeGenerator(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean encoder,
                                                 ViBoolean scrambler);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeEmphasisEnabled(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeEmphasisCursor(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViReal64  preCursor,
                                                      ViReal64  postCursor1,
                                                      ViReal64  postCursor2);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeNoise(ViSession instrSession,
                                             ViInt32   channel,
                                             ViBoolean enabled,
                                             ViReal64  RMS);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeDUTMode(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   DUTMode);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeEqualizationEnabled(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeEqualizationCTLEEqualizer(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  DC,
                            ViReal64  zero,
                            ViReal64  pole1,
                            ViReal64  pole2);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeTransferRateUnits(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   units);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeJitterEnabled(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeJitterRandom(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean enabled,
                                                    ViReal64  standardDeviation);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeJitterPeriodic(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViBoolean enabled,
                                                      ViReal64  magnitude,
                                                      ViReal64  frequency,
                                                      ViReal64  phase);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeJitterDirac(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean enabled,
                                                   ViReal64  delta,
                                                   ViReal64  probability);

ViStatus _VI_FUNC rsvna_TDREyeLoadUserJitter(ViSession instrSession,
                                             ViString  traceName,
                                             ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeJitterUserEnabled(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskShow(ViSession instrSession,
                                                ViInt32   channel,
                                                ViBoolean show);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskEnabled(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean enabled);

ViStatus _VI_FUNC rsvna_TDREyeMaskAuto(ViSession instrSession,
                                       ViInt32   channel);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskFailBeep(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean beep);

ViStatus _VI_FUNC rsvna_QueryTDREyeMaskFailStatus(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32*  status);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskPolygon(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean enabled,
                                                   ViInt32   shape,
                                                   ViReal64  widthMain,
                                                   ViReal64  heightMain,
                                                   ViReal64  widthMinor,
                                                   ViReal64  heightMinor);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskTop(ViSession instrSession,
                                               ViInt32   channel,
                                               ViBoolean enabled,
                                               ViReal64  width,
                                               ViReal64  height);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskBottom(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViBoolean enabled,
                                                  ViReal64  width,
                                                  ViReal64  height);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskCenter(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViReal64  width,
                                                  ViReal64  height);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskFailCondition(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   failCondition,
                                                         ViInt32   toleranceRate);

ViStatus _VI_FUNC rsvna_QueryTDREyeMaskData(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   bufferSize,
                                            ViChar    data[]);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskLoad(ViSession instrSession,
                                                ViString  traceName,
                                                ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskStore(ViSession instrSession,
                                                 ViString  traceName,
                                                 ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMaskStoreResults(ViSession instrSession,
                                                        ViString  traceName,
                                                        ViString  fileName);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMeasurementEnabled(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryTDREyeMeasurementData(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   bufferSize,
                                                   ViChar    data[]);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMeasurementThreshold(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViInt32   threshold);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMeasurementThresholdCustom(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   lower,
                            ViInt32   upper);

ViStatus _VI_FUNC rsvna_ConfigureTDREyeMeasurementExport(ViSession instrSession,
                                                         ViString  traceName,
                                                         ViString  fileName,
                                                         ViInt32   decimalSeparator,
                                                         ViInt32   fieldSeparator);

ViStatus _VI_FUNC rsvna_ConfigureRiseTimeEnabled(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureRiseTimeThreshold(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   threshold);

ViStatus _VI_FUNC rsvna_ConfigureRiseTimeThresholdCustom(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   lower,
                                                         ViInt32   upper);

ViStatus _VI_FUNC rsvna_QueryRiseTimeData(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   arraySize,
                                          ViReal64  data[],
                                          ViInt32*  returnedValues);

ViStatus _VI_FUNC rsvna_ConfigureSkew(ViSession instrSession,
                                      ViInt32   channel,
                                      ViBoolean enabled,
                                      ViString  referenceTrace,
                                      ViInt32   position,
                                      ViBoolean failBeep);

ViStatus _VI_FUNC rsvna_ConfigureSkewLimitCheck(ViSession instrSession,
                                                ViInt32   channel,
                                                ViBoolean checkOn,
                                                ViReal64  skewLimit);

ViStatus _VI_FUNC rsvna_QuerySkewFail(ViSession  instrSession,
                                      ViInt32    channel,
                                      ViBoolean* skewFail);

ViStatus _VI_FUNC rsvna_QuerySkewResult(ViSession instrSession,
                                        ViInt32   channel,
                                        ViReal64* skewResult);

ViStatus _VI_FUNC rsvna_QuerySkewAllResults(ViSession instrSession,
                                            ViInt32   channel,
                                            ViInt32   arraySize,
                                            ViReal64  results[],
                                            ViInt32*  returnedValues);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorEnabled(ViSession instrSession,
                                                             ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorFrequency(
                            ViSession instrSession,
                            ViReal64  frequency);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorPower(ViSession instrSession,
                                                           ViReal64  power);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorRFOff(ViSession instrSession,
                                                           ViBoolean RFOff);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorPortEnabled(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorPortFrequency(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViReal64  frequency);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorPortPower(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViReal64  power);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorPortRFOff(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViBoolean RFOff);

ViStatus _VI_FUNC rsvna_ConfigureIndependentGeneratorPortStepAttenuator(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViReal64  stepAttenuator);

ViStatus _VI_FUNC rsvna_ConfigurePresetScope(ViSession instrSession,
                                             ViInt32   presetScope);

ViStatus _VI_FUNC rsvna_ConfigureRemotePreset(ViSession instrSession,
                                              ViBoolean alignRSTToUserDefPreset);

ViStatus _VI_FUNC rsvna_ConfigureStartInPreset(ViSession instrSession,
                                               ViBoolean startInPreset);

ViStatus _VI_FUNC rsvna_FactoryPreset(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureSetAllTracesto0(ViSession instrSession,
                                                 ViBoolean setAllTracesTo0);

ViStatus _VI_FUNC rsvna_ConfigureUserDefinedPreset(ViSession instrSession,
                                                   ViBoolean userDefinedPreset,
                                                   ViString  userDefinedPresetFile);

ViStatus _VI_FUNC rsvna_ConfigureUserDefinedPresetCalibration(
                            ViSession instrSession,
                            ViString  userDefinedPresetCalibration);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationAutoAveraging(ViSession instrSession,
                                                          ViInt32   averaging);

ViStatus _VI_FUNC rsvna_ConfigureKeyLock(ViSession instrSession,
                                         ViBoolean keyLockout);

ViStatus _VI_FUNC rsvna_ConfigureTouchscreenLock(ViSession instrSession,
                                                 ViInt32   type);

ViStatus _VI_FUNC rsvna_ConfigureUserKeyFunction(ViSession instrSession,
                                                 ViInt32   key,
                                                 ViString  label,
                                                 ViString  function);

ViStatus _VI_FUNC rsvna_ConfigureRemoteLanguage(ViSession instrSession,
                                                ViInt32   language);

ViStatus _VI_FUNC rsvna_ConfigureChannelBits(ViSession instrSession,
                                             ViInt32   channel,
                                             ViString  channelBits,
                                             ViInt32   pin1617);

ViStatus _VI_FUNC rsvna_ConfigureUserPortOptionalColumns(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViInt32   segment,
                                                         ViBoolean enabled,
                                                         ViInt32   value);

ViStatus _VI_FUNC rsvna_SaveReport(ViSession instrSession,
                                   ViInt32   timeout,
                                   ViString  reportFileName);

ViStatus _VI_FUNC rsvna_ConfigureRemoteLogging(ViSession instrSession,
                                               ViBoolean remoteLogging);

ViStatus _VI_FUNC rsvna_ConfigureRemoteEncoding(ViSession instrSession,
                                                ViInt32   encoding);

ViStatus _VI_FUNC rsvna_ConfigureCalibrationManagerState(ViSession instrSession,
                                                         ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureDisplayTitle(ViSession instrSession,
                                              ViString  displayTitle);

ViStatus _VI_FUNC rsvna_ConfigureDisplayNumberofTraceColours(ViSession instrSession,
                                                             ViInt32   numberOfTraceColours);

ViStatus _VI_FUNC rsvna_ConfigureDisplayUpdate(ViSession instrSession,
                                               ViBoolean displayInRemote);

ViStatus _VI_FUNC rsvna_DisplayUpdateOnce(ViSession instrSession);

ViStatus _VI_FUNC rsvna_QuerySystemSCPICommands(ViSession instrSession,
                                                ViString  parser,
                                                ViInt32   bufferSize,
                                                ViChar    SCPICommands[],
                                                ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_QueryImplementedSCPICommands(ViSession instrSession,
                                                     ViInt32   bufferSize,
                                                     ViChar    SCPICommands[],
                                                     ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_SystemHelpSyntax(ViSession instrSession,
                                         ViString  command,
                                         ViInt32   bufferSize,
                                         ViChar    syntax[],
                                         ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_ConfigureGPIBSupresses(ViSession instrSession,
                                               ViBoolean GPIBSupresses);

ViStatus _VI_FUNC rsvna_ConfigureHostname(ViSession instrSession,
                                          ViString  hostname);

ViStatus _VI_FUNC rsvna_DisplayCorrectionWizardDialog(ViSession instrSession,
                                                      ViInt32   dialog);

ViStatus _VI_FUNC rsvna_ConfigureReferenceOscillatorSMAInputFrequency(
                            ViSession instrSession,
                            ViInt32   inputFrequency);

ViStatus _VI_FUNC rsvna_ConfigureReferenceOscillatorSMAOutputFrequency(
                            ViSession instrSession,
                            ViInt32   outputFrequency);

ViStatus _VI_FUNC rsvna_ConfigureUserDefinedLedColor(ViSession instrSession,
                                                     ViInt32   ledColor);

ViStatus _VI_FUNC rsvna_ConfigureRFOffBehavior(ViSession instrSession,
                                               ViInt32   RFOffBehavior);

ViStatus _VI_FUNC rsvna_ConfigureSystemPresetMode(ViSession instrSession,
                                                  ViInt32   systemPresetMode);

ViStatus _VI_FUNC rsvna_QueryOPCCheckSelftest(ViSession instrSession,
                                              ViInt32*  selftest);

ViStatus _VI_FUNC rsvna_QueryLevelOPCCheckPort(ViSession  instrSession,
                                               ViInt32    physicalPort,
                                               ViReal64   power,
                                               ViString   limitLines,
                                               ViBoolean* checkPort);

ViStatus _VI_FUNC rsvna_ConfigureFormatIdentify(ViSession instrSession,
                                                ViInt32   formatIdentify);

ViStatus _VI_FUNC rsvna_ConfigureUIData(ViSession instrSession,
                                        ViInt32   port,
                                        ViInt32   data);

ViStatus _VI_FUNC rsvna_ConfigureUIDataFlowDirection(ViSession instrSession,
                                                     ViInt32   port,
                                                     ViInt32   mode,
                                                     ViInt32   data);

ViStatus _VI_FUNC rsvna_ConfigureUIOutputPortsState(ViSession instrSession,
                                                    ViInt32   outputPortNumber,
                                                    ViInt32   data,
                                                    ViInt32   state);

ViStatus _VI_FUNC rsvna_ConfigureUIDataPortsLogic(ViSession instrSession,
                                                  ViInt32   logic);

ViStatus _VI_FUNC rsvna_ConfigureUIPassFail(ViSession instrSession,
                                            ViInt32   logic,
                                            ViInt32   defaultState,
                                            ViInt32   scope,
                                            ViInt32   policy);

ViStatus _VI_FUNC rsvna_ConfigureUISweepEndEvent(ViSession instrSession,
                                                 ViInt32   event);

ViStatus _VI_FUNC rsvna_ConfigureUIDigitalSignal(ViSession instrSession,
                                                 ViInt32   signal,
                                                 ViBoolean digitalSignal);

ViStatus _VI_FUNC rsvna_QueryUIPassFailStatus(ViSession instrSession,
                                              ViInt32*  status);

ViStatus _VI_FUNC rsvna_QueryUITransitionStatus(ViSession instrSession,
                                                ViInt32*  status);

ViStatus _VI_FUNC rsvna_ConfigureReference(ViSession instrSession,
                                           ViInt32   reference);

ViStatus _VI_FUNC rsvna_SetStatusRegister(ViSession instrSession,
                                          ViInt32   registerOperation,
                                          ViInt32   questionableRegister,
                                          ViInt32   enable,
                                          ViInt32   PTransition,
                                          ViInt32   NTransition);

ViStatus _VI_FUNC rsvna_GetStatusRegister(ViSession instrSession,
                                          ViInt32   questionableRegister,
                                          ViInt32   registerPart,
                                          ViInt32*  registerValue);

ViStatus _VI_FUNC rsvna_ConfigureExternalGenerator(ViSession instrSession,
                                                   ViInt32   generatorNumber,
                                                   ViString  generatorName,
                                                   ViString  generatorType,
                                                   ViString  interfaceType,
                                                   ViString  interfaceAddress,
                                                   ViBoolean fastSweepMode,
                                                   ViBoolean referenceFrequency);

ViStatus _VI_FUNC rsvna_QueryExternalGenerator(ViSession  instrSession,
                                               ViInt32    generatorNumber,
                                               ViChar     generatorName[],
                                               ViChar     generatorType[],
                                               ViChar     interfaceType[],
                                               ViChar     interfaceAddress[],
                                               ViBoolean* fastSweepMode,
                                               ViBoolean* referenceFrequency);

ViStatus _VI_FUNC rsvna_ConfigureExternalGeneratorPower(ViSession instrSession,
                                                        ViInt32   generatorNumber,
                                                        ViInt32   mode,
                                                        ViReal64  power);

ViStatus _VI_FUNC rsvna_QueryExternalGeneratorCount(ViSession instrSession,
                                                    ViInt32*  generatorCount);

ViStatus _VI_FUNC rsvna_QueryExternalGeneratorNumbers(ViSession instrSession,
                                                      ViInt32   arraySize,
                                                      ViChar    generatorNumbers[]);

ViStatus _VI_FUNC rsvna_DeleteExternalGenerator(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureExternalGeneratorState(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   generator,
                                                        ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigurePermanentSignalGeneratorState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   generator,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureExternalPowerMeter(ViSession instrSession,
                                                    ViInt32   powerMeterNumber,
                                                    ViString  powerMeterName,
                                                    ViString  powerMeterType,
                                                    ViString  interfaceType,
                                                    ViString  interfaceAddress);

ViStatus _VI_FUNC rsvna_ConfigureExternalPowerSParameterCorrection(
                            ViSession instrSession,
                            ViInt32   powerMeterNumber,
                            ViBoolean state);

ViStatus _VI_FUNC rsvna_QueryExternalPowerMeter(ViSession instrSession,
                                                ViInt32   powerMeterNumber,
                                                ViChar    powerMeterName[],
                                                ViChar    powerMeterType[],
                                                ViChar    interfaceType[],
                                                ViChar    interfaceAddress[]);

ViStatus _VI_FUNC rsvna_QueryExternalPowerMeterCount(ViSession instrSession,
                                                     ViInt32*  powerMeterCount);

ViStatus _VI_FUNC rsvna_QueryExternalPowerMeterNumbers(ViSession instrSession,
                                                       ViInt32   bufferSize,
                                                       ViChar    powerMeterNumbers[]);

ViStatus _VI_FUNC rsvna_DeleteExternalPowerMeter(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigurePortPowerLimitEnabled(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   physicalPort,
                                                       ViBoolean enabled,
                                                       ViReal64  portPowerLimit);

ViStatus _VI_FUNC rsvna_ConfigureExternalGeneratorLimit(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   generatorNumber,
                                                        ViBoolean enabled,
                                                        ViReal64  externalGeneratorLimit);

ViStatus _VI_FUNC rsvna_ConfigureSwitchMatrixState(ViSession instrSession,
                                                   ViBoolean switchMatrixState);

ViStatus _VI_FUNC rsvna_ConfigureSwitchMatrix(ViSession instrSession,
                                              ViInt32   switchMatrixNumber,
                                              ViString  switchMatrixName,
                                              ViString  switchMatrixType,
                                              ViString  interfaceType,
                                              ViString  interfaceAddress);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrix(ViSession instrSession,
                                          ViInt32   switchMatrixNumber,
                                          ViChar    switchMatrixName[],
                                          ViChar    switchMatrixType[],
                                          ViChar    interfaceType[],
                                          ViChar    interfaceAddress[]);

ViStatus _VI_FUNC rsvna_DeteleAllSwitchMatrixes(ViSession instrSession);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixCount(ViSession instrSession,
                                               ViInt32*  switchMatrixCount);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixNumbers(ViSession instrSession,
                                                 ViInt32   bufferSize,
                                                 ViChar    switchMatrixNumbers[]);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixTestPortsCount(ViSession instrSession,
                                                        ViInt32*  testPortsCount);

ViStatus _VI_FUNC rsvna_SwitchMatrixConfigurationOperation(ViSession instrSession,
                                                           ViInt32   operation);

ViStatus _VI_FUNC rsvna_ConfigureVNAPortToMatrixVNAPort(ViSession instrSession,
                                                        ViInt32   switchMatrixNumber,
                                                        ViInt32   numberOfPorts,
                                                        ViInt32   matrixVNAPorts[],
                                                        ViInt32   VNAPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureVNAPortToMatrixVNAPortLabel(ViSession instrSession,
                                                             ViInt32   switchMatrixNumber,
                                                             ViInt32   numberOfPorts,
                                                             ViChar    matrixVNAPortLabels[],
                                                             ViInt32   VNAPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureMatrixVNAPortToTestPort(ViSession instrSession,
                                                         ViInt32   switchMatrixNumber,
                                                         ViInt32   numberOfPorts,
                                                         ViInt32   matrixTestPorts[],
                                                         ViInt32   testPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureMatrixVNAPortToTestPortLabel(
                            ViSession instrSession,
                            ViInt32   switchMatrixNumber,
                            ViInt32   numberOfPorts,
                            ViChar    matrixTestPortLabels[],
                            ViInt32   testPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureVNAPortToTestPort(ViSession instrSession,
                                                   ViInt32   switchMatrixNumber,
                                                   ViInt32   numberOfPorts,
                                                   ViInt32   testPorts[],
                                                   ViInt32   VNAPorts[]);

ViStatus _VI_FUNC rsvna_ConfigureSwitchMatrixOptimization(ViSession instrSession,
                                                          ViInt32   optimization);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixRelayCount(ViSession instrSession,
                                                    ViInt32   switchMatrixNumber,
                                                    ViInt32   arraySize,
                                                    ViInt32   relayCounts[]);

ViStatus _VI_FUNC rsvna_QueryExternalSwitchMatrices(ViSession instrSession,
                                                    ViInt32   bufferSize,
                                                    ViChar    switchMatrices[]);

ViStatus _VI_FUNC rsvna_QuerySwitchMatrixSerialNumber(ViSession instrSession,
                                                      ViInt32   switchMatrixNumber,
                                                      ViInt32   bufferSize,
                                                      ViChar    serialNumber[]);

ViStatus _VI_FUNC rsvna_ConfigureUserDefinedPhysicalPort(ViSession instrSession,
                                                         ViInt32   physicalPort,
                                                         ViString  portRedefinition);

ViStatus _VI_FUNC rsvna_DeleteUserDefinedPhysicalPorts(ViSession instrSession);

ViStatus _VI_FUNC rsvna_DataMemoryInitializeNewSharedBuffer(ViSession instrSession);

ViStatus _VI_FUNC rsvna_DataMemoryAddTrace(ViSession instrSession,
                                           ViString  traceName,
                                           ViInt32   dataFormat,
                                           ViInt32   numberOfPoints);

ViStatus _VI_FUNC rsvna_QueryDataMemoryNamesCatalog(ViSession instrSession,
                                                    ViInt32   bufferSize,
                                                    ViChar    catalog[],
                                                    ViInt32*  responseLength);

ViStatus _VI_FUNC rsvna_DataMemoryCommitName(ViSession instrSession,
                                             ViString  name);

ViStatus _VI_FUNC rsvna_DataMemoryDelete(ViSession instrSession,
                                         ViString  name);

ViStatus _VI_FUNC rsvna_QueryDataMemoryOffset(ViSession instrSession,
                                              ViInt32*  offset);

ViStatus _VI_FUNC rsvna_QueryDataMemorySize(ViSession instrSession,
                                            ViInt32*  size);

ViStatus _VI_FUNC rsvna_DataMemoryDeleteAllSharedBuffers(ViSession instrSession);

ViStatus _VI_FUNC rsvna_ConfigureHUMSEnabled(ViSession instrSession,
                                             ViBoolean HUMSEnabled,
                                             ViBoolean RESTEnabled,
                                             ViInt32   SNMPVersion);

ViStatus _VI_FUNC rsvna_HUMSExportHistoryToFile(ViSession instrSession,
                                                ViString  filename);

ViStatus _VI_FUNC rsvna_HUMSDeleteHistory(ViSession instrSession);

ViStatus _VI_FUNC rsvna_HUMSDeleteDeviceHistory(ViSession instrSession);

ViStatus _VI_FUNC rsvna_HUMSDeleteUtilizationHistory(ViSession instrSession);

ViStatus _VI_FUNC rsvna_HUMSQueryDataFromInstrument(ViSession instrSession,
                                                    ViInt32   format,
                                                    ViInt32   data,
                                                    ViInt32   bufferSize,
                                                    ViChar    buffer[],
                                                    ViInt32*  returnedBytes);

ViStatus _VI_FUNC rsvna_HUMSSaveDataFromInstrumentToFile(ViSession instrSession,
                                                         ViInt32   format,
                                                         ViInt32   data,
                                                         ViString  filename);

ViStatus _VI_FUNC rsvna_ConfigureHUMSSNMPCommunityStrings(ViSession instrSession,
                                                          ViString  readOnly,
                                                          ViString  readWrite);

ViStatus _VI_FUNC rsvna_ConfigureHUMSSNMPSystemInfo(ViSession instrSession,
                                                    ViString  systemContact,
                                                    ViString  systemLocation);

ViStatus _VI_FUNC rsvna_QueryHUMSSNMPSystemInfo(ViSession instrSession,
                                                ViInt32   buffersSize,
                                                ViChar    systemContact[],
                                                ViChar    systemLocation[]);

ViStatus _VI_FUNC rsvna_HUMSSNMPDefineUserProfile(ViSession instrSession,
                                                  ViString  username,
                                                  ViInt32   access,
                                                  ViInt32   security,
                                                  ViString  authPwd,
                                                  ViString  privPwd);

ViStatus _VI_FUNC rsvna_QueryHUMSSNMPUserList(ViSession instrSession,
                                              ViInt32   bufferSize,
                                              ViChar    userList[],
                                              ViInt32*  returnedBytes);

ViStatus _VI_FUNC rsvna_HUMSSNMPDeleteUser(ViSession instrSession,
                                           ViString  user);

ViStatus _VI_FUNC rsvna_HUMSSNMPDeleteAllUsers(ViSession instrSession);

ViStatus _VI_FUNC rsvna_HUMSTagDefineKeyValuePair(ViSession instrSession,
                                                  ViInt32   tagID,
                                                  ViString  key,
                                                  ViString  value);

ViStatus _VI_FUNC rsvna_QueryHUMSTagKeyValuePair(ViSession instrSession,
                                                 ViInt32   tagID,
                                                 ViInt32   bufferSize,
                                                 ViChar    keyValuePair[],
                                                 ViInt32*  returnedBytes);

ViStatus _VI_FUNC rsvna_QueryHUMSAllDeviceTags(ViSession instrSession,
                                               ViInt32   bufferSize,
                                               ViChar    deviceTags[],
                                               ViInt32*  returnedBytes);

ViStatus _VI_FUNC rsvna_HUMSTagDelete(ViSession instrSession,
                                      ViInt32   tagID);

ViStatus _VI_FUNC rsvna_HUMSTagDeleteAll(ViSession instrSession);

ViStatus _VI_FUNC rsvna_HUMSUtilizationTrackingEnable(ViSession instrSession,
                                                      ViInt32   utilizationID,
                                                      ViBoolean trackingEnabled);

ViStatus _VI_FUNC rsvna_QueryHUMSDeviceStatusSummary(ViSession instrSession,
                                                     ViInt32*  deviceStatus);

ViStatus _VI_FUNC rsvna_HUMSAddEvent(ViSession instrSession,
                                     ViInt32   severity,
                                     ViString  message,
                                     ViString  detail);

ViStatus _VI_FUNC rsvna_HUMSUpdateCustomEvent(ViSession instrSession,
                                              ViInt32   ID,
                                              ViString  value,
                                              ViInt32   mode,
                                              ViBoolean state);

ViStatus _VI_FUNC rsvna_HUMSUtilizationCreatesCustom(ViSession instrSession,
                                                     ViInt32   ID,
                                                     ViString  name,
                                                     ViString  description,
                                                     ViString  unit,
                                                     ViBoolean enabled);

ViStatus _VI_FUNC rsvna_QueryHUMSUtilizationAll(ViSession instrSession,
                                                ViInt32   bufferSize,
                                                ViChar    customUtilization[]);

ViStatus _VI_FUNC rsvna_HUMSUtilizationDeleteCustomAll(ViSession instrSession);

ViStatus _VI_FUNC rsvna_HUMSUtilizationDeleteCustom(ViSession instrSession,
                                                    ViInt32   deleteUtilization);

ViStatus _VI_FUNC rsvna_QueryHUMSUtilizationHistoryDetailed(ViSession instrSession,
                                                            ViInt32   ID,
                                                            ViString  startDate,
                                                            ViString  endDate,
                                                            ViReal64  granularity,
                                                            ViChar    history[]);

ViStatus _VI_FUNC rsvna_QueryServiceCalibrationLastDate(ViSession instrSession,
                                                        ViInt32   bufferSize,
                                                        ViChar    date[]);

ViStatus _VI_FUNC rsvna_QueryServiceCalibrationDueDate(ViSession instrSession,
                                                       ViInt32   bufferSize,
                                                       ViChar    date[]);

ViStatus _VI_FUNC rsvna_QueryServiceCalibrationDueState(ViSession instrSession,
                                                        ViInt32*  calibrationDue);

ViStatus _VI_FUNC rsvna_QueryServiceCalibrationInterval(ViSession instrSession,
                                                        ViInt32   bufferSize,
                                                        ViChar    interval[]);

ViStatus _VI_FUNC rsvna_QueryServiceLastServiceDate(ViSession instrSession,
                                                    ViInt32   bufferSize,
                                                    ViChar    lastServiceDate[]);

ViStatus _VI_FUNC rsvna_QueryServiceDueState(ViSession instrSession,
                                             ViInt32*  serviceDue);

ViStatus _VI_FUNC rsvna_QueryGenericDeviceCatalog(ViSession instrSession,
                                                  ViInt32   bufferSize,
                                                  ViChar    generators[]);

ViStatus _VI_FUNC rsvna_GenericDeviceDeleteAll(ViSession instrSession);

ViStatus _VI_FUNC rsvna_GenericDeviceDelete(ViSession instrSession,
                                            ViInt32   genericDevice);

ViStatus _VI_FUNC rsvna_ConfigureGenericDevice(ViSession instrSession,
                                               ViInt32   genericDevice,
                                               ViString  interfaceType,
                                               ViString  address);

ViStatus _VI_FUNC rsvna_QueryGenericDeviceAddresses(ViSession instrSession,
                                                    ViInt32   bufferSize,
                                                    ViChar    addresses[]);

ViStatus _VI_FUNC rsvna_ConfigureGenericDevicesDeleteAddress(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViString  deleteAddress);

ViStatus _VI_FUNC rsvna_SelectGenericDevice(ViSession instrSession,
                                            ViInt32   channel,
                                            ViString  address,
                                            ViString  commandFilepath,
                                            ViString  commandId);

ViStatus _VI_FUNC rsvna_ConfigureGenericDevicesCharacterization(
                            ViSession instrSession,
                            ViString  characterization);

ViStatus _VI_FUNC rsvna_QueryGenericDeviceConfiguredDevices(ViSession instrSession,
                                                            ViInt32   bufferSize,
                                                            ViChar    devices[]);

ViStatus _VI_FUNC rsvna_GenericDeviceCalibrationMixer(ViSession instrSession,
                                                      ViInt32   genericDevice,
                                                      ViString  address,
                                                      ViString  commandFilePath,
                                                      ViString  mixerCommandId);

ViStatus _VI_FUNC rsvna_QueryConverterTypesCatalog(ViSession instrSession,
                                                   ViInt32   bufferSize,
                                                   ViChar    catalog[]);

ViStatus _VI_FUNC rsvna_QueryConverterCount(ViSession instrSession,
                                            ViInt32*  count);

ViStatus _VI_FUNC rsvna_DefineConverterNewType(ViSession instrSession,
                                               ViString  define);

ViStatus _VI_FUNC rsvna_DeleteConverterDefinition(ViSession instrSession,
                                                  ViString  converterDefinition);

ViStatus _VI_FUNC rsvna_ConfigureConverterStartStopFrequency(ViSession instrSession,
                                                             ViString  type,
                                                             ViReal64  startFrequency,
                                                             ViReal64  stopFrequency);

ViStatus _VI_FUNC rsvna_QueryConverterStartStopFrequency(ViSession instrSession,
                                                         ViString  type,
                                                         ViReal64* startFrequency,
                                                         ViReal64* stopFrequency);

ViStatus _VI_FUNC rsvna_ConfigureConverterIFFrequency(ViSession instrSession,
                                                      ViString  type,
                                                      ViReal64  IFFrequency);

ViStatus _VI_FUNC rsvna_QueryConverterIFFrequency(ViSession instrSession,
                                                  ViString  type,
                                                  ViReal64* IFFrequency);

ViStatus _VI_FUNC rsvna_ConfigureConverterLocalMultiplicator(ViSession instrSession,
                                                             ViString  type,
                                                             ViInt32   localMultiplicator);

ViStatus _VI_FUNC rsvna_QueryConverterLocalMultiplicator(ViSession instrSession,
                                                         ViString  type,
                                                         ViInt32*  localMultiplicator);

ViStatus _VI_FUNC rsvna_ConfigureConverterLocalPower(ViSession instrSession,
                                                     ViString  type,
                                                     ViReal64  localPower,
                                                     ViReal64  offset);

ViStatus _VI_FUNC rsvna_QueryConverterLocalPower(ViSession instrSession,
                                                 ViString  type,
                                                 ViReal64* localPower,
                                                 ViReal64* offset);

ViStatus _VI_FUNC rsvna_ConfigureConverterMaximumInputPower(ViSession instrSession,
                                                            ViString  type,
                                                            ViReal64  maximumInputPower);

ViStatus _VI_FUNC rsvna_QueryConverterMaximumInputPower(ViSession instrSession,
                                                        ViString  type,
                                                        ViReal64* maximumInputPower);

ViStatus _VI_FUNC rsvna_ConfigureConverterSourceMultiplicator(
                            ViSession instrSession,
                            ViString  type,
                            ViInt32   multiplicator);

ViStatus _VI_FUNC rsvna_QueryConverterSourceMultiplicator(ViSession instrSession,
                                                          ViString  type,
                                                          ViInt32*  multiplicator);

ViStatus _VI_FUNC rsvna_ConfigureConverterSourceNominalPower(ViSession instrSession,
                                                             ViString  type,
                                                             ViReal64  nominalPower,
                                                             ViReal64  nominalPowerOffset);

ViStatus _VI_FUNC rsvna_QueryConverterSourceNominalPower(ViSession instrSession,
                                                         ViString  type,
                                                         ViReal64* nominalPower,
                                                         ViReal64* nominalPowerOffset);

ViStatus _VI_FUNC rsvna_ConfigureFrequencyConversionVNAPort(ViSession instrSession,
                                                            ViInt32   physicalPort,
                                                            ViString  VNAPort);

ViStatus _VI_FUNC rsvna_ConfigureConverterSplitterLOPort(ViSession instrSession,
                                                         ViInt32   splitter,
                                                         ViInt32   localPort,
                                                         ViInt32   localIndex);

ViStatus _VI_FUNC rsvna_ConfigureConverterSplitterLoss(ViSession instrSession,
                                                       ViInt32   splitter,
                                                       ViReal64  splitterLoss);

ViStatus _VI_FUNC rsvna_ConfigureConverterDefault(ViSession instrSession,
                                                  ViInt32   physicalPort,
                                                  ViReal64  cableLoss,
                                                  ViReal64  slopeFactor);

ViStatus _VI_FUNC rsvna_ConfigureConverterIF(ViSession instrSession,
                                             ViInt32   physicalPort,
                                             ViInt32   IFInput,
                                             ViReal64  IFFrequency);

ViStatus _VI_FUNC rsvna_ConfigureConverterLOLoss(ViSession instrSession,
                                                 ViInt32   physicalPort,
                                                 ViReal64  loss);

ViStatus _VI_FUNC rsvna_ConfigureConverterLOPort(ViSession instrSession,
                                                 ViInt32   physicalPort,
                                                 ViInt32   local,
                                                 ViInt32   index);

ViStatus _VI_FUNC rsvna_ConfigureConverterSourcePortLoss(ViSession instrSession,
                                                         ViInt32   physicalPort,
                                                         ViReal64  portLoss);

ViStatus _VI_FUNC rsvna_QueryConverterSourcePort(ViSession instrSession,
                                                 ViInt32   physicalPort,
                                                 ViInt32*  portType,
                                                 ViInt32*  sourcePort);

ViStatus _VI_FUNC rsvna_ConfigureConverterPortEnabled(ViSession instrSession,
                                                      ViInt32   physicalPort,
                                                      ViBoolean enabled);

ViStatus _VI_FUNC rsvna_ConfigureConverterPortRFSlope(ViSession instrSession,
                                                      ViInt32   physicalPort,
                                                      ViReal64  RFSlope);

ViStatus _VI_FUNC rsvna_ConfigureLevelingDatasetPowerControlMode(
                            ViSession instrSession,
                            ViInt32   physicalPort,
                            ViInt32   controlMode);

ViStatus _VI_FUNC rsvna_ConfigureLevelingDatasetMode(ViSession instrSession,
                                                     ViInt32   physicalPort,
                                                     ViInt32   mode);

ViStatus _VI_FUNC rsvna_ConfigureLevelingDatasetFile(ViSession instrSession,
                                                     ViInt32   physicalPort,
                                                     ViString  filename);

ViStatus _VI_FUNC rsvna_ConfigureLevelingDatasetPath(ViSession instrSession,
                                                     ViInt32   physicalPort,
                                                     ViString  path);

ViStatus _VI_FUNC rsvna_reset(ViSession instrSession);

ViStatus _VI_FUNC rsvna_SystemShutdown(ViSession instrSession);

ViStatus _VI_FUNC rsvna_self_test(ViSession instrSession,
                                  ViInt16*  selfTestResult,
                                  ViChar    selfTestMessage[]);

ViStatus _VI_FUNC rsvna_revision_query(ViSession instrSession,
                                       ViChar    instrumentDriverRevision[],
                                       ViChar    firmwareRevision[]);

ViStatus _VI_FUNC rsvna_error_query(ViSession instrSession,
                                    ViInt32*  errorCode,
                                    ViChar    errorMessage[]);

ViStatus _VI_FUNC rsvna_error_message(ViSession instrSession,
                                      ViStatus  errorCode,
                                      ViChar    errorMessage[]);

ViStatus _VI_FUNC rsvna_ErrorCount(ViSession instrSession,
                                   ViInt32*  systemErrorCount);

ViStatus _VI_FUNC rsvna_SystemVersion(ViSession instrSession,
                                      ViInt32   length,
                                      ViChar    systemVersion[]);

ViStatus _VI_FUNC rsvna_setCheckStatus(ViSession instrSession,
                                       ViBoolean state);

ViStatus _VI_FUNC rsvna_ConfigureErrorChecking(ViSession instrSession,
                                               ViBoolean optionChecking,
                                               ViBoolean rangeChecking,
                                               ViBoolean statusChecking);

ViStatus _VI_FUNC rsvna_TemporaryDisableOption(ViSession instrSession,
                                               ViString  optionName,
                                               ViInt32   disableOption);

ViStatus _VI_FUNC rsvna_ClearStatus(ViSession instrSession);

ViStatus _VI_FUNC rsvna_IDQueryResponse(ViSession instrSession,
                                        ViUInt32  bufferSize,
                                        ViChar    IDQueryResponse[]);

ViStatus _VI_FUNC rsvna_ProcessAllPreviousCommands(ViSession instrSession);

ViStatus _VI_FUNC rsvna_QueryOPC(ViSession instrSession,
                                 ViInt32*  OPC);

ViStatus _VI_FUNC rsvna_SetOPCTimeout(ViSession instrSession,
                                      ViInt32   OPCTimeout);

ViStatus _VI_FUNC rsvna_GetOPCTimeout(ViSession instrSession,
                                      ViInt32*  timeout);

ViStatus _VI_FUNC rsvna_SetVISATimeout(ViSession instrSession,
                                       ViUInt32  VISATimeout);

ViStatus _VI_FUNC rsvna_GetVISATimeout(ViSession instrSession,
                                       ViUInt32* VISATimeout);

ViStatus _VI_FUNC rsvna_ConfigureAutoSystemErrQuery(ViSession instrSession,
                                                    ViBoolean autoSystErrQuery);

ViStatus _VI_FUNC rsvna_ConfigureMultiThreadLocking(ViSession instrSession,
                                                    ViBoolean multiThreadLocking);

ViStatus _VI_FUNC rsvna_ExecuteProgram(ViSession instrSession,
                                       ViString  directory,
                                       ViString  fileToExecute,
                                       ViString  fileParameters,
                                       ViBoolean waitToFinish);

ViStatus _VI_FUNC rsvna_QueryProgramReturnValue(ViSession instrSession,
                                                ViInt32   bufferSize,
                                                ViChar    returnValue[]);

ViStatus _VI_FUNC rsvna_GetError(ViSession instrSession,
                                 ViStatus* code,
                                 ViInt32   bufferSize,
                                 ViChar    description[]);

ViStatus _VI_FUNC rsvna_ClearError(ViSession instrSession);

ViStatus _VI_FUNC rsvna_WriteInstrData(ViSession instrSession,
                                       ViString  writeBuffer);

ViStatus _VI_FUNC rsvna_QueryViBoolean(ViSession  instrSession,
                                       ViString   command,
                                       ViBoolean* value);

ViStatus _VI_FUNC rsvna_QueryViInt32(ViSession instrSession,
                                     ViString  command,
                                     ViInt32*  value);

ViStatus _VI_FUNC rsvna_QueryViReal64(ViSession instrSession,
                                      ViString  command,
                                      ViReal64* value);

ViStatus _VI_FUNC rsvna_QueryViString(ViSession instrSession,
                                      ViString  command,
                                      ViInt32   bufferSize,
                                      ViChar    value[]);

ViStatus _VI_FUNC rsvna_SetAttributeViInt32(ViSession instrSession,
                                            ViString  channelName,
                                            ViAttr    attributeId,
                                            ViInt32   attributeValue);

ViStatus _VI_FUNC rsvna_SetAttributeViInt64(ViSession instrSession,
                                            ViString  channelName,
                                            ViAttr    attributeId,
                                            ViInt64   attributeValue);

ViStatus _VI_FUNC rsvna_SetAttributeViReal64(ViSession instrSession,
                                             ViString  channelName,
                                             ViAttr    attributeId,
                                             ViReal64  attributeValue);

ViStatus _VI_FUNC rsvna_SetAttributeViString(ViSession instrSession,
                                             ViString  channelName,
                                             ViAttr    attributeId,
                                             ViString  attributeValue);

ViStatus _VI_FUNC rsvna_SetAttributeViBoolean(ViSession instrSession,
                                              ViString  channelName,
                                              ViAttr    attributeId,
                                              ViBoolean attributeValue);

ViStatus _VI_FUNC rsvna_GetAttributeViInt32(ViSession instrSession,
                                            ViString  channelName,
                                            ViAttr    attributeId,
                                            ViInt32*  attributeValue);

ViStatus _VI_FUNC rsvna_GetAttributeViInt64(ViSession instrSession,
                                            ViString  channelName,
                                            ViAttr    attributeId,
                                            ViInt64*  attributeValue);

ViStatus _VI_FUNC rsvna_GetAttributeViReal64(ViSession instrSession,
                                             ViString  channelName,
                                             ViAttr    attributeId,
                                             ViReal64* attributeValue);

ViStatus _VI_FUNC rsvna_GetAttributeViString(ViSession instrSession,
                                             ViString  channelName,
                                             ViAttr    attributeId,
                                             ViInt32   bufferSize,
                                             ViChar    attributeValue[]);

ViStatus _VI_FUNC rsvna_GetAttributeViBoolean(ViSession  instrSession,
                                              ViString   channelName,
                                              ViAttr     attributeId,
                                              ViBoolean* attributeValue);

ViStatus _VI_FUNC rsvna_SynchronizeDynamicRepeatedCapability(ViSession instrSession,
                                                             ViInt32   repeatedCapability);

ViStatus _VI_FUNC rsvna_GetAttributeRepCapName(ViSession instrSession,
                                               ViAttr    attributeId,
                                               ViString  repCapNameID,
                                               ViInt32   index,
                                               ViInt32   bufferSize,
                                               ViChar    repCapName[]);

ViStatus _VI_FUNC rsvna_GetAttributeRepeatedCapabilityIds(ViSession instrSession,
                                                          ViAttr    attributeId,
                                                          ViInt32   bufferSize,
                                                          ViChar    repeatedCapabilityId_s[]);

ViStatus _VI_FUNC rsvna_GetAttributeRepeatedCapabilityIdNames(
                            ViSession instrSession,
                            ViAttr    attributeId,
                            ViString  repeatedCapabilityId,
                            ViInt32   bufferSize,
                            ViChar    repeatedCapabilityIdName_s[]);

ViStatus _VI_FUNC rsvna_close(ViSession instrSession);

    /****************************************************************************
     *------------------------ Error And Completion Codes ----------------------*
     ****************************************************************************/

#define RSVNA_WARN_MEASURE_UNCALIBRATED    (RS_CLASS_WARN_BASE + 0x0001L)
#define RSVNA_WARN_OVER_RANGE              (RS_CLASS_WARN_BASE + 0x0002L)

#define RSVNA_ERROR_MARKER_NOT_ENABLED     (RS_CLASS_ERROR_BASE + 0x0003L)
#define RSVNA_ERROR_NOT_DELTA_MARKER       (RS_CLASS_ERROR_BASE + 0x0002L)
#define RSVNA_ERROR_MAX_TIME_EXCEEDED      (RS_CLASS_ERROR_BASE + 0x0001L)

#define RSVNA_WARNMSG_MEASURE_UNCALIBRATED "Uncalibrated measurement"
#define RSVNA_WARNMSG_OVER_RANGE           "Measurement overrange"

#define RSVNA_ERRMSG_MARKER_NOT_ENABLED    "Marker Not Enabled"
#define RSVNA_ERRMSG_NOT_DELTA_MARKER      "Not Delta Marker"
#define RSVNA_ERRMSG_MAX_TIME_EXCEEDED     "Max Time Exceeded"

#define RSVNA_ERROR_CODES_AND_MSGS \
    RSVNA_WARN_MEASURE_UNCALIBRATED,  RSVNA_WARNMSG_MEASURE_UNCALIBRATED  ,\
    RSVNA_WARN_OVER_RANGE,            RSVNA_WARNMSG_OVER_RANGE  ,\
    {RSVNA_ERROR_MARKER_NOT_ENABLED,   RSVNA_ERRMSG_MARKER_NOT_ENABLED},\
    {RSVNA_ERROR_NOT_DELTA_MARKER,     RSVNA_ERRMSG_NOT_DELTA_MARKER},\
    {RSVNA_ERROR_MAX_TIME_EXCEEDED,    RSVNA_ERRMSG_MAX_TIME_EXCEEDED}

    /****************************************************************************/
    /*---------------------------- End Include File ----------------------------*/
    /****************************************************************************/

#if defined(__cplusplus) || defined(__cplusplus__)
}
#endif
#endif /* __RSVNA_HEADER */
